﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Scryber
{
    /// <summary>
    /// Defines a complete list of all the data items
    /// in a data source schema for a specific path.
    /// Allows runtime analysis and evaluation of a data source.
    /// </summary>
    public class PDFDataSchema
    {
        //
        // properties
        //

        #region public string RootPath { get; private set; }

        /// <summary>
        /// Gets the path from the root where this data schema refers to.
        /// </summary>
        public string RootPath { get; private set; }

        #endregion

        #region public PDFDataItemCollection Items { get; private set; }

        /// <summary>
        /// Gets the collection of data items in this schema
        /// </summary>
        public PDFDataItemCollection Items { get; private set; }

        #endregion

        //
        // .ctor
        //

        #region public PDFDataSchema(string rootpath, PDFDataItemCollection items)

        /// <summary>
        /// Creates a new instance of the data schema with the specified root path and items
        /// </summary>
        /// <param name="rootpath"></param>
        /// <param name="items"></param>
        public PDFDataSchema(string rootpath, PDFDataItemCollection items)
        {
            this.RootPath = rootpath;
            this.Items = items;
        }

        #endregion
    }

    /// <summary>
    /// Represents a single chunck of selectable data returned from a data source.
    /// This can either be an item with inner child items, or a discreet value source.
    /// </summary>
    public class PDFDataItem
    {

        #region public string Path { get; private set; }

        /// <summary>
        /// Gets the path relative to the container for the data item
        /// </summary>
        public string Path { get; private set; }

        #endregion

        #region public string Name { get; private set; }

        /// <summary>
        /// Gets the human readable name for this data item
        /// </summary>
        public string Name { get; private set; }

        #endregion

        #region public Type DataType { get; private set; }

        /// <summary>
        /// Gets the underlying data type for this data item
        /// </summary>
        public Type DataType { get; private set; }

        #endregion

        #region public bool HasChildItems { get; private set; }

        /// <summary>
        /// Returns true if this data item has inner items
        /// </summary>
        public bool HasChildItems { get; private set; }

        #endregion

        #region public PDFDataItemCollection Children { get; private set; }

        /// <summary>
        /// Gets the inner collection of child data items. 
        /// Will be null if this instance has no children
        /// </summary>
        public PDFDataItemCollection Children { get; private set; }

        #endregion

        //
        // .ctor
        //

        #region public PDFDataItem(string path, string name, Type datatype)

        /// <summary>
        /// Creates a new simple data item (with no children)
        /// </summary>
        /// <param name="path"></param>
        /// <param name="name"></param>
        /// <param name="datatype"></param>
        public PDFDataItem(string path, string name, Type datatype)
            : this(path, name, datatype, false, null)
        {
        }

        #endregion

        #region public PDFDataItem(string path, string name, PDFDataItemCollection children)

        /// <summary>
        /// Creates a new container data item (has children and the DataType is Array)
        /// </summary>
        /// <param name="path"></param>
        /// <param name="name"></param>
        /// <param name="children"></param>
        public PDFDataItem(string path, string name, PDFDataItemCollection children)
            : this(path, name, typeof(Array), true, children)
        {
        }

        #endregion

        #region protected PDFDataItem(string path, string name, Type datatype, bool hasChildren, PDFDataItemCollection children)

        /// <summary>
        /// Protected constructor that initializes all the parameters.
        /// </summary>
        /// <param name="path"></param>
        /// <param name="name"></param>
        /// <param name="datatype"></param>
        /// <param name="hasChildren"></param>
        /// <param name="children"></param>
        protected PDFDataItem(string path, string name, Type datatype, bool hasChildren, PDFDataItemCollection children)
        {
            if (null == datatype)
                throw new ArgumentNullException("datatype");
            if (hasChildren && (null == children || children.Count == 0))
                throw new ArgumentNullException("children");
            if (string.IsNullOrEmpty(path))
                throw new ArgumentNullException("path");
            if (string.IsNullOrEmpty(name))
                throw new ArgumentNullException("name");

            this.Path = path;
            this.Name = name;
            this.DataType = datatype;
            this.HasChildItems = hasChildren;
            this.Children = children;
        }

        #endregion
    }

    /// <summary>
    /// Represents a read only collection of data items (either as children of 
    /// another data item or as the root items in the schema)
    /// </summary>
    public class PDFDataItemCollection : IEnumerable<PDFDataItem>
    {

        #region ivars

        private List<PDFDataItem> _inneritems;

        #endregion

        //
        // properties
        //

        #region public int Count {get;}

        /// <summary>
        /// Gets the number of items in this collection
        /// </summary>
        public int Count
        {
            get { return _inneritems.Count; }
        }

        #endregion

        #region public bool IsReadOnly {get;}

        /// <summary>
        /// Returns true - as this collection IS readonly
        /// </summary>
        public bool IsReadOnly
        {
            get { return true; }
        }

        #endregion

        #region public PDFDataItem this[int index] {get;}

        /// <summary>
        /// Gets the data item at the specified index.
        /// </summary>
        /// <param name="index"></param>
        /// <returns></returns>
        public PDFDataItem this[int index]
        {
            get { return _inneritems[index]; }
        }

        #endregion

        #region public PDFDataItem this[string path] {get;}

        /// <summary>
        /// Gets the data item with the specified path (case sensitive)
        /// </summary>
        /// <param name="path"></param>
        /// <returns></returns>
        public PDFDataItem this[string path]
        {
            get
            {
                for (int i = 0; i < _inneritems.Count; i++)
                {
                    if (string.Equals(_inneritems[i].Path, path, StringComparison.Ordinal))
                        return _inneritems[i];
                }
                return null;
            }
        }

        #endregion

        //
        // .ctor
        //

        #region public PDFDataItemCollection(IEnumerable<PDFDataItem> items)

        /// <summary>
        /// Creates a new instace of the PDFDataItemCollection with a read only collection of items specified
        /// </summary>
        /// <param name="items">The PDFDataItes in this collection</param>
        public PDFDataItemCollection(IEnumerable<PDFDataItem> items)
        {
            this._inneritems = new List<PDFDataItem>(items);
        }

        #endregion

        //
        // methods
        //

        #region public void CopyTo(PDFDataItem[] array, int arrayIndex)

        /// <summary>
        /// Copies the items in this collection to an array starting at the specified index.
        /// </summary>
        /// <param name="array"></param>
        /// <param name="arrayIndex"></param>
        public void CopyTo(PDFDataItem[] array, int arrayIndex)
        {
            _inneritems.CopyTo(array, arrayIndex);
        }

        #endregion

        #region public IEnumerator<PDFDataItem> GetEnumerator()

        /// <summary>
        /// Returns an enumerator for the data item collection
        /// </summary>
        /// <returns></returns>
        public IEnumerator<PDFDataItem> GetEnumerator()
        {
            return _inneritems.GetEnumerator();
        }

        #endregion

        #region System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion
    }

}
