﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using Scryber.Drawing;
using Scryber.Styles;

namespace Scryber.Components.Layout
{
    internal class DocumentLayoutEngine : PDFBlockLayoutEngineBase
    {

        //
        // properties
        //

        #region public PDFDocument Document {get;}

        /// <summary>
        /// Gets the document associated with this layout engine
        /// </summary>
        public PDFDocument Document
        {
            get { return base.Component as PDFDocument; }
        }

        #endregion

        //
        // ctor(s)
        //

        #region internal DocumentLayoutEngine(PDFDocument doc, IPDFLayoutEngine parent, PDFLayoutContext context)

        /// <summary>
        /// Creates a new instance of the document layout engine which can arrange all the pages and the components for this document into a PDFDocumentLayout
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="parent"></param>
        /// <param name="context"></param>
        internal DocumentLayoutEngine(PDFDocument doc, IPDFLayoutEngine parent, PDFLayoutContext context)
            : base(doc, parent)
        {
        }

        #endregion

        //
        // implementation methods
        //

        
        /// <summary>
        /// Main calling method that lays out each of the pages in this document
        /// </summary>
        /// <param name="avail">Available size (ignored)</param>
        /// <param name="startPageIndex">The starting page index</param>
        /// <returns>an empty size</returns>
        protected override void DoLayoutComponent()
        {
#if MEASURE_PERFORMANCE
            _fullstyleduration.Reset();
            _textlayoutduration.Reset();
            _measurestringducration.Reset();
            _fullstylecount = 0;
            _fulltextcount = 0;
            _measurecount = 0;
#endif

            this.Context.DocumentLayout = CreateDocumentLayout();
            PDFStyle style = this.FullStyle;
            this.ApplyDefaultPageNumbering(style);

            //Do the registration of top level document artefacts
            PDFArtefactRegistrationSet artefacts = this.Document.RegisterLayoutArtefacts(this.Context, style);

            foreach (PDFPage page in this.Document.Pages)
            {
                page.ValidateLicense();
                if (page.Visible)
                    this.LayoutPage(page);
            }

            if (null != artefacts)
                this.Document.CloseLayoutArtefacts(Context, artefacts, style);
            
#if MEASURE_PERFORMANCE

            this.Context.TraceLog.Add(TraceLevel.Message, "Timings", "Total for styles : " + _fullstyleduration.Elapsed.ToString() + " for " + _fullstylecount + " calls");
            this.Context.TraceLog.Add(TraceLevel.Message, "Timings", "Total for text : " + _textlayoutduration.Elapsed.ToString() + " for " + _fulltextcount + " calls");
            this.Context.TraceLog.Add(TraceLevel.Message, "Timings", "Total for measure string : " + _measurestringducration.Elapsed.ToString() + " for " + _measurecount + " calls");
            System.Diagnostics.Stopwatch sw = System.Diagnostics.Stopwatch.StartNew();
            this.Context.DocumentLayout.PushComponentLayout(this.Context, 0, PDFUnit.Zero, PDFUnit.Zero);
            sw.Stop();
            this.Context.TraceLog.Add(TraceLevel.Message, "Timings", "Total for pushing component layout : " + sw.Elapsed.ToString());
#else
            this.Context.DocumentLayout.PushComponentLayout(this.Context, 0, PDFUnit.Zero, PDFUnit.Zero);

#endif


        }


        protected virtual void ApplyDefaultPageNumbering(PDFStyle full)
        {
            PDFPageStyle pg;
            full.TryGetPaper(out pg);
            this.DocumentLayout.InitPageNumbering(pg);
        }

        private PDFPageNumbering _assignedNumbers;
       
        private void RegisterPageNumbering(PDFLayoutPage page, PDFPageStyle numbers)
        {
             this.DocumentLayout.RegisterPageNumbering(page, numbers, out _assignedNumbers);
        }

        private void UnRegisterPageNumbering(PDFLayoutPage last, PDFPageStyle numbers)
        {
            this.DocumentLayout.UnRegisterPageNumbering(-1, last, _assignedNumbers);
        }

        #region protected virtual PDFLayoutDocument CreateDocumentLayout()

        /// <summary>
        /// Creates a new PDFLayoutDocument and returns. Inheritors can override this implementation to return their own instance
        /// </summary>
        protected virtual PDFLayoutDocument CreateDocumentLayout()
        {
            return new PDFLayoutDocument(this.Document, this);
        }

        #endregion

        #region protected virtual PDFSize LayoutPage(PDFPage pg)

        /// <summary>
        /// Lays out the individual page specified within the document - by calling the pages IPDFViewPortComponent interface
        /// </summary>
        /// <param name="pg">The page to perform the layout on</param>
        protected virtual void LayoutPage(PDFPage pg)
        {
            PDFStyle style = pg.GetAppliedStyle();
            
            if(null != style)
                this.StyleStack.Push(style);

            PDFStyle full = this.Context.StyleStack.GetFullStyle(pg);

            PDFArtefactRegistrationSet artefacts = pg.RegisterLayoutArtefacts(this.Context, full);

            using (IPDFLayoutEngine engine = pg.GetEngine(this, this.Context, full))
            {
                if(this.TraceLog.ShouldLog(TraceLevel.Message))
                    this.TraceLog.Begin(TraceLevel.Message, "Document Layout", "Starting the layout of page " + pg.ID);

                engine.Layout(this.Context, full);

                if (this.TraceLog.ShouldLog(TraceLevel.Message))
                    this.TraceLog.End(TraceLevel.Message, "Document Layout", "Completed the layout of page " + pg.ID + " and now on page index " + this.DocumentLayout.CurrentPageIndex.ToString());
            }

            if (null != artefacts)
                pg.CloseLayoutArtefacts(this.Context, artefacts, full);

            if (null != style)
                this.Context.StyleStack.Pop();
        }

        #endregion


    }
}
