﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Scryber.Drawing;
using Scryber.Styles;

namespace Scryber.Components.Layout
{
    public abstract class PDFBlockLayoutEngineBase : IPDFLayoutEngine
    {

        internal const string LOG_CATEGORY = "Layout Engine";

        #region ivars

        /// <summary>
        /// An instance reference to the current block that was created and opened by this Panel
        /// Layout engine and therefore should be closed at the end. Can be updated by moving regions
        /// and pages as we close and open new blocks.
        /// </summary>
        private PDFLayoutBlock _currentBlock;

        /// <summary>
        /// An instance reference to the Component this engine is laying out
        /// </summary>
        private PDFContainerComponent _component;

        /// <summary>
        /// An instance reference to the current context
        /// </summary>
        private PDFLayoutContext _context;

        /// <summary>
        /// An instance reference to the FullStyle
        /// </summary>
        private PDFStyle _style;

        #endregion

        #region protected PDFContainerComponent Component

        /// <summary>
        /// Gets the component this engine is laying out
        /// </summary>
        protected PDFContainerComponent Component
        {
            get { return _component; }
        }

        #endregion

        #region protected PDFTraceLog TraceLog {get;}

        /// <summary>
        /// Gets the trace log associated with this layout engine
        /// </summary>
        protected PDFTraceLog TraceLog
        {
            get
            {
                if (null == this._context)
                    return null;
                else
                    return this._context.TraceLog;
            }
        }

        #endregion

        #region protected PDFLayoutContext Context {get;}

        /// <summary>
        /// Gets the layout context associated with this layout engine
        /// </summary>
        public PDFLayoutContext Context
        {
            get { return _context; }
        }

        #endregion

        #region protected PDFStyle FullStyle {get;}

        /// <summary>
        /// Gets or sets the full style for the current component
        /// </summary>
        protected PDFStyle FullStyle
        {
            get { return _style; }
            set { _style = value; }
        }

        #endregion

        #region protected PDFLayoutDocument DocumentLayout

        /// <summary>
        /// Gets the document as the root of this layout context
        /// </summary>
        protected PDFLayoutDocument DocumentLayout
        {
            get
            {
                if (null == _context)
                    return null;
                else
                    return _context.DocumentLayout;
            }
        }

        #endregion

        #region protected PDFStyleStack StyleStack {get;}

        /// <summary>
        /// Gets the StyleStack associated with this layout
        /// </summary>
        protected PDFStyleStack StyleStack
        {
            get { return this.Context.StyleStack; }
        }

        #endregion

        #region protected bool ContinueLayout {get;set;}
        
        /// <summary>
        /// Returns false if we have come to the end of the available space and cannot continue. Inheritors can set this value
        /// </summary>
        protected bool ContinueLayout
        {
            get;
            set;
        }

        #endregion

        #region protected PDFGraphics Graphics

        private PDFGraphics _graphics;

        /// <summary>
        /// Gets the graphics context associated with this layout
        /// </summary>
        protected PDFGraphics Graphics
        {
            get
            {
                if (null == _graphics)
                    _graphics = this.Component.CreateGraphics(this.StyleStack, this.Context);
                return _graphics;
            }
        }

        #endregion

        #region protected PDFLayoutBlock CurrentBlock

        /// <summary>
        /// Gets or sets the current layout block
        /// </summary>
        protected PDFLayoutBlock CurrentBlock
        {
            get { return _currentBlock; }
            set { _currentBlock = value; }
        }

        #endregion

        #region protected IPDFLayoutEngine ParentEngine {get;set;}

        /// <summary>
        /// Gets  the parent engine that called this one
        /// </summary>
        protected IPDFLayoutEngine ParentEngine
        {
            get;
            private set;
        }

        #endregion

        //
        // ctor(s)
        //

        #region protected PDFBlockLayoutEngineBase(IPDFComponent component, IPDFLayoutEngine parent)

        /// <summary>
        /// protected constructor that accepts the component this engine should layout
        /// </summary>
        /// <param name="component"></param>
        protected PDFBlockLayoutEngineBase(PDFContainerComponent component, IPDFLayoutEngine parent)
        {
            this._component = component;
            this.ParentEngine = parent;
        }

        #endregion

        //
        // iIPDFLayoutEngine interface implementation
        //

        #region public void Layout(PDFLayoutContext context, PDFStyle fullstyle)

        /// <summary>
        /// Interface implementation of the LayoutBlock method
        /// </summary>
        /// <param name="context"></param>
        /// <param name="fullstyle"></param>
        public void Layout(PDFLayoutContext context, PDFStyle fullstyle)
        {
            this._context = context;
            this._style = fullstyle;
            this.DoLayoutComponent();
        }

        #endregion

        //
        // abstract methods
        //

        /// <summary>
        /// Abstract method that inheritors must implement in order to actually 
        /// layout the component and any child components
        /// </summary>
        protected abstract void DoLayoutComponent();

        //
        // layout children methods
        //

        #region protected virtual void DoLayoutChildren()

        /// <summary>
        /// enumerates over any / each of the engines component children and performs the correct layout option for these.
        /// The component this engine is laying out should implement the IPDFContainerComponent
        /// </summary>
        protected virtual void DoLayoutChildren()
        {
            //Set this at the top
            this.ContinueLayout = true;
            PDFComponentList children;
            if (TryGetComponentChildren(this.Component, out children))
            {
                DoLayoutChildren(children);
            }
        }

        /// <summary>
        /// Gets the list of child components and returns true if there are some
        /// </summary>
        /// <param name="children"></param>
        /// <returns></returns>
        private static bool TryGetComponentChildren(IPDFComponent parent, out PDFComponentList children)
        {
            children = GetComponentChildren(parent);
            return null != children && children.Count > 0;
        }

        /// <summary>
        /// Returns the child components of the IPDFcomponent if any.
        /// </summary>
        /// <param name="parent"></param>
        /// <returns></returns>
        private static PDFComponentList GetComponentChildren(IPDFComponent parent)
        {
            if (parent is IPDFContainerComponent)
            {
                IPDFContainerComponent contaner = parent as IPDFContainerComponent;
                if (contaner.HasContent)
                    return contaner.Content;
            }
            return null;
        }

        #endregion

        #region protected virtual void DoLayoutChildren(PDFComponentList children)

        /// <summary>
        /// enumerates over each of the components in the specified list 
        /// and performs the layout of each individually
        /// </summary>
        /// <param name="children"></param>
        private void DoLayoutChildren(PDFComponentList children)
        {
            
            foreach (PDFComponent comp in children)
            {
                if (comp.Visible)
                    this.DoLayoutAChild(comp);

                if (this.ContinueLayout == false 
                    || this.DocumentLayout.CurrentPage.IsClosed 
                    || this.DocumentLayout.CurrentPage.CurrentBlock == null)
                    break;
            }
            
        }

        

        #endregion

        #region protected virtual void DoLayoutAChild(PDFComponent comp)

#if MEASURE_PERFORMANCE
        protected static System.Diagnostics.Stopwatch _fullstyleduration = new System.Diagnostics.Stopwatch();
        protected static int _fullstylecount = 0;
        protected static System.Diagnostics.Stopwatch _textlayoutduration = new System.Diagnostics.Stopwatch();
        protected static int _fulltextcount = 0;
        protected static System.Diagnostics.Stopwatch _measurestringducration = new System.Diagnostics.Stopwatch();
        protected static int _measurecount = 0;
#endif

        /// <summary>
        /// Lays out an individual component
        /// </summary>
        /// <param name="comp"></param>
        /// <remarks>Gets the applied style, and pushes this onto the stack.
        /// Extracts the full style for this component based on the style stack
        /// Calls the explict overload DoLayoutAChild(comp,full). </remarks>
        protected void DoLayoutAChild(PDFComponent comp)
        {

#if MEASURE_PERFORMANCE
            _fullstylecount++;
            _fullstyleduration.Start();
#endif

            PDFStyle applied = comp.GetAppliedStyle();
            if (null != applied)
                this.StyleStack.Push(applied);

            PDFStyle full = this.StyleStack.GetFullStyle(comp);

#if MEASURE_PERFORMANCE
            _fullstyleduration.Stop();
#endif
            
            PDFArtefactRegistrationSet artefacts = comp.RegisterLayoutArtefacts(this.Context, full);

            this.DoLayoutAChild(comp, full);
            
            if (null != applied)
                this.StyleStack.Pop();

            //We want to add the child only if it should be rendered and we are a full block
            if (this.ContinueLayout)
                RegisterChildLayout(comp);

            if(null != artefacts)
                comp.CloseLayoutArtefacts(this.Context, artefacts, full);
        }

        #endregion

        #region protected virtual void DoLayoutAChild(IPDFComponent comp, PDFStyle full)

        /// <summary>
        /// Based on the interface implementation of the component (Text, Image, Path , etc)
        /// - calls the appropriate layout method
        /// </summary>
        /// <param name="comp"></param>
        /// <param name="full"></param>
        protected virtual void DoLayoutAChild(IPDFComponent comp, PDFStyle full)
        {
            //Here we can set up any required regions and then do the layout for the explicit types
            PDFLayoutRegion postioned = null;
            PDFPositionOptions pos = full.CreatePostionOptions();

            if (pos.PositionMode == PositionMode.Absolute)
                postioned = this.BeginNewAbsoluteRegionForChild(pos, comp, full);
            
            else if (pos.PositionMode == PositionMode.Relative)
                postioned = this.BeginNewRelativeRegionForChild(pos, comp, full);
            

            if (comp is IPDFViewPortComponent)
                this.DoLayoutViewPortComponent(comp as IPDFViewPortComponent, full);

            else if (comp is IPDFPageBreak)
                this.DoLayoutPageBreak(comp as IPDFPageBreak, full);
            
            else if (comp is IPDFColumnBreak)
                this.DoLayoutColumnBreak(comp as IPDFColumnBreak, full);
            else if (comp is IPDFLineBreak)
            {
                this.DoLayoutLineBreak(comp as IPDFLineBreak, full);
            }
            else if (comp is IPDFTextComponent)
            {
#if MEASURE_PERFORMANCE
                _fulltextcount++;
                _textlayoutduration.Start();
                this.DoLayoutTextComponent(comp as IPDFTextComponent, full);
                _textlayoutduration.Stop();
#else
                this.DoLayoutTextComponent(comp as IPDFTextComponent, full);
#endif


            }

            else if (comp is IPDFImageComponent)
                this.DoLayoutImageComponent(comp as IPDFImageComponent, full);

            else if (comp is IPDFGraphicPathComponent)
                this.DoLayoutPathComponent(comp as IPDFGraphicPathComponent, full);

            else if (comp is IPDFInvisibleContainer)
                this.DoLayoutInvisibleComponent(comp as IPDFInvisibleContainer, full);
            

            //If we did create a positioned region then we need to close it now, so that it does not get added to.
            if (null != postioned)
                postioned.Close();
            
        }

        protected virtual PDFLayoutRegion BeginNewRelativeRegionForChild(PDFPositionOptions pos, IPDFComponent comp, PDFStyle full)
        {
            PDFLayoutPage page = this.Context.DocumentLayout.CurrentPage;
            PDFLayoutBlock last = page.LastOpenBlock();
            PDFLayoutRegion rel = last.BeginNewPositionedRegion(pos, page, comp, full);
            return rel;
        }

        protected virtual PDFLayoutRegion BeginNewAbsoluteRegionForChild(PDFPositionOptions pos, IPDFComponent comp, PDFStyle full)
        {
            PDFLayoutPage page = this.Context.DocumentLayout.CurrentPage;
            PDFLayoutBlock last = page.LastOpenBlock();
            PDFLayoutRegion abs = last.BeginNewPositionedRegion(pos, page, comp, full);
            return abs;
        }

        #endregion

        //
        // DoLayoutXXX
        //

        #region protected virtual void DoLayoutPageBreak(IPDFPageBreak pgbreak, PDFStyle style)

        /// <summary>
        /// Breaks the current page layout and attempts to create a new page.
        /// </summary>
        /// <param name="pgbreak">The declared page break</param>
        /// <param name="style">The full style of the page break</param>
        protected virtual void DoLayoutPageBreak(IPDFPageBreak pgbreak, PDFStyle style)
        {
            PDFLayoutBlock block = this.DocumentLayout.CurrentPage.LastOpenBlock();
            PDFLayoutRegion region = block.CurrentRegion;
            Stack<PDFLayoutBlock> depth = BuildLayoutStack(block);
            if (this.MoveToNextPage(depth, ref region, ref block) == false)
            {
                this.ContinueLayout = false;
            }
            else
            {
                Stack<PDFLayoutBlock> closeorder = new Stack<PDFLayoutBlock>(depth.Reverse());

                //We need to close from the inner to the outer
                while (closeorder.Count > 0)
                {
                    PDFLayoutBlock toclose = closeorder.Pop();
                    if (!toclose.IsClosed)
                        toclose.Close();
                }

                //and then start new blocks on the next page from outer to inner
                PDFLayoutBlock current = block;
                while (depth.Count > 0)
                {
                    PDFLayoutBlock child = depth.Pop();
                    if (child.Owner is PDFPage)
                        continue;
                    else
                        current = child.Engine.CloseCurrentBlockAndStartNewInRegion(child, current.CurrentRegion);
                }
                region = current.CurrentRegion;
                this.CurrentBlock = current;
                block = current;
            }
        }

        /// <summary>
        /// Just builds the stack of layouts from our 
        /// current position to the top of the layout.
        /// </summary>
        /// <param name="block"></param>
        /// <returns></returns>
        private static Stack<PDFLayoutBlock> BuildLayoutStack(PDFLayoutBlock block)
        {
            Stack<PDFLayoutBlock> depth = new Stack<PDFLayoutBlock>();
            PDFLayoutBlock curr = block;
            while (null != curr)
            {
                depth.Push(curr);
                curr = curr.Parent as PDFLayoutBlock;
            }
            return depth;
        }

        #endregion

        #region protected virtual void DoLayoutColumnBreak(IPDFColumnBreak colbreak, PDFStyle style)

        /// <summary>
        /// 
        /// </summary>
        /// <param name="colbreak"></param>
        /// <param name="style"></param>
        protected virtual void DoLayoutColumnBreak(IPDFColumnBreak colbreak, PDFStyle style)
        {
            PDFLayoutBlock block = this.DocumentLayout.CurrentPage.LastOpenBlock();
            PDFLayoutRegion region = block.CurrentRegion;
            bool newpage;
            
            if (!this.MoveToNextRegion(ref region, ref block, out newpage))
            {
                this.ContinueLayout = false;
            }
        }

        #endregion

        #region protected virtual void DoLayoutLineBreak(IPDFLineBreak colbreak, PDFStyle style)

        /// <summary>
        /// 
        /// </summary>
        /// <param name="linebreak"></param>
        /// <param name="style"></param>
        protected virtual void DoLayoutLineBreak(IPDFLineBreak linebreak, PDFStyle style)
        {
            PDFLayoutBlock block = this.DocumentLayout.CurrentPage.LastOpenBlock();
            PDFLayoutRegion region = block.CurrentRegion;
            if (!region.HasOpenItem)
            {
                PDFTextRenderOptions txtopts = style.CreateTextOptions();
                PDFPositionOptions posopts = style.CreatePostionOptions();
                PDFUnit height;
                if(txtopts.Leading.HasValue)
                    height = txtopts.Leading.Value;
                else
                    height = txtopts.Font.Size * 1.2;//Bit of a hack for the leading size

                PDFLayoutLine line = region.BeginNewLine();
                line.AddTextRun(" ", linebreak, new PDFSize((PDFUnit)1, height), txtopts, posopts);
                
            }
            region.CloseCurrentItem();
        }

        #endregion


        #region protected virtual void DoLayoutInvisibleComponent(IPDFInvisibleContainer invisible, PDFStyle style)

        /// <summary>
        /// Laysout the individual components inside the invisible component
        /// </summary>
        /// <param name="invisible"></param>
        /// <param name="style"></param>
        protected virtual void DoLayoutInvisibleComponent(IPDFInvisibleContainer invisible, PDFStyle style)
        {
            PDFComponentList children;
            if (TryGetComponentChildren(invisible, out children))
            {
                this.DoLayoutChildren(children);
            }
        }

        #endregion

        #region protected virtual void DoLayoutTextComponent(IPDFTextComponent textComponent, PDFStyle style)

        /// <summary>
        /// Checks that the splitting has not occurred too near the start
        /// </summary>
        /// <param name="charCount"></param>
        /// <param name="characters"></param>
        /// <returns></returns>
        private static bool CanSplitWord(int charCount, string characters)
        {
            if (characters.Length <= charCount)
                return true;

            int minWordLength = 5;
            if (charCount < minWordLength)
            {
                if (characters.Length > charCount)
                {
                    if (!char.IsWhiteSpace(characters, charCount) && charCount > 0 && !char.IsWhiteSpace(characters,charCount-1))
                        //the word has been forced to be split too short.
                        return false;
                }
            }

            return true;
        }

        private static int GetClosestWordBoundary(int offset, string characters)
        {
            while (offset > 0 && !char.IsWhiteSpace(characters, offset))
                offset--;
            return offset;
        }

        /// <summary>
        /// Lays out all the characters for the text component into a series of blocks
        /// </summary>
        /// <param name="textComponent"></param>
        /// <param name="style"></param>
        protected virtual void DoLayoutTextComponent(IPDFTextComponent textComponent, PDFStyle style)
        {
            //Start by getting all the measurements together
            PDFTextRenderOptions txtOpts = style.CreateTextOptions();
            PDFPositionOptions posOpts = style.CreatePostionOptions();


            //PDFThickness margins = PDFThickness.Empty();
            //PDFThickness padding = PDFThickness.Empty();
            PDFRect bounds = PDFRect.Empty;
            PDFRect content = PDFRect.Empty;
            PDFRect border = PDFRect.Empty;
            PDFSize required = PDFSize.Empty;

            bool lastwasbreak = false;
            bool hasreadtext = false;
            PDFUnit ascentheight = txtOpts.Font.FontMetrics.Ascent;

            Text.WordWrap wraptext = txtOpts.WrapText.HasValue? txtOpts.WrapText.Value : Text.WordWrap.Auto;
            
            PDFUnit lineheight;
            if (txtOpts.Leading.HasValue)
                lineheight = txtOpts.Leading.Value;
            else
                lineheight = txtOpts.Font.FontMetrics.LineHeight;
            
            PDFLayoutPage firstpage = this.DocumentLayout.CurrentPage;
            PDFLayoutPage lastpage = firstpage;

            PDFLayoutBlock firstblock = firstpage.LastOpenBlock();
            PDFLayoutBlock lastblock = firstblock;
            if (null == firstblock || firstblock.IsClosed || firstblock.CurrentRegion.IsClosed)
            {
                this.ContinueLayout = false;
                //There is no open block on this page
                return;
            }
            
            PDFLayoutRegion firstregion = firstblock.CurrentRegion;
            PDFLayoutRegion lastregion = firstregion;

            //Make sure we have an open line
            PDFLayoutLine firstline = firstregion.HasOpenItem ? firstregion.CurrentItem as PDFLayoutLine : firstregion.BeginNewLine();
            PDFLayoutLine lastline = firstline;

            PDFUnit xoffset = firstregion.OffsetX;
            PDFLayoutTextRun firstrun = null;
            PDFLayoutTextRun lastrun = null;


            Text.PDFTextReader reader = textComponent.CreateReader(this.Context, style);
            int readcount = 0;
            PDFUnit lastinset = firstline.Width;
            PDFUnit offsetX = lastline.Width;

            if (offsetX == PDFUnit.Zero && txtOpts.FirstLineInset.HasValue) //no measurable items then use the first line inset from the style.
            {
                offsetX = txtOpts.FirstLineInset.Value;
                lastline.SetMaxWidth(lastline.FullWidth - txtOpts.FirstLineInset.Value);
            }
            
            
            Text.PDFTextLayout layout = new Text.PDFTextLayout(txtOpts, posOpts);
            Text.TextBlock textblock = layout.BeginNewBlock(firstpage.PageIndex);
            textblock.YOffset = firstregion.Height.PointsValue;
            Text.TextParagraph para = new Text.TextParagraph();
            textblock.AddParagraph(para);
            Text.TextLine textline = textblock.LastParagraph.AddLine();
            textline.LineInset = offsetX.PointsValue;

            int zerolinecheck = 0;

            while (reader.Read())
            {
                switch (reader.OpType)
                {
                    case Scryber.Text.PDFTextOpType.None:
                        lastwasbreak = false;
                        //Skip no-op
                        break;

                    case Scryber.Text.PDFTextOpType.LineBreak:

                        if (readcount == 0 || !hasreadtext)
                        {
                            textblock.YOffset += lineheight.PointsValue;
                            lastregion.CloseCurrentItem();
                            offsetX = txtOpts.FirstLineInset.HasValue ? txtOpts.FirstLineInset.Value : PDFUnit.Zero;
                            textline.LineInset = offsetX.PointsValue;
                            lastline.SetMaxWidth(lastline.FullWidth - txtOpts.FirstLineInset.Value);
                            lastline = lastregion.BeginNewLine();
                        }
                        else
                        {
                            if (lastline.IsEmpty && readcount > 0) 
                            {
                                //account for double line breaks by adding empty text to the line
                                textline.Height = lineheight.PointsValue;
                                lastline.AddTextRun(String.Empty, null, new PDFSize(0, lineheight), txtOpts, posOpts);
                            }
                               

                            textline = textblock.LastParagraph.AddLine();
                            lastregion.CloseCurrentItem();
                            lastline = lastregion.BeginNewLine();
                            if (txtOpts.FirstLineInset.HasValue)
                            {
                                textline.LineInset = txtOpts.FirstLineInset.Value.PointsValue;
                                lastline.SetMaxWidth(lastline.FullWidth - txtOpts.FirstLineInset.Value);
                            }
                        }
                        lastwasbreak = true;
                        break;

                    case Scryber.Text.PDFTextOpType.TextContent:
                        if (lastline.IsClosed)
                        {
                            lastline = lastregion.BeginNewLine();
                        }
                        string chars = ((Text.PDFTextDrawOp)reader.Value).Characters;
                        string used = string.Empty;
                        int blocklinecount = 0;
                        bool buildnewblock = false;
                        int nextchar = 0;

                        while (nextchar < chars.Length)
                        {
                            //Check if we can fit the next line. If not then try to move to a new region
                            PDFSize avail = new PDFSize(lastline.AvailableWidth, lastregion.AvailableHeight);

                            if (avail.Height < lineheight)
                            {
                                PDFUnit lastheight = lastregion.Height;
                                bool newpage;
                                buildnewblock = true;
                                PDFLayoutBlock prevblock = lastblock;
                                PDFLayoutRegion prevrevion = lastregion;
                                PDFLayoutBlock prevouter = prevblock.Parent as PDFLayoutBlock;
                                PDFLayoutRegion prevouterRegion = (null == prevouter) ? null : prevouter.CurrentRegion;
                                PDFUnit prevouterHeight = (null == prevouter) ? (PDFUnit)(-1) : prevouter.Height;
                                

                                if (null != firstrun && textblock != null && firstrun.TextBlock == null)
                                {
                                    firstrun.TextBlock = textblock;
                                    firstrun.TextOptions = txtOpts;
                                }
                                if (!this.MoveToNextRegion(ref lastregion, ref lastblock, out newpage))
                                {
                                    if (readcount > 0)
                                    {
                                        prevblock.ChildComponents.Add(textComponent);
                                        textComponent.TextLayout = layout;
                                    }
                                    this.ContinueLayout = false;
                                    return; //No region or bock or page to move to
                                }
                                else if (lastblock == prevblock && lastregion == prevrevion)
                                {
                                    buildnewblock = false;
                                    //HACK: Add the height of the line that doesn't fit to the region.
                                    if (lastregion.CurrentItem != null && lastregion.CurrentItem.IsClosed == false)
                                        lastregion.CloseCurrentItem();
                                    //lastregion.UsedSize = new PDFSize(lastregion.UsedSize.Width, lastregion.UsedSize.Height + lineheight);
                                }
                                else if (nextchar > 0)
                                {
                                    //HACK: If the previous block has not been added to the height of it's parent then need to add it.
                                    if (prevouter != null && prevouter.Height <= prevouterHeight)
                                    {
                                        if (prevouterRegion != null)
                                            prevouterRegion.AddToSize(prevblock);
                                    }
                                    textblock = layout.BeginNewBlock(lastpage.PageIndex);
                                    lastheight = 0;//reset the height as we are now on a new page at point zero
                                    //lastrun.TextBlock = textblock;
                                    //lastrun.TextOptions = txtOpts;

                                }
                                else
                                {
                                    if (readcount == 0)
                                    {
                                        this.TraceLog.Add(TraceLevel.Debug, LOG_CATEGORY, "Rolled over to a new region with an empty previous region");
                                        textblock.RemoveLastParagraph();
                                    }
                                    //we have rolled everything over to a new region (no need for a new block)
                                    buildnewblock = false;
                                }
                                avail.Height = lastregion.AvailableHeight;
                                
                                //para.ResetY = true;
                                //para.YOffset = -lastheight.PointsValue;

                                //The very first line in a block may be flipped to the new region, so we need to
                                //do the ascent height again.
                                if (readcount == 0 && blocklinecount == 0)
                                    para.YOffset += ascentheight.PointsValue;

                                if (buildnewblock || readcount == 0)
                                {
                                    para = new Text.TextParagraph();
                                    textblock.AddParagraph(para);
                                    textline = para.AddLine();
                                    lastline = lastregion.BeginNewLine();
                                }
                                if (lastwasbreak && txtOpts.FirstLineInset.HasValue)
                                {
                                    textline.LineInset += txtOpts.FirstLineInset.Value.PointsValue;
                                    lastline.SetMaxWidth(lastline.FullWidth - txtOpts.FirstLineInset.Value);
                                }

                            }

                            int fitted;

#if MEASURE_PERFORMANCE
                            _measurecount++;
                            _measurestringducration.Start();
                            PDFSize measured = this.Graphics.MeasureString(chars, nextchar, avail, txtOpts, out fitted);

                            _measurestringducration.Stop();
#else
                            PDFSize measured = this.Graphics.MeasureString(chars, nextchar, avail, txtOpts, out fitted);
#endif
                            //If we cannot wrap on a character and we are not splitting on a word
                            if (wraptext != Text.WordWrap.Character && !CanSplitWord(fitted, chars))
                            {
                                //then provided we are not the only thing on the line
                                if (lastline.IsEmpty == false)
                                    fitted = 0;
                            }

                            //if (measured.Height < lineheight)
                            measured.Height = lineheight;
                            
                            if (fitted == 0)
                            {
                                zerolinecheck++;
                                if (zerolinecheck > 100)
                                    throw new InvalidOperationException("Error: Fell into a suspected never ending loop of laying out zero characters per line. Breaking out ungracefully");

                                lastregion.CloseCurrentItem();
                                lastline = lastregion.BeginNewLine();
                                lastinset = txtOpts.FirstLineInset.HasValue ? txtOpts.FirstLineInset.Value : 0;
                                if (readcount == 0)
                                {
                                    textblock.YOffset = lastregion.Height.PointsValue;
                                    textline.LineInset = 0;
                                }
                            }
                            else if (fitted == chars.Length)
                            {
                                zerolinecheck = 0;
                                used = chars.Substring(nextchar);
                                chars = string.Empty;

                                lastrun = lastline.AddTextRun(used, textComponent, measured, txtOpts, posOpts);

                                textline = textblock.LastParagraph.LastLine;
                                textline.AddSpan(lastrun.Span);
                                textline.Height = lastline.Height == PDFUnit.Zero ? lineheight.PointsValue : lastline.Height.PointsValue;
                                textline.AscentHeight = ascentheight.PointsValue;
                                if (textline.ContentCount == 1)
                                {
                                    lastrun.StartLine = textline;
                                    if (para.LineCount == 1)
                                        lastrun.StartParagraph = para;
                                }
                            }
                            else
                            {
                                zerolinecheck = 0;
                                used = chars.Substring(nextchar, fitted);
                                //Increment the next offset
                                nextchar += fitted;
                                while (nextchar < chars.Length && char.IsWhiteSpace(chars, nextchar))
                                    nextchar++;

                                //chars = chars.Substring(fitted);

                                lastrun = lastline.AddTextRun(used, textComponent, measured, txtOpts, posOpts);
                                lastregion.CloseCurrentItem();

                                //Add the span to the textblock
                                textline = textblock.LastParagraph.LastLine;
                                textline.AddSpan(lastrun.Span);
                                textline.Height = lastline.Height == PDFUnit.Zero ? lineheight.PointsValue : lastline.Height.PointsValue;
                                textline.AscentHeight = ascentheight.PointsValue;
                                
                                if (textline.ContentCount == 1)
                                {
                                    lastrun.StartLine = textline;
                                    if (para.LineCount == 1)
                                        lastrun.StartParagraph = para;
                                }

                                textline = textblock.LastParagraph.AddLine();
                                lastline = lastregion.BeginNewLine();

                                offsetX = PDFUnit.Zero;
                            }
                            if (null == firstrun && null != lastrun)
                            {
                                firstrun = lastrun;
                                firstrun.TextBlock = textblock;
                                firstrun.TextOptions = txtOpts;
                            }
                            else if (buildnewblock && lastrun != null)
                            {
                                lastrun.TextBlock = textblock;
                                lastrun.TextOptions = txtOpts;
                                buildnewblock = false;
                            }
                            blocklinecount++;
                            lastwasbreak = false;
                            hasreadtext = true;
                        }
                        
                        break;
                    default:
                        lastwasbreak = false;
                        break;
                }
                readcount++;
                
            }

            textComponent.TextLayout = layout;

            if (posOpts.PositionMode != PositionMode.Inline)
            {
                //We are a block, so close the current line
                if (null != lastregion && lastregion.HasOpenItem)
                {
                    lastregion.CloseCurrentItem();
                }
            }
        }

        /// <summary>
        /// Checks to see if this line is empty (does not have anything on it)
        /// </summary>
        /// <param name="lastline"></param>
        /// <returns></returns>
        private bool IsEmptyLine(PDFLayoutLine line)
        {
            return line.IsEmpty;
        }

        #endregion

        #region protected virtual void DoLayoutImageComponent(IPDFImageComponent image, PDFStyle style)

        /// <summary>
        /// Lays out an image component on the current page
        /// </summary>
        /// <param name="image">The image reference to layout</param>
        /// <param name="style">The image style</param>
        protected virtual void DoLayoutImageComponent(IPDFImageComponent image, PDFStyle style)
        {
            PDFLayoutBlock block = this.DocumentLayout.CurrentPage.LastOpenBlock();
            PDFLayoutRegion region = block.CurrentRegion;
            
            Resources.PDFImageXObject imgx = image.GetImageObject(this.Context, style);

            if (null == imgx)
            {
                this.TraceLog.Add(TraceLevel.Message, LOG_CATEGORY, "No image data for component '" + image.ToString() + "'. No layout required.");
                return;
            }

            PDFPositionOptions options = style.CreatePostionOptions();
            
            //If the image does not fit across the width of the region
            //Then reduce the required size (including the images margins and padding)
            //This does not affect the expicit sizes - which are caclulated later
            PDFSize sz = imgx.GetImageSize();
            PDFUnit availWidth = region.UnusedBounds.Width - ((options.Padding.Left + options.Padding.Right) + (options.Margins.Left + options.Margins.Right));

            if (sz.Width > availWidth)
            {
                PDFSize newsize = new PDFSize(availWidth, sz.Height * (availWidth.PointsValue / sz.Width.PointsValue));
                sz = newsize;
            }
            //Rects 
            PDFRect border, content, total;
            bool hasmargins, haspadding;
            PDFThickness marginThick, padThick;


            sz = BuildContentSizes(style, options, sz, out border, out content, out total, out hasmargins, out haspadding, out marginThick, out padThick);

            
            
            PDFLayoutLine linetoAddTo;
            PositionMode mode = options.PositionMode;

            linetoAddTo = GetOpenLine(total.Width, region, mode);

            //Add the image layout run to the current line.
            
            //Check that the new line height will fit of the page
            PDFUnit linebottom = total.Height + linetoAddTo.OffsetY;
            PDFUnit maxbottom = region.TotalBounds.Height;
            bool canfitVertical = linebottom < maxbottom;

            //If not - check that we can overflow onto new pages
            if (!canfitVertical)
            {
                region.CloseCurrentItem();
                bool newPage;
                //If we can overflow then move the line onto a new column page.
                if (this.MoveToNextRegion(ref region, ref block, out newPage))
                {
                    linetoAddTo = region.BeginNewLine();
                }
                else
                {
                    if (options.OverflowAction == OverflowAction.Truncate)
                        region.AssertRemoveLastItem(linetoAddTo);

                    this.ContinueLayout = false;
                    return;
                }
            }

            linetoAddTo.AddComponentRun(image, total, border, content, options, style);

            if (mode != PositionMode.Inline)
            {
                //Close the current line because we are not inline
                region.CloseCurrentItem();
                
            }
            
            

            //if (mode == PositionMode.Relative || mode == PositionMode.Absolute)
            //{
            //    block.Close();
            //}
        }

        #endregion

        #region protected virtual void DoLayoutPathComponent(IPDFGraphicPathComponent comp, PDFStyle style)

        protected virtual void DoLayoutPathComponent(IPDFGraphicPathComponent comp, PDFStyle style)
        {
            PDFPositionOptions options = style.CreatePostionOptions();
            
            PDFSize sz = new PDFSize(0,0);
            //Rects 
            
            PDFLayoutBlock block = this.DocumentLayout.CurrentPage.LastOpenBlock();
            PDFLayoutRegion region = block.CurrentRegion;
            PDFLayoutLine linetoAddTo = null;
            PositionMode mode = options.PositionMode;

            if (region.HasOpenItem)
            {
                if (mode == PositionMode.Block)
                {
                    region.CloseCurrentItem();
                    linetoAddTo = region.BeginNewLine();
                }
                else if (!(region.CurrentItem is PDFLayoutLine))
                {
                    region.CloseCurrentItem();
                    linetoAddTo = region.BeginNewLine();
                }
                else
                    linetoAddTo = region.CurrentItem as PDFLayoutLine;
            }
            if (null == linetoAddTo)
                linetoAddTo = region.BeginNewLine();

            sz = new PDFSize(linetoAddTo.AvailableWidth, region.AvailableHeight);
            if (options.HasWidth)
                sz.Width = options.Width;
            if (options.HasHeight)
                sz.Height = options.Height;
            sz.Width -= options.Margins.Left + options.Margins.Right + options.Padding.Left + options.Padding.Right;
            sz.Height -= options.Margins.Top + options.Margins.Bottom + options.Padding.Top + options.Padding.Bottom;


            PDFGraphicsPath path = comp.CreatePath(sz, style);
            comp.Path = path;
            sz = path.Bounds.Size;

            PDFRect content = new PDFRect(
                options.Padding.Top + options.Margins.Top,      
                options.Padding.Left + options.Margins.Left,
                sz.Width,
                sz.Height);

            PDFRect border = new PDFRect(
                options.Margins.Top,
                options.Margins.Left,
                sz.Width + options.Padding.Left + options.Padding.Right,
                sz.Height + options.Padding.Top + options.Padding.Bottom);

            PDFRect total = new PDFRect(
                0,
                0,
                border.Width + options.Margins.Left + options.Margins.Right,
                border.Height + options.Margins.Top + options.Margins.Bottom);

            //Add the image layout run to the current line.

            //Check that the new line height will fit of the page
            bool canfitVertical = region.AvailableHeight > total.Height;

            //If not - check that we can overflow onto new pages
            if (!canfitVertical)
            {
                linetoAddTo.Close();
                bool newPage;
                //If we can overflow then move the line onto a new column page.
                if (this.MoveToNextRegion(ref region, ref block, out newPage))
                {
                    linetoAddTo = region.BeginNewLine();
                }
                else
                {
                    if (options.OverflowAction == OverflowAction.Truncate)
                        region.AssertRemoveLastItem(linetoAddTo);

                    this.ContinueLayout = false;
                    return;
                }
            }

            linetoAddTo.AddComponentRun(comp, total, border, content, options, style);

            if (mode != PositionMode.Inline)
            {
                //Close the current line because we are not inline
                region.CloseCurrentItem();

            }

            if (mode == PositionMode.Relative || mode == PositionMode.Absolute)
            {
                //TODO: Close the current block
                block.Close();
            }
        }

        #endregion

        #region protected virtual void DoLayoutViewPortComponent(IPDFViewPortComponent viewPort, PDFStyle style)

        /// <summary>
        /// Laysout a ViewPort component using the engine returned from the GetEngine invocation
        /// </summary>
        /// <param name="viewPort"></param>
        /// <param name="style"></param>
        protected virtual void DoLayoutViewPortComponent(IPDFViewPortComponent viewPort, PDFStyle style)
        {
            try
            {
                using (IPDFLayoutEngine engine = viewPort.GetEngine(this, this.Context, style))
                {
                    engine.Layout(this.Context, style);
                }
                PDFLayoutBlock block = this.Context.DocumentLayout.CurrentPage.LastOpenBlock();
                if (null == block)
                    this.ContinueLayout = false;
            }
            catch (PDFLayoutException)
            {
                throw;
            }
            catch (Exception ex)
            {
                string msg = String.Format(Errors.LayoutFailedForComponent, viewPort.ID, ex.Message);
                throw new PDFLayoutException(msg, ex);
            }
        }

        #endregion


        //
        // support methods
        //

        #region protected void RegisterChildLayout(IPDFComponent comp)

        /// <summary>
        /// When a child component has completed it's layout then this is called to register it's completion
        /// </summary>
        /// <param name="comp">The component that was laid out</param>
        protected virtual void RegisterChildLayout(IPDFComponent comp)
        {
            if (this.CurrentBlock != null)
                this.CurrentBlock.ChildComponents.Add(comp);
        }

        #endregion

        #region private bool MoveToNextRegion(ref PDFLayoutRegion region, ref PDFLayoutBlock block, out bool newPage)

        /// <summary>
        /// Attempts to move the cursor to the next available region. Either in the current block, 
        /// or another block in a new page
        /// </summary>
        /// <param name="region">Pass in the current region. Will be set to the next region</param>
        /// <param name="block">Pass in the current block. Will be set to the next block if it has changed</param>
        /// <param name="newPage">Set to true if we are now going to be working on another page</param>
        /// <returns>True is we can move to another page.</returns>
        protected virtual bool MoveToNextRegion(ref PDFLayoutRegion region, ref PDFLayoutBlock block, out bool newPage)
        {
            PDFLayoutRegion origRegion = region;
            PDFLayoutBlock origBlock = block;

            TraceLevel overflowTraceLevel = TraceLevel.Debug;
            string logCategory = "Region Overflow";

            if (!this.CanOverflowFromCurrentRegion(region))
            {
                //We do not overflow relative or absolute regions
                newPage = false;
                return false;
            }
            this.TraceLog.Begin(overflowTraceLevel, logCategory, "Starting the overflow to a new page");

            Stack<PDFLayoutBlock> reverseBlocks = new Stack<PDFLayoutBlock>();
            PDFLayoutBlock tomove = null;

            PDFLayoutBlock current = block;
            while (current != null)
            {
                if (current.MoveToNextRegion())
                {
                    this.TraceLog.Add(overflowTraceLevel, logCategory, "Current block " + current + " has another region it can flow into. Moving to this region");

                    if (region.IsClosed == false && (null == tomove || region != tomove.CurrentRegion))
                        region.Close();

                    while (reverseBlocks.Count > 0)
                    {
                        PDFLayoutBlock child = reverseBlocks.Pop();
                        this.TraceLog.Add(overflowTraceLevel, logCategory, "Closing inner block '" + child + "' and starting a new region");
                        current = child.Engine.CloseCurrentBlockAndStartNewInRegion(child, current.CurrentRegion);
                    }
                    region = current.CurrentRegion;
                    block = current;
                    newPage = false;

                    if (null != tomove)
                    {
                        this.TraceLog.Add(overflowTraceLevel, logCategory, "Have existing block '" + tomove + "' to move to the next region");
                        region.AddExistingItem(tomove);
                        tomove.Offset(PDFUnit.Zero, PDFUnit.Zero - tomove.TotalBounds.Y);
                        block = tomove;
                        PDFRect total = region.TotalBounds;
                        total.Width -= block.Position.Margins.Left + block.Position.Margins.Right;
                        block.AvailableBounds = total;
                        region = tomove.CurrentRegion;

                        if (region.IsClosed == false)
                        {
                            total = region.TotalBounds;
                            total.Height = block.AvailableBounds.Height;
                            region.TotalBounds = total;
                        }
                    }
                    this.TraceLog.End(overflowTraceLevel, logCategory, "Completed the overflow to a new region");
                    return true;
                }
                else if (current.OverflowSplit == OverflowSplit.Any)
                {
                    reverseBlocks.Push(current);
                    if (current.CurrentRegion.IsClosed == false)
                        current.CurrentRegion.Close();
                    //current.Close();
                    current = current.Parent as PDFLayoutBlock;
                }
                else
                {
                    tomove = current;
                    (current.Parent as PDFLayoutBlock).CurrentRegion.RemoveLastItem();
                    reverseBlocks.Clear();
                    //if (current.CurrentRegion.IsClosed == false)
                    //    current.CurrentRegion.Close();
                    //current.Close();
                    current = current.Parent as PDFLayoutBlock;
                }
            }

            if (null == tomove && region.HasOpenItem)
                region.CloseCurrentItem();

            if (null == tomove)
            {
                if (region.IsClosed == false)
                    region.Close();
                if (block.IsClosed == false)
                    block.Close();
            }
            PDFLayoutBlock top = reverseBlocks.Pop(); //Get the top level block.
            newPage = top.Engine.MoveToNextPage(reverseBlocks, ref region, ref block);
            if (newPage)
            {
                //OverflowSplit parentSplit = OverflowSplit.Any;
                while (reverseBlocks.Count > 0)
                {
                    current = reverseBlocks.Pop();
                    PDFLayoutBlock outer = current;
                    //If this component should not be split, or the parent specified that it can only be split on components
                    current = current.Engine.CloseCurrentBlockAndStartNewInRegion(current, region);
                    region = current.CurrentRegion;
                    block = current;
                    
                    //parentSplit = current.OverflowSplit;
                }

                //We have a block that needs to move (en mass) to the new page
                if (null != tomove)
                {
                    region.AddExistingItem(tomove);
                    tomove.Offset(PDFUnit.Zero, region.Height - tomove.TotalBounds.Y);
                    block = tomove;
                    PDFRect total = region.TotalBounds;
                    total.Width -= block.Position.Margins.Left + block.Position.Margins.Right;
                    block.AvailableBounds = total;
                    for (int i = 0; i < tomove.ColumnCount; i++)
                    {
                        region = tomove.Columns[i];

                        if (region.IsClosed == false)
                        {
                            region.ResetAvailableHeight(block.AvailableBounds.Height, true);
                        }
                    }

                    //As we have moved the block and region, we can restore them to the original values.
                    region = origRegion;
                    block = origBlock;                    
                }
            }

            if (newPage)
                this.TraceLog.End(overflowTraceLevel, logCategory, "Completed the overflow to a new page");
            else
                this.TraceLog.End(overflowTraceLevel, logCategory, "No new page available to overflow");

            return newPage;

        }

        /// <summary>
        /// Checks the provided region to see if there is support for overflow
        /// </summary>
        /// <param name="region"></param>
        /// <returns></returns>
        /// <remarks>The base method returns false if the region is set as absolute or relative</remarks>
        protected virtual bool CanOverflowFromCurrentRegion(PDFLayoutRegion region)
        {
            if (region.PositionMode == PositionMode.Absolute || region.PositionMode == PositionMode.Relative)
                return false;
            else
                return true;
        }

        #endregion 

        
        #region private bool MoveToNextRegion_Prev(ref PDFLayoutRegion region, ref PDFLayoutBlock block, out bool newPage)

        /// <summary>
        /// Attempts to move the cursor to the next available region. Either in the current block, 
        /// or another block in a new page
        /// </summary>
        /// <param name="region">Pass in the current region. Will be set to the next region</param>
        /// <param name="block">Pass in the current block. Will be set to the next block if it has changed</param>
        /// <param name="newPage">Set to true if we are now going to be working on another page</param>
        /// <returns>True is we can move to another page.</returns>
        protected virtual bool MoveToNextRegion_Prev(ref PDFLayoutRegion region, ref PDFLayoutBlock block, out bool newPage)
        {
            if (!this.CanOverflowFromCurrentRegion_prev(region))
            {
                //We do not overflow relative or absolute regions
                newPage = false;
                return false;
            }

            Stack<PDFLayoutBlock> reverseBlocks = new Stack<PDFLayoutBlock>();
            PDFLayoutBlock current = block;
            while (current != null)
            {
                if (current.MoveToNextRegion())
                {

                    if (region.IsClosed == false)
                        region.Close();

                    while (reverseBlocks.Count > 0)
                    {
                        PDFLayoutBlock child = reverseBlocks.Pop();
                        current = child.Engine.CloseCurrentBlockAndStartNewInRegion(child, current.CurrentRegion);
                    }
                    region = current.CurrentRegion;
                    block = current;
                    newPage = false;
                    return true;
                }
                else if (current.OverflowSplit == OverflowSplit.Any)
                {
                    reverseBlocks.Push(current);
                    if (current.CurrentRegion.IsClosed == false)
                        current.CurrentRegion.Close();
                    //current.Close();
                    current = current.Parent as PDFLayoutBlock;
                }
                else
                {

                    reverseBlocks.Push(current);
                    if (current.CurrentRegion.IsClosed == false)
                        current.CurrentRegion.Close();
                    //current.Close();
                    current = current.Parent as PDFLayoutBlock;
                }
            }

            if (region.HasOpenItem)
                region.CloseCurrentItem();
            if (region.IsClosed == false)
                region.Close();
            if (block.IsClosed == false)
                block.Close();
            PDFLayoutBlock top = reverseBlocks.Pop(); //Get the top level block.
            newPage = top.Engine.MoveToNextPage(reverseBlocks, ref region, ref block);
            if (newPage)
            {
                //OverflowSplit parentSplit = OverflowSplit.Any;
                while (reverseBlocks.Count > 0)
                {
                    current = reverseBlocks.Pop();
                    PDFLayoutBlock outer = current;
                    //If this component should not be split, or the parent specified that it can only be split on components
                    current = current.Engine.CloseCurrentBlockAndStartNewInRegion(current, region);
                    region = current.CurrentRegion;


                    //parentSplit = current.OverflowSplit;

                }
            }
            return newPage;

        }

        /// <summary>
        /// Checks the provided region to see if there is support for overflow
        /// </summary>
        /// <param name="region"></param>
        /// <returns></returns>
        /// <remarks>The base method returns false if the region is set as absolute or relative</remarks>
        protected virtual bool CanOverflowFromCurrentRegion_prev(PDFLayoutRegion region)
        {
            if (region.PositionMode == PositionMode.Absolute || region.PositionMode == PositionMode.Relative)
                return false;
            else
                return true;
        }

        #endregion 

        #region public virtual PDFLayoutBlock CloseCurrentBlockAndStartNewInRegion(PDFLayoutBlock blockToClose, PDFLayoutRegion joinToRegion)

        /// <summary>
        /// Closes the current block and starts a new block in the current region
        /// </summary>
        /// <param name="blockToClose"></param>
        /// <param name="joinToRegion"></param>
        /// <returns></returns>
        public virtual PDFLayoutBlock CloseCurrentBlockAndStartNewInRegion(PDFLayoutBlock blockToClose, PDFLayoutRegion joinToRegion)
        {
            if (!blockToClose.IsClosed)
                blockToClose.Close();

            int repeatindex = blockToClose.BlockRepeatIndex + 1;

            PDFPositionOptions position = this.FullStyle.CreatePostionOptions();
            PDFRect total = new PDFRect(new PDFPoint(0, joinToRegion.Height), joinToRegion.UnusedBounds.Size);

            if (position.HasWidth)
                total.Width = position.Width;
            else if (position.FillWidth)
                total.Width -= position.Margins.Left + position.Margins.Right;

            PDFUnit alley;
            int colcount = this.GetBlockColumnCount(out alley);

            PDFLayoutBlock block = ((PDFLayoutBlock)joinToRegion.Parent).BeginNewBlock(this.Component, this, this.FullStyle, position.PositionMode);
            block.InitRegions(total, position, colcount, alley);
            block.BlockRepeatIndex = repeatindex;
            this.CurrentBlock = block;

            return block;
        }

        #endregion

        #region public virtual bool MoveToNextPage(Stack<PDFLayoutBlock> depth, ref PDFLayoutRegion region, ref PDFLayoutBlock block)

        /// <summary>
        /// Simply calls Move to next page on the parent engine
        /// </summary>
        /// <param name="depth"></param>
        /// <param name="region"></param>
        /// <param name="block"></param>
        /// <returns></returns>
        public virtual bool MoveToNextPage(Stack<PDFLayoutBlock> depth, ref PDFLayoutRegion region, ref PDFLayoutBlock block)
        {
            return this.ParentEngine.MoveToNextPage(depth, ref region, ref block);
        }

        #endregion

        #region private PDFSize BuildContentSizes(......)

        /// <summary>
        /// Based on the style and position options this method generates the required size 
        /// of an element along with - 1. the total, border, content rectangles. 
        /// 2. The margins and padding around the element
        /// </summary>
        /// <param name="style"></param>
        /// <param name="options"></param>
        /// <param name="origsize"></param>
        /// <param name="border"></param>
        /// <param name="content"></param>
        /// <param name="total"></param>
        /// <param name="hasmargins"></param>
        /// <param name="haspadding"></param>
        /// <param name="marginThick"></param>
        /// <param name="padThick"></param>
        /// <returns></returns>
        protected PDFSize BuildContentSizes(PDFStyle style, PDFPositionOptions options, PDFSize origsize, out PDFRect border, out PDFRect content, out PDFRect total, out bool hasmargins, out bool haspadding, out PDFThickness marginThick, out PDFThickness padThick)
        {
            PDFPaddingStyle padding;
            PDFMarginsStyle margins;

            //Extract the margins and padding from the style

            if (style.TryGetMargins(out margins))
            {
                marginThick = margins.GetThickness();
                hasmargins = true;
            }
            else
            {
                marginThick = PDFThickness.Empty();
                hasmargins = false;
            }

            if (style.TryGetPadding(out padding))
            {
                padThick = padding.GetThickness();
                haspadding = true;
            }
            else
            {
                padThick = PDFThickness.Empty();
                haspadding = false;
            }

            //Create the image size based on any defined width and/or height.
            if (options.HasWidth && options.HasHeight)
                origsize = new PDFSize(options.Width, options.Height);

            else if (options.HasWidth)
            {
                if (origsize.Width > 0)
                    origsize = new PDFSize(options.Width, origsize.Height * (options.Width.PointsValue / origsize.Width.PointsValue));
                else
                    origsize = new PDFSize(options.Height, 0);
            }
            else if (options.HasHeight)
            {
                if (origsize.Height > 0)
                    origsize = new PDFSize(origsize.Width * (options.Height.PointsValue / origsize.Height.PointsValue), options.Height);
                else
                    origsize = new PDFSize(0, options.Height);
            }
            //else sz = sz

            content = new PDFRect(marginThick.Top + padThick.Top, marginThick.Left + padThick.Left, origsize.Width, origsize.Height);
            border = new PDFRect(marginThick.Top, marginThick.Left, origsize.Width + padThick.Left + padThick.Right, origsize.Height + padThick.Top + padThick.Right);
            total = new PDFRect(0, 0, border.Width + marginThick.Left + marginThick.Right, border.Height + marginThick.Top + marginThick.Bottom);
            return origsize;

        }

        #endregion

        #region private PDFLayoutLine GetOpenLine(PDFUnit requiredwidth, PDFLayoutRegion region, PositionMode mode)

        /// <summary>
        /// Gets the next open line to append content to
        /// </summary>
        /// <param name="requiredwidth"></param>
        /// <param name="region"></param>
        /// <param name="mode"></param>
        /// <returns></returns>
        private PDFLayoutLine GetOpenLine(PDFUnit requiredwidth, PDFLayoutRegion region, PositionMode mode)
        {
            PDFLayoutLine linetoAddTo;
            switch (mode)
            {
                case PositionMode.Absolute:
                case PositionMode.Relative:
                case PositionMode.Block:
                    // Close the current line and start a new one
                    if (region.HasOpenItem)
                        region.CloseCurrentItem();
                    linetoAddTo = region.BeginNewLine();
                    break;

                case PositionMode.Inline:
                    //Check that the image fits on the current line
                    if (region.HasOpenItem && region.CurrentItem is PDFLayoutLine)
                    {
                        linetoAddTo = region.CurrentItem as PDFLayoutLine;

                        //If not then close the current line and add start a new line
                        if (requiredwidth > PDFUnit.Zero && linetoAddTo.CanFitWidth(requiredwidth) == false)
                        {
                            region.CloseCurrentItem();
                            linetoAddTo = region.BeginNewLine();
                        }
                    }
                    else
                        linetoAddTo = region.BeginNewLine();

                    break;
                default:
                    throw new ArgumentOutOfRangeException("PositionMode");
            }
            return linetoAddTo;
        }

        #endregion

        #region protected void InitBlock(PDFLayoutBlock block, PDFRect bounds, PDFStyle style)

        /// <summary>
        /// Initializes the margins, padding etc for a block
        /// </summary>
        /// <param name="block"></param>
        /// <param name="bounds"></param>
        /// <param name="style"></param>
        protected void InitBlock(PDFLayoutBlock block, PDFRect bounds, PDFStyle style)
        {
            PDFPositionOptions options = style.CreatePostionOptions();

            int count = 1;
            PDFUnit alley = PDFUnit.Zero;

            PDFColumnsStyle columns;
            if (style.TryGetColumns(out columns))
            {
                count = columns.ColumnCount;
                alley = columns.AlleyWidth;
            }

            block.InitRegions(bounds, options, count, alley);
        }

        #endregion

        #region protected PDFRect GetContentRectFromBounds(PDFRect bounds ...)

        /// <summary>
        /// Calculates the available content rect of this block based on the bounds
        /// </summary>
        /// <param name="bounds">The total available size</param>
        /// <param name="margthick">set to the applied margins</param>
        /// <param name="padthick">set to the applied padding</param>
        /// <returns>The available content rectangle</returns>
        protected PDFRect GetContentRectFromBounds(PDFRect bounds, out PDFThickness margthick, out PDFThickness padthick)
        {
            PDFMarginsStyle marg;
            PDFPaddingStyle pad;
            margthick = PDFThickness.Empty();
            padthick = PDFThickness.Empty();

            PDFRect contentrect = bounds;
            if (this.FullStyle.TryGetMargins(out marg))
            {
                margthick = marg.GetThickness();
                contentrect = contentrect.Inset(margthick);
            }

            
            if (this.FullStyle.TryGetPadding(out pad))
            {
                padthick = pad.GetThickness();
                contentrect = contentrect.Inset(padthick);
            }
            return contentrect;
        }

        #endregion

        #region protected int GetBlockColumnCount(ref PDFUnit alleywidth)

        /// <summary>
        /// Gets the number of columns and also the alley width for this block
        /// </summary>
        /// <param name="alleywidth"></param>
        /// <returns></returns>
        protected int GetBlockColumnCount(out PDFUnit alleywidth)
        {
            int count;
            PDFColumnsStyle cols;
            if (this.FullStyle.TryGetColumns(out cols))
            {
                alleywidth = cols.AlleyWidth;
                count = cols.ColumnCount;
            }
            else
            {
                alleywidth = 0;
                count = 1;
            }
            return count;
        }

        #endregion


        //
        // disposal
        //

        #region public void Dispose()

        /// <summary>
        /// Interaface implementation of the IDisposable interface
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
        }

        #endregion

        #region protected virtual void Dispose(bool disposing)

        /// <summary>
        /// Inheritors can override this method to dispose of any unmanaged resources
        /// </summary>
        /// <param name="disposing"></param>
        protected virtual void Dispose(bool disposing)
        {
            if (null != this._graphics)
                this._graphics.Dispose();
        }

        #endregion
    }
}
