﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Scryber.Drawing;
using Scryber.Styles;
using Scryber.Native;

namespace Scryber.Components.Layout
{

    /// <summary>
    /// Represents a block of content on a page. Each block has an array of available regions to add content to (other blocks and lines).
    /// </summary>
    public class PDFLayoutBlock : PDFLayoutContainerItem
    {
        //
        // properties
        //

        #region public PDFLayoutRegion CurrentRegion { get;}

        private PDFLayoutRegion _current;

        /// <summary>
        /// Gets the current region in the block
        /// </summary>
        /// <remarks>Is there is a current open
        /// positioned region (absolute or relative) then this will be returned, 
        /// otherwise it will be the last open column if any</remarks>
        public PDFLayoutRegion CurrentRegion
        {
            get
            {
                if (this.HasPositionedRegions)
                {
                    PDFLayoutRegion last = this.PositionedRegions[this.PositionedRegions.Count - 1];
                    if (last.IsClosed == false)
                        return last;
                }
                return _current;
            }
            protected set
            {
                _current = value;
            }
        }

        #endregion

        #region public PDFLayoutRegion[] Columns { get; private set; }

        /// <summary>
        /// Gets all the regions in the block.
        /// </summary>
        public PDFLayoutRegion[] Columns { get; private set; }

        #endregion

        #region private PDFLayoutRegionCollection PositionedRegions {get;} + HasPositionedRegions {get;}

        private PDFLayoutRegionCollection _positioned;

        /// <summary>
        /// Gets the collection of regions in this block that have relative or absolute positioning.
        /// </summary>
        private PDFLayoutRegionCollection PositionedRegions
        {
            get
            {
                if (null == _positioned)
                    _positioned = new PDFLayoutRegionCollection();
                return _positioned;
            }
        }

        /// <summary>
        /// True if this block has positioned regions
        /// </summary>
        public bool HasPositionedRegions
        {
            get { return null != _positioned && _positioned.Count > 0; }
        }

        #endregion

        #region public int ColumnCount { get; private set; }

        /// <summary>
        /// Gets the total number of regions in this block
        /// </summary>
        public int ColumnCount { get; private set; }

        #endregion

        #region public PDFUnit AlleyWidth { get; private set; }

        /// <summary>
        /// Gets the alley width between regions
        /// </summary>
        public PDFUnit AlleyWidth { get; private set; }

        #endregion

        #region public PDFUnit ColumnWidth { get; private set; }

        /// <summary>
        /// Gets the width of each column
        /// </summary>
        public PDFUnit ColumnWidth { get; private set; }

        #endregion

        #region public PDFRect AvailableBounds { get; private set; }

        private PDFRect _avail;

        /// <summary>
        /// Gets or sets the available bounds of this block relative to the Absolute bounds
        /// </summary>
        public PDFRect AvailableBounds
        {
            get { return _avail; }
            set { _avail = value; }
        }

        #endregion

        #region public PDFRect TotalBounds { get;}

        private PDFRect _total = PDFRect.Empty;

        /// <summary>
        /// Gets the absolute bounds (inc margins and padding) for this block
        /// </summary>
        public PDFRect TotalBounds 
        {
            get { return _total; }
            set { _total = value; }
        }

        #endregion

        #region public PDFPositionOptions Position { get; }

        /// <summary>
        /// Gets the position options for this block
        /// </summary>
        public PDFPositionOptions Position { get; protected set; }

        #endregion

        #region public PDFSize Size { get; set; }

        /// <summary>
        /// Gets or sets the total used size
        /// </summary>
        public PDFSize Size { get; set; }

        #endregion

        #region public OverflowSplit OverflowSplit {get; set;}

        /// <summary>
        /// If overflow split is Never then the content in this block should be kept together at all times.
        /// If it it component then each child component must be kept together in one block, otherwise any component can split across regions and pages.
        /// </summary>
        public OverflowSplit OverflowSplit
        {
            get;
            set;
        }

        #endregion

        #region  protected bool IsContainer { get; set; }

        /// <summary>
        /// Marks this block explicitly as a container block 
        /// (holds references to its contents)
        /// </summary>
        protected bool IsContainer { get; set; }

        #endregion

        #region public override PDFUnit Width {get;}

        /// <summary>
        /// Gets the width of this block
        /// </summary>
        public override PDFUnit Width
        {
            get
            {
                if (this.IsClosed)
                    return this.TotalBounds.Width;
                else
                {
                    if (this.ColumnCount == 0)
                        return PDFUnit.Zero;
                    else if (this.ColumnCount == 1)
                        return this.Columns[0].Width;
                    else
                    {
                        return this.TotalBounds.Width;
                    }
                }
            }
        }

        #endregion

        #region public override PDFUnit Height {get;}

        /// <summary>
        /// Gets the height of this block
        /// </summary>
        public override PDFUnit Height
        {
            get
            {
                if (this.IsClosed)
                    return this.TotalBounds.Height;
                else
                {
                    if (this.ColumnCount == 0)
                        return PDFUnit.Zero;
                    else if (this.ColumnCount == 1)
                        return this.Columns[0].Height;
                    else
                    {
                        PDFUnit h = PDFUnit.Zero;
                        for (int i = 0; i < this.ColumnCount; i++)
                        {
                            h = PDFUnit.Max(h, this.Columns[i].Height);
                        }
                        return h;
                    }
                }
            }
        }

        #endregion

        #region public int BlockRepeatIndex { get; set; }

        /// <summary>
        /// Gets or sets the repeat index for this block.
        /// </summary>
        /// <remarks>PanelLayoutEngine uses this to count the number 
        /// of individual blocks generated for each panel</remarks>
        public int BlockRepeatIndex { get; set; }

        #endregion

        #region public List<IPDFComponent> ChildComponents {get;} + HasChildComponents

        private List<IPDFComponent> _children;

        /// <summary>
        /// Gets a list of all the child components that are part of this block
        /// </summary>
        public List<IPDFComponent> ChildComponents
        {
            get
            {
                if (null == _children)
                    _children = new List<IPDFComponent>();
                return _children;
            }
        }

        /// <summary>
        /// Returns true if this block has components that can rendered in it
        /// </summary>
        public bool HasChildComponents
        {
            get { return null != _children && _children.Count > 0; }
        }

        #endregion

        //
        // ctor(s)
        //

        #region public PDFLayoutBlock(PDFLayoutItem parent, IPDFComponent owner, IPDFLayoutEngine engine, PDFStyle fullstyle, OverflowSplit split)

        /// <summary>
        /// Creates a new layout block
        /// </summary>
        /// <param name="parent"></param>
        /// <param name="owner"></param>
        /// <param name="fullstyle"></param>
        public PDFLayoutBlock(PDFLayoutItem parent, IPDFComponent owner, IPDFLayoutEngine engine, PDFStyle fullstyle, OverflowSplit split)
            : base(parent, owner, engine, fullstyle)
        {
            this.OverflowSplit = split;
        }

        #endregion

        //
        // methods
        //

        #region protected override bool DoClose(ref string msg)

        /// <summary>
        /// Closes the entire block
        /// </summary>
        /// <param name="msg"></param>
        /// <returns></returns>
        protected override bool DoClose(ref string msg)
        {
            if (this.CurrentRegion != null && !this.CurrentRegion.IsClosed)
                this.CurrentRegion.Close();
           
            //If the owner is a page then we don't want to resize
            //as it's fixed.
            if (!(Owner is PDFPage)) 
            {
                PDFSize sz = this.GetRequiredSize();
                this.ShrinkToFit(sz.Width, sz.Height);
            }

            //if (this.Size == PDFSize.Empty && this.IsContainer 
            //    && this.HasChildComponents == false
            //    && this.HasPositionedRegions == false)
            //{
            //    this.TotalBounds = PDFRect.Empty;
            //    this.Size = PDFSize.Empty;
            //}
            

            return base.DoClose(ref msg);
        }

        #endregion

        #region public virtual void InitRegions(PDFRect totalbounds, PDFPositionOptions position, int columncount, PDFUnit alley)

        public virtual void InitRegions(PDFRect totalbounds, PDFPositionOptions position, int columncount, PDFUnit alley)
        {
            bool flow = true;
            this.InitRegions(totalbounds, position, columncount, alley, flow);
        }

        /// <summary>
        /// Initializes and sets up the regions for a block. 
        /// </summary>
        /// <param name="available">The available space inorder to set inner content placement.</param>
        /// <param name="totalbounds">The total bounds including any margins and padding</param>
        /// <param name="margins">The margins associated with this content</param>
        /// <param name="padding">The padding associated with this content</param>
        /// <param name="columncount"></param>
        /// <param name="autoflow">If true and there is more than one column, 
        /// they will be linked so that content overflows from one to the next. Otherwise the MoveToNextRegion(explicit) must be used.</param>
        /// <param name="alley"></param>
        public virtual void InitRegions(PDFRect totalbounds, PDFPositionOptions position, int columncount, PDFUnit alley, bool autoflow)
        {
            PDFTraceLog log = PDFTraceContext.GetLog();
            if (log.ShouldLog(TraceLevel.Debug))
                log.Add(TraceLevel.Debug, PDFBlockLayoutEngineBase.LOG_CATEGORY, "Initialized layout block for component " + (null == this.Owner ? "'NONE'" : this.Owner.ToString()));

            this.AlleyWidth = alley;
            this.TotalBounds = totalbounds;

            PDFRect avail = new PDFRect(position.Padding.Left, position.Padding.Top, totalbounds.Width - position.Padding.Left - position.Padding.Right, totalbounds.Height - position.Padding.Top - position.Padding.Bottom);
            //avail.Location = PDFPoint.Empty;

            this.AvailableBounds = avail;

            this.Size = PDFSize.Empty;
            this.Position = position;

            this.Columns = new PDFLayoutRegion[columncount];
            if (columncount == 1)
            {
                avail = new PDFRect(PDFPoint.Empty, avail.Size);
                
                this.ColumnCount = 1;
                this.Columns[0] = new PDFLayoutRegion(this, this.Owner, avail, 0, position.HAlign, position.VAlign);
            }
            else
            {
                this.ColumnCount = columncount;
                PDFUnit totalAlley = alley * (columncount - 1);
                PDFUnit totalAvailableWidth = this.AvailableBounds.Width - totalAlley;
                PDFUnit columnWidth = totalAvailableWidth / columncount;
                this.ColumnWidth = columnWidth;
                PDFUnit x = 0;// position.Padding.Left;
                PDFUnit y = 0;// position.Padding.Top;
                PDFUnit h = this.AvailableBounds.Height;
                PDFLayoutRegion previous = null;

                for (int i = 0; i < columncount; i++)
                {
                    PDFRect regionbounds = new PDFRect(x, y, columnWidth, h);
                    PDFLayoutRegion column = new PDFLayoutRegion(this, this.Owner, regionbounds, i, position.HAlign, position.VAlign);

                    //Set up the linked regions
                    if (null != previous)
                    {
                        previous.NextRegion = column;
                        previous.AutoOverflow = autoflow;
                    }

                    this.Columns[i] = column;
                    x += columnWidth + alley;
                    previous = column;
                }
            }

            //Set up the start column
            this.CurrentRegion = this.Columns[0];
        }

        #endregion

        #region public PDFLayoutBlock LastOpenBlock()

        /// <summary>
        /// Gets the last open block in the hierarchy. 
        /// If this block does not contain any open inner blocks, then it returns itself
        /// If this block is closed then an exception is raised.
        /// </summary>
        /// <returns></returns>
        public PDFLayoutBlock LastOpenBlock()
        {
            if (this.IsClosed)
                throw new InvalidOperationException(Errors.CannotLayoutABlockThatHasBeenClosed);

            PDFLayoutBlock last = this;

            PDFLayoutRegion reg = this.CurrentRegion;
            if (!reg.IsClosed)
            {
                PDFLayoutBlock childlast = reg.LastOpenBlock();
                if (null != childlast)
                    last = childlast;
            }
            return last;
        }

        #endregion

        #region public override bool MoveToNextRegion()

        /// <summary>
        /// Attempts to move to the next region, returning true if possible, otherwise false.
        /// </summary>
        /// <returns></returns>
        public override bool MoveToNextRegion()
        {
            bool force = false;
            return this.MoveToNextRegion(force);
        }

        /// <summary>
        /// Attempts to move to the next region, returning true if possible, otherwise false.
        /// </summary>
        /// <param name="force">If false then the regions must have been set up to AutoOverflow from one to the next (soft flow),
        /// if force is true then if there is a next region it will be used.</param>
        /// <returns></returns>
        public bool MoveToNextRegion(bool force)
        {
            PDFLayoutRegion region = this.CurrentRegion;
            if (region.NextRegion != null && (region.AutoOverflow || force))
            {
                this.CurrentRegion = region.NextRegion;
                return true;
            }
            else
                return false;
        }

        #endregion

        #region public override void PushComponentLayout(PDFLayoutContext context, PDFUnit xoffset, PDFUnit yoffset)

        /// <summary>
        /// Overrides the default implementation to push the component layouts on this blocks regions
        /// </summary>
        /// <param name="context"></param>
        protected override void DoPushComponentLayout(PDFLayoutContext context, int pageIndex, PDFUnit xoffset, PDFUnit yoffset)
        {
            bool logdebug = context.TraceLog.ShouldLog(TraceLevel.Debug);

            if (logdebug)
                context.TraceLog.Begin(TraceLevel.Debug, "Layout Block", "Starting to push the component arrangements for block " + this.ToString());
            
            
            //Offset our total bounds
            if (this.MovesWithLayout())
            {
                PDFRect total = this.TotalBounds;
                total = new PDFRect(xoffset + total.X, yoffset + total.Y, total.Width, total.Height);
                this.TotalBounds = total;
            }

            this.Size = this.TotalBounds.Size;
            this.Size = this.Size.Subtract(this.Position.Padding);

            xoffset = 0;
            yoffset = 0;

            



            for (int i = 0; i < this.Columns.Length; i++)
            {
                this.Columns[i].PushComponentLayout(context, pageIndex, xoffset, yoffset);
                //xoffset += this.AlleyWidth + this.ColumnWidth;
            }

            if (logdebug)
                context.TraceLog.End(TraceLevel.Debug, "Layout Block", "Completed pushing the component arrangements for block " + this.ToString());

        }

        /// <summary>
        /// returns true if this block should move and offset with the layout of other items
        /// </summary>
        /// <returns></returns>
        private bool MovesWithLayout()
        {
            PositionMode mode = this.Position.PositionMode;
            return mode != PositionMode.Absolute;
        }

        #endregion

        #region private PDFSize GetRequiredSize()

        /// <summary>
        /// Gets the required size of the block to fill it's content (or the explicit size)
        /// </summary>
        /// <returns></returns>
        private PDFSize GetRequiredSize()
        {
            PDFUnit w = this.Width;
            PDFUnit h = this.Height;
            if (this.Position.HasWidth)
            {
                w = this.Position.Width;
            }
            else if (this.Position.FillWidth)
            {
                w = this.AvailableBounds.Width;
            }

            if (this.Position.HasHeight)
            {
                h = this.Position.Height;
            }
            
            return new PDFSize(w, h);
        }

        #endregion

        #region private void ShrinkToFit(PDFUnit width, PDFUnit height)

        /// <summary>
        /// Reduces the height and width of this block and it's contents
        /// </summary>
        /// <param name="width"></param>
        /// <param name="height"></param>
        private void ShrinkToFit(PDFUnit width, PDFUnit height)
        {

            PDFRect full = this.TotalBounds;
            full.Height = height;
            full.Width = width;

            //Add padding back in for the region sizes
            //If there was explict heights on this block
            if (this.Position.Padding.IsEmpty == false)
            {
                if (this.Position.HasHeight)
                    full.Height -= (this.Position.Padding.Top + this.Position.Padding.Bottom);
                if (this.Position.HasWidth)
                    full.Width -= (this.Position.Padding.Left + this.Position.Padding.Right);
            }

            if (this.ColumnCount > 1)
            {
                for (int i = 0; i < this.ColumnCount; i++)
                {
                    //with columns we only shrink the height, not the width
                    PDFRect one = this.Columns[i].TotalBounds;
                    one.Height = full.Height;
                    this.Columns[i].TotalBounds = one;
                }
            }
            else
            {
                PDFRect content = this.Columns[0].TotalBounds;
                content.Height = full.Height;
                content.Width = full.Width;
                this.Columns[0].TotalBounds = content;
            }
            if (this.HasPositionedRegions)
            {
                foreach (PDFLayoutPositionedRegion reg in this.PositionedRegions)
                {
                    if (reg.PositionMode == PositionMode.Relative)
                    {
                        PDFRect content = reg.TotalBounds;
                        PDFSize bottomright = new PDFSize(content.X + content.Width, content.Y + content.Height);
                        
                        if (full.Width < bottomright.Width)
                            full.Width = bottomright.Width;
                        if (full.Height < bottomright.Height)
                            full.Height = bottomright.Height;
                    }
                }
            }

            //TODO: Use a separate rect for updating regions and setting this blocks content bounds

            //Add padding back in for setting this blocks bounds
            if (this.Position.Padding.IsEmpty == false)
            {
                full.Width += this.Position.Padding.Left + this.Position.Padding.Right;
                full.Height += this.Position.Padding.Top + this.Position.Padding.Bottom;
            }
            if (this.Position.Margins.IsEmpty == false)
            {
                full.Width += this.Position.Margins.Left + this.Position.Margins.Right;
                full.Height += this.Position.Margins.Top + this.Position.Margins.Bottom;
            }
            this.TotalBounds = full;
            this.Size = full.Size;
        }

        #endregion

        #region public void SetContentSize(PDFUnit width, PDFUnit height)

        /// <summary>
        /// Changes the content size of this Block to the specified values. THis should exclude any margins or padding on the block,
        /// which will automatically be added to the TotalBounds
        /// </summary>
        /// <param name="width"></param>
        /// <param name="height"></param>
        public void SetContentSize(PDFUnit width, PDFUnit height)
        {
            this.Size = new PDFSize(width, height);
            PDFRect total = this.TotalBounds;
            total.Width = width;
            total.Height = height;
            if (this.Position.Margins.IsEmpty == false)
            {
                total.Width += this.Position.Margins.Left + this.Position.Margins.Right;
                total.Height += this.Position.Margins.Top + this.Position.Margins.Bottom;
            }
            if (this.Position.Padding.IsEmpty == false)
            {
                total.Width += this.Position.Padding.Left + this.Position.Padding.Right;
                total.Height += this.Position.Padding.Top + this.Position.Padding.Bottom;
            }
            this.TotalBounds = total;
        }

        #endregion

        #region public override string ToString()

        /// <summary>
        /// Overrides the default implementation to return a string representation of this block
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            if (this.Owner != null)
                return this.Owner.ToString() + ":Block" + this.BlockRepeatIndex;
            else
                return "Orphaned:Block" + this.BlockRepeatIndex;
        }

        #endregion


        #region public PDFLayoutBlock BeginNewBlock(IPDFComponent owner, PDFStyle fullstyle)

        /// <summary>
        /// Begins a new block on the current region of this block
        /// </summary>
        /// <param name="owner"></param>
        /// <param name="fullstyle"></param>
        /// <returns></returns>
        public PDFLayoutBlock BeginNewBlock(IPDFComponent owner, IPDFLayoutEngine engine, PDFStyle fullstyle, PositionMode mode)
        {
            if (mode == PositionMode.Inline)
                throw RecordAndRaise.ArgumentOutOfRange("mode", Errors.CannotBeginABlockThatIsInline);
            //else if (mode == PositionMode.Absolute)
            //    throw RecordAndRaise.ArgumentOutOfRange("mode",Errors.AbsolutePositioningBlocksRegisterWithTheLayoutPage);
            //else if (mode == PositionMode.Relative) //TODO:Relative positioned blocks
            //    throw RecordAndRaise.NotSupported("Need to do relative positioning");

            PDFLayoutBlock container = this.LastOpenBlock();
            if (null == container)
                container = this;

            OverflowSplit split = OverflowSplit.Any;

            PDFOverflowStyle overflow;
            if (fullstyle.TryGetOverflow(out overflow))
                split = overflow.Split;
            PDFLayoutRegion region = container.CurrentRegion;

            PDFLayoutBlock block = new PDFLayoutBlock(container, owner, engine, fullstyle, split);
            region.Contents.Add(block);
            return block;
        }

        #endregion

        #region public PDFLayoutBlock BeginNewContainerBlock(PDFContainerComponent owner, IPDFLayoutEngine engine, PDFStyle fullstyle, PositionMode mode)

        /// <summary>
        /// Creates and returns a new open container block 
        /// </summary>
        /// <param name="owner"></param>
        /// <param name="engine"></param>
        /// <param name="fullstyle"></param>
        /// <param name="mode"></param>
        /// <returns></returns>
        public PDFLayoutBlock BeginNewContainerBlock(PDFContainerComponent owner, IPDFLayoutEngine engine, PDFStyle fullstyle, PositionMode mode)
        {
            PDFLayoutBlock block = this.BeginNewBlock(owner, engine, fullstyle, mode);
            block.IsContainer = true;
            return block;
        }

        #endregion

        #region public PDFLayoutRegion BeginNewPositionedRegion(PDFPositionOptions pos, PDFLayoutPage page, IPDFComponent comp, PDFStyle full)

        /// <summary>
        /// Starts a new layout region within this block that is either relatively or absolutely positioned.
        /// </summary>
        /// <param name="pos"></param>
        /// <param name="page"></param>
        /// <param name="comp"></param>
        /// <param name="full"></param>
        /// <returns></returns>
        public PDFLayoutRegion BeginNewPositionedRegion(PDFPositionOptions pos, PDFLayoutPage page, IPDFComponent comp, PDFStyle full)
        {
            PDFLayoutRegion before = this.CurrentRegion;
            PDFLayoutLine beforeline = before.CurrentItem as PDFLayoutLine;
            if (null == beforeline)
                beforeline = before.BeginNewLine();

            PDFRect space;
            if (pos.PositionMode == PositionMode.Absolute)
                //Page sizing
                space = new PDFRect(PDFUnit.Zero, PDFUnit.Zero, page.Width, page.Height);
            else
                //Block sizing
                space = new PDFRect(PDFPoint.Empty, this.TotalBounds.Size);

            //Calculate the available space
            if (pos.HasX)
            {
                space.X += pos.X;
                space.Width -= pos.X;
            }

            if (pos.HasY)
            {
                space.Y += pos.Y;
                space.Height -= pos.Y;
            }

            if (pos.HasWidth)
            {
                space.Width = pos.Width;
            }

            if (pos.HasHeight)
            {
                space.Height = pos.Height;
            }

            int index = -(this.PositionedRegions.Count + 1); //use a negative value for the positioned items

            PDFLayoutRegion created = new PDFLayoutPositionedRegion(this, comp, space, index, pos);
            this.PositionedRegions.Add(created);
            PDFLayoutPositionedRegionRun posRun = beforeline.AddPositionedRun(created, comp);

            return created;
        }

        #endregion

        public override void ResetAvailableHeight(PDFUnit height, bool includeChildren)
        {
            this._avail.Height = height;
            if (includeChildren && null != this.Columns)
            {
                foreach (PDFLayoutRegion region in this.Columns)
                {
                    region.ResetAvailableHeight(height, includeChildren);
                }
            }
            base.ResetAvailableHeight(height, includeChildren);
        }

        #region public override PDFObjectRef DoOutputToPDF(PDFRenderContext context, PDFWriter writer)

        /// <summary>
        /// Renders this block to the PDFWriter
        /// </summary>
        /// <param name="context"></param>
        /// <param name="writer"></param>
        /// <returns></returns>
        protected override PDFObjectRef DoOutputToPDF(PDFRenderContext context, PDFWriter writer)
        {
            bool logdebug = context.TraceLog.ShouldLog(TraceLevel.Debug);
            PDFComponent component = this.Owner as PDFComponent;
            
            if (logdebug)
            {
                if (null != component)
                {
                    context.TraceLog.Begin(TraceLevel.Debug, "Layout Block", "Writing Block for: " + this.Owner.ID);
                }
                else
                    context.TraceLog.Begin(TraceLevel.Debug, "Layout Block", "Writing un-owned Block");
            }
            
            
            PDFStyle fullstyle = this.FullStyle;
            if (null == fullstyle)
                throw new NullReferenceException(Errors.ThisBlockDoesNotHaveAnyStyle);
            PDFPositionOptions position = this.Position;
            if (null == position)
                throw new NullReferenceException(Errors.ThisBlockDoesNotHaveAnyPostionOptions);

            bool render = this.ShouldOutput(context);
            //Check that this content should be rendered
            if (!render)
            {
                if (logdebug)
                    context.TraceLog.Add(TraceLevel.Debug, "PDFLayoutBlock", "Skipping the output of block " + this.ToString() + " because it has height '0'");
                return null;
            }

            PDFStyle prevStyle = context.FullStyle;
            PDFSize prevSize = context.Space;
            PDFPoint prevLoc = context.Offset;
            try
            {
                context.FullStyle = fullstyle;

                PDFRect total = this.TotalBounds;
                total = total.Offset(context.Offset);

                PDFRect borderRect = total.Inset(this.Position.Margins);
                PDFRect contentRect = borderRect.Inset(this.Position.Padding);

                

                //Get the border to draw
                PDFBorderStyle border = null;
                bool hasborder = null != fullstyle && fullstyle.TryGetBorder(out border);

                //Get any background
                PDFBackgroundStyle bg = null;
                bool hasBg = null != fullstyle && fullstyle.TryGetBackground(out bg);

                if (hasBg && render)
                {
                    this.OutputBackground(bg, border, context, borderRect);
                }

                context.Offset = contentRect.Location;
                context.Space = contentRect.Size;

                //Perform the atual wrting of this blocks inner conntent
                OutputInnerContent(context, writer);

                
                if (hasborder && render)
                {
                    this.OutputBorder(bg, border, context, borderRect);
                }

                if (render && null != component)
                {
                    component.SetArrangement(context, fullstyle, contentRect);
                }

                PDFOverlayGridStyle grid;
                if (this.FullStyle.TryGetOverlayGrid(out grid))
                {
                    if (grid.ShowGrid)
                        this.OutputOverlayGrid(grid, context, total);
                    if (grid.HighlightColumns)
                        this.OutputRegionOverlay(grid, context, contentRect);
                }
            }
            catch (PDFRenderException)
            {
                throw;
            }
            catch (Exception ex)
            {
                throw new PDFRenderException("Could not output the block '" + this.ToString() + "'. " + ex.Message, ex);
            }

            context.Space = prevSize;
            context.Offset = prevLoc;
            context.FullStyle = prevStyle;

            if (logdebug)
            {
                if (null != this.Owner)
                    context.TraceLog.End(TraceLevel.Debug, "Layout Block", "Finished writing Block for: " + this.Owner.ID);
                else
                    context.TraceLog.End(TraceLevel.Debug, "Layout Block", "Finished writing un-owned Block");
            }

            return null;
        }

        
        #endregion

        #region protected virtual void OutputInnerContent(PDFRenderContext context, PDFWriter writer)

        /// <summary>
        /// Renders the inner content in this block
        /// </summary>
        /// <param name="context"></param>
        /// <param name="writer"></param>
        protected virtual void OutputInnerContent(PDFRenderContext context, PDFWriter writer)
        {
            
            if (this.ColumnCount > 1)
            {
                foreach (PDFLayoutRegion region in this.Columns)
                {
                    region.OutputToPDF(context, writer);
                }
            }
            else
            {
                this.Columns[0].OutputToPDF(context, writer);
            }
        }

        #endregion

        #region protected virtual bool ShouldOutput(PDFRenderContext context)

        /// <summary>
        /// Returns true if this block should actually be written.
        /// </summary>
        /// <param name="context"></param>
        /// <returns></returns>
        protected virtual bool ShouldOutput(PDFRenderContext context)
        {
            if (this.Position.Visibility == Visibility.Visible)
            {
                if (this.Height > 0)
                    return true;
                
                foreach (PDFLayoutRegion reg in this.Columns)
                {
                    if (reg.Height > 0)
                        return true;
                }

                if (this.HasPositionedRegions)
                    return true;
                
                //No height or content with any height so do not render.
                return false;
            }
            else
                return false;
        }

        #endregion

        #region private void OutputRegionOverlay(PDFOverlayGridStyle grid, PDFRenderContext context, PDFRect contentRect)

        private const double ColumnOverlayOpacity = 0.4;

        /// <summary>
        /// Writes a full block for each of the regions in this block
        /// </summary>
        /// <param name="grid">The grid style to use for rendering</param>
        /// <param name="context">The current render context</param>
        /// <param name="contentRect">The content rectagle that encloses all the columns</param>
        private void OutputRegionOverlay(PDFOverlayGridStyle grid, PDFRenderContext context, PDFRect contentRect)
        {
            if (null == grid)
                throw new ArgumentNullException("grid");

            PDFGraphics graphics = context.Graphics;

            int count = this.ColumnCount;
            PDFUnit w = contentRect.Width;
            PDFUnit alley = this.AlleyWidth;
            w -= alley * (count - 1);
            w /= count;

            PDFRect r = new PDFRect(contentRect.Location, new PDFSize(w, contentRect.Height));
            PDFSolidBrush solid = PDFSolidBrush.Create(grid.GridColor);
            solid.Opacity = ColumnOverlayOpacity;

            for (int i = 0; i < count; i++)
            {
                graphics.FillRectangle(solid, r);
                r.X += w + alley;
            }

        }

        #endregion

        #region private void OutputOverlayGrid(PDFOverlayGridStyle grid, PDFRenderContext context)

        /// <summary>
        /// Renders any overlay grid and column highlights
        /// </summary>
        /// <param name="grid">The grid to render</param>
        /// <param name="context"></param>
        /// <param name="rect">The rectangle to use to render the grid within</param>
        private void OutputOverlayGrid(PDFOverlayGridStyle grid, PDFRenderContext context, PDFRect rect)
        {
            if (null == grid)
                throw new ArgumentNullException("grid");

            PDFGraphics graphics = context.Graphics;
            PDFPen pen = grid.GetPen();
            if (null == pen)
                return;

            //Will paint a 50% opacity colour over the columns in a panel
            //so they can be identified
            graphics.SaveGraphicsState();

            

            PDFRect absolutebounds = rect;
            pen.SetUpGraphics(graphics, absolutebounds);
            
            //vertical lines first
            PDFUnit space = grid.GridSpacing;
            if (space <= 0)
                throw new ArgumentOutOfRangeException("Cannot render a grid with zero or less spacing");

            PDFUnit x1 = absolutebounds.X + grid.GridXOffset;
            PDFUnit y1 = absolutebounds.Y;
            PDFUnit y2 = absolutebounds.Y + absolutebounds.Height;
            PDFUnit x2 = x1 + absolutebounds.Width;

            while (x1 < x2)
            {
                graphics.DrawLine(x1, y1, x1, y2);
                x1 += space;
            }

            x1 = absolutebounds.X;
            y1 = absolutebounds.Y + grid.GridYOffset;

            while (y1 < y2)
            {
                graphics.DrawLine(x1, y1, x2, y1);
                y1 += space;
            }

            pen.ReleaseGraphics(graphics, absolutebounds);
            graphics.RestoreGraphicsState();
        }

        #endregion

        #region internal void Offset(PDFUnit x, PDFUnit y)

        /// <summary>
        /// Offsets this block by the specifed amounts
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        internal void Offset(PDFUnit x, PDFUnit y)
        {
            PDFRect content = this.AvailableBounds;
            content.Y += y;
            content.X += x;
            this.AvailableBounds = content;
            
            PDFRect total = this.TotalBounds;
            total.Y += y;
            total.X += x;
            this.TotalBounds = total;
        }

        #endregion

        #region internal void Shrink(PDFUnit width, PDFUnit height)

        /// <summary>
        /// Reduces the size of this layout block bythe specifed amount
        /// </summary>
        /// <param name="width"></param>
        /// <param name="height"></param>
        internal void Shrink(PDFUnit width, PDFUnit height)
        {
            PDFRect content = this.AvailableBounds;
            content.Width -= width;
            content.Height -= height;
            this.AvailableBounds = content;

            PDFRect total = this.TotalBounds;
            total.Width -= width;
            total.Height -= height;
            this.TotalBounds = total;

            if (this.Columns != null && this.Columns.Length > 0)
            {
                for (int i = 0; i < this.Columns.Length; i++)
                {
                    PDFLayoutRegion region = this.Columns[i];
                    content = region.TotalBounds;
                    content.Width -= width;
                    content.Height -= height;
                    region.TotalBounds = content;
                }
            }
        }

        #endregion

    }

    /// <summary>
    /// Represents a collection of blocks
    /// </summary>
    public class PDFLayoutBlockCollection : List<PDFLayoutBlock>
    {

    }
}
