﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Scryber.Drawing;
using Scryber.Styles;
using Scryber.Native;
using Scryber.Drawing.Imaging;
using Scryber.Resources;

namespace Scryber.Components.Layout
{
    /// <summary>
    /// Represents the layout of a document
    /// </summary>
    public class PDFLayoutDocument : PDFLayoutContainerItem
    {
        //
        // properties
        //

        #region public PDFDocument DocumentComponent { get; }

        /// <summary>
        /// Gets or sets the document that started the layout
        /// </summary>
        public PDFDocument DocumentComponent { get; private set; }

        #endregion

        #region public int CurrentPageIndex { get; set; }

        /// <summary>
        /// Gets or sets the current page index in this layout
        /// </summary>
        public int CurrentPageIndex { get; set; }

        #endregion

        #region public PDFLayoutPage CurrentPage { get; }

        /// <summary>
        /// Gets the current page layout (based on the CurrentPageIndex)
        /// </summary>
        public PDFLayoutPage CurrentPage
        {
            get { return this.AllPages[this.CurrentPageIndex]; }
        }

        #endregion

        #region public PDFLayoutPageCollection AllPages { get;}

        /// <summary>
        /// Gets all the laid out pages in this document
        /// </summary>
        public PDFLayoutPageCollection AllPages { get; private set; }

        #endregion


        #region public int TotalPageCount {get;}

        /// <summary>
        /// Gets the total number of document pages in this document after layout
        /// </summary>
        public int TotalPageCount
        {
            get
            {
                return this.Numbers.PageCount;
            }
        }

        #endregion

        #region public PageNumberingCollection Numbers {get;}

        private PageNumberingCollection _numbers;
        /// <summary>
        /// Gets the numbering collection. Readonly, but value is set with the InitPageNumbers method
        /// </summary>
        protected PageNumberingCollection Numbers
        {
            get
            {
                return _numbers;
            }
        }

        #endregion

        #region protected PDFPageNumberingStack NumberStyleStack

        private PDFPageNumberingStack _numberstack;

        protected PDFPageNumberingStack NumberStyleStack
        {
            get
            {
                if (null == _numberstack)
                    _numberstack = new PDFPageNumberingStack();
                return _numberstack;
            }
        }

        #endregion

        //
        // ctor(s)
        //

        #region public PDFLayoutDocument(PDFDocument doc)

        /// <summary>
        /// Creates a new PDFLayoutDocument
        /// </summary>
        public PDFLayoutDocument(PDFDocument doc, IPDFLayoutEngine engine)
            : base(null, doc, engine, new PDFStyle())
        {
            AllPages = new PDFLayoutPageCollection();
            CurrentPageIndex = -1;
            this.DocumentComponent = doc;
        }

        #endregion

        //
        // public methods
        //

        #region public PDFLayoutPage BeginNewContinuationPage()

        /// <summary>
        /// Begins a new page based on the current page's size and content rect. This will then be the current page
        /// </summary>
        /// <returns></returns>
        public PDFLayoutPage BeginNewContinuationPage()
        {
            if (CurrentPageIndex < 0)
                throw new ArgumentOutOfRangeException("Cannot begin a new page based on previous page if there are no previous pages");
            PDFLayoutPage pg = this.CurrentPage;

            PDFTraceLog log = PDFTraceContext.GetLog();
            if (log.ShouldLog(TraceLevel.Debug))
                log.Add("LAYOUT", "Beginning a new continuation page for '" + pg + "'");

            if (!pg.IsClosed)
                pg.Close();

            PDFSize size = pg.Size;
            PDFStyle style = pg.FullStyle;
            PDFPage owner = pg.Owner as PDFPage;
            OverflowAction overflow = pg.OverflowAction;

            PDFLayoutPage newpg = this.BeginNewPage(owner, this.Engine, style, overflow);
            return newpg;
        }

        #endregion

        #region public PDFLayoutPage BeginNewPage(PDFPage owner, PDFSize size, PDFStyle full)

        /// <summary>
        /// Begins a new page of the requested size and with the specified bounds. This will then be the current page
        /// </summary>
        /// <param name="size">The actual size of the page</param>
        /// <param name="full">The full style of the page</param>
        /// <returns></returns>
        public PDFLayoutPage BeginNewPage(PDFPage owner, IPDFLayoutEngine engine, PDFStyle full, OverflowAction action)
        {

            PDFLayoutPage pg = new PDFLayoutPage(this, owner, engine, full, action);
            this.CurrentPageIndex = this.AllPages.Count;
            this.AllPages.Add(pg);
            return this.CurrentPage;
        }

        #endregion

        
        //
        // abstract overrides
        //

        #region public override PDFUnit Height {get;}

        /// <summary>
        /// Returns Zero for the height of a document
        /// </summary>
        public override PDFUnit Height
        {
            get { return PDFUnit.Zero; }
        }

        #endregion

        #region public override PDFUnit Width

        /// <summary>
        /// Returns zero for the width of a document
        /// </summary>
        public override PDFUnit Width
        {
            get { return PDFUnit.Zero; }
        }

        #endregion

        #region protected override void DoPushComponentLayout(PDFLayoutContext context, int pageIndex, PDFUnit xoffset, PDFUnit yoffset)

        /// <summary>
        /// Overrides the base abstract method to push the arrangements for each page
        /// </summary>
        /// <param name="context"></param>
        protected override void DoPushComponentLayout(PDFLayoutContext context, int pageIndex, PDFUnit xoffset, PDFUnit yoffset)
        {
            if (context.TraceLog.ShouldLog(TraceLevel.Debug))
                context.TraceLog.Begin(TraceLevel.Debug, "Layout Document", "Starting to push the layouts back onto the components");

            foreach (PDFLayoutPage page in this.AllPages)
            {
                page.PushComponentLayout(context, pageIndex, xoffset, yoffset);
            }

            if (context.TraceLog.ShouldLog(TraceLevel.Debug))
                context.TraceLog.End(TraceLevel.Debug, "Layout Document", "Completed pushing the layouts onto the components");

        }

        #endregion

        //
        // Artefacts
        //

        #region public PDFArtefactCollectionSet Artefacts {get;}

        private PDFArtefactCollectionSet _artefacts = new PDFArtefactCollectionSet();

        /// <summary>
        /// Gets the set of artefact collections
        /// </summary>
        public PDFArtefactCollectionSet Artefacts
        {
            get
            {
                return _artefacts;
            }
        }

        #endregion

        #region public object RegisterCatalogEntry(PDFRegistrationContext context, string catalogtype, ICatalogEntry entry)

        /// <summary>
        /// Registers the IArtefactEntry values with the artefact collection of type 'catalogtype'.
        /// The method returns a reference object that must be passed back to CloseArtefactEntry when the entry should be closed.
        /// 
        /// </summary>
        /// <param name="context"></param>
        /// <param name="catalogtype"></param>
        /// <param name="entry"></param>
        /// <returns></returns>
        public object RegisterCatalogEntry(PDFLayoutContext context, string catalogtype, IArtefactEntry entry)
        {
            IArtefactCollection col;
            if (!this._artefacts.TryGetCollection(catalogtype, out col))
            {
                col = this.CreateArtefactCollection(catalogtype);
                _artefacts.Add(col);
            }
            return col.Register(entry);
        }

        #endregion

        #region protected virtual ICatalogCollection CreateArtefactCollection(string type)

        /// <summary>
        /// Overridable method that creates the specific collections for individual artefact types
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        internal protected virtual IArtefactCollection CreateArtefactCollection(string type)
        {
            switch (type)
            {
                case (PDFArtefactTypes.Annotations):
                    return new PDFAnnotationCollection(type);
                case (PDFArtefactTypes.Names):
                    return new PDFNameDictionary(type);
                case (PDFArtefactTypes.Outlines):
                    return new PDFOutlineStack(type);

                default:
                    throw RecordAndRaise.NotSupported("The catalog type {0} is not a known catalog type", type);

            }
        }

        #endregion

        #region public void CloseArtefactEntry(string catalogtype, object entry)

        /// <summary>
        /// Closes the last artefact entry that was started with the 'RegisterCatalogEntry'
        /// </summary>
        /// <param name="catalogtype"></param>
        /// <param name="entry"></param>
        public void CloseArtefactEntry(string catalogtype, object entry)
        {
            IArtefactCollection col;
            if (this._artefacts.TryGetCollection(catalogtype, out col))
            {
                col.Close(entry);
            }
        }

        #endregion

        //
        // Outputting
        //

        #region protected override PDFObjectRef DoOutputToPDF(PDFRenderContext context, PDFWriter writer)

        /// <summary>
        /// Overrides the base implementation to output the document
        /// </summary>
        /// <param name="context"></param>
        /// <param name="writer"></param>
        /// <returns></returns>
        protected override PDFObjectRef DoOutputToPDF(PDFRenderContext context, PDFWriter writer)
        {
            writer.Open();
            PDFObjectRef catalog = this.WriteCatalog(context, writer);

            this.WriteInfo(context, writer);

            PDFDocumentID id = this.DocumentComponent.DocumentID;
            if (null == id)
                id = PDFDocumentID.Create();

            writer.Close(id);

            return catalog;
        }

        #endregion

        #region private PDFObjectRef WriteCatalog(PDFRenderContext context, PDFWriter writer)

        /// <summary>
        /// Outputs the document catalog (usually the first Component) and then calls output on each of the documents 
        /// </summary>
        /// <param name="context"></param>
        /// <param name="writer"></param>
        /// <returns></returns>
        protected virtual PDFObjectRef WriteCatalog(PDFRenderContext context, PDFWriter writer)
        {
            PDFObjectRef catalog = writer.BeginObject("Catalog");
            writer.BeginDictionary();
            writer.WriteDictionaryNameEntry("Type", "Catalog");

            // Pages
            context.TraceLog.Begin(TraceLevel.Debug,"Document", "Writing Pages");
            PDFObjectRef pglist = this.OutputPages(context, writer);
            writer.WriteDictionaryObjectRefEntry("Pages", pglist);
            context.TraceLog.End(TraceLevel.Debug, "Document", "Writing Pages");

            // Page Labels
            context.TraceLog.Begin(TraceLevel.Debug, "Document", "Writing Page Labels");
            PDFObjectRef pglabels = this.WritePageLabels(context, writer);
            if (null != pglabels)
                writer.WriteDictionaryObjectRefEntry("PageLabels", pglabels);
            context.TraceLog.End(TraceLevel.Debug, "Document", "Writing Page Labels");

            // Artefacts
            context.TraceLog.Begin(TraceLevel.Debug, "Document", "Writing Artefacts");
            this.WriteArtefacts(context, writer);
            context.TraceLog.End(TraceLevel.Debug, "Document", "Writing Artefacts");

            //Viewer Preferences
            context.TraceLog.Begin(TraceLevel.Debug, "Document", "Writing Viewer Preferences");
            this.WriteViewerPreferences(context, writer);
            context.TraceLog.End(TraceLevel.Debug, "Document", "Writing Viewer Preferences");
            writer.EndDictionary();
            writer.EndObject();

            return catalog;
        }

        #endregion

        #region protected virtual void WriteViewerPreferences(PDFRenderContext context, PDFWriter writer)

        /// <summary>
        /// Outputs the positioning and visibility of UI elements for the viewer
        /// </summary>
        /// <param name="context"></param>
        /// <param name="writer"></param>
        protected virtual void WriteViewerPreferences(PDFRenderContext context, PDFWriter writer)
        {
            PDFDocumentViewPreferences docview = this.DocumentComponent.ViewPreferences;

            if (null != docview)
            {
                PDFObjectRef view = docview.OutputToPDF(context, writer);
                if (null != view)
                {
                    writer.WriteDictionaryObjectRefEntry("ViewerPreferences", view);
                }

                string value = docview.GetPageDisplayName(docview.PageDisplay);
                if (!string.IsNullOrEmpty(value))
                {
                    writer.WriteDictionaryNameEntry("PageMode", value);
                }

                value = docview.GetPageLayoutName(docview.PageLayout);
                if (!string.IsNullOrEmpty(value))
                {
                    writer.WriteDictionaryNameEntry("PageLayout", value);
                }
            }
        }

        #endregion

        #region private void WriteArtefacts(PDFRenderContext context, PDFWriter writer)

        /// <summary>
        /// If this document has artefacts then the collection is output. 
        /// </summary>
        /// <param name="context"></param>
        /// <param name="writer"></param>
        private void WriteArtefacts(PDFRenderContext context, PDFWriter writer)
        {
            PDFArtefactCollectionSet artefacts = this.Artefacts;
            if (artefacts != null && artefacts.Count > 0)
            {
                foreach (IArtefactCollection col in artefacts)
                {
                    context.TraceLog.Begin(TraceLevel.Debug, "Layout Document", "Outputting artefact catalog entry collection " + col.CollectionName);
                    writer.BeginDictionaryEntry(col.CollectionName);

                    PDFObjectRef entry = col.OutputToPDF(context, writer);
                    if (entry != null)
                        writer.WriteObjectRef(entry);
                    else
                        writer.WriteNull();
                    writer.EndDictionaryEntry();

                    context.TraceLog.End(TraceLevel.Debug, "Layout Document", "Finished artefact catalog entry collection " + col.CollectionName);
                }
            }
        }

        #endregion

        #region private void WriteInfo(PDFRenderContext context, PDFWriter writer) + support methods

        /// <summary>
        /// writes the collection of document info entries
        /// </summary>
        /// <param name="context"></param>
        /// <param name="writer"></param>
        /// <returns></returns>
        protected virtual void WriteInfo(PDFRenderContext context, PDFWriter writer)
        {
            PDFDocumentInfo info = this.DocumentComponent.Info;
            if (null != info)
                info.OutputToPDF(context, writer);            
        }


        #endregion


        #region private PDFObjectRef WritePageLabels(PDFRenderContext context, PDFWriter writer)

        /// <summary>
        /// Writes the collection of page labels and returns a reference to this collection
        /// </summary>
        /// <param name="context"></param>
        /// <param name="writer"></param>
        /// <returns></returns>
        private PDFObjectRef WritePageLabels(PDFRenderContext context, PDFWriter writer)
        {
            PageNumberingCollection nums = this.Numbers;
            PDFObjectRef labels = writer.BeginObject("PageLabels");
            writer.BeginDictionary();
            writer.BeginDictionaryEntry("Nums");
            writer.BeginArrayS();
            foreach (PageNumberingEntry entry in this.Numbers)
            {
                writer.BeginArrayEntry();
                writer.WriteNumberS(entry.PageIndex);
                writer.BeginDictionaryS();
                string type;
                switch (entry.NumberingFormat.NumberStyle)
                {
                    case PageNumberStyle.Decimals:
                        type = "D";
                        break;
                    case PageNumberStyle.UppercaseRoman:
                        type = "R";
                        break;
                    case PageNumberStyle.LowercaseRoman:
                        type = "r";
                        break;
                    case PageNumberStyle.UppercaseLetters:
                        type = "A";
                        break;
                    case PageNumberStyle.LowercaseLetters:
                        type = "a";
                        break;
                    default:
                        type = "";
                        break;
                }
                if (!string.IsNullOrEmpty(type))
                    writer.WriteDictionaryNameEntry("S", type);
                if (!string.IsNullOrEmpty(entry.NumberingFormat.Prefix))
                    writer.WriteDictionaryStringEntry("P", entry.NumberingFormat.Prefix);
                if (entry.NumberingFormat.StartIndex > 1)
                    writer.WriteDictionaryNumberEntry("St", entry.NumberingFormat.StartIndex);
                writer.EndDictionary();
                writer.EndArrayEntry();
            }
            writer.EndArray();
            writer.EndDictionaryEntry();
            writer.EndDictionary();
            writer.EndObject();
            return labels;

        }

        #endregion

        #region protected PDFObjectRef OutputPages(PDFRenderContext context, PDFWriter writer)

        /// <summary>
        /// Outputs the first Page tree Component and calls Output on each of the layout pages.
        /// </summary>
        /// <param name="context">The current context</param>
        /// <param name="writer">The current writer</param>
        /// <returns>A reference to the current page tree root Component</returns>
        protected PDFObjectRef OutputPages(PDFRenderContext context, PDFWriter writer)
        {
            //Begin the Pages object and dictionary
            PDFObjectRef pgs = writer.BeginObject(Const.PageTreeName);
            writer.BeginDictionary();
            writer.WriteDictionaryNameEntry("Type", "Pages");

            List<PDFObjectRef> pagerefs = new List<PDFObjectRef>(this.AllPages.Count);
            context.PageIndex = 0;

            //allow each page to output itself, and add the returned PDFObjectRef to the list for the kids dictionary entry below. 
            foreach (PDFLayoutPage ppe in this.AllPages)
            {
                PDFObjectRef oref = ppe.OutputToPDF(context, writer);
                if (oref != null)
                {
                    pagerefs.Add(oref);
                    context.PageIndex += 1;
                }
            }

            //write the kids array entry in the dictionary
            writer.BeginDictionaryEntry("Kids");
            writer.BeginArray();
            foreach (PDFObjectRef kid in pagerefs)
            {
                writer.BeginArrayEntry();
                writer.WriteFileObject(kid);
                writer.EndArrayEntry();
            }
            writer.EndArray();
            //Write the total number of pages to the dictionary
            writer.EndDictionaryEntry();
            writer.BeginDictionaryEntry("Count");
            writer.WriteNumber(pagerefs.Count);
            writer.EndDictionaryEntry();

            //close the ditionary and the object
            writer.EndDictionary();

            writer.EndObject();

            return pgs;
        }

        #endregion


        //
        // page numbers
        //

        /// <summary>
        /// Called at the start of document layout to initialize the numbering collection
        /// </summary>
        /// <param name="style"></param>
        public void InitPageNumbering(PDFPageStyle style)
        {
            this._numbers = new PageNumberingCollection(this, style);
            this._numberstack = new PDFPageNumberingStack();
            this._numberstack.Push(this._numbers.First().NumberingFormat);

        }

        #region public void RegisterPageNumbering(int pageindex, PDFPage page, PDFPageStyle num)

        /// <summary>
        /// Registers the current page numbering options for the specified page at the specified index
        /// </summary>
        /// <param name="pageindex"></param>
        /// <param name="page"></param>
        /// <param name="num"></param>
        public int RegisterPageNumbering(PDFLayoutPage page, PDFPageStyle style, out PDFPageNumbering numbering)
        {
            int index = this.TotalPageCount;
            this.Numbers.Register(index, page, style, out numbering);
            if (null != numbering)
                this.NumberStyleStack.Push(numbering);
            return index;
        }

        #endregion

        public void UnRegisterPageNumbering(int currentPageIndex, PDFLayoutPage page, PDFPageNumbering numbering)
        {
            if (null != numbering)
            {
                this.NumberStyleStack.Pop();
                PDFPageNumbering previous = this.NumberStyleStack.Peek();
                this.Numbers.Restore(currentPageIndex + 1, previous);
            }
        }

        /// <summary>
        /// Gets the page numbering for the requested page
        /// </summary>
        /// <param name="pageindex"></param>
        /// <param name="index">The translated ofset based on the requested page index</param>
        /// <returns></returns>
        public PDFPageNumbering GetNumbering(int pageindex, out int offset, out int max)
        {
            if (this.Numbers == null)
            {
                offset = 0;
                max = 0;
                return null;
            }
            else
            {
                return this.Numbers.GetNumbering(pageindex, out offset, out max);
            }
        }


        //
        // inner classes
        //


        #region protected class PageNumberingStack + protected class PageNumberingEntry + private class PageNumberingCollection

        protected class PDFPageNumberingStack : Stack<PDFPageNumbering>
        {
        }

        /// <summary>
        /// Defines a single page number format that starts at the specified index
        /// </summary>
        protected class PageNumberingEntry
        {
            public int PageIndex { get; set; }

            public PDFLayoutPage Page { get; set; }

            public PDFPageNumbering NumberingFormat { get; set; }

            public string GetPageString(int forpageindex)
            {
                forpageindex = forpageindex - this.PageIndex;
                return NumberingFormat.GetPageNumber(forpageindex);
            }
        }

        /// <summary>
        /// Retains a sorted list of the Page Number format entries
        /// </summary>
        protected class PageNumberingCollection : IEnumerable<PageNumberingEntry>
        {
            private PDFLayoutDocument _ownerdoc;
            private int _pageCount;
            private List<PageNumberingEntry> _entries;

            public PDFLayoutDocument OwnerDocument
            {
                get { return _ownerdoc; }
            }

            public int PageCount
            {
                get { return _pageCount; }
            }

            public PageNumberingCollection(PDFLayoutDocument owner):
                this(owner, new PDFPageNumbering() { NumberStyle = PageNumberStyle.Decimals, Prefix = "", StartIndex = 0 })
            {
            }

            public PageNumberingCollection(PDFLayoutDocument owner, PDFPageStyle style)
                : this(owner, CreateDefaultNumbering(style))
            {
            }

            public PageNumberingCollection(PDFLayoutDocument owner, PDFPageNumbering defaultNumbering)
            {
                this._ownerdoc = owner;
                this._entries = new List<PageNumberingEntry>();
                this._pageCount = 0;
                this._entries.Add(new PageNumberingEntry() { NumberingFormat = defaultNumbering, Page = null, PageIndex = 0 });

            }

            /// <summary>
            /// restores a previous page numbering style onto the list so that it is re-used in the hierarchy
            /// </summary>
            /// <param name="nextPageIndex"></param>
            /// <param name="num"></param>
            public void Restore(int nextPageIndex, PDFPageNumbering num)
            {
                PageNumberingEntry entry = new PageNumberingEntry();
                entry.NumberingFormat = num;

                entry.Page = null;
                entry.PageIndex = nextPageIndex;
                this._entries.Add(entry);
            }

            public void Register(int pageindex, PDFLayoutPage page, PDFPageStyle style, out PDFPageNumbering num)
            {
                this._pageCount++;
                num = null;
                if (null == style)
                    return;

                for (int i = _entries.Count - 1; i >= 0; i--)
                {
                    PageNumberingEntry entry = _entries[i];
                    if (entry.PageIndex == pageindex)
                    {
                        if (null != page && entry.Page != null && entry.Page != page)
                            throw new ArgumentException("Duplicate page indexes for different pages. Cannot be registered");
                        num = MergeNumberingFormat(style, entry.NumberingFormat);
                        entry.NumberingFormat = num;
                    }
                    if (entry.PageIndex < pageindex)
                    {
                        if (entry.NumberingFormat.Equals(style) == false)
                        {
                            PDFPageNumbering origformat = entry.NumberingFormat;
                            entry = new PageNumberingEntry();
                            num = MergeNumberingFormat(style, origformat);
                            entry.NumberingFormat = num;
                            entry.Page = page;
                            entry.PageIndex = pageindex;
                            _entries.Add(entry);
                            break;
                        }
                        else
                            break;//we have already registered this type
                    }
                }

            }

            

            public PDFPageNumbering GetNumbering(int pageindex, out int numberindex, out int numberingmax)
            {
                if (_entries.Count == 0)
                {
                    numberindex = 0;
                    //PDFPageNumbering num = new PDFPageNumbering() { NumberStyle = PageNumberStyle.Decimals, Prefix = string.Empty, StartIndex = 1 };
                    //return num;
                    throw new ArgumentOutOfRangeException("No pages registered");
                }
                for (int i = this._entries.Count - 1; i >= 0; i--)
                {
                    PageNumberingEntry entry = this._entries[i];
                    if (entry.PageIndex <= pageindex)
                    {
                        numberindex = pageindex - entry.PageIndex;
                        if (i == this._entries.Count - 1)
                            numberingmax = this._ownerdoc.TotalPageCount - entry.PageIndex - 1;
                        else
                            numberingmax = _entries[i + 1].PageIndex - entry.PageIndex - 1;

                        //Add the numbering start index to the offsets if it's set.
                        if (entry.NumberingFormat.StartIndex > 1)
                        {
                            numberindex += entry.NumberingFormat.StartIndex - 1;
                            numberingmax += entry.NumberingFormat.StartIndex - 1;
                        }
                        return entry.NumberingFormat;
                    }
                }
                numberindex = pageindex;
                numberingmax = this._ownerdoc.TotalPageCount;
                return _entries[0].NumberingFormat;
            }

            #region IEnumerable<PageNumberingEntry> Members

            public IEnumerator<PageNumberingEntry> GetEnumerator()
            {
                return this._entries.GetEnumerator();
            }

            #endregion

            #region IEnumerable Members

            System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
            {
                return this.GetEnumerator();
            }

            #endregion

            private static PDFPageNumbering CreateDefaultNumbering(PDFPageStyle style)
            {
                PDFPageNumbering def = new PDFPageNumbering() { NumberStyle = PageNumberStyle.Decimals, Prefix = null, StartIndex = 0 };
                return MergeNumberingFormat(style, def);
            }

            protected static PDFPageNumbering MergeNumberingFormat(PDFPageStyle newStyle, PDFPageNumbering baseformat)
            {
                PDFPageNumbering newformat = new PDFPageNumbering();
                if (newStyle.IsDefined(StyleKeys.PageNumberStart))
                    newformat.StartIndex = newStyle.NumberStartIndex;
                else
                    newformat.StartIndex = baseformat.StartIndex;

                if (newStyle.IsDefined(StyleKeys.PageNumberPrefix))
                    newformat.Prefix = newStyle.NumberPrefix;
                else
                    newformat.Prefix = baseformat.Prefix;

                if (newStyle.IsDefined(StyleKeys.PageNumberStyleAttr))
                    newformat.NumberStyle = newStyle.NumberStyle;
                else
                    newformat.NumberStyle = baseformat.NumberStyle;

                return newformat;
            }
        }

        #endregion

    }
}
