﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Scryber.Styles;
using Scryber.Drawing;
using Scryber.Native;

namespace Scryber.Components.Layout
{
    /// <summary>
    /// A distinct layout item - block, line, table, anything
    /// </summary>
    public abstract class PDFLayoutItem
    {
        protected const int NULL_PAGEINDEX = -1;
        protected const int NULL_COLUMNINDEX = -1;
        protected const string LOG_CATEGORY = "PDFLayout";

        //
        // properties
        //

        #region public PDFLayoutItem Parent { get; private set; }

        /// <summary>
        /// Gets the parent of this layout item 
        /// </summary>
        public PDFLayoutItem Parent { get; private set; }

        #endregion
        
        #region public bool IsClosed { get; private set; }

        /// <summary>
        /// Returns true if this item is closed and cannot be appended to
        /// </summary>
        public bool IsClosed { get; private set; }

        #endregion

        public IPDFComponent Owner
        {
            get;
            set;
        }
        //
        // ctor(s)
        //

        #region protected PDFLayoutItem(PDFLayoutItem parent)

        /// <summary>
        /// Creates a new PDFLayout item
        /// </summary>
        /// <param name="parent"></param>
        protected PDFLayoutItem(PDFLayoutItem parent, IPDFComponent owner)
        {
            this.Owner = owner;
            this.Parent = parent;
            this.IsClosed = false;
        }

        #endregion

        //
        // abstract methods
        //

        /// <summary>
        /// Gets the height of this item
        /// </summary>
        public abstract PDFUnit Height { get; }

        /// <summary>
        /// Gets the width of this item
        /// </summary>
        public abstract PDFUnit Width { get; }


        public void PushComponentLayout(PDFLayoutContext context, int pageIndex, PDFUnit xoffset, PDFUnit yoffset)
        {
            PDFComponent comp = this.Owner as PDFComponent;
            if (null != comp)
                comp.SetPageLayoutIndex(pageIndex);

            this.DoPushComponentLayout(context, pageIndex, xoffset, yoffset);
        }

        /// <summary>
        /// Updates position data and layout after everything has been placed
        /// </summary>
        /// <param name="context">The current layout context</param>
        /// <param name="xoffset">Any horizontal adjustment to be applied when the component layout is set</param>
        /// <param name="yoffset">Any vertical adjustment to be applied when the component layout is set</param>
        protected abstract void DoPushComponentLayout(PDFLayoutContext context, int pageIndex, PDFUnit xoffset, PDFUnit yoffset);

        //
        // Methods
        // 

        #region public virtual bool MoveToNextRegion()

        /// <summary>
        /// Checks if this layout item supports splitting of components across pages
        /// </summary>
        /// <returns></returns>
        public virtual bool MoveToNextRegion()
        {
            return false;
        }

        #endregion

        
        #region public void Close()

        /// <summary>
        /// Closes the Layout Item.
        /// </summary>
        /// <returns>True if the item was successfully closed.</returns>
        /// <remarks>Internally this calls the virtual DoClose method.
        /// Inheritors have an opportunity to perform some cleanup. 
        /// If the item should not be closed then inheritors can return false 
        /// from the DoClose method and an exception will be raised.</remarks>
        public bool Close()
        {
            string msg = "";

            //Only perform the actual closing if we are not already closed
            if (this.IsClosed == false)
            {
                bool perform = this.DoClose(ref msg);
                PDFTraceLog log = PDFTraceContext.GetLog();
                if (perform)
                {
                    if(this.Parent is PDFLayoutRegion)
                        (this.Parent as PDFLayoutRegion).AddToSize(this);

                    if (log.ShouldLog(TraceLevel.Debug))
                        log.Add(TraceLevel.Debug, PDFLayoutItem.LOG_CATEGORY, "Marking layout item " + this.ToString() + " as closed");
                    this.IsClosed = true;
                }
                else
                {
                    log.Add(TraceLevel.Error, PDFLayoutItem.LOG_CATEGORY, "Could not close layout " + this.ToString() + ". Its an error to attempt to close an item that is already closed");
                    if (null == msg)
                        msg = String.Empty;
                    throw new InvalidOperationException(Errors.LayoutItemCouldNotBeClosed + msg);
                }
                return perform;
            }
            else
                return true;
        }

        #endregion 

        #region protected virtual bool DoClose(ref string msg)

        /// <summary>
        /// Base implementation simply returns true. 
        /// Inheritors should override this method to perform their own clean up.
        /// If the item should not be closed then return false.
        /// </summary>
        /// <param name="msg">Set this to any message to pass on why the item could not be closed.</param>
        /// <returns>True unless a sub-class decides that it should not be closed and returns an alternative value.</returns>
        protected virtual bool DoClose(ref string msg)
        {
            return true;
        }

        #endregion

        #region protected void AssertIsOpen()

        /// <summary>
        /// Checks to see if this item is open. If not then it raises an exception.
        /// </summary>
        protected void AssertIsOpen()
        {
            if (this.IsClosed)
                throw new InvalidOperationException(Errors.LayoutItemIsClosed);
        }

        #endregion

        #region internal void SetParent(PDFLayoutItem item)

        /// <summary>
        /// Internal method that changes the parent of this item - cannot be null
        /// </summary>
        /// <param name="item"></param>
        internal void SetParent(PDFLayoutItem item)
        {
            if (null == item)
                throw new ArgumentNullException("item");

            this.Parent = item;
        }

        #endregion

        #region protected int GetPageIndex()

        /// <summary>
        /// Gets the index of the page this item is positioned on
        /// </summary>
        /// <returns></returns>
        protected int GetPageIndex()
        {
            PDFLayoutItem item = this;
            while (null != item)
            {
                if (item is PDFLayoutPage)
                    return (item as PDFLayoutPage).PageIndex;
                else
                    item = item.Parent;
            }
            return NULL_PAGEINDEX;
        }

        #endregion

        #region protected int GetColumnIndex()

        /// <summary>
        /// Gets the index of the column this item is positioned on
        /// </summary>
        /// <returns></returns>
        protected int GetColumnIndex()
        {
            PDFLayoutItem item = this;
            while (null != item)
            {
                if (item is PDFLayoutRegion)
                    return (item as PDFLayoutRegion).ColumnIndex;
                else if (item is PDFLayoutBlock)
                    return (item as PDFLayoutBlock).CurrentRegion.ColumnIndex;
                else
                    item = item.Parent;
            }
            return NULL_COLUMNINDEX;
        }

        #endregion

        #region public virtual void SetMaxWidth(PDFUnit width)

        /// <summary>
        /// Adjusts the width of this component so that it does not extend beyond the boundaries of its container.
        /// The base implementation does nothing. Inheritors should override this to perform their own re-sizing
        /// </summary>
        /// <param name="width">The maximum width allowed</param>
        public virtual void SetMaxWidth(PDFUnit width)
        {
            
        }

        #endregion

        public virtual void ResetAvailableHeight(PDFUnit height, bool includeChildren)
        {
        }

        #region public virtual PDFObjectRef OutputToPDF(PDFRenderContext context, PDFWriter writer)

        /// <summary>
        /// Outputs the contents of this item to the writer
        /// </summary>
        /// <param name="context"></param>
        /// <param name="writer"></param>
        /// <returns></returns>
        public PDFObjectRef OutputToPDF(PDFRenderContext context, PDFWriter writer)
        {
            return this.DoOutputToPDF(context, writer);
        }

        protected virtual PDFObjectRef DoOutputToPDF(PDFRenderContext context, PDFWriter writer)
        {
            return null;
        }

        #endregion

        #region protected void OutputBackground(PDFBackgroundStyle bg, PDFBorderStyle border, PDFRenderContext context, PDFRect rect)

        /// <summary>
        /// Outputs the background for this block in the specified rect with the background style.
        /// </summary>
        /// <param name="bg">The backgrond style</param>
        /// <param name="border">The border style (which indicates the corner radius)</param>
        /// <param name="context">The current render context</param>
        /// <param name="rect">The rectangle to be output</param>
        protected virtual void OutputBackground(PDFBackgroundStyle bg, PDFBorderStyle border, PDFRenderContext context, PDFRect rect)
        {
            PDFGraphics g = context.Graphics;
            PDFBrush brush = bg.CreateBrush();
            if (null != brush)
            {
                if (null != border && border.CornerRadius != PDFUnit.Zero)
                    g.FillRoundRectangle(brush, rect, border.CornerRadius);
                else
                    g.FillRectangle(bg.CreateBrush(), rect);
            }
        }

        #endregion

        #region protected void OutputBorder(PDFBackgroundStyle bg, PDFBorderStyle border, PDFRenderContext context, PDFRect rect)

        /// <summary>
        /// Outputs the border for this block in the specified rect with the border style
        /// </summary>
        /// <param name="bg">The background style</param>
        /// <param name="border">The border style</param>
        /// <param name="context">The current render context</param>
        /// <param name="rect">The rectangle that shoud be rendered as the border</param>
        protected virtual void OutputBorder(PDFBackgroundStyle bg, PDFBorderStyle border, PDFRenderContext context, PDFRect rect)
        {
            PDFGraphics g = context.Graphics;
            PDFPen pen = border.CreatePen();

            if (null != pen)
            {
                if (border.CornerRadius != PDFUnit.Zero)
                    g.DrawRoundRectangle(pen, rect, border.Sides, border.CornerRadius);
                else
                    g.DrawRectangle(pen, rect, border.Sides);
            }
        }

        #endregion


    }


    /// <summary>
    /// A collection of layout items
    /// </summary>
    public class PDFLayoutItemCollection : List<PDFLayoutItem>
    {
    }
}
