﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Scryber.Drawing;

namespace Scryber.Components.Layout
{

    /// <summary>
    /// A run in a line that contains textual content
    /// </summary>
    public class PDFLayoutTextRun : PDFLayoutRun
    {
        #region public Text.TextSpan Span { get; set; }

        /// <summary>
        /// Gets or sets the Text Span for this text run
        /// </summary>
        public Text.TextSpan Span { get; set; }

        #endregion

        #region public PDFUnit OffsetY { get; set; }

        /// <summary>
        /// Gets or sets the Y offset of this run 
        /// </summary>
        public PDFUnit OffsetY { get; set; }

        #endregion

        #region public PDFSize MeasuredSize { get; set; }

        /// <summary>
        /// Gets or sets the measured size of the text in this run with the correct font
        /// </summary>
        public PDFSize MeasuredSize { get; set; }

        #endregion

        #region public Text.TextParagraph StartParagraph {get; set;}

        /// <summary>
        /// Gets or sets the starting paragraph this layout run is associated with
        /// </summary>
        /// <remarks>If set then this line marks the first line of a new paragraph in a text block</remarks>
        public Text.TextParagraph StartParagraph
        {
            get;
            set;
        }

        #endregion

        #region public Text.TextLine StartLine {get; set;}

        /// <summary>
        /// Gets or sets the starting text line this layout run is associated with
        /// </summary>
        /// <remarks>If set this line marks the start of a new text line in a block</remarks>
        public Text.TextLine StartLine
        {
            get;
            set;
        }

        #endregion

        #region public Text.TextBlock TextBlock { get; set; }

        /// <summary>
        /// Gets or sets the TextBlock this Run initiates
        /// </summary>
        public Text.TextBlock TextBlock { get; set; }

        #endregion

        #region public PDFTextRenderOptions TextOptions { get; set; }

        /// <summary>
        /// Gets or sets the text rendering options for this run
        /// </summary>
        public PDFTextRenderOptions TextOptions { get; set; }

        #endregion

        #region public override PDFUnit Height {get;}

        /// <summary>
        /// Returns the Height of this run (based on the measured size)
        /// </summary>
        public override PDFUnit Height
        {
            get { return this.MeasuredSize.Height; }
        }

        #endregion

        #region public override PDFUnit Width {get;}

        /// <summary>
        /// Returns the Width of this run (based on the measured size)
        /// </summary>
        public override PDFUnit Width
        {
            get { return this.MeasuredSize.Width; }
        }

        #endregion

        //
        // ctor
        //

        #region public PDFLayoutTextRun(Text.TextSpan span, PDFSize size, PDFLayoutLine line): base(line)

        public PDFLayoutTextRun(Text.TextSpan span, IPDFComponent owner, PDFSize size, PDFLayoutLine line)
            : base(line, owner)
        {
            this.Span = span;
            this.MeasuredSize = size;

        }

        #endregion

        //
        // methods
        //

        #region public override void SetOffsetY(PDFUnit y)

        /// <summary>
        /// Adjusts the vertical offset of this run wrt the line
        /// </summary>
        /// <param name="y"></param>
        public override void SetOffsetY(PDFUnit y)
        {
            this.OffsetY = y;
        }

        #endregion

        #region protected override void DoPushComponentLayout(PDFLayoutContext context, int pageIndex, PDFUnit xoffset, PDFUnit yoffset)

        /// <summary>
        /// Pushes the layout dimensions onto the component. In this case if we are the starting paragraph - set up the YOffset, and if we are the starting line set up the line inset
        /// </summary>
        /// <param name="context"></param>
        /// <param name="xoffset"></param>
        /// <param name="yoffset"></param>
        protected override void DoPushComponentLayout(PDFLayoutContext context, int pageIndex, PDFUnit xoffset, PDFUnit yoffset)
        {
            if (null != this.TextBlock)
            {
                this.TextBlock.YOffset += yoffset.PointsValue;
            }

            if (null != this.StartLine)
            {
                this.StartLine.LineInset += xoffset.PointsValue;
                if (context.TraceLog.ShouldLog(TraceLevel.Debug))
                    context.TraceLog.Add(TraceLevel.Debug, PDFLayoutItem.LOG_CATEGORY, "Modified the layout text run " + this.ToString() + " and updated the line x offset by " + xoffset.ToString() + " to " + this.StartLine.LineInset.ToString());
            }
        }

        #endregion

        #region public override string ToString()

        /// <summary>
        /// Returns a string representation of this text run
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            if (null == this.Span)
                return "Empty text run";
            else
            {
                PDFRect rect = new PDFRect(this.StartLine == null ? PDFUnit.Zero : this.StartLine.LineInset, this.StartParagraph == null ? PDFUnit.Zero : this.StartParagraph.YOffset, this.Width, this.Height);
                return "Layout text run for '" + this.Span.ToString() + "' (bounds:" + rect.ToString() + ")";
            }
        }

        #endregion

        #region public override Native.PDFObjectRef DoOutputToPDF(PDFRenderContext context, PDFWriter writer)

        /// <summary>
        /// Overrides the default implementation to output any text block associated with this run
        /// </summary>
        /// <param name="context"></param>
        /// <param name="writer"></param>
        /// <returns></returns>
        protected override Native.PDFObjectRef DoOutputToPDF(PDFRenderContext context, PDFWriter writer)
        {
            //We only need to render the text block that is built.
            //If we don't have a block there is nothing to do.
            if (null != this.TextBlock)
            {
                PDFComponent owner = this.Owner as PDFComponent;
                if (null != owner)
                {
                    Text.TextBlock block = this.TextBlock;
                    PDFRect bounds = new PDFRect(block.XOffset, block.YOffset, block.Width, block.Height);
                    bounds.X += context.Offset.X;
                    bounds.Y += context.Offset.Y;

                    owner.SetArrangement(context, context.FullStyle, bounds);
                }
                if (context.TraceLog.ShouldLog(TraceLevel.Debug))
                    context.TraceLog.Add(TraceLevel.Debug, PDFLayoutItem.LOG_CATEGORY, "Rendering text '" + this.TextBlock.ToString(100) + "' at offset " + context.Offset.ToString() + " with block offset of (" + TextBlock.XOffset + "," + TextBlock.YOffset + ")");

                
                context.Graphics.FillText(this.TextBlock, this.TextOptions, context.Offset);
            }
            return base.DoOutputToPDF(context, writer);
        }

        #endregion
    }
}
