﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Scryber.Drawing;
using Scryber.Styles;

namespace Scryber.Components.Layout
{
    public class PDFTableLayoutEngineOverflow : PDFBlockLayoutEngineBase
    {
        private const string TableEngineLogCategory = "Table Layout Engine";
        private const TraceLevel TableEngineLogLevel = TraceLevel.Debug;

        private PDFTableGrid _tbl;
        private CellDimension[] _widths;
        private CellDimension[] _heights;
        private TableReference _tblRef;
        private PDFLayoutBlock _rowblock;
        private int _rowIndex = -1;

        PDFUnit _rowOffset = PDFUnit.Zero;

        protected PDFTableGrid Table
        {
            get { return _tbl; }
        }


        private TableReference AllCells
        {
            get { return _tblRef; }
        }

        public PDFTableLayoutEngineOverflow(PDFTableGrid table, IPDFLayoutEngine parent)
            : base(table, parent)
        {
            this._tbl = table;
        }

        protected override void DoLayoutComponent()
        {
            try
            {
                if (this.TraceLog.ShouldLog(TableEngineLogLevel))
                    this.TraceLog.Begin(TableEngineLogLevel, TableEngineLogCategory, string.Format("Starting the layout of the table", this.Table.ID));

                this.ContinueLayout = true;
                this.CurrentBlock = this.Context.DocumentLayout.CurrentPage.LastOpenBlock();
                if (this.CurrentBlock.CurrentRegion != null && this.CurrentBlock.CurrentRegion.HasOpenItem)
                    this.CurrentBlock.CurrentRegion.CloseCurrentItem();

                PDFPositionOptions tablepos = this.FullStyle.CreatePostionOptions();

                int rows, columns;
                this.CalcGridCounts(out rows, out columns);

                PDFLayoutBlock tableBlock = this.CurrentBlock.BeginNewContainerBlock(this.Table, this, this.FullStyle, tablepos.PositionMode);

                PDFRect available = this.CalculateTableSpace(tableBlock, tablepos);
                available.X = 0;
                
                this.BuildReferenceGrid(rows, columns, tablepos, available, tableBlock);

                this.CalcExplicitSizes(rows, columns);
                this.CalculateUnassignedWidths();
                this.DoLayoutTableRows();
                this.PushConsistentCellWidths();


                if (this.TraceLog.ShouldLog(TableEngineLogLevel))
                    this.TraceLog.End(TableEngineLogLevel, TableEngineLogCategory, string.Format("Completed the layout of the table", this.Table.ID));
            }
            catch (PDFLayoutException)
            {
                throw;
            }
            catch (Exception ex)
            {
                throw new PDFLayoutException(string.Format(Errors.CouldNotLayoutGridCells, this.Table.ID, ex.Message), ex);
            }
        }

        //
        // actual table layout
        //

        #region private void DoLayoutTableRows()

        /// <summary>
        /// For each of the rows in this engines table - starts it's layout
        /// </summary>
        private void DoLayoutTableRows()
        {
            for (int i = 0; i < this.Table.Rows.Count; i++)
            {
                _rowIndex = i;
                PDFTableRow row = this.Table.Rows[i];
                this.DoLayoutTableRow(row, i);

                //check that we should keep going, or we have just run out of space
                if (!this.ContinueLayout)
                    return;
            }
        }

        #endregion

        #region private PDFUnit DoLayoutTableRow(PDFTableRow row, int index)

        /// <summary>
        /// Lays out an individual table row
        /// </summary>
        /// <param name="row">The row to lay out</param>
        /// <param name="index">The index of the row in the table</param>
        /// <returns></returns>
        private PDFUnit DoLayoutTableRow(PDFTableRow row, int index)
        {
            PDFLayoutBlock block = this.AllCells.CurrentGrid.TableBlock;
            //Check we can continue
            if (block.IsClosed)
                throw new InvalidOperationException("Table block already closed, has the cell overflowed");


            RowReference rowRef = _tblRef.AllRows[index];
            PDFStyle rowStyle = rowRef.FullStyle;
            this.Context.StyleStack.Push(rowRef.AppliedStyle);

            PDFArtefactRegistrationSet artefacts = row.RegisterLayoutArtefacts(this.Context, rowStyle);


            //create the table row block and region

            PDFPositionOptions rowpos = rowStyle.CreatePostionOptions();

            this._rowblock = block.BeginNewContainerBlock(row, this, rowStyle, rowpos.PositionMode);
            rowRef.Block = this._rowblock;

            //Set the height to the page height and then we can move it after if it's too big for the height, unless it's explicit.
            PDFUnit pageHeight = this.Context.DocumentLayout.CurrentPage.Height;
            PDFUnit h = pageHeight;// block.AvailableBounds.Height;

            if (_heights[index].Explicit)
                h = _heights[index].Size;
            PDFUnit w = block.AvailableBounds.Width;
            PDFUnit y = _rowOffset;

            PDFRect totalbounds = new PDFRect(PDFUnit.Zero, y, w, h);

            bool autoflow = false;
            this._rowblock.InitRegions(totalbounds, rowpos, this.AllCells.TotalColumnCount, PDFUnit.Zero, autoflow);



            //Layout the inner cells
            PDFUnit rowHeight = DoLayoutRowCells(row, index);

            //Refresh the reference to the current block (just in case)
            block = this.AllCells.CurrentGrid.TableBlock;
            block.CurrentRegion.CloseCurrentItem();

            //Have we extended beyond the boundaries available
            if (rowHeight > this.AllCells.CurrentGrid.AvailableSpace.Height - _rowOffset)
            {
                PDFLayoutBlock origparent = this.CurrentBlock;
                PDFLayoutRegion origregion = origparent.CurrentRegion;
                PDFLayoutBlock origtable = block;

                if (this.StartNewTableInAnotherRegion(this._rowblock, index))
                {
                    origregion.AddToSize(origtable);
                    _rowOffset = this._rowblock.Height;
                }
                else
                {
                    this.AllCells.SetRowCount(index, true);
                    this.ContinueLayout = false;
                }
            }
            else
            {
                //block.CurrentRegion.AddToSize(this._rowblock);
                _rowOffset += rowHeight;
            }


            this._rowblock = null;

            //complete the row layout with registration, closing the artefacts and poping the style stack

            if (this.ContinueLayout)
                RegisterChildLayout(row);

            if (null != artefacts)
                row.CloseLayoutArtefacts(this.Context, artefacts, rowStyle);


            this.Context.StyleStack.Pop();

            return rowHeight;
        }

        #endregion

        #region private void DoLayoutRowCells(PDFTableRow row, int rowindex)

        /// <summary>
        /// Lays out the cells in a single row
        /// </summary>
        /// <param name="row">The cells to layout</param>
        /// <param name="rowindex">The index of the row we are laying out</param>
        /// <returns>The maximum height of the row</returns>
        private PDFUnit DoLayoutRowCells(PDFTableRow row, int rowindex)
        {
            PDFUnit offsetX = PDFUnit.Zero;

            for (int cellindex = 0; cellindex < this.AllCells.TotalColumnCount; cellindex++)
            {
                PDFLayoutRegion cellRegion = this._rowblock.CurrentRegion;
                _lastRowWasMovedToANewTable = false;
                //set the total bounds for the region, and make the height the 
                PDFRect total = new PDFRect(offsetX, cellRegion.TotalBounds.Y, _widths[cellindex].Size, cellRegion.TotalBounds.Height);
                cellRegion.TotalBounds = total;
                this._rowblock.CurrentRegion.SetMaxWidth(_widths[cellindex].Size);

                if (cellindex < row.Cells.Count)
                {
                    this.DoLayoutARowCell(row.Cells[cellindex], cellindex, rowindex);
                }
                if (_lastRowWasMovedToANewTable)
                {
                    GridReference gref = AllCells.CurrentGrid;
                    RowReference currrow = gref[0];
                    PDFLayoutBlock rowBlock = gref.TableBlock.LastOpenBlock();
                    currrow.Block = rowBlock;
                    this._rowblock = rowBlock;
                }
                
                offsetX += cellRegion.TotalBounds.Width;

                bool forced = true;
                this._rowblock.MoveToNextRegion(forced);
            }

            PDFUnit maxh = this._tblRef.GetMaxCellHeightForRow(rowindex);
            this._tblRef.SetCellHeightForRow(rowindex, maxh);

            return maxh;
        }

        #endregion

        #region protected void DoLayoutARowCell(PDFTableCell cell, int cellindex, int rowindex)

        /// <summary>
        /// Sets up the cell for laying out, with the styles and artefacts before calling 
        /// DoLayoutAcell - that lays out the content of the cell.
        /// </summary>
        /// <param name="cell">The cell to layout the content for</param>
        /// <param name="colindex">The column index of the cell in the row</param>
        /// <param name="rowindex">the index of the row the cell is on.</param>
        protected void DoLayoutARowCell(PDFTableCell cell, int colindex, int rowindex)
        {
            CellReference cref = _tblRef.AllCells[rowindex, colindex];

            Scryber.Styles.PDFStyle applied = cref.AppliedStyle;
            if (null != applied)
                this.StyleStack.Push(applied);

            Scryber.Styles.PDFStyle full = cref.FullStyle;

            PDFArtefactRegistrationSet artefacts = cell.RegisterLayoutArtefacts(this.Context, full);

            //Call the base implementation to actually layout the cell.
            this.DoLayoutACell(cell, full, colindex, rowindex);

            if (null != applied)
                this.StyleStack.Pop();

            //We want to add the child only if it should be rendered and we are a full block
            if (this.ContinueLayout)
                RegisterChildLayout(cell);

            if (null != artefacts)
                cell.CloseLayoutArtefacts(this.Context, artefacts, full);


        }

        #endregion

        #region private void DoLayoutACell(PDFTableCell cell, Styles.PDFStyle full, int colindex, int rowindex)

        /// <summary>
        /// Performs the actual content layout of the cell
        /// </summary>
        /// <param name="cell">The cell to layout</param>
        /// <param name="full">The full style of the cell</param>
        /// <param name="colindex">The index of the column the cell is in</param>
        /// <param name="rowindex">The index of the row the cell is in</param>
        private void DoLayoutACell(PDFTableCell cell, Styles.PDFStyle full, int colindex, int rowindex)
        {
            try
            {
                using (IPDFLayoutEngine engine = cell.GetEngine(this, this.Context, full))
                {
                    engine.Layout(this.Context, full);
                }

                PDFLayoutRegion rowregion = this._rowblock.CurrentRegion;

                if (rowregion.Contents.Count > 0)
                {
                    PDFLayoutItem last = rowregion.Contents[rowregion.Contents.Count - 1];
                    if (last is PDFLayoutBlock)
                        this._tblRef.AllCells[rowindex, colindex].Block = (PDFLayoutBlock)last;
                    else if (this.Context.Conformance == ConformanceMode.Strict)
                        throw new PDFLayoutException("The last item in the region was not a block. A block should be the last item as we have just laid out a cell");
                    else
                        this.TraceLog.Add(TraceLevel.Failure, TableEngineLogCategory, "The last item in the region was not a block. A block should be the last item as we have just laid out a cell");
                }
            }
            catch (PDFLayoutException)
            {
                throw;
            }
            catch (Exception ex)
            {
                string msg = String.Format(Errors.LayoutFailedForComponent, cell.ID, ex.Message);
                throw new PDFLayoutException(msg, ex);
            }
        }


        #endregion

        //
        // table calculations
        //

        #region private void CalculateTableSpace()

        /// <summary>
        /// Calculates the available space in the current block to 
        /// fit the table, and initializes it's region to this.
        /// </summary>
        private PDFRect CalculateTableSpace(PDFLayoutBlock tableblock, PDFPositionOptions tablepos)
        {
            PDFRect space = this.CurrentBlock.CurrentRegion.TotalBounds;
            space.Y = this.CurrentBlock.CurrentRegion.Height;
            space.X = 0;
            space.Height -= space.Y;


            if (tablepos.HasWidth)
                space.Width = tablepos.Width;

            if (tablepos.HasHeight)
                space.Height = tablepos.Height;

            if (tablepos.Margins.IsEmpty == false)
            {
                //We only account for the margins if we don't have an explicit height
                if (tablepos.HasWidth == false)
                    space.Width -= tablepos.Margins.Left + tablepos.Margins.Right;

                if (tablepos.HasHeight == false)
                    space.Height -= tablepos.Margins.Top + tablepos.Margins.Bottom;
            }

            tableblock.InitRegions(space, tablepos, 1, PDFUnit.Zero);

            if (tablepos.Padding.IsEmpty == false)
            {
                //Don't alter the position for the padding
                space.Width -= tablepos.Padding.Left + tablepos.Padding.Right;
                space.Height -= tablepos.Padding.Top + tablepos.Padding.Bottom;
            }

            if (this.TraceLog.ShouldLog(TraceLevel.Debug))
                this.TraceLog.Add(TraceLevel.Debug, TableEngineLogCategory, string.Format("Table grid available space is {0}", space.ToString()));


            return space;
        }

        #endregion

        #region protected virtual void PushConsistentSizes()

        /// <summary>
        /// Makes sure all the cells in the same column have the same (maximum) width across all rows.
        /// </summary>
        protected virtual void PushConsistentCellWidths()
        {
            if (TraceLog.ShouldLog(TableEngineLogLevel))
                this.TraceLog.Add(TableEngineLogLevel, TableEngineLogCategory, "Starting to push the component sizes");


            //Set the width of the cells in each column
            PDFUnit xOffset = PDFUnit.Zero;
            for (int colIndex = 0; colIndex < this.AllCells.TotalColumnCount; colIndex++)
            {
                PDFUnit maxColWidth = _tblRef.GetMaxCellWidthForColumn(colIndex);
                if (_widths[colIndex].Explicit)
                    maxColWidth = _widths[colIndex].Size;
                else if (this.AllCells.CurrentGrid.Position.FillWidth)
                    maxColWidth = PDFUnit.Max(maxColWidth, _widths[colIndex].Size);

                if (this.TraceLog.ShouldLog(TableEngineLogLevel))
                    this.TraceLog.Add(TableEngineLogLevel, TableEngineLogCategory, "Max column width for #" + colIndex + " is " + maxColWidth);


                _tblRef.SetCellWidthForColumn(colIndex, maxColWidth);
                _tblRef.SetRowOffsetAndWidthForColumn(colIndex, xOffset, maxColWidth);

                xOffset += maxColWidth;
            }

            if (this.TraceLog.ShouldLog(TableEngineLogLevel))
                this.TraceLog.Add(TableEngineLogLevel, TableEngineLogCategory, "All cells set to the maximum width of their column.");


            int index = 0;
            int last = this.AllCells.AllGrids.Count - 1;

            foreach (GridReference grid in this.AllCells.AllGrids)
            {
                if (grid.StartRowIndex < 0)
                    //this is a grid that has no rows - automatically flowed onto a new nape
                    continue;

                PDFLayoutBlock tableblock = grid.TableBlock;
                //Assumption: The table block contains only one region that contains all the row blocks
                PDFLayoutRegion tableregion = tableblock.Columns[0];

                for (int rowindex = grid.StartRowIndex; rowindex < grid.EndRowIndex + 1; rowindex++)
                {
                    RowReference rowref = this.AllCells.AllRows[rowindex];
                    PDFLayoutBlock row = rowref.Block;

                    PDFRect rowbounds = row.TotalBounds;
                    rowbounds.Width = xOffset; // +row.Position.Padding.Left + row.Position.Padding.Right + row.Position.Margins.Left + row.Position.Margins.Right;
                    row.TotalBounds = rowbounds;
                    tableregion.AddToSize(row);

                    if (this.TraceLog.ShouldLog(TableEngineLogLevel))
                        this.TraceLog.Add(TableEngineLogLevel, TableEngineLogCategory, "Added row index " + rowindex + " to the table grid region " + tableregion.ToString());

                }

                if (!tableblock.IsClosed)
                    tableblock.Close();
                tableblock.SetContentSize(xOffset, tableregion.Height);

                //Should only add the size of the last grid to the current region.
                if (index == last)
                {
                    PDFLayoutBlock parent = tableblock.Parent as PDFLayoutBlock;
                    parent.CurrentRegion.AddToSize(tableblock);
                }

                index++;

            }
            if (this.TraceLog.ShouldLog(TableEngineLogLevel))
                this.TraceLog.Add(TableEngineLogLevel, TableEngineLogCategory, "All row sizes added to the final table height");
        }

        #endregion

        #region protected virtual void CalcGridCounts()

        /// <summary>
        /// Calculates the total number of columns and rows in the current table.
        /// Stored in _rowcount, and _columncount.
        /// </summary>
        private void CalcGridCounts(out int rowCount, out int columnCount)
        {
            int column = 0;
            int rows = this.Table.Rows.Count;

            foreach (PDFTableRow row in this.Table.Rows)
            {
                int maxRowColumns = row.Cells.Count;
                foreach (PDFTableCell cell in row.Cells)
                {
                    //TODO:Use Cell full style for getting the column count.
                    if (cell.ColumnSpan <= 0)
                        maxRowColumns++;
                    else
                        maxRowColumns += cell.ColumnSpan;
                }
                if (maxRowColumns > column)
                    column = maxRowColumns;
            }
            rowCount = rows;
            columnCount = column;

            if (this.TraceLog.ShouldLog(TableEngineLogLevel))
                this.TraceLog.Add(TableEngineLogLevel, TableEngineLogCategory, string.Format("Table grid calculated at {0} rows by {1} columns", rowCount, columnCount));

        }

        #endregion

        #region private void BuildReferenceGrid()

        /// <summary>
        /// Builds a complete grid of references to all rows and cells within the table inc stlyes
        /// </summary>
        private void BuildReferenceGrid(int rowcount, int columncount, PDFPositionOptions tablepos, PDFRect availableSpace, PDFLayoutBlock tableblock)
        {
            TableReference tbl = new TableReference(rowcount, columncount, this.FullStyle, tablepos);
            GridReference grid = tbl.CurrentGrid;

            grid.AvailableSpace = availableSpace;
            grid.TableBlock = tableblock;

            for (int rowIndex = 0; rowIndex < tbl.TotalRowCount; rowIndex++)
            {
                PDFTableRow row = this.Table.Rows[rowIndex];
                PDFStyle applied = row.GetAppliedStyle();
                this.StyleStack.Push(applied);
                PDFStyle full = this.StyleStack.GetFullStyle(row);
                RowReference rref = tbl.AddRowReference(row, grid, applied, full, rowIndex);

                this.BuildReferenceCells(tbl, rref, rowIndex);

                this.StyleStack.Pop();
            }

            if (this.TraceLog.ShouldLog(TableEngineLogLevel))
                this.TraceLog.Add(TableEngineLogLevel, TableEngineLogCategory, "Table grid references generated, inc. all styles");

            this._tblRef = tbl;
        }

        private void BuildReferenceCells(TableReference tbl, RowReference rref, int rowIndex)
        {
            for (int column = 0; column < tbl.TotalColumnCount; column++)
            {
                if (column < rref.Row.Cells.Count)
                {
                    PDFTableCell cell = rref.Row.Cells[column];
                    PDFStyle applied = cell.GetAppliedStyle();

                    //If we have a table that is full width we need the cells to be full width too.
                    if (tbl.CurrentGrid.Position.FillWidth)
                        applied.Position.FullWidth = true;
                    this.StyleStack.Push(applied);
                    PDFStyle full = this.StyleStack.GetFullStyle(cell);
                    tbl.AddCellReference(cell, rref, applied, full, column);
                    this.StyleStack.Pop();
                }
                else
                {
                    ;
                }
            }
        }

        #endregion

        #region private void CalcExplicitSizes(int rowcount, int columncount)

        /// <summary>
        /// Calculates the explicit widths and heights of each of the rows and columns.
        /// Stored in the instance variables _columnWidths, _rowHeights
        /// </summary>
        private void CalcExplicitSizes(int rowcount, int columncount)
        {
            CellDimension[] widths = new CellDimension[columncount];
            CellDimension[] heights = new CellDimension[rowcount];
            int explicitCount = 0;
            for (int v = 0; v < rowcount; v++)
            {
                for (int h = 0; h < columncount; h++)
                {
                    CellReference cref = _tblRef.AllCells[v, h];
                    if (null != cref)
                    {

                        if (cref.HasExplicitWidth)
                        {
                            widths[h].Size = PDFUnit.Max(widths[h].Size, cref.TotalWidth);
                            widths[h].Explicit = true;
                            explicitCount++;
                        }

                        if (cref.HasExplicitHeight)
                        {
                            heights[v].Size = PDFUnit.Max(heights[v].Size, cref.TotalHeight);
                            heights[v].Explicit = true;
                            explicitCount++;
                        }
                    }
                }
                RowReference rowref = _tblRef.AllRows[v];
                if (rowref.HasExplicitHeight)
                    heights[v].Size = PDFUnit.Max(heights[v].Size, rowref.ExplicitHeight);
            }

            _widths = widths;
            _heights = heights;

            if (this.TraceLog.ShouldLog(TableEngineLogLevel))
                this.TraceLog.Add(TableEngineLogLevel, TableEngineLogCategory, "All explicit cell sizes accounted for");
        }

        #endregion

        #region  private void CalculateUnassignedWidths()

        /// <summary>
        /// Based on the available space left after each explicit width. 
        /// This calculates the remaining available for each column and splits it up between them.  
        /// </summary>
        private void CalculateUnassignedWidths()
        {

            PDFUnit tablewidth;
            PDFUnit totalexplicit = PDFUnit.Zero;

            tablewidth = this.AllCells.CurrentGrid.AvailableSpace.Width;

            int nonExplicitCount = 0;
            for (int i = 0; i < this.AllCells.TotalColumnCount; i++)
            {
                CellDimension w = _widths[i];
                if (w.Explicit == false)
                {
                    nonExplicitCount++;
                }
                else
                {
                    tablewidth -= w.Size;
                    totalexplicit += w.Size;
                }
            }

            if (this.TraceLog.ShouldLog(TableEngineLogLevel))
                this.TraceLog.Add(TableEngineLogLevel, TableEngineLogCategory, "Available width for non explicit sizes is " + tablewidth.ToString());

            //make sure we actually have some widths
            if (nonExplicitCount > 0)
            {
                PDFUnit unassignedwidth = tablewidth / nonExplicitCount;
                for (int i = 0; i < _widths.Length; i++)
                {
                    if (_widths[i].Explicit == false)
                        _widths[i].Size = unassignedwidth;
                }

                if (this.TraceLog.ShouldLog(TableEngineLogLevel))
                    this.TraceLog.Add(TableEngineLogLevel, TableEngineLogCategory, "Avialable width split between " + nonExplicitCount + " columns, with each given " + unassignedwidth);
            }


        }

        #endregion

        //
        // overflow methods
        //

        #region protected virtual void CloseTableAndReleaseRowsFrom(int rowindex)

        /// <summary>
        /// Closes the tabe reference and removes all rows after and including the specified index
        /// </summary>
        /// <param name="rowindex"></param>
        protected virtual void CloseTableAndReleaseRowsFrom(int rowindex)
        {
            this._tblRef.SetRowCount(rowindex, true);
        }

        #endregion

        #region protected virtual bool StartNewTableInAnotherRegion(PDFLayoutBlock row, int rowindex)

        /// <summary>
        /// Called when we have run out of available space in the current region to contine 
        /// </summary>
        /// <param name="row"></param>
        /// <param name="rowindex"></param>
        /// <returns></returns>
        protected virtual bool StartNewTableInAnotherRegion(PDFLayoutBlock row, int rowindex)
        {
            GridReference origGrid = this.AllCells.CurrentGrid;
            origGrid.EndRowIndex = rowindex - 1;
            PDFSize origGridSize = origGrid.CalculateContentSize();

            PDFLayoutBlock origTblblock = origGrid.TableBlock;
            if (origTblblock.IsClosed == false)
                origTblblock.Close();
            origTblblock.SetContentSize(origGridSize.Width, origGridSize.Height);


            PDFLayoutRegion origTblRegion = origTblblock.CurrentRegion;
            PDFLayoutRegion tblparent = (origTblblock.Parent as PDFLayoutBlock).CurrentRegion;
            tblparent.AddToSize(origTblblock);
            this._rowIndex = rowindex;
            PDFLayoutRegion region = row.CurrentRegion;
            bool newpage;

            bool started = this.MoveToNextRegion(ref region, ref row, out newpage);
            if (started)
            {
                PDFLayoutBlock newTableBlock = this.AllCells.CurrentGrid.TableBlock;
                PDFLayoutRegion newTableRegion = newTableBlock.CurrentRegion;
                newTableRegion.UsedSize = PDFSize.Empty;

                if (newpage)
                    row.Offset(0, 0 - row.Height);
                
            }



            return started;
        }

        #endregion

        #region public override PDFLayoutBlock CloseCurrentBlockAndStartNewInRegion(PDFLayoutBlock blockToClose, PDFLayoutRegion joinToRegion)


        bool _lastRowWasMovedToANewTable = false;

        /// <summary>
        /// Overrides the default behaviour to beging a new GridReference to hold the next rows.
        /// </summary>
        /// <param name="blockToClose"></param>
        /// <param name="joinToRegion"></param>
        /// <returns></returns>
        public override PDFLayoutBlock CloseCurrentBlockAndStartNewInRegion(PDFLayoutBlock blockToClose, PDFLayoutRegion joinToRegion)
        {

            PDFLayoutBlock orig = this.CurrentBlock;
            GridReference curr = this.AllCells.CurrentGrid;
            PDFRect avail = curr.AvailableSpace;
            avail.Height = joinToRegion.AvailableHeight;

            //TODO: Use this to caclulate the space
            PDFLayoutBlock newTable = base.CloseCurrentBlockAndStartNewInRegion(blockToClose, joinToRegion);

            this.CurrentBlock = (PDFLayoutBlock)newTable.Parent;

            GridReference nextgrid = this.AllCells.BeginNewGrid(curr.TableStyle, curr.Position, this._rowIndex);

            nextgrid.TableBlock = newTable;
            avail.Y = 0;

            

            if (nextgrid.Position.Margins.IsEmpty == false)
            {
                PDFThickness margins = nextgrid.Position.Margins;
                avail.Width -= margins.Left + margins.Right;
                avail.Height -= margins.Top + margins.Right;
                avail.X += margins.Left;
                avail.Y += margins.Top;

            }
            nextgrid.AvailableSpace = avail;

            
            this._rowOffset = 0;

            //this._lastRowWasMovedToANewTable = true;

            return newTable;
        }

        #endregion

        //
        // inner classes
        //

        #region private class CellReference

        /// <summary>
        /// A reference to a single table cell and it's associated data in this table
        /// </summary>
        private class CellReference
        {
            #region ivars

            private PDFTableCell _cell;
            private PDFStyle _fullStyle;
            private PDFStyle _appliedStyle;
            private int _rowindex;
            private int _colindex;
            private PDFLayoutBlock _block;
            private PDFUnit _totalWidth;
            private PDFUnit _totalHeight;
            private PDFThickness _margins;
            private RowReference _row;

            #endregion

            #region public PDFTableCell Cell {get;}

            /// <summary>
            /// Gets the table cell associated with this reference
            /// </summary>
            public PDFTableCell Cell
            {
                get { return _cell; }
            }

            #endregion

            #region public PDFStyle FullStyle {get;}

            /// <summary>
            /// Gets the full style of the cell
            /// </summary>
            public PDFStyle FullStyle
            {
                get { return _fullStyle; }
            }

            #endregion

            #region public PDFStyle AppliedStyle {get;}

            /// <summary>
            /// Gets the style that is Applied to the table cell
            /// </summary>
            public PDFStyle AppliedStyle
            {
                get { return _appliedStyle; }
            }

            #endregion

            #region public int RowIndex {get;}

            /// <summary>
            /// Gets the (global) row index of this cell
            /// </summary>
            public int RowIndex
            {
                get { return _rowindex; }
            }

            #endregion

            #region public int ColumnIndex {get;}

            /// <summary>
            /// Gets the column index of this cell
            /// </summary>
            public int ColumnIndex
            {
                get { return _colindex; }
            }

            #endregion

            #region public int ColumnSpan {get;set;}

            private int _colspan;

            /// <summary>
            /// Gets or sets the column span associated with this cell
            /// </summary>
            public int ColumnSpan
            {
                get { return _colspan; }
                set { _colspan = value; }
            }

            #endregion

            #region public PDFThickness Margins {get;}

            /// <summary>
            /// Gets any margins associated with this cell's style
            /// </summary>
            public PDFThickness Margins
            {
                get { return _margins; }
            }

            #endregion

            #region public PDFUnit TotalWidth {get} + public bool HasExplicitWidth {get;}

            /// <summary>
            /// Gets the total explicit width specified for the cell
            /// </summary>
            public PDFUnit TotalWidth
            {
                get { return _totalWidth; }
            }

            /// <summary>
            /// Returns true if there is an explicit width to this cell
            /// </summary>
            public bool HasExplicitWidth
            {
                get
                {
                    return _totalWidth != PDFUnit.Zero;
                }
            }

            #endregion

            #region public PDFUnit TotalHeight {get;} + public bool HasExplicitHeight

            /// <summary>
            /// Gets the total explicit height specified for the cell
            /// </summary>
            public PDFUnit TotalHeight
            {
                get { return _totalHeight; }
            }

            public bool HasExplicitHeight
            {
                get { return _totalHeight != PDFUnit.Zero; }
            }

            #endregion

            #region public PDFLayoutBlock Block {get;set;}

            /// <summary>
            /// Gets or sets the associated block for this reference
            /// </summary>
            public PDFLayoutBlock Block
            {
                get { return _block; }
                set { _block = value; }
            }

            #endregion

            // .ctor

            #region public CellReference(PDFTableCell cell, RowReference row,  PDFStyle applied, PDFStyle fullstyle, int rowindex, int colindex)

            /// <summary>
            /// Creates a new cell reference
            /// </summary>
            /// <param name="cell"></param>
            /// <param name="row"></param>
            /// <param name="applied"></param>
            /// <param name="fullstyle"></param>
            /// <param name="rowindex"></param>
            /// <param name="colindex"></param>
            public CellReference(PDFTableCell cell, RowReference row, PDFStyle applied, PDFStyle fullstyle, int rowindex, int colindex)
            {
                this._cell = cell;
                this._appliedStyle = applied;
                this._fullStyle = fullstyle;
                this._rowindex = rowindex;
                this._colindex = colindex;
                this._block = null;
                this._row = row;

                PopulateStyleValues(fullstyle);
            }

            #endregion

            #region private void PopulateStyleValues(PDFStyle fullstyle)

            /// <summary>
            /// Populates all the style related
            /// ivars of this instance with the values from the full style
            /// </summary>
            /// <param name="fullstyle"></param>
            private void PopulateStyleValues(PDFStyle fullstyle)
            {
                PDFPositionStyle pos;
                PDFMarginsStyle margins;
                if (!fullstyle.TryGetMargins(out margins))
                    margins = new PDFMarginsStyle();
                _margins = margins.GetThickness();
                if (fullstyle.TryGetPosition(out pos))
                {
                    if (pos.IsDefined(StyleKeys.WidthAttr))
                        _totalWidth = pos.Width + margins.Left + margins.Right;
                    else
                        _totalWidth = PDFUnit.Zero;

                    if (pos.IsDefined(StyleKeys.HeightAttr))
                        _totalHeight = pos.Height + margins.Top + margins.Bottom;
                    else
                        _totalHeight = PDFUnit.Zero;
                }
            }

            #endregion
        }

        #endregion

        #region private class RowReference

        /// <summary>
        /// A reference to a single row in a table
        /// </summary>
        private class RowReference
        {
            #region ivars

            private PDFTableRow _row;
            private PDFStyle _fullStyle;
            private PDFStyle _appliedStyle;
            private int _rowindex;
            private PDFLayoutBlock _block;
            private PDFUnit _explicitHeight;
            private PDFThickness _margins;
            private GridReference _grid;

            #endregion

            // properties

            #region public PDFTableRow Row {get;}

            /// <summary>
            /// Gets the table row associated with this reference
            /// </summary>
            public PDFTableRow Row
            {
                get { return _row; }
            }

            #endregion

            #region public PDFStyle FullStyle {get;}

            /// <summary>
            /// Gets the full style associated with the row
            /// </summary>
            public PDFStyle FullStyle
            {
                get { return _fullStyle; }
            }

            #endregion

            #region public int RowIndex {get;}

            /// <summary>
            /// Gets the absolute row index of this row in the whole table
            /// </summary>
            public int RowIndex
            {
                get { return _rowindex; }
            }

            #endregion

            #region public PDFStyle AppliedStyle {get;}

            /// <summary>
            /// Gets the applied style for the row (rather than the full style)
            /// </summary>
            public PDFStyle AppliedStyle
            {
                get { return _appliedStyle; }
            }

            #endregion

            #region public PDFThickness Margins {get;}

            /// <summary>
            /// Gets the margins for this Row
            /// </summary>
            public PDFThickness Margins
            {
                get { return this._margins; }
            }

            #endregion

            #region public PDFUnit ExplicitHeight {get} + public bool HasExplicitHeight {get;}

            /// <summary>
            /// Gets the total explicit height specified for the cell
            /// </summary>
            public PDFUnit ExplicitHeight
            {
                get { return _explicitHeight; }
            }

            /// <summary>
            /// Returns true if this row has an explicit height
            /// </summary>
            public bool HasExplicitHeight
            {
                get { return _explicitHeight != PDFUnit.Zero; }
            }

            #endregion

            #region public PDFLayoutBlock Block {get; set;}

            /// <summary>
            /// Gets or sets the associated block for this reference
            /// </summary>
            public PDFLayoutBlock Block
            {
                get { return _block; }
                set { _block = value; }
            }

            #endregion

            #region public GridReference OwnerGrid {get;}

            /// <summary>
            /// Gets the grid that contains this row reference
            /// </summary>
            public GridReference OwnerGrid
            {
                get { return _grid; }
            }

            #endregion


            #region public CellReference this[int columnIndex] {get;}

            /// <summary>
            /// Gets this rows cell reference at the specified column index
            /// </summary>
            /// <param name="columnIndex"></param>
            /// <returns></returns>
            public CellReference this[int columnIndex]
            {
                get { return this.OwnerGrid.OwnerTable.AllCells[this.RowIndex, columnIndex]; }
            }

            #endregion

            // .ctor

            #region public RowReference(PDFTableRow row, PDFStyle applied, PDFStyle fullstyle,  int rowindex)

            /// <summary>
            /// Creates a new instance of a reference to a table row
            /// </summary>
            /// <param name="row"></param>
            /// <param name="applied"></param>
            /// <param name="fullstyle"></param>
            /// <param name="posOpts"></param>
            /// <param name="rowindex"></param>
            public RowReference(PDFTableRow row, GridReference grid, PDFStyle applied, PDFStyle fullstyle, int rowindex)
            {
                this._row = row;
                this._rowindex = rowindex;
                this._fullStyle = fullstyle;
                this._appliedStyle = applied;
                this._grid = grid;
                this.PopulateStyleValues(fullstyle);
            }

            #endregion

            // methods

            #region private void PopulateStyleValues(PDFStyle fullstyle)

            /// <summary>
            /// Fills the values of this row reference based on the required style values
            /// </summary>
            /// <param name="fullstyle"></param>
            private void PopulateStyleValues(PDFStyle fullstyle)
            {
                PDFPositionStyle pos;
                PDFMarginsStyle margins;
                if (!fullstyle.TryGetMargins(out margins))
                    margins = new PDFMarginsStyle();
                _margins = margins.GetThickness();

                if (fullstyle.TryGetPosition(out pos))
                {
                    if (pos.IsDefined(StyleKeys.HeightAttr))
                        _explicitHeight = pos.Height + margins.Top + margins.Bottom;
                    else
                        _explicitHeight = PDFUnit.Zero;
                }
            }

            #endregion

        }

        #endregion

        #region private class GridReference

        /// <summary>
        /// A reference to a single continuous table grid in one region.
        /// </summary>
        private class GridReference
        {

            #region ivars

            private int _startRowIndex;
            private int _endRowIndex;
            private PDFLayoutBlock _tableBlock;
            private PDFStyle _tableStyle;
            private PDFPositionOptions _posOpts;
            private TableReference _ownerTable;
            private PDFRect _availspace;

            #endregion

            // properties

            #region public int StartRowIndex {get;set;}

            /// <summary>
            /// Gets or sets the starting index of rows in this table
            /// </summary>
            public int StartRowIndex
            {
                get { return _startRowIndex; }
                set { this._startRowIndex = value; }
            }

            #endregion

            #region  public int EndRowIndex { get; set;}

            /// <summary>
            /// Gets or sets the last index of rows in this table
            /// </summary>
            public int EndRowIndex
            {
                get { return _endRowIndex; }
                set { this._endRowIndex = value; }
            }

            #endregion

            #region public PDFLayoutBlock TableBlock {get;}

            /// <summary>
            /// Gets or sets the layout block for this table
            /// </summary>
            public PDFLayoutBlock TableBlock
            {
                get { return _tableBlock; }
                set { _tableBlock = value; }
            }

            #endregion

            #region public PDFStyle TableStyle {get;}

            /// <summary>
            /// Gets the style associated with the table
            /// </summary>
            public PDFStyle TableStyle
            {
                get { return _tableStyle; }
            }

            #endregion

            #region public PDFPositionOptions Position {get;}

            /// <summary>
            /// Gets the position options for the table
            /// </summary>
            public PDFPositionOptions Position
            {
                get { return _posOpts; }
            }

            #endregion

            #region public TableReference OwnerTable {get;}

            /// <summary>
            /// Gets the owner of this grid
            /// </summary>
            public TableReference OwnerTable
            {
                get { return _ownerTable; }
            }

            #endregion

            #region public RowReference this[int index] {get;}

            /// <summary>
            /// Gets the Row reference for this grid
            /// (offset from the whole table by this grids starting row index)
            /// </summary>
            /// <param name="index"></param>
            /// <returns></returns>
            public RowReference this[int index]
            {
                get
                {
                    return this.OwnerTable.AllRows[index];
                }
            }

            #endregion

            #region public PDFRect AvailableSpace {get;set;}

            /// <summary>
            /// Gets or sets the available space with in the outer table block for this grid
            /// </summary>
            public PDFRect AvailableSpace
            {
                get { return _availspace; }
                set { _availspace = value; }
            }

            #endregion

            // ctor

            #region public GridReference(TableReference owner, PDFStyle style, PDFPositionOptions opts, int startRow, int endRow)

            /// <summary>
            /// Creates a new instance of a GridReference
            /// </summary>
            /// <param name="owner"></param>
            /// <param name="style"></param>
            /// <param name="opts"></param>
            /// <param name="startRow"></param>
            /// <param name="endRow"></param>
            public GridReference(TableReference owner, PDFStyle style, PDFPositionOptions opts, int startRow, int endRow)
            {
                this._ownerTable = owner;
                this._tableStyle = style;
                this._posOpts = opts;
                this._startRowIndex = startRow;
                this._endRowIndex = endRow;
            }

            #endregion

            #region internal PDFSize CalculateContentSize()

            internal PDFSize CalculateContentSize()
            {
                PDFUnit h = PDFUnit.Zero;
                PDFUnit w = PDFUnit.Zero;
                for (int i = this.StartRowIndex; i <= this.EndRowIndex; i++)
                {
                    RowReference row = this[i];
                    if (null != row.Block)
                    {
                        h += row.Block.Height;
                        w = PDFUnit.Max(w, row.Block.Width);
                    }
                }

                return new PDFSize(w, h);
            }

            #endregion
        }

        #endregion

        #region private class TableReference

        /// <summary>
        /// A table of references to the grids, rows and cells in this table
        /// </summary>
        private class TableReference
        {

            #region ivars

            private RowReference[] _rows;
            private CellReference[,] _cells;
            private List<GridReference> _grids;
            private int _columncount;
            private int _rowcount;

            #endregion

            // properties

            #region public RowReference[] AllRows

            /// <summary>
            /// Gets all the row references in all the grids
            /// </summary>
            public RowReference[] AllRows
            {
                get { return _rows; }
            }

            #endregion

            #region public CellReference[,] AllCells

            /// <summary>
            /// Gets the 2 dimensional array of all the cells in all grids
            /// </summary>
            public CellReference[,] AllCells
            {
                get { return _cells; }
            }

            #endregion

            #region public ICollection<GridReference> AllGrids

            /// <summary>
            /// Gets the collection of contiguous grids in this table
            /// </summary>
            public ICollection<GridReference> AllGrids
            {
                get { return _grids; }
            }

            #endregion

            #region public GridReference CurrentGrid {get;}

            /// <summary>
            /// Gets the current grid in this table
            /// </summary>
            public GridReference CurrentGrid
            {
                get
                {
                    int last = _grids.Count - 1;
                    if (last < 0)
                        return null;
                    else
                        return _grids[last];
                }
            }

            #endregion

            #region public int TotalRowCount {get;}

            /// <summary>
            /// Gets the total count of all rows in this referenced table
            /// </summary>
            public int TotalRowCount
            {
                get { return _rowcount; }
            }

            #endregion

            #region public int TotalColumnCount {get;}

            /// <summary>
            /// Gets the total count of all columns in this referenced table
            /// </summary>
            public int TotalColumnCount
            {
                get { return _columncount; }
            }

            #endregion

            // .ctors

            #region public TableReference(int rowcount, int columncount, PDFStyle tableStyle, PDFPositionOptions posOpts)

            /// <summary>
            /// Create a new Table reference, initializing the Rows, Cells and one Grid based on the row and column counts
            /// </summary>
            /// <param name="rowcount"></param>
            /// <param name="columncount"></param>
            /// <param name="tableStyle"></param>
            /// <param name="posOpts"></param>
            public TableReference(int rowcount, int columncount, PDFStyle tableStyle, PDFPositionOptions posOpts)
            {
                _rows = new RowReference[rowcount];
                _cells = new CellReference[rowcount, columncount];
                _columncount = columncount;
                _rowcount = rowcount;
                _grids = new List<GridReference>();


                this.BeginNewGrid(tableStyle, posOpts, 0);
            }

            #endregion

            // methods

            #region public GridReference BeginNewGrid(PDFStyle fulltableStyle, PDFPositionOptions posOpts, int startRow)

            /// <summary>
            /// Begins a new contiguous grid for the table closing off any previous grid at theprevious row
            /// </summary>
            /// <param name="fulltableStyle"></param>
            /// <param name="posOpts"></param>
            /// <param name="startRow"></param>
            /// <returns></returns>
            public GridReference BeginNewGrid(PDFStyle fulltableStyle, PDFPositionOptions posOpts, int startRow)
            {
                if (null != this.CurrentGrid)
                {
                    //Truncate the pre
                    this.CurrentGrid.EndRowIndex = startRow - 1;
                }

                GridReference grid = new GridReference(this, fulltableStyle, posOpts, startRow, this._rowcount - 1); //one less for the end row index
                _grids.Add(grid);
                return grid;

            }

            #endregion




            #region public RowReference AddRowReference(PDFTableRow row, GridReference grid, PDFStyle applied, PDFStyle fullstyle, int rowIndex) + 1 overload

            /// <summary>
            /// Adds a new row reference at the specified index and returns the instance
            /// </summary>
            /// <param name="row">The associated table row for the reference</param>
            /// <param name="grid">The grid this row is a part of</param>
            /// <param name="applied">The applied style of the row</param>
            /// <param name="fullstyle">The full style of the row</param>
            /// <param name="rowIndex">The index of the row to insert. There cannot be an existing row reference at this index</param>
            public RowReference AddRowReference(PDFTableRow row, GridReference grid, PDFStyle applied, PDFStyle fullstyle, int rowIndex)
            {
                RowReference rref = new RowReference(row, grid, applied, fullstyle, rowIndex);
                this.AddRowReference(rref);
                return rref;
            }


            /// <summary>
            /// Adds the row reference to the grid at the index specified in the reference's RowIndex. 
            /// NOTE: There cannot be an existing row reference at this index
            /// </summary>
            /// <param name="rref"></param>
            public void AddRowReference(RowReference rref)
            {
                if (null != _rows[rref.RowIndex])
                    throw new ArgumentException("Row reference has already been set");
                this._rows[rref.RowIndex] = rref;
            }

            #endregion

            #region public void AddCellReference(PDFTableCell cell, RowReference rowRef, PDFStyle applied, PDFStyle fullstyle, int column) + 1 overload

            /// <summary>
            /// Adds a new cell reference at the specified index and returns the instance
            /// </summary>
            /// <param name="cell">The associated table cell</param>
            /// <param name="rowRef">The row reference for this cell</param>
            /// <param name="applied">The applied style of the cell</param>
            /// <param name="fullstyle">The full style of the cell</param>
            /// <param name="column">The column index of the cell. NOTE: There cannot be an existing cell reference in the rows column</param>
            public CellReference AddCellReference(PDFTableCell cell, RowReference rowRef, PDFStyle applied, PDFStyle fullstyle, int column)
            {
                CellReference cref = new CellReference(cell, rowRef, applied, fullstyle, rowRef.RowIndex, column);
                this.AddCellReference(cref);
                return cref;
            }



            /// <summary>
            /// Adds a cell reference to the grid at the row and column index in the reference
            /// NOTE: There cannot be an existing cell reference in the rows column
            /// </summary>
            /// <param name="reference"></param>
            public void AddCellReference(CellReference reference)
            {
                if (null != this._cells[reference.RowIndex, reference.ColumnIndex])
                    throw new ArgumentException("Cell reference has already been set");

                this._cells[reference.RowIndex, reference.ColumnIndex] = reference;
            }


            #endregion

            #region public void SetRowCount(int count, bool clear)

            /// <summary>
            /// Alters the total row count in this table, 
            /// and optionally removes / clears any references to rows beyond this.
            /// </summary>
            /// <param name="count">The new row count</param>
            /// <param name="clear">Option to clear rows that have been created but are no longer required.</param>
            public void SetRowCount(int count, bool clear)
            {
                int orig = _rowcount;
                this._rowcount = count;

                if (orig < count)
                {
                    RowReference[] newRows = _rows;
                    Array.Resize<RowReference>(ref newRows, count);
                    _rows = newRows;
                    CellReference[,] newcells = _cells;
                    Resize2DArray<CellReference>(ref newcells, count, _columncount);
                    _cells = newcells;
                }
                else if (clear)
                    this.ClearAllRowReferencesFrom(count);

            }

            #endregion

            #region public void ClearAllRowReferencesFrom(int rowindex)

            /// <summary>
            /// Removes all the cell and row references for all rows after and including the row at the specified index.
            /// And also sets any associated layout blocks for the rows and cells to invisible
            /// </summary>
            /// <param name="rowindex"></param>
            public void ClearAllRowReferencesFrom(int rowindex)
            {
                for (int index = _rowcount - 1; index >= rowindex; index--)
                {
                    this.ClearRowReference(index);
                }
            }

            #endregion

            #region public void ClearRowReference(int rowindex)

            /// <summary>
            /// Clears the row and cell references for the specified index, and also sets any associated layout block to invisible.
            /// </summary>
            /// <param name="rowindex"></param>
            public void ClearRowReference(int rowindex)
            {
                PDFLayoutBlock block = _rows[rowindex].Block;
                if (null != block)
                    block.Position.Visibility = Visibility.None;
                for (int col = 0; col < _columncount; col++)
                {
                    block = this._cells[rowindex, col].Block;
                    if (null != block)
                        block.Position.Visibility = Visibility.None;
                    this._cells[rowindex, col] = null;
                }
                this._rows[rowindex] = null;

            }

            #endregion

            #region public PDFUnit GetMaxCellHeightForRow(int rowindex)

            /// <summary>
            /// Gets the maximum required height of all cells in a particular row
            /// </summary>
            /// <param name="rowindex"></param>
            /// <returns></returns>
            public PDFUnit GetMaxCellHeightForRow(int rowindex)
            {
                PDFUnit maxH = PDFUnit.Zero;
                for (int col = 0; col < _columncount; col++)
                {
                    CellReference cref = this._cells[rowindex, col];
                    if (null != cref && null != cref.Block)
                    {
                        PDFUnit h = cref.Block.TotalBounds.Height;
                        maxH = PDFUnit.Max(h, maxH);
                    }
                }
                return maxH;
            }

            #endregion

            #region public PDFUnit GetMaxCellWidthForColumn(int colIndex)

            /// <summary>
            /// Gets the maximum required width of all cells in a particular column
            /// </summary>
            /// <param name="colIndex"></param>
            /// <returns></returns>
            public PDFUnit GetMaxCellWidthForColumn(int colIndex)
            {
                PDFUnit maxW = PDFUnit.Zero;
                for (int row = 0; row < _rowcount; row++)
                {
                    CellReference cref = this._cells[row, colIndex];
                    if (null != cref && null != cref.Block)
                    {
                        PDFUnit w = cref.Block.TotalBounds.Width;
                        maxW = PDFUnit.Max(w, maxW);
                    }
                }
                return maxW;
            }

            #endregion

            #region public void SetCellHeightForRow(int rowIndex, PDFUnit h)

            /// <summary>
            /// Sets the height of all cells in a row to the specified value. 
            /// Also sets the height of any block associcated with the cell
            /// </summary>
            /// <param name="rowIndex"></param>
            /// <param name="h"></param>
            public void SetCellHeightForRow(int rowIndex, PDFUnit h)
            {
                for (int col = 0; col < _columncount; col++)
                {
                    CellReference cref = this._cells[rowIndex, col];
                    if (null != cref && null != cref.Block)
                    {
                        PDFRect total = cref.Block.TotalBounds;
                        total.Height = h; //- cref.Margins.Top - cref.Margins.Bottom;
                        cref.Block.TotalBounds = total;
                    }
                }
            }

            #endregion

            #region public void SetCellWidthForColumn(int colIndex, PDFUnit w)

            /// <summary>
            /// Sets the width of all the cell in a column to the specified value. 
            /// Also sets the width of any block associated with the cell
            /// </summary>
            /// <param name="colIndex"></param>
            /// <param name="w"></param>
            public void SetCellWidthForColumn(int colIndex, PDFUnit w)
            {
                for (int row = 0; row < _rowcount; row++)
                {
                    CellReference cref = this._cells[row, colIndex];
                    if (null != cref && null != cref.Block)
                    {
                        PDFRect total = cref.Block.TotalBounds;
                        total.Width = w; //- cref.Margins.Left - cref.Margins.Right;
                        cref.Block.TotalBounds = total;
                    }
                }
            }

            #endregion

            #region internal void SetRowOffsetAndWidthForColumn(int colIndex, PDFUnit xOffset, PDFUnit maxColWidth)

            /// <summary>
            /// Sets the offset and width of a column region in each of the rows
            /// </summary>
            /// <param name="colIndex">The column region</param>
            /// <param name="xOffset">The horizontal offset from the start of the row</param>
            /// <param name="maxColWidth">The width of the column</param>
            internal void SetRowOffsetAndWidthForColumn(int colIndex, PDFUnit xOffset, PDFUnit maxColWidth)
            {
                for (int rowindex = 0; rowindex < _rowcount; rowindex++)
                {
                    RowReference rowref = _rows[rowindex];
                    if (null != rowref && null != rowref.Block)
                    {
                        PDFLayoutBlock rowblock = rowref.Block;
                        PDFLayoutRegion column = rowblock.Columns[colIndex];
                        PDFRect colbounds = column.TotalBounds;
                        colbounds.X = xOffset;
                        colbounds.Width = maxColWidth;
                        column.TotalBounds = colbounds;
                    }
                }
            }

            #endregion

            #region private static void Resize2DArray<T>(ref T[,] original, int rows, int cols)

            /// <summary>
            /// Helper method that will resize a 2 dimensional array
            /// </summary>
            /// <typeparam name="T"></typeparam>
            /// <param name="original"></param>
            /// <param name="rows"></param>
            /// <param name="cols"></param>
            private static void Resize2DArray<T>(ref T[,] original, int rows, int cols)
            {
                var newArray = new T[rows, cols];
                int minRows = Math.Min(rows, original.GetLength(0));
                int minCols = Math.Min(cols, original.GetLength(1));
                for (int row = 0; row < minRows; row++)
                    for (int col = 0; col < minCols; col++)
                        newArray[row, col] = original[row, col];
                original = newArray;
            }

            #endregion
        }

        #endregion


        #region private struct CellDimension

        /// <summary>
        /// Holds a single dimension for a cell
        /// </summary>
        private struct CellDimension
        {
            private PDFUnit _sz;
            private bool _explicit;

            public PDFUnit Size { get { return _sz; } set { _sz = value; } }
            public bool Explicit { get { return _explicit; } set { _explicit = value; } }

            public CellDimension(PDFUnit size)
                : this(size, true)
            {
            }

            public CellDimension(PDFUnit size, bool isExplicit)
            {
                _sz = size;
                _explicit = isExplicit;
            }

        }

        #endregion

    }
}
