﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using Scryber.Styles;
using Scryber.Drawing;
using Scryber.Native;

namespace Scryber.Components.Layout
{
    internal class PageLayoutEngine : PDFBlockLayoutEngineBase
    {

        private PDFStyleStack _origStyleStack;

        //
        // properties
        //

        #region protected PDFStyleStack PageStyleStack {get;set;}

        /// <summary>
        /// Gets or sets the style stack for the page
        /// (so continuation headers and footers can use this when content is overflowed, rather than the full stack).
        /// </summary>
        protected PDFStyleStack PageStyleStack
        {
            get { return _origStyleStack; }
            set { _origStyleStack = value; }
        }

        #endregion

        #region protected PDFPage Page {get;}

        /// <summary>
        /// Gets the page definition associated with this engine
        /// </summary>
        protected PDFPage Page
        {
            get { return this.Component as PDFPage; }
        }

        #endregion

        //
        // ctor(s)
        //

        #region public PageLayoutEngine(PDFPage pg, IPDFLayoutEngine parent)

        /// <summary>
        /// Creates a new instance of the page layout engine that laysout the content of the page.
        /// </summary>
        /// <param name="pg">The page definition that contains all the components to layout</param>
        /// <param name="context">The layout context</param>
        /// <param name="style">The full style associated with the page </param>
        public PageLayoutEngine(PDFPage pg, IPDFLayoutEngine parent)
            : base(pg, parent)
        {
        }

        #endregion


        //
        // overrides
        //


        /// <summary>
        /// Main overridden method
        /// </summary>
        protected override void DoLayoutComponent()
        {
            //Get the page size
            PDFPageSize pgsize;
            PDFPageStyle paper;
            if (this.FullStyle.TryGetPaper(out paper))
                pgsize = paper.CreatePageSize();
            else
                pgsize = PDFPageSize.A4;

            PDFPositionOptions options = this.FullStyle.CreatePostionOptions();

            //Get the page numbering
            
            


            //Size, border, margins
            PDFRect bounds = new PDFRect(PDFPoint.Empty, pgsize.Size);
            PDFThickness margthick;
            PDFThickness padthick;
            PDFRect contentrect = GetContentRectFromBounds(bounds, out margthick, out padthick);
            
            
            //Columns
            PDFUnit alley = PDFUnit.Zero;
            int colcount = GetBlockColumnCount(out alley);

            //Overflow
            OverflowAction action = OverflowAction.None;
            PDFOverflowStyle overflow;
            if (this.FullStyle.TryGetOverflow(out overflow))
                action = overflow.Action;

            this.PageStyleStack = this.Context.StyleStack.Clone();


            PDFLayoutPage pg = BuildNewPage(pgsize.Size, options, alley, colcount, action);

            PDFPageStyle numbers = this.GetPageNumbering(this.FullStyle);
            this.RegisterPageNumbering(pg, numbers);

            this.LayoutPageContent();
            


            //close the last page
            PDFLayoutPage last = this.DocumentLayout.CurrentPage;
            if(last.IsClosed == false)
                last.Close();

            this.UnRegisterPageNumbering(last, numbers);
            
        }

        private int _registeredIndex;
        private PDFPageNumbering _assigned;

        private void RegisterPageNumbering(PDFLayoutPage page, PDFPageStyle numbers)
        {
            _registeredIndex = this.DocumentLayout.RegisterPageNumbering(page, numbers, out _assigned);
            this.Context.TraceLog.Add(TraceLevel.Message, PageLayoutEngine.LOG_CATEGORY, "Registered the page numbering");
        }

        private void UnRegisterPageNumbering(PDFLayoutPage last, PDFPageStyle numbers)
        {
            this.DocumentLayout.UnRegisterPageNumbering(last.PageIndex, last, _assigned);
            this.Context.TraceLog.Add(TraceLevel.Message, PageLayoutEngine.LOG_CATEGORY, "Un-registered the page numbering");
        }


        /// <summary>
        /// Checks the overflow style and if new pages are supported closes the current page layout and
        /// creates a new page layout (becomming the current page) and returns true. 
        /// If overflow is not supported - returns false
        /// </summary>
        /// <param name="region">If there is a change in current page, this is set to the new region</param>
        /// <param name="block">If there is a change in current page, this is set to the new block</param>
        /// <returns></returns>
        public override bool MoveToNextPage(Stack<PDFLayoutBlock> depth, ref PDFLayoutRegion region, ref PDFLayoutBlock block)
        {
            PDFOverflowStyle overflow;
            
            if (this.FullStyle.TryGetOverflow(out overflow) && overflow.Action == OverflowAction.NewPage)
            {
                PDFLayoutPage lastpage = this.DocumentLayout.CurrentPage;
                PDFLayoutBlock open = lastpage.ContentBlock;
                if (open.IsClosed)
                    open = null;
                else
                    open = open.LastOpenBlock();

                while (null != open)
                {
                    
                    PDFLayoutBlock parent = open.Parent as PDFLayoutBlock;
                    if (null != parent)
                    {
                        PDFLayoutRegion openregion = parent.CurrentRegion;
                        if (null != openregion)
                        {
                            open.Close();
                            openregion.AddToSize(open);
                            openregion.Close();
                        }
                    }
                    open = parent;


                }
                lastpage.Close();
                PDFLayoutPage page = BuildContinuationPage(lastpage);

                block = page.CurrentBlock;
                region = block.CurrentRegion;

                return true;
            }
            else
            {
                return false; //Cannot overflow
            }
        }

        #region private PDFLayoutPage BuildContinuationPage(PDFLayoutPage copyfrom)

        /// <summary>
        /// Based on the sizeing and style of the last page buids a continuation page that content can flow into
        /// </summary>
        /// <param name="copyfrom"></param>
        /// <returns></returns>
        private PDFLayoutPage BuildContinuationPage(PDFLayoutPage copyfrom)
        {
            //Take a reference of the current stack and replace with the page stack from first page
            PDFStyleStack orig = this.Context.StyleStack;
            this.Context.StyleStack = this.PageStyleStack;

            PDFUnit alley = PDFUnit.Zero;
            int colcount = GetBlockColumnCount(out alley);

            PDFLayoutPage page = this.BuildNewPage(copyfrom.Size, copyfrom.PositionOptions, alley, colcount, copyfrom.OverflowAction);

            int pgIndex = this.DocumentLayout.CurrentPageIndex;

            //PDFPageStyle numStyle;
            //this.FullStyle.TryGetPaper(out numStyle);
            //becasue we are a continuation page, we have the same number style, so let's just register it with null
            PDFPageNumbering nums;
            this.DocumentLayout.RegisterPageNumbering(page, null, out nums);

            this.LayoutPageHeaderAndFooter();

            //Restore the context stack with the reference
            this.Context.StyleStack = orig;

            return page;
        }

        #endregion
        
        //
        // support methods
        //

        #region private void BuildNewPage(PDFPageSize pgsize, PDFPositionOptions options ...)
        
        /// <summary>
        /// Creates a new page with the specified options and adds it to the current layout
        /// </summary>
        /// <param name="pgsize"></param>
        /// <param name="options"></param>
        /// <param name="alley"></param>
        /// <param name="colcount"></param>
        /// <param name="action"></param>
        private PDFLayoutPage BuildNewPage(PDFSize pgsize, PDFPositionOptions options, PDFUnit alley, int colcount, OverflowAction action)
        {
            PDFLayoutDocument doclayout = this.DocumentLayout;
            PDFLayoutPage pg = doclayout.BeginNewPage(this.Page, this, this.FullStyle, action);

            pg.InitPage(doclayout.CurrentPageIndex, pgsize, options, colcount, alley);

            return pg;
        }

        #endregion

        #region protected void LayoutPageContent()

        /// <summary>
        /// Lays out all the content for the page inc Headers and Footers
        /// </summary>
        protected void LayoutPageContent()
        {
            this.LayoutPageHeaderAndFooter();
            //We have to push the arangement for this engine onto the page
            this.DoLayoutChildren();
        }

        #endregion

        #region private void LayoutPageHeaderAndFooter()

        /// <summary>
        /// Performs the layout of a single page, with header and footer generated and included as per the definition.
        /// Does not adjust the content size of the page
        /// </summary>
        private void LayoutPageHeaderAndFooter()
        {
            IPDFTemplate headtemplate = this.GetCurrentHeaderTemplate(this.DocumentLayout.CurrentPageIndex);
            IPDFTemplate foottemplate = this.GetCurrentFooterTemplate(this.DocumentLayout.CurrentPageIndex);

            if (headtemplate != null)
            {
                PDFPageHeader header = new PDFPageHeader();
                this.Page.AddGeneratedHeader(header, this.DocumentLayout.CurrentPageIndex);
                //Create the content inside the header component
                InstantiateTemplateForPage(header, headtemplate);
                
                this.DocumentLayout.CurrentPage.BeginHeader(header);
                this.DoLayoutAChild(header);
                this.DocumentLayout.CurrentPage.EndHeader();

            }

            if (foottemplate != null)
            {
                PDFPageFooter footer = new PDFPageFooter();
                this.Page.AddGeneratedFooter(footer, this.DocumentLayout.CurrentPageIndex);

                InstantiateTemplateForPage(footer, foottemplate);
                
                this.DocumentLayout.CurrentPage.BeginFooter(footer);
                this.DoLayoutAChild(footer);
                this.DocumentLayout.CurrentPage.EndFooter();

            }
        }

        #endregion

        #region protected virtual IPDFTemplate GetCurrentHeaderTemplate(int pageIndex)

        /// <summary>
        /// Gets the IPDFTemplate for the Header definition on this page
        /// </summary>
        /// <param name="pageIndex">The current page index in the document layout</param>
        /// <returns></returns>
        protected virtual IPDFTemplate GetCurrentHeaderTemplate(int pageIndex)
        {
            return this.Page.Header;
        }

        #endregion

        #region protected virtual IPDFTemplate GetCurrentFooterTemplate(int pageIndex)

        /// <summary>
        /// Gets the IPDFTemplate for the Footer definition on the page
        /// </summary>
        /// <param name="pageIndex">The current page index in the document layout</param>
        /// <returns></returns>
        protected virtual IPDFTemplate GetCurrentFooterTemplate(int pageIndex)
        {
            return this.Page.Footer;
        }

        #endregion

        #region protected virtual void InstantiateTemplateForPage(PDFLayoutTemplateComponent container, IPDFTemplate template)

        /// <summary>
        /// Creates a concrete instance of the correct template into the PDFLayoutTemplate component
        /// </summary>
        /// <param name="container">A non null instance of the layout template component to contain the generated template components</param>
        protected virtual void InstantiateTemplateForPage(PDFLayoutTemplateComponent container, IPDFTemplate template)
        {
            if (null == template)
                throw new ArgumentNullException("template");
            if (null == container)
                throw new ArgumentNullException("container");

            PDFRect avail =  this.DocumentLayout.CurrentPage.LastOpenBlock().CurrentRegion.TotalBounds;
            int pg = this.DocumentLayout.CurrentPageIndex;

            container.InstantiateTemplate(template, this.Context, avail, pg);
        }

        #endregion

        #region protected PDFPageNumbering GetPageNumbering(PDFStyle style)

        /// <summary>
        /// Gets the current page numbering style for this page
        /// </summary>
        /// <param name="style"></param>
        /// <returns></returns>
        protected PDFPageStyle GetPageNumbering(PDFStyle style)
        {
            PDFPageStyle pgstyle;
            style.TryGetPaper(out pgstyle);
            return pgstyle;
        }

        
        #endregion

        #region protected PDFPageSize GetPageSize(PDFStyle style)

        /// <summary>
        /// Gets the current page size style for this page
        /// </summary>
        /// <param name="style"></param>
        /// <returns></returns>
        protected PDFPageSize GetPageSize(PDFStyle style)
        {
            PDFPageStyle pg;
            if (!style.TryGetPaper(out pg))
                return PDFPageSize.A4;
            else
                return pg.CreatePageSize();
        }

        #endregion



    }
}
