﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;

namespace Scryber.Components
{
    public abstract class PDFBindingTemplateComponent : PDFVisualComponent
    {
        #region public event TemplateItemDataBoundHandler ItemDataBound

        [PDFAttribute("on-item-databound")]
        public event TemplateItemDataBoundHandler ItemDataBound;

        protected virtual void OnItemDataBound(TemplateItemDataBoundArgs args)
        {
            if (null != this.ItemDataBound)
                this.ItemDataBound(this, args);
        }

        protected virtual void OnItemDataBound(PDFDataContext context, IPDFComponent item)
        {
            TemplateItemDataBoundArgs args = new TemplateItemDataBoundArgs(item, context);
            this.OnItemDataBound(args);
        }

        #endregion


        /// <summary>
        /// The list of Components that were added to the
        /// parents Component collection when the template was bound.
        /// </summary>
        private List<IPDFComponent> _addedonbind;

        /// <summary>
        /// The parent Component the items were added to.
        /// </summary>
        private IPDFContainerComponent _toparent;
        

        //
        // ctor(s)
        //

        protected PDFBindingTemplateComponent(PDFObjectType type)
            : base(type)
        {
        }


        //
        // methods
        //
        
        /// <summary>
        /// Main override so that the data can be bound to the any templates
        /// </summary>
        /// <param name="context"></param>
        /// <param name="includeChildren"></param>
        protected override void DoDataBind(PDFDataContext context, bool includeChildren)
        {
            if (_addedonbind != null && _addedonbind.Count > 0)
                this.ClearPreviousBoundComponents(_addedonbind, _toparent);
            else if(null == _addedonbind)
                _addedonbind = new List<IPDFComponent>();

            //call the base method first
            base.DoDataBind(context, includeChildren);

            if (includeChildren)
            {
                IPDFContainerComponent container = GetContainerParent();
                DoDataBindToContainer(context, container);
            }
        }


        /// <summary>
        /// 
        /// </summary>
        /// <param name="context"></param>
        /// <param name="container"></param>
        protected virtual void DoDataBindToContainer(PDFDataContext context, IPDFContainerComponent container)
        {
            //If we have a template and should be binding on it
            if (context.TraceLog.ShouldLog(TraceLevel.Message))
                context.TraceLog.Begin(TraceLevel.Message, "Binding Template", "Starting template binding with datasource on component " + container.ToString());

            int oldindex = context.CurrentIndex;
            int index = container.Content.IndexOf(this);
            DoBindDataIntoContainer(container, index, context);
            _toparent = container;

            if (context.TraceLog.ShouldLog(TraceLevel.Message))
                context.TraceLog.End(TraceLevel.Message, "Binding Template", "Bound " + (context.CurrentIndex + 1) + " items with datasource on component " + container.ToString());

            context.CurrentIndex = oldindex;

        }

        protected IPDFContainerComponent GetContainerParent()
        {
            PDFComponent ele = this;
            while (null != ele)
            {
                PDFComponent par = ele.Parent;
                if (par == null)
                    throw RecordAndRaise.ArgumentNull(Errors.TemplateComponentParentMustBeContainer);

                else if ((par is IPDFContainerComponent) == false)
                    ele = par;
                else
                    return par as IPDFContainerComponent;
            }

            //If we get this far then we haven't got a viable container to add our items to.
            throw RecordAndRaise.ArgumentNull(Errors.TemplateComponentParentMustBeContainer);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="container"></param>
        /// <param name="containerposition"></param>
        /// <param name="template"></param>
        /// <param name="context"></param>
        protected virtual void DoBindDataIntoContainer(IPDFContainerComponent container, int containerposition, PDFDataContext context)
        {
            int prevcount = context.CurrentIndex;
            object data = context.DataStack.Current;

            int count = 0;
            int added = 0;
            IEnumerator enumerator = null;
            
            if (data is System.Xml.XPath.XPathNodeIterator)
            {
                System.Xml.XPath.XPathNodeIterator itter = (System.Xml.XPath.XPathNodeIterator)data;
                enumerator = this.CreateEnumerator(itter);
                
                while (enumerator.MoveNext())
                {
                    System.Xml.XPath.XPathNavigator nav = (System.Xml.XPath.XPathNavigator)enumerator.Current;
                    
                    context.CurrentIndex = count++;
                    context.DataStack.Push(nav);
                    added += InstantiateAndAddFromTemplate(context, count, added + containerposition, container);

                    context.DataStack.Pop();
                }
            }
            else if (data is IEnumerable)
            {
                IEnumerable ienum = data as IEnumerable;
                enumerator = this.CreateEnumerator(ienum);
                while (enumerator.MoveNext())
                {
                    context.CurrentIndex = count++;
                    context.DataStack.Push(enumerator.Current);
                    added += InstantiateAndAddFromTemplate(context, count, added + containerposition, container);

                    context.DataStack.Pop();
                }
            }
            else if (data != null)
            {
                context.CurrentIndex = 1;
                context.DataStack.Push(data);
                added += InstantiateAndAddFromTemplate(context, count, added + containerposition, container);
                context.DataStack.Pop();
                
            }

            context.CurrentIndex = prevcount;
            
        }

        /// <summary>
        /// Abstract method that all inheritors must override to return the template required for instaniating and binding.
        /// </summary>
        /// <param name="context"></param>
        /// <param name="index"></param>
        /// <param name="count"></param>
        /// <returns></returns>
        protected abstract IPDFTemplate GetTemplateForBinding(PDFDataContext context, int index, int count);


        #region protected virtual IEnumerator CreateEnumerator(IEnumerable enumerable)

        /// <summary>
        /// Returns an enumerator that will loop over all the items in the IEnumerable instance.
        /// </summary>
        /// <param name="enumerable"></param>
        /// <returns></returns>
        protected virtual IEnumerator CreateEnumerator(IEnumerable enumerable)
        {
            return enumerable.GetEnumerator();
        }

        #endregion

        #region protected virtual IEnumerator CreateSingleDataEnumerator(object data)

        /// <summary>
        /// returns an enumerator that will enumerate once over a single item
        /// </summary>
        /// <param name="data"></param>
        /// <returns></returns>
        protected virtual IEnumerator CreateSingleDataEnumerator(object data)
        {
            return new SingleItemEnumerator(data);
        }

        #endregion

        #region private int InstantiateAndAddFromTemplate(PDFDataContext context, int count, int index, IPDFContainerComponent container, IPDFTemplate template)

        /// <summary>
        /// Creates a new instance of the template and adds it to this components content
        /// </summary>
        /// <param name="context"></param>
        /// <param name="count"></param>
        /// <param name="index"></param>
        /// <param name="container"></param>
        /// <param name="template"></param>
        /// <returns></returns>
        private int InstantiateAndAddFromTemplate(PDFDataContext context, int count, int index, IPDFContainerComponent container)
        {
            IPDFTemplate template = this.GetTemplateForBinding(context, count, index);

            if (null == template)
                return 0;
            PDFTraceLog log = context.TraceLog;
            PDFInitContext init = new PDFInitContext(context.Items, log);
            PDFLoadContext load = new PDFLoadContext(context.Items, log);
            
            IEnumerable<IPDFComponent> created = template.Instantiate(count);
            int added = 0;
            if (created != null)
            {
                foreach (IPDFComponent ele in ((IEnumerable)created))
                {
                    InsertComponentInContainer(container, index, ele, init, load);
                    if (ele is IPDFBindableComponent)
                    {
                        ((IPDFBindableComponent)ele).DataBind(context);
                    }
                    index++;
                    added++;
                    this.OnItemDataBound(context, ele);
                }

            }
            return added;
        }

        #endregion

        #region private void ClearPreviousBoundComponents(ICollection<IPDFComponent> all, IPDFContainerComponent container)

        /// <summary>
        /// Removes any previously bound Components
        /// </summary>
        /// <param name="all"></param>
        /// <param name="container"></param>
        private void ClearPreviousBoundComponents(ICollection<IPDFComponent> all, IPDFContainerComponent container)
        {
            //dispose and clear
            foreach (PDFComponent ele in all)
            {
                container.Content.Remove(ele);
                ele.Dispose();
            }
            all.Clear();
        }

        #endregion

        #region private void InsertComponentInContainer(IPDFContainerComponent container, int index, IPDFComponent ele, PDFInitContext init, PDFLoadContext load)

        /// <summary>
        /// Inserts a new Component in the container
        /// </summary>
        /// <param name="container"></param>
        /// <param name="index"></param>
        /// <param name="ele"></param>
        private void InsertComponentInContainer(IPDFContainerComponent container, int index, IPDFComponent ele, PDFInitContext init, PDFLoadContext load)
        {
            ele.Init(init);
            IPDFComponentList list = container.Content as IPDFComponentList;
            list.Insert(index, ele);
            _addedonbind.Add(ele);

            if (ele is IPDFVisualComponent)
            {
                ((IPDFVisualComponent)ele).Load(load);
            }
        }

        #endregion

        #region private class SingleItemEnumerator : IEnumerator


        /// <summary>
        /// Implements an IEnumerator for a single piece of data
        /// </summary>
        private class SingleItemEnumerator : IEnumerator
        {
            private int _index;
            private object _data;

            public object Current
            {
                get
                {
                    if (_index == 0)
                        return _data;
                    else
                        throw new ArgumentOutOfRangeException("index");
                }
            }

            public SingleItemEnumerator(object data)
            {
                this._data = data;
                this.Reset();
            }

            public bool MoveNext()
            {
                _index++;
                return _index == 0;
            }

            public void Reset()
            {
                _index = -1;
            }
        }

        #endregion
    }
}
