﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Scryber.Resources;
using Scryber.Native;
using Scryber.Drawing;
using Scryber.Styles;

namespace Scryber.Components
{
    public abstract class PDFImageBase : PDFVisualComponent, IPDFImageComponent
    {

        private PDFImageXObject _xobj = null;

        /// <summary>
        /// Gets or sets the image data associated with this instance.
        /// </summary>
        protected PDFImageXObject XObject
        {
            get { return _xobj; }
            set { _xobj = value; }
        }


        protected PDFImageBase(PDFObjectType type)
            : base(type)
        {
        }


        /// <summary>
        /// Implementation method for the IPDFImageComponent to return the image data in a PDFImageXObject instance.
        /// </summary>
        /// <returns></returns>
        public virtual PDFImageXObject GetImageObject(PDFContextBase context, PDFStyle fullstyle)
        {
            if (null == this.XObject)
            {
                bool log = context.TraceLog.ShouldLog(TraceLevel.Debug);
                if (log)
                    context.TraceLog.Begin("Image", "Initializing and loading image for component " + this.UniqueID);

                try
                {
                    this.XObject = this.InitImageXObject(context, fullstyle);
                }
                catch (PDFException)
                {
                    throw;
                }
                catch (Exception ex)
                {
                    throw new PDFException(string.Format(Errors.CouldNotInitializeTheImageForComponent, this.UniqueID, ex.Message), ex);
                }


                if (log)
                {
                    if(null == this.XObject)
                        context.TraceLog.End("Image", "No image loaded for component " + this.UniqueID);

                    context.TraceLog.End("Image", "Initialized and loaded image '" + this.XObject.ToString() + "' for component " + this.UniqueID);
                }
            }
            return this.XObject;
        }

        /// <summary>
        /// Releases any stored resources required for generating and rendering the image.
        /// </summary>
        public virtual void ResetImageObject()
        {
            this._xobj = null;
        }


        /// <summary>
        /// Inheritors must implement this method to actually load the required image data.
        /// </summary>
        /// <returns></returns>
        protected abstract Resources.PDFImageXObject InitImageXObject(PDFContextBase context, PDFStyle fullstyle);

        /// <summary>
        /// Overrides the base implementation to specify that images are by default blocks.
        /// </summary>
        /// <returns></returns>
        protected override PDFStyle GetBaseStyle()
        {
            PDFStyle s = base.GetBaseStyle();
            s.Position.PositionMode = Scryber.Drawing.PositionMode.Block;
            return s;
        }

        /// <summary>
        /// Overrides the base implementation to registed the image data as a resource on the current container
        /// </summary>
        /// <param name="context"></param>
        /// <param name="set"></param>
        /// <param name="fullstyle"></param>
        protected override void DoRegisterArtefacts(PDFLayoutContext context, PDFArtefactRegistrationSet set, PDFStyle fullstyle)
        {
            IPDFResourceContainer resources = this.GetResourceContainer();
            if (null == resources)
                throw RecordAndRaise.NullReference(Errors.ResourceContainerOfComponnetNotFound, "Image", this.ID);
            PDFImageXObject xobj = this.GetImageObject(context, fullstyle);

            if (null != xobj)
                resources.Register(xobj);

            base.DoRegisterArtefacts(context, set, fullstyle);
        }


        public virtual PDFObjectRef OutputToPDF(PDFRenderContext context, PDFWriter writer)
        {
            PDFGraphics graphics = context.Graphics;
            PDFStyle full = null;

            PDFComponentArrangement arrange = this.GetFirstArrangement();
            if (null != arrange)
                full = arrange.FullStyle;

            if (null == full)
                full = context.FullStyle;

            PDFImageXObject img = this.GetImageObject(context, full);
            if (img != null)
            {
                PDFPoint pos = context.Offset;


                PDFSize imgsize = context.Space;

                //the pictures are drawn from their bottom left corner, so take off the height.
                //if (context.DrawingOrigin == DrawingOrigin.TopLeft)
                //    pos.Y = pos.Y + imgsize.Height;

                graphics.SaveGraphicsState();

                PDFFillStyle fill = null;
                if (null == fill && full.TryGetFill(out fill))
                {
                    PDFReal op = fill.Opacity;
                    if (op < 1.0)
                    {
                        graphics.SetFillOpacity(op);
                    }
                }
                PDFObjectRef imgref = img.EnsureRendered(context, writer);
                graphics.PaintImageRef(img, imgsize, pos);

                graphics.RestoreGraphicsState();
                return imgref;
            }
            else
                return null;
        }


        /// <summary>
        /// Helper method that adjusts the image size (proportionally) based on the defined size of the image in the image data, and any styleing that is applied.
        /// </summary>
        /// <param name="style">The current full style of the image component</param>
        /// <param name="imgsize">The actual size of the image in the data</param>
        /// <returns>An appropriate size</returns>
        internal static PDFSize AdjustImageSize(PDFStyle style, PDFSize imgsize)
        {
            PDFSize rendersize = imgsize;
            PDFPositionStyle pos;

            if (style.TryGetPosition(out pos))
            {
                bool scaleWidth = false; //proportional flags
                bool scaleHeight = false;

                if (pos.IsDefined(StyleKeys.WidthAttr))
                {
                    rendersize.Width = pos.Width;
                    scaleHeight = true;
                }
                if (pos.IsDefined(StyleKeys.HeightAttr))
                {
                    rendersize.Height = pos.Height;
                    scaleWidth = true;
                }


                if (scaleWidth && scaleHeight)
                {
                    //Do nothing as the size is set for both height and width.
                }
                else if (scaleWidth)
                {
                    double val = rendersize.Width.PointsValue;
                    double scale = rendersize.Height.PointsValue / imgsize.Height.PointsValue;
                    rendersize.Width = (PDFUnit)(val * scale);
                }
                else if (scaleHeight)
                {
                    double val = rendersize.Height.PointsValue;
                    double scale = rendersize.Width.PointsValue / imgsize.Width.PointsValue;
                    rendersize.Height = (PDFUnit)(val * scale);
                }

                imgsize = rendersize;
            }
            return imgsize;
        }
    }
}
