﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Scryber.Styles;

namespace Scryber.Components
{

    /// <summary>
    /// Encloses a group of pages or sections. Can also be styled and outlined to provide structure to document pages
    /// </summary>
    [PDFParsableComponent("PageGroup")]
    public class PDFPageGroup : PDFPage
    {

        #region public PDFPageList Pages {get;set;}

        private PDFPageList _pages;

        /// <summary>
        /// Gets or sets the list of Pages / Sections and also nested PageGroups in this group
        /// </summary>
        [PDFArray(typeof(PDFPage))]
        [PDFElement("Pages")]
        public PDFPageList Pages
        {
            get
            {
                if (null == _pages)
                    _pages = new PDFPageList(this.InnerContent);
                return _pages;
            }
        }

        #endregion

        //
        // .ctor
        //

        #region public PDFPageGroup()

        public PDFPageGroup()
            : this(PDFObjectTypes.PageGroup)
        {
        }

        #endregion

        #region protected PDFPageGroup(PDFObjectType type)

        protected PDFPageGroup(PDFObjectType type)
            : base(type)
        {
        }

        #endregion

        //
        // methods
        //

        #region public override IPDFLayoutEngine GetEngine(IPDFLayoutEngine parent, PDFLayoutContext context, Styles.PDFStyle style)

        /// <summary>
        /// Override the base implementation to return a PageGroupLayoutEngine
        /// </summary>
        /// <param name="parent"></param>
        /// <param name="context"></param>
        /// <param name="style"></param>
        /// <returns></returns>
        public override IPDFLayoutEngine GetEngine(IPDFLayoutEngine parent, PDFLayoutContext context, Styles.PDFStyle style)
        {
            return new PageGroupLayoutEngine(this, parent, context, style);
        }

        #endregion

        #region public override PDFComponentArrangement GetFirstArrangement()

        /// <summary>
        /// Override the GetFirstArrangement to return the arrangement of the first page
        /// </summary>
        /// <returns></returns>
        public override PDFComponentArrangement GetFirstArrangement()
        {
            if (this.Pages.Count > 0)
                return this.Pages[0].GetFirstArrangement();
            else
                return null;
        }

        #endregion

        //
        // inner classes
        //

        #region private class PageGroupLayoutEngine : IPDFLayoutEngine

        private class PageGroupLayoutEngine : IPDFLayoutEngine
        {
            private PDFPageGroup _group;
            private IPDFLayoutEngine _parent;
            private PDFLayoutContext _context;
            private PDFStyle _full;

            public PageGroupLayoutEngine(PDFPageGroup group, IPDFLayoutEngine parent, PDFLayoutContext context, PDFStyle full)
            {
                if (null == group)
                    throw new ArgumentNullException("group");

                this._group = group;
                this._parent = parent;
                this._context = context;
                this._full = full;
            }


            public PDFLayoutContext Context
            {
                get { return _context ; }
            }

            public void Layout(PDFLayoutContext context, Styles.PDFStyle fullstyle)
            {
                foreach (PDFPage pg in this._group.Pages)
                {
                    this.LayoutAPage(pg);
                }
            }

            public void LayoutAPage(PDFPage pg)
            {
                PDFStyle style = pg.GetAppliedStyle();

                if (null != style)
                    this.Context.StyleStack.Push(style);

                PDFStyle full = this.Context.StyleStack.GetFullStyle(pg);

                PDFArtefactRegistrationSet artefacts = pg.RegisterLayoutArtefacts(this.Context, full);

                using (IPDFLayoutEngine engine = pg.GetEngine(this, this.Context, full))
                {
                    if (this.Context.TraceLog.ShouldLog(TraceLevel.Message))
                        this.Context.TraceLog.Begin(TraceLevel.Message, "Page Group Layout", "Starting the layout of page " + pg.ID);

                    engine.Layout(this.Context, full);

                    if (this.Context.TraceLog.ShouldLog(TraceLevel.Message))
                        this.Context.TraceLog.End(TraceLevel.Message, "Page Group Layout", "Completed the layout of page " + pg.ID + " and now on page index " + this.Context.DocumentLayout.CurrentPageIndex.ToString());
                }

                if (null != artefacts)
                    pg.CloseLayoutArtefacts(this.Context, artefacts, full);

                if (null != style)
                    this.Context.StyleStack.Pop();
            }

            public bool MoveToNextPage(Stack<Layout.PDFLayoutBlock> depth, ref Layout.PDFLayoutRegion region, ref Layout.PDFLayoutBlock block)
            {
                return this._parent.MoveToNextPage(depth, ref region, ref block);
            }

            public Layout.PDFLayoutBlock CloseCurrentBlockAndStartNewInRegion(Layout.PDFLayoutBlock blockToClose, Layout.PDFLayoutRegion joinToRegion)
            {
                return _parent.CloseCurrentBlockAndStartNewInRegion(blockToClose, joinToRegion);
            }

            public void Dispose()
            {
                
            }
        }

        #endregion
    }
}
