﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using Scryber.Native;

namespace Scryber.Components
{
    [PDFParsableComponent("PageNumber")]
    public class PDFPageNumberLabel : PDFTextBase
    {

        #region public string PageLabelFormat 

        private string _format;
        /// <summary>
        /// Gets or sets the format of the string to be displayed. Using the indexes of the DisplayFields
        /// </summary>
        [PDFAttribute("format")]
        public string PageLabelFormat 
        { 
            get { return this._format; }
            set { this._format = value; }
        }

        #endregion

        #region public int TotalPageCountHint {get;set;}

        private int _pgcounthint = 99;

        /// <summary>
        /// A unique string representation of a string that is similar in 
        /// size to the total page count so that the length can be measured during layout.
        /// </summary>
        /// <remarks>Becasue we do not know the total page count until layout has completed 
        /// we need something to tell us what length the total page count will be. 
        /// This is the hint and the default value is %%</remarks>
        [PDFAttribute("page-count-hint")]
        public int TotalPageCountHint
        {
            get { return _pgcounthint; }
            set { _pgcounthint = value; }
        }

        #endregion

        //local reference to the layout document
        private Layout.PDFLayoutDocument _doc;

        //local value for the page index of this label
        private int _renderpageindex = -1;

        

        protected override string BaseText
        {
            get
            {
                return this.GetDisplayText(false);
            }
            set
            {
                throw new InvalidOperationException(Errors.CannotSetBaseTextOfPageNumber);
                //Do Nothing
            }
        }


        public PDFPageNumberLabel()
            : base(PDFObjectTypes.Text)
        {
        }


        protected override Text.PDFTextReader CreateReader(PDFLayoutContext context, Styles.PDFStyle style)
        {
            _doc = context.DocumentLayout;
            return base.CreateReader(context, style);
        }

        /// <summary>
        /// Once layout is comlete then we can replace the text that was used when not rendering
        /// with the text that was.
        /// </summary>
        internal override void RegisterLayoutComplete(PDFLayoutContext context)
        {
            base.RegisterLayoutComplete(context);
            string text = this.GetDisplayText(this._renderpageindex, true);

            //We assume there is only one span on one line in one paragraph
            if (null != this.TextLayout)
            {
                Text.PDFTextLayout layout = this.TextLayout;
                Text.TextBlock block = layout.InnerBlock;
                foreach (Text.TextParagraph para in block)
                {
                    if (para.LineCount == 1)
                    {
                        foreach (Text.TextLine line in para)
                        {
                            foreach (Text.TextSpan span in line)
                            {
                                span.Characters = text;
                                break;
                            }
                            break;
                        }
                    }
                    else
                    {
                        foreach (Text.TextLine line in para)
                        {
                            foreach (Text.TextSpan span in line)
                            {
                                int len = Math.Min(text.Length, span.Characters.Length);
                                string toreplace = text.Substring(0, len);
                                span.Characters = toreplace;
                                text = text.Substring(len);
                                break;
                            }
                            if (text.Length == 0)
                                break;
                        }
                    }
                    break;
                }
            }
        }


        #region private string GetDisplayText(bool rendering)

        /// <summary>
        /// Gets the full text for the page label
        /// </summary>
        /// <param name="context"></param>
        /// <returns></returns>
        /// <remarks>
        /// 1. If there is no PageFormat, then the current page index in the label style for this page is used
        /// 2. If there is a format then the format is used with the following fields...
        ///                     {0} = page label, {1} = total label for this pages numbering group, 
        ///                     {2} global page index, {3} global page count
        /// </remarks>
        private string GetDisplayText(bool rendering)
        {
            
            int pageindex = _doc.CurrentPageIndex;
            this._renderpageindex = pageindex;
            return this.GetDisplayText(pageindex, rendering);
        }

        private string GetDisplayText(int pageindex, bool rendering)
        {
            string total;
            

            int sectionindex, sectioncount;
            PDFPageNumbering num = this._doc.GetNumbering(pageindex, out sectionindex, out sectioncount);
            if (rendering)
                total = _doc.TotalPageCount.ToString();
            else
            {
                total = this.TotalPageCountHint.ToString();
                sectioncount = this.TotalPageCountHint;
            }
            pageindex++; // convert to '1' based index
            string numname = num.GetPageNumber(sectionindex);
            string sectname = num.GetPageNumber(sectioncount);

            if (string.IsNullOrEmpty(this.PageLabelFormat))
                return numname;
            else
                return String.Format(this.PageLabelFormat, numname, sectname, pageindex, total);
            
            
        }

        #endregion

        

        #region private string GetDisplayNameValue(int index, int total, PDFPageNumbering num, PageNumberDisplay numtype)
        
        /// <summary>
        /// Gets the actual string value for an individual PageNumberDisplay field
        /// </summary>
        /// <param name="index"></param>
        /// <param name="total"></param>
        /// <param name="num"></param>
        /// <param name="numtype"></param>
        /// <returns></returns>
        private string GetDisplayNameValue(int index, int total, PDFPageNumbering num, PageNumberDisplay numtype)
        {
            string val;
            switch (numtype)
            {
                case PageNumberDisplay.PageIndex:
                    val = index.ToString();
                    break;
                case PageNumberDisplay.PageCount:
                    val = total.ToString();
                    break;
                case PageNumberDisplay.PageIndexLabel:
                    val = num.GetPageNumber(index);
                    break;
                case PageNumberDisplay.PageCountLabel:
                    val = num.GetPageNumber(total);
                    break;
                case PageNumberDisplay.SectionOffset:
                    val = this.Page.GetPageIndex(this).ToString();
                    break;
                case PageNumberDisplay.SectionOffsetLabel:
                    val = num.GetPageNumber(this.Page.GetPageIndex(this));
                    break;
                case PageNumberDisplay.SectionCount:
                case PageNumberDisplay.SectionCountLabel:
                default:
                    //TODO: Support the declaration of sections and their labels in PageLabel
                    throw new NotSupportedException(numtype.ToString());
            }
            return val;
        }

        #endregion

    }
}
