﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Scryber.Native;

namespace Scryber.Components
{
    /// <summary>
    /// Renders the page index of a referenced component.
    /// </summary>
    [PDFParsableComponent("PageOf")]
    public class PDFPageOfLabel: PDFTextBase
    {
        private PDFComponent _found;
        private Layout.PDFLayoutDocument _doc;
        private bool _haspagenumber = false;
        private const string PageReplacementText = "%%";
        private string _toreplace = null; //the text to replace at the end of the layout

        #region protected override string BaseText {get; set = Exception;}

        /// <summary>
        /// Overrides the base implementation to return the display text for the page the referenced component is on.
        /// Cannot be set.
        /// </summary>
        protected override string BaseText
        {
            get
            {
                _toreplace = this.GetDisplayText();
                return _toreplace;
            }
            set
            {
                throw new InvalidOperationException(Errors.CannotSetBaseTextOfPageNumber);
                //Do Nothing
            }
        }

        #endregion

        #region public string ComponentName

        private string _compname;

        /// <summary>
        /// Gets or sets the name or ID of component to get the page for. To specify an ID prefix with #
        /// </summary>
        [PDFAttribute("component")]
        public string ComponentName
        {
            get { return _compname; }
            set { _compname = value; }
        }

        #endregion

        #region public string NotFoundText {get;set;}

        private string _notfound = string.Empty;

        /// <summary>
        /// Gets or sets the text that is used if the referenced component
        /// is not found (Lax only) or does not have an arrangement
        /// </summary>
        [PDFAttribute("not-found-replacement")]
        public string NotFoundText
        {
            get { return _notfound; }
            set { _notfound = value; }
        }

        #endregion

        #region public string PageLabelFormat {get;set;}

        private string _format = string.Empty;

        /// <summary>
        /// Gets or Sets the format of the label 
        /// </summary>
        [PDFAttribute("label-format")]
        public string PageLabelFormat
        {
            get { return _format; }
            set { _format = value; }
        }

        #endregion

        //
        // ctor(s)
        //

        public PDFPageOfLabel()
            : base((PDFObjectType)"pgof")
        {
        }


        #region private string GetDisplayText()

        /// <summary>
        /// Gets the full text for the page label that the referenced component is placed on
        /// </summary>
        /// <param name="context"></param>
        /// <returns></returns>
        private string GetDisplayText()
        {
            string empty = this.NotFoundText;
            if (string.IsNullOrEmpty(empty))
                empty = PageReplacementText;

            //This is caled when we begin to build the text block
            string pg = GetPageNumberString(empty);
            

            //If we have a format then apply it.
            if (!string.IsNullOrEmpty(this.PageLabelFormat))
                pg = string.Format(this.PageLabelFormat, pg);
            return pg;
        }

        private string GetPageNumberString(string emptyreplacement)
        {
            string pg = emptyreplacement;
            int index = -1;
            
            PDFComponentArrangement arrange = null;
            if (null != _found)
            {
                arrange = _found.GetFirstArrangement();
                if (null == arrange)
                {
                    index = _found.PageLayoutIndex;
                }
                else
                {
                    index = arrange.PageIndex;
                }
            }

            if(index >= 0)
            {
                int offset, max;
                PDFPageNumbering nums = _doc.GetNumbering(index, out offset, out max);
                if (null == nums)
                    pg = (index + 1).ToString();
                else
                    pg = nums.GetPageNumber(offset);
                _haspagenumber = true;
            }
            

            return pg;
        }

        #endregion

        #region protected override void DoCloseLayoutArtefacts(PDFLayoutContext context, PDFArtefactRegistrationSet artefacts, Styles.PDFStyle fullstyle)

        /// <summary>
        /// Overrides base implementation to lookup components required for the rendering of a page locator label
        /// </summary>
        /// <param name="context"></param>
        /// <param name="artefacts"></param>
        /// <param name="fullstyle"></param>
        protected override void DoCloseLayoutArtefacts(PDFLayoutContext context, PDFArtefactRegistrationSet artefacts, Styles.PDFStyle fullstyle)
        {
            if (!string.IsNullOrEmpty(this.ComponentName))
            {
                LookupExternalComponent(context, this.ComponentName);
            }
            base.DoCloseLayoutArtefacts(context, artefacts, fullstyle);
        }

        #endregion

        /// <summary>
        /// We know that all components now have the blocks laid out and complete
        /// </summary>
        internal override void RegisterLayoutComplete(PDFLayoutContext context)
        {
            base.RegisterLayoutComplete(context);
            if (null != this.TextLayout)
            {
                string actual = GetDisplayText();
                Text.PDFTextLayout layout = this.TextLayout;
                Text.TextBlock block = layout.InnerBlock;
                foreach (Text.TextParagraph para in block)
                {
                    foreach (Text.TextLine line in para)
                    {
                        foreach (Text.TextSpan span in line)
                        {
                            if (span.Characters.IndexOf(_toreplace) > -1)
                            {
                                //This time we call the GetPageNumberString with the NotFoundText.
                                span.Characters = span.Characters = actual;
                                break;
                            }
                        }
                    }
                }
            }
        }

        

        #region private void LookupExternalComponent(PDFLayoutContext context, string name)

        /// <summary>
        /// Looks for the component with the specified name or ID and sets instance variables appropriately
        /// </summary>
        /// <param name="context"></param>
        /// <param name="name"></param>
        private void LookupExternalComponent(PDFLayoutContext context, string name)
        {
            PDFComponent comp;

            if (name.StartsWith(Const.ComponentIDPrefix))
            {
                name = name.Substring(Const.ComponentIDPrefix.Length);
                comp = this.Document.FindAComponentById(name);

                if (null == comp && string.IsNullOrEmpty(this.NotFoundText))
                {
                    if (context.Conformance == ConformanceMode.Strict)
                        throw new PDFException(string.Format(Errors.CouldNotFindControlWithID, name));
                    else
                        context.TraceLog.Add(TraceLevel.Error, "Page Locator", string.Format(Errors.CouldNotFindControlWithID, this.ComponentName));

                }
            }
            else
            {
                comp = Document.FindAComponentByName(this.ComponentName);

                if (null == comp && string.IsNullOrEmpty(this.NotFoundText))
                {
                    if (context.Conformance == ConformanceMode.Strict)
                        throw new PDFException(string.Format(Errors.CouldNotFindControlWithName, name));
                    else
                        context.TraceLog.Add(TraceLevel.Error, "Page Locator", string.Format(Errors.CouldNotFindControlWithName, this.ComponentName));

                }
            }

            if (null != comp)
            {
                this._found = comp;
                this._doc = context.DocumentLayout;
            }
            else
            {
                _found = null;
                _doc = null;
            }
        }

        #endregion
    }
}
