﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using Scryber.Styles;
using Scryber.Drawing;
using Scryber.Native;

namespace Scryber.Components
{
    /// <summary>
    /// Represents a complete row in a table grid
    /// </summary>
    [PDFParsableComponent("Row")]
    public class PDFTableRow : PDFContainerComponent, IPDFStyledComponent
    {

        //
        // style attributes - row does not support positioning attributes
        //


        #region public PDFColor BackgroundColor

        /// <summary>
        /// Gets or sets the background color of this component
        /// </summary>
        [PDFAttribute("bg-color", Const.PDFStylesNamespace)]
        public PDFColor BackgroundColor
        {
            get
            {
                PDFBackgroundStyle bg;
                if (this.HasStyle && this.Style.TryGetBackground(out bg))
                    return bg.Color;
                else
                    return PDFColors.Transparent;
            }
            set
            {
                this.Style.Background.Color = value;
            }
        }

        #endregion

        #region public string BackgroundImage

        /// <summary>
        /// Gets or sets the background image of this component
        /// </summary>
        [PDFAttribute("bg-image", Const.PDFStylesNamespace)]
        public string BackgroundImage
        {
            get
            {
                PDFBackgroundStyle bg;
                if (this.HasStyle && this.Style.TryGetBackground(out bg))
                    return bg.ImageSource;
                else
                    return string.Empty;
            }
            set
            {
                this.Style.Background.ImageSource = value;
            }
        }

        #endregion

        #region public PatternRepeat BackgroundRepeat

        /// <summary>
        /// Gets or sets the background image repeat of this component
        /// </summary>
        [PDFAttribute("bg-repeat", Const.PDFStylesNamespace)]
        public PatternRepeat BackgroundRepeat
        {
            get
            {
                PDFBackgroundStyle bg;
                if (this.HasStyle && this.Style.TryGetBackground(out bg))
                    return bg.PatternRepeat;
                else
                    return PatternRepeat.RepeatBoth;
            }
            set
            {
                this.Style.Background.PatternRepeat = value;
            }
        }

        #endregion

        #region public PDFReal BackgroundOpacity

        /// <summary>
        /// Gets or sets the background fill color opacity of this component
        /// </summary>
        [PDFAttribute("bg-opacity", Const.PDFStylesNamespace)]
        public PDFReal BackgroundOpacity
        {
            get
            {
                PDFBackgroundStyle bg;
                if (this.HasStyle && this.Style.TryGetBackground(out bg))
                    return bg.Opacity;
                else
                    return PDFReal.Zero;
            }
            set
            {
                this.Style.Background.Opacity = value;
            }
        }

        #endregion

        #region public PDFUnit BorderWidth

        /// <summary>
        /// Gets or sets the border width of this component
        /// </summary>
        [PDFAttribute("border-width", Const.PDFStylesNamespace)]
        public PDFUnit BorderWidth
        {
            get
            {
                PDFBorderStyle b;
                if (this.HasStyle && this.Style.TryGetBorder(out b))
                    return b.Width;
                else
                    return PDFUnit.Empty;
            }
            set
            {
                this.Style.Border.Width = value;
            }
        }

        #endregion

        #region public PDFColor BorderColor

        /// <summary>
        /// Gets or sets the border color of this component
        /// </summary>
        [PDFAttribute("border-color", Const.PDFStylesNamespace)]
        public PDFColor BorderColor
        {
            get
            {
                PDFBorderStyle b;
                if (this.HasStyle && this.Style.TryGetBorder(out b))
                    return b.Color;
                else
                    return PDFColors.Transparent;
            }
            set
            {
                this.Style.Border.Color = value;
            }
        }

        #endregion

        #region public PDFDash BorderDashPattern

        /// <summary>
        /// Gets or sets the border dash pattern of this component
        /// </summary>
        [PDFAttribute("border-dash", Const.PDFStylesNamespace)]
        public PDFDash BorderDashPattern
        {
            get
            {
                PDFBorderStyle b;
                if (this.HasStyle && this.Style.TryGetBorder(out b))
                    return b.Dash;
                else
                    return PDFDash.None;
            }
            set
            {
                this.Style.Border.Dash = value;
            }
        }

        #endregion

        #region public double BorderOpacity

        /// <summary>
        /// Gets or sets the border opacity of this component
        /// </summary>
        [PDFAttribute("border-opacity", Const.PDFStylesNamespace)]
        public double BorderOpacity
        {
            get
            {
                PDFBorderStyle b;
                if (this.HasStyle && this.Style.TryGetBorder(out b))
                    return b.Opacity;
                else
                    return 1;
            }
            set
            {
                this.Style.Border.Opacity = value;
            }
        }

        #endregion

        #region public Sides BorderSides

        /// <summary>
        /// Gets or sets the border sides of this component
        /// </summary>
        [PDFAttribute("border-sides", Const.PDFStylesNamespace)]
        public Sides BorderSides
        {
            get
            {
                PDFBorderStyle b;
                if (this.HasStyle && this.Style.TryGetBorder(out b))
                    return b.Sides;
                else
                    return (Sides)0;
            }
            set
            {
                this.Style.Border.Sides = value;
            }
        }

        #endregion

        #region public PDFUnit BorderCornerRadius

        /// <summary>
        /// Gets or sets the border corner radii of this component
        /// </summary>
        [PDFAttribute("border-corner-radius", Const.PDFStylesNamespace)]
        public PDFUnit BorderCornerRadius
        {
            get
            {
                PDFBorderStyle b;
                if (this.HasStyle && this.Style.TryGetBorder(out b))
                    return b.CornerRadius;
                else
                    return PDFUnit.Zero;
            }
            set
            {
                this.Style.Border.CornerRadius = value;
            }
        }

        #endregion

        #region public PDFColor FillColor

        /// <summary>
        /// Gets or sets the Fill color of this component
        /// </summary>
        [PDFAttribute("fill-color", Const.PDFStylesNamespace)]
        public PDFColor FillColor
        {
            get
            {
                PDFFillStyle bg;
                if (this.HasStyle && this.Style.TryGetFill(out bg))
                    return bg.Color;
                else
                    return PDFColors.Transparent;
            }
            set
            {
                this.Style.Fill.Color = value;
            }
        }

        #endregion

        #region public string FillImage

        /// <summary>
        /// Gets or sets the Fill image of this component
        /// </summary>
        [PDFAttribute("fill-image", Const.PDFStylesNamespace)]
        public string FillImage
        {
            get
            {
                PDFFillStyle bg;
                if (this.HasStyle && this.Style.TryGetFill(out bg))
                    return bg.ImageSource;
                else
                    return string.Empty;
            }
            set
            {
                this.Style.Fill.ImageSource = value;
            }
        }

        #endregion

        #region public PatternRepeat FillRepeat

        /// <summary>
        /// Gets or sets the Fill image repeat of this component
        /// </summary>
        [PDFAttribute("fill-repeat", Const.PDFStylesNamespace)]
        public PatternRepeat FillRepeat
        {
            get
            {
                PDFFillStyle bg;
                if (this.HasStyle && this.Style.TryGetFill(out bg))
                    return bg.PatternRepeat;
                else
                    return PatternRepeat.RepeatBoth;
            }
            set
            {
                this.Style.Fill.PatternRepeat = value;
            }
        }

        #endregion

        #region public PDFReal FillOpacity

        /// <summary>
        /// Gets or sets the Fill color opacity of this component
        /// </summary>
        [PDFAttribute("fill-opacity", Const.PDFStylesNamespace)]
        public PDFReal FillOpacity
        {
            get
            {
                PDFFillStyle bg;
                if (this.HasStyle && this.Style.TryGetFill(out bg))
                    return bg.Opacity;
                else
                    return PDFReal.Zero;
            }
            set
            {
                this.Style.Fill.Opacity = value;
            }
        }

        #endregion

        #region public PDFUnit StrokeWidth

        /// <summary>
        /// Gets or sets the Stroke width of this component
        /// </summary>
        [PDFAttribute("stroke-width", Const.PDFStylesNamespace)]
        public PDFUnit StrokeWidth
        {
            get
            {
                PDFStrokeStyle b;
                if (this.HasStyle && this.Style.TryGetStroke(out b))
                    return b.Width;
                else
                    return PDFUnit.Empty;
            }
            set
            {
                this.Style.Stroke.Width = value;
            }
        }

        #endregion

        #region public PDFColor StrokeColor

        /// <summary>
        /// Gets or sets the Stroke color of this component
        /// </summary>
        [PDFAttribute("stroke-color", Const.PDFStylesNamespace)]
        public PDFColor StrokeColor
        {
            get
            {
                PDFStrokeStyle b;
                if (this.HasStyle && this.Style.TryGetStroke(out b))
                    return b.Color;
                else
                    return PDFColors.Transparent;
            }
            set
            {
                this.Style.Stroke.Color = value;
            }
        }

        #endregion

        #region public PDFDash StrokeDashPattern

        /// <summary>
        /// Gets or sets the Stroke dash pattern of this component
        /// </summary>
        [PDFAttribute("stroke-dash", Const.PDFStylesNamespace)]
        public PDFDash StrokeDashPattern
        {
            get
            {
                PDFStrokeStyle b;
                if (this.HasStyle && this.Style.TryGetStroke(out b))
                    return b.Dash;
                else
                    return PDFDash.None;
            }
            set
            {
                this.Style.Stroke.Dash = value;
            }
        }

        #endregion

        #region public double StrokeOpacity

        /// <summary>
        /// Gets or sets the Stroke opacity of this component
        /// </summary>
        [PDFAttribute("stroke-opacity", Const.PDFStylesNamespace)]
        public double StrokeOpacity
        {
            get
            {
                PDFStrokeStyle b;
                if (this.HasStyle && this.Style.TryGetStroke(out b))
                    return b.Opacity;
                else
                    return 1;
            }
            set
            {
                this.Style.Stroke.Opacity = value;
            }
        }

        #endregion

        #region public string FontFamily

        /// <summary>
        /// Gets or sets the Font Family of this component
        /// </summary>
        [PDFAttribute("font-family", Const.PDFStylesNamespace)]
        public string FontFamily
        {
            get
            {
                PDFFontStyle b;
                if (this.HasStyle && this.Style.TryGetFont(out b))
                    return b.FontFamily;
                else
                    return string.Empty;
            }
            set
            {
                this.Style.Font.FontFamily = value;
            }
        }

        #endregion

        #region public string FontSize

        /// <summary>
        /// Gets or sets the Font Size of this component
        /// </summary>
        [PDFAttribute("font-size", Const.PDFStylesNamespace)]
        public PDFUnit FontSize
        {
            get
            {
                PDFFontStyle b;
                if (this.HasStyle && this.Style.TryGetFont(out b))
                    return b.FontSize;
                else
                    return PDFUnit.Zero;
            }
            set
            {
                this.Style.Font.FontSize = value;
            }
        }

        #endregion

        #region public bool FontBold

        /// <summary>
        /// Gets or sets the Font bold value of this component
        /// </summary>
        [PDFAttribute("font-bold", Const.PDFStylesNamespace)]
        public bool FontBold
        {
            get
            {
                PDFFontStyle b;
                if (this.HasStyle && this.Style.TryGetFont(out b))
                    return b.FontBold;
                else
                    return false;
            }
            set
            {
                this.Style.Font.FontBold = value;
            }
        }

        #endregion

        #region public bool FontItalic

        /// <summary>
        /// Gets or sets the Font italic value of this component
        /// </summary>
        [PDFAttribute("font-italic", Const.PDFStylesNamespace)]
        public bool FontItalic
        {
            get
            {
                PDFFontStyle b;
                if (this.HasStyle && this.Style.TryGetFont(out b))
                    return b.FontItalic;
                else
                    return false;
            }
            set
            {
                this.Style.Font.FontItalic = value;
            }
        }

        #endregion

        //
        // end of style attributes
        //


        #region PDFTableCellList Cells {get;}

        private PDFTableCellList _cells;

        /// <summary>
        /// Gets the list of cells in this row. Returned value will never be null
        /// </summary>
        [PDFArray(typeof(PDFTableCell))]
        [PDFElement("")]
        public PDFTableCellList Cells
        {
            get 
            {
                if (_cells == null)
                    _cells = new PDFTableCellList(this.InnerContent);

                return _cells;
            }
        }

        #endregion

        #region public PDFTable ContainingTable {get;}

        /// <summary>
        /// Gets the table that contains this Row (or null)
        /// </summary>
        /// <remarks>We cannot directly return the parent as this row 
        /// could be within a binding invisible container. So we must walk up the hierarchy 
        /// until we find a table (or not a table).</remarks>
        public PDFTableGrid ContainingTable
        {
            get
            {
                PDFComponent Component = this.Parent;
                
                while (Component != null)
                {
                    if (Component is PDFTableGrid)
                        return Component as PDFTableGrid;
                    else if (Component is IPDFInvisibleContainer)
                        Component = Component.Parent;
                    else
                        Component = null;
                }
                return null; //not found.
            }
        }

        #endregion

        #region public PDFStyle Style {get;set;} + bool HasStyle {get;}

        private PDFStyle _style;

        /// <summary>
        /// Gets or sets the style data for this TableRow
        /// </summary>
        public PDFStyle Style
        {
            get
            {
                if (null == _style)
                    _style = new PDFStyle();
                return _style;
            }
            set
            {
                _style = value;
            }
        }

        /// <summary>
        /// Returns true if this row has any style data associated with it.
        /// </summary>
        public bool HasStyle
        {
            get { return null != _style && _style.HasItems; }
        }

        #endregion


        #region public string OutlineTitle {get;set;}

        /// <summary>
        /// Gets or sets the title of the outline for this component.
        /// </summary>
        [PDFAttribute("outline-title")]
        public string OutlineTitle
        {
            get
            {
                if (this.HasOutline)
                    return this.Outline.Title;
                else
                    return string.Empty;
            }
            set
            {
                if (!this.HasOutline && string.IsNullOrEmpty(value))
                {
                    //Nothing to do as we don't have an outline and the value is empty
                    return;
                }
                this.Outline.Title = value;
            }
        }

        #endregion


        #region .ctor() + .ctor(PDFObjectType)

        /// <summary>
        /// Creates a new instance of the table row
        /// </summary>
        public PDFTableRow()
            : this(PDFObjectTypes.TableRow)
        {
        }


        /// <summary>
        /// Protected constructor that sub classes 
        /// can use to create an instance of their class using a different ObjectType
        /// </summary>
        /// <param name="type">The type identifier</param>
        protected PDFTableRow(PDFObjectType type)
            : base(type)
        {

        }

        #endregion

        //
        // override methods
        //

        #region protected override Styles.PDFStyle GetBaseStyle()

        /// <summary>
        /// Overrides the base method to apply the overflow split of never to the base style
        /// so they will not break apart.
        /// </summary>
        /// <returns></returns>
        protected override Styles.PDFStyle GetBaseStyle()
        {
            PDFStyle style = base.GetBaseStyle();
            style.Overflow.Split = Drawing.OverflowSplit.Never;
            return style;
        }

        #endregion

        #region public override PDFStyle GetAppliedStyle(PDFComponent forComponent, PDFStyle baseStyle)

        /// <summary>
        /// Overrides the base method to remove any inappropriate styles before returning. 
        /// The Grid row does not support position, margin and padding options.
        /// </summary>
        /// <param name="forComponent"></param>
        /// <param name="baseStyle"></param>
        /// <returns></returns>
        protected override void MergeDeclaredStyles(PDFStyle applied)
        {
            base.MergeDeclaredStyles(applied);
            this.RemoveInapplicableStyles(applied);
        }

        /// <summary>
        /// Removes all the styles that cannot be applied to a row explicitly
        /// </summary>
        /// <param name="applied"></param>
        private void RemoveInapplicableStyles(PDFStyle applied)
        {
            bool modified = false;
            PDFMarginsStyle margins;
            if (applied.TryGetMargins(out margins))
            {
                applied.Remove(margins.Type);
                modified = true;
            }
            PDFPaddingStyle padding;
            if (applied.TryGetPadding(out padding))
            {
                applied.Remove(padding.Type);
                modified = true;
            }
            PDFPositionStyle pos;
            if (applied.TryGetPosition(out pos))
            {
                pos.RemovePositionMode();
                pos.RemoveHeight();
                pos.RemoveWidth();
                pos.RemoveX();
                pos.RemoveY();
                modified = true;
            }



            if (modified)
            {
                PDFTraceLog log = PDFTraceContext.GetLog();
                if (log.ShouldLog(TraceLevel.Message))
                    log.Add(TraceLevel.Message, "PDFTableRow", "Removed all unsupported Margins, Padding and Postion style items that are not supported on a table row");
            }
        }

        #endregion
    }




    /// <summary>
    /// Represents a complete header row in table grid. 
    /// Overrides the default style to return an emboldened font style
    /// </summary>
    [PDFParsableComponent("Header-Row")]
    public class PDFTableHeaderRow : PDFTableRow
    {

        #region .ctor() and .ctor(PDFObjectType)

        /// <summary>
        /// Creates a new header row
        /// </summary>
        public PDFTableHeaderRow()
            : this(PDFObjectTypes.TableHeaderRow)
        {
        }

        /// <summary>
        /// Creates a new header row with the specified object type value
        /// </summary>
        /// <param name="type"></param>
        public PDFTableHeaderRow(PDFObjectType type)
            : base(type)
        {
        }

        #endregion

        #region protected override Styles.PDFStyle GetBaseStyle()

        /// <summary>
        /// Overrides the base style to apply a bold font to any text in the header
        /// </summary>
        /// <returns></returns>
        protected override Styles.PDFStyle GetBaseStyle()
        {
            Styles.PDFStyle style = base.GetBaseStyle();
            style.Font.FontBold = true;
            return style;
        }

        #endregion
    }




    /// <summary>
    /// Represents a complete footer row in a table grid
    /// </summary>
    [PDFParsableComponent("Footer-Row")]
    public class PDFTableFooterRow : PDFTableRow
    {

        #region .ctor() and .ctor(PDFObjectType)

        /// <summary>
        /// Creates a new instance of a table row
        /// </summary>
        public PDFTableFooterRow()
            : this(PDFObjectTypes.TableFooterRow)
        {
        }

        /// <summary>
        /// Creates a new instance of a table row with the specified object type
        /// </summary>
        /// <param name="type">The object type of this instance</param>
        public PDFTableFooterRow(PDFObjectType type)
            : base(type)
        {
        }

        #endregion

    }





    /// <summary>
    /// A list of table rows. As this is a component wrapping list, if there are inner invisible items, 
    /// they will also be enumerated over to be included in the collection
    /// </summary>
    public class PDFTableRowList : PDFComponentWrappingList<PDFTableRow>
    {

        #region .ctor(PDFComponentList)

        /// <summary>
        /// Creates the list of rows based on the provided component list. 
        /// </summary>
        /// <param name="inner"></param>
        public PDFTableRowList(PDFComponentList inner)
            : base(inner)
        {
        }

        #endregion

    }
}
