﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using Scryber.Native;
using System.Drawing;
using Scryber.Styles;
using Scryber.Resources;
using Scryber.Drawing;

namespace Scryber.Components
{
    public abstract class PDFVisualComponent : PDFContainerComponent, IPDFVisualComponent, IPDFStyledComponent
    {

        #region public PDFStyle Style {get;set;} + public bool HasStyle{get;}

        private PDFStyle _style;

        /// <summary>
        /// Gets the applied style for this page Component
        /// </summary>
        [PDFElement("Style")]
        public PDFStyle Style
        {
            get 
            {
                if (_style == null)
                    _style = new PDFStyle();
                return _style; 
            }
        }

        

        /// <summary>
        /// Gets the flag to indicate if this page Component has style 
        /// information associated with it.
        /// </summary>
        public bool HasStyle
        {
            get { return this._style != null && this._style.HasItems; }
        }

        #endregion

        #region public PDFUnit X {get;set;} + public bool HasX {get;}

        /// <summary>
        /// Gets or Sets the X (Horizontal) position of this page Component
        /// </summary>
        [PDFAttribute("x", Const.PDFStylesNamespace)]
        public PDFUnit X
        {
            get
            {
                return this.Style.Position.X;
            }
            set
            {
                this.Style.Position.X = value;
            }
        }

        /// <summary>
        /// Gets the flag to identify if the X position has been set for this Page Component
        /// </summary>
        public bool HasX
        {
            get
            {
                PDFPositionStyle pos;
                return this._style != null && this._style.TryGetPosition(out pos) && pos.IsDefined(StyleKeys.XAttr);
            }
        }

        #endregion

        #region public PDFUnit Y {get;set;} + public bool HasY {get;}

        /// <summary>
        /// Gets or sets the Y (vertical) position of the Page Component
        /// </summary>
        [PDFAttribute("y", Const.PDFStylesNamespace)]
        public PDFUnit Y
        {
            get
            {
                return this.Style.Position.Y;
            }
            set
            {
                this.Style.Position.Y = value;
            }
        }

        /// <summary>
        /// Gets the flag to identifiy is the Y value has been set on this page Component
        /// </summary>
        public bool HasY
        {
            get
            {
                PDFPositionStyle pos;
                return this._style != null && this._style.TryGetPosition(out pos) && pos.IsDefined(StyleKeys.YAttr);
            }
        }

        #endregion

        #region public PDFUnit Width {get;set;} + public bool HasWidth {get;}

        /// <summary>
        /// Gets or Sets the Width of this page Component
        /// </summary>
        [PDFAttribute("width", Const.PDFStylesNamespace)]
        public PDFUnit Width
        {
            get
            {
                return this.Style.Position.Width;
            }
            set
            {
                this.Style.Position.Width = value;
            }
        }

        

        /// <summary>
        /// Gets the flag to identify if the Width has been set for this Page Component
        /// </summary>
        public bool HasWidth
        {
            get
            {
                PDFPositionStyle pos;
                return this._style != null && this._style.TryGetPosition(out pos) && pos.IsDefined(StyleKeys.WidthAttr);
            }
        }

        #endregion

        #region public PDFUnit Height {get;set;} + public bool HasHeight {get;}

        /// <summary>
        /// Gets or sets the Height of the Page Component
        /// </summary>
        [PDFAttribute("height", Const.PDFStylesNamespace)]
        public PDFUnit Height
        {
            get
            {
                return this.Style.Position.Height;
            }
            set
            {
                this.Style.Position.Height = value;
            }
        }

        /// <summary>
        /// Gets the flag to identifiy is the Height has been set on this page Component
        /// </summary>
        public bool HasHeight
        {
            get
            {
                PDFPositionStyle pos;
                return this._style != null && this._style.TryGetPosition(out pos) && pos.IsDefined(StyleKeys.HeightAttr);
            }
        }

        #endregion

        //
        // Style qualified attributes
        //

        #region public PDFThickness Margins

        /// <summary>
        /// Gets or sets the margins for this component
        /// </summary>
        [PDFAttribute("margins",Const.PDFStylesNamespace)]
        public PDFThickness Margins
        {
            get
            {
                PDFMarginsStyle margins;
                if (this.HasStyle && this.Style.TryGetMargins(out margins))
                    return margins.GetThickness();
                else
                    return PDFThickness.Empty();
            }
            set
            {
                this.Style.Margins.SetThickness(value);
            }
        }

        #endregion

        #region public PDFThickness Padding

        /// <summary>
        /// Gets or sets the padding on this component
        /// </summary>
        [PDFAttribute("padding",Const.PDFStylesNamespace)]
        public PDFThickness Padding
        {
            get
            {
                PDFPaddingStyle padding;
                if (this.HasStyle && this.Style.TryGetPadding(out padding))
                    return padding.GetThickness();
                else
                    return PDFThickness.Empty();
            }
            set
            {
                this.Style.Padding.SetThickness(value);
            }
        }

        #endregion


        #region public PDFColor BackgroundColor

        /// <summary>
        /// Gets or sets the background color of this component
        /// </summary>
        [PDFAttribute("bg-color",Const.PDFStylesNamespace)]
        public PDFColor BackgroundColor
        {
            get
            {
                PDFBackgroundStyle bg;
                if (this.HasStyle && this.Style.TryGetBackground(out bg))
                    return bg.Color;
                else
                    return PDFColors.Transparent;
            }
            set
            {
                this.Style.Background.Color = value;
            }
        }

        #endregion

        #region public string BackgroundImage

        /// <summary>
        /// Gets or sets the background image of this component
        /// </summary>
        [PDFAttribute("bg-image", Const.PDFStylesNamespace)]
        public string BackgroundImage
        {
            get
            {
                PDFBackgroundStyle bg;
                if (this.HasStyle && this.Style.TryGetBackground(out bg))
                    return bg.ImageSource;
                else
                    return string.Empty;
            }
            set
            {
                this.Style.Background.ImageSource = value;
            }
        }

        #endregion

        #region public PatternRepeat BackgroundRepeat

        /// <summary>
        /// Gets or sets the background image repeat of this component
        /// </summary>
        [PDFAttribute("bg-repeat", Const.PDFStylesNamespace)]
        public PatternRepeat BackgroundRepeat
        {
            get
            {
                PDFBackgroundStyle bg;
                if (this.HasStyle && this.Style.TryGetBackground(out bg))
                    return bg.PatternRepeat;
                else
                    return PatternRepeat.RepeatBoth;
            }
            set
            {
                this.Style.Background.PatternRepeat = value;
            }
        }

        #endregion

        #region public PDFReal BackgroundOpacity

        /// <summary>
        /// Gets or sets the background fill color opacity of this component
        /// </summary>
        [PDFAttribute("bg-opacity", Const.PDFStylesNamespace)]
        public PDFReal BackgroundOpacity
        {
            get
            {
                PDFBackgroundStyle bg;
                if (this.HasStyle && this.Style.TryGetBackground(out bg))
                    return bg.Opacity;
                else
                    return PDFReal.Zero;
            }
            set
            {
                this.Style.Background.Opacity = value;
            }
        }

        #endregion



        #region public PDFUnit BorderWidth

        /// <summary>
        /// Gets or sets the border width of this component
        /// </summary>
        [PDFAttribute("border-width", Const.PDFStylesNamespace)]
        public PDFUnit BorderWidth
        {
            get
            {
                PDFBorderStyle b;
                if (this.HasStyle && this.Style.TryGetBorder(out b))
                    return b.Width;
                else
                    return PDFUnit.Empty;
            }
            set
            {
                this.Style.Border.Width = value;
            }
        }

        #endregion

        #region public PDFColor BorderColor

        /// <summary>
        /// Gets or sets the border color of this component
        /// </summary>
        [PDFAttribute("border-color", Const.PDFStylesNamespace)]
        public PDFColor BorderColor
        {
            get
            {
                PDFBorderStyle b;
                if (this.HasStyle && this.Style.TryGetBorder(out b))
                    return b.Color;
                else
                    return PDFColors.Transparent;
            }
            set
            {
                this.Style.Border.Color = value;
            }
        }

        #endregion

        #region public PDFDash BorderDashPattern

        /// <summary>
        /// Gets or sets the border dash pattern of this component
        /// </summary>
        [PDFAttribute("border-dash", Const.PDFStylesNamespace)]
        public PDFDash BorderDashPattern
        {
            get
            {
                PDFBorderStyle b;
                if (this.HasStyle && this.Style.TryGetBorder(out b))
                    return b.Dash;
                else
                    return PDFDash.None;
            }
            set
            {
                this.Style.Border.Dash = value;
            }
        }

        #endregion

        #region public double BorderOpacity

        /// <summary>
        /// Gets or sets the border opacity of this component
        /// </summary>
        [PDFAttribute("border-opacity", Const.PDFStylesNamespace)]
        public double BorderOpacity
        {
            get
            {
                PDFBorderStyle b;
                if (this.HasStyle && this.Style.TryGetBorder(out b))
                    return b.Opacity;
                else
                    return 1;
            }
            set
            {
                this.Style.Border.Opacity = value;
            }
        }

        #endregion

        #region public Sides BorderSides

        /// <summary>
        /// Gets or sets the border sides of this component
        /// </summary>
        [PDFAttribute("border-sides", Const.PDFStylesNamespace)]
        public Sides BorderSides
        {
            get
            {
                PDFBorderStyle b;
                if (this.HasStyle && this.Style.TryGetBorder(out b))
                    return b.Sides;
                else
                    return (Sides)0;
            }
            set
            {
                this.Style.Border.Sides = value;
            }
        }

        #endregion

        #region public PDFUnit BorderCornerRadius

        /// <summary>
        /// Gets or sets the border corner radii of this component
        /// </summary>
        [PDFAttribute("border-corner-radius", Const.PDFStylesNamespace)]
        public PDFUnit BorderCornerRadius
        {
            get
            {
                PDFBorderStyle b;
                if (this.HasStyle && this.Style.TryGetBorder(out b))
                    return b.CornerRadius;
                else
                    return PDFUnit.Zero;
            }
            set
            {
                this.Style.Border.CornerRadius = value;
            }
        }

        #endregion


        #region public PDFColor FillColor

        /// <summary>
        /// Gets or sets the Fill color of this component
        /// </summary>
        [PDFAttribute("fill-color", Const.PDFStylesNamespace)]
        public PDFColor FillColor
        {
            get
            {
                PDFFillStyle bg;
                if (this.HasStyle && this.Style.TryGetFill(out bg))
                    return bg.Color;
                else
                    return PDFColors.Transparent;
            }
            set
            {
                this.Style.Fill.Color = value;
            }
        }

        #endregion

        #region public string FillImage

        /// <summary>
        /// Gets or sets the Fill image of this component
        /// </summary>
        [PDFAttribute("fill-image", Const.PDFStylesNamespace)]
        public string FillImage
        {
            get
            {
                PDFFillStyle bg;
                if (this.HasStyle && this.Style.TryGetFill(out bg))
                    return bg.ImageSource;
                else
                    return string.Empty;
            }
            set
            {
                this.Style.Fill.ImageSource = value;
            }
        }

        #endregion

        #region public PatternRepeat FillRepeat

        /// <summary>
        /// Gets or sets the Fill image repeat of this component
        /// </summary>
        [PDFAttribute("fill-repeat", Const.PDFStylesNamespace)]
        public PatternRepeat FillRepeat
        {
            get
            {
                PDFFillStyle bg;
                if (this.HasStyle && this.Style.TryGetFill(out bg))
                    return bg.PatternRepeat;
                else
                    return PatternRepeat.RepeatBoth;
            }
            set
            {
                this.Style.Fill.PatternRepeat = value;
            }
        }

        #endregion

        #region public PDFReal FillOpacity

        /// <summary>
        /// Gets or sets the Fill color opacity of this component
        /// </summary>
        [PDFAttribute("fill-opacity", Const.PDFStylesNamespace)]
        public PDFReal FillOpacity
        {
            get
            {
                PDFFillStyle bg;
                if (this.HasStyle && this.Style.TryGetFill(out bg))
                    return bg.Opacity;
                else
                    return PDFReal.Zero;
            }
            set
            {
                this.Style.Fill.Opacity = value;
            }
        }

        #endregion



        #region public PDFUnit StrokeWidth

        /// <summary>
        /// Gets or sets the Stroke width of this component
        /// </summary>
        [PDFAttribute("stroke-width", Const.PDFStylesNamespace)]
        public PDFUnit StrokeWidth
        {
            get
            {
                PDFStrokeStyle b;
                if (this.HasStyle && this.Style.TryGetStroke(out b))
                    return b.Width;
                else
                    return PDFUnit.Empty;
            }
            set
            {
                this.Style.Stroke.Width = value;
            }
        }

        #endregion

        #region public PDFColor StrokeColor

        /// <summary>
        /// Gets or sets the Stroke color of this component
        /// </summary>
        [PDFAttribute("stroke-color", Const.PDFStylesNamespace)]
        public PDFColor StrokeColor
        {
            get
            {
                PDFStrokeStyle b;
                if (this.HasStyle && this.Style.TryGetStroke(out b))
                    return b.Color;
                else
                    return PDFColors.Transparent;
            }
            set
            {
                this.Style.Stroke.Color = value;
            }
        }

        #endregion

        #region public PDFDash StrokeDashPattern

        /// <summary>
        /// Gets or sets the Stroke dash pattern of this component
        /// </summary>
        [PDFAttribute("stroke-dash", Const.PDFStylesNamespace)]
        public PDFDash StrokeDashPattern
        {
            get
            {
                PDFStrokeStyle b;
                if (this.HasStyle && this.Style.TryGetStroke(out b))
                    return b.Dash;
                else
                    return PDFDash.None;
            }
            set
            {
                this.Style.Stroke.Dash = value;
            }
        }

        #endregion

        #region public double StrokeOpacity

        /// <summary>
        /// Gets or sets the Stroke opacity of this component
        /// </summary>
        [PDFAttribute("stroke-opacity", Const.PDFStylesNamespace)]
        public double StrokeOpacity
        {
            get
            {
                PDFStrokeStyle b;
                if (this.HasStyle && this.Style.TryGetStroke(out b))
                    return b.Opacity;
                else
                    return 1;
            }
            set
            {
                this.Style.Stroke.Opacity = value;
            }
        }

        #endregion



        #region public string FontFamily

        /// <summary>
        /// Gets or sets the Font Family of this component
        /// </summary>
        [PDFAttribute("font-family", Const.PDFStylesNamespace)]
        public string FontFamily
        {
            get
            {
                PDFFontStyle b;
                if (this.HasStyle && this.Style.TryGetFont(out b))
                    return b.FontFamily;
                else
                    return string.Empty;
            }
            set
            {
                this.Style.Font.FontFamily = value;
            }
        }

        #endregion

        #region public string FontSize

        /// <summary>
        /// Gets or sets the Font Size of this component
        /// </summary>
        [PDFAttribute("font-size", Const.PDFStylesNamespace)]
        public PDFUnit FontSize
        {
            get
            {
                PDFFontStyle b;
                if (this.HasStyle && this.Style.TryGetFont(out b))
                    return b.FontSize;
                else
                    return PDFUnit.Zero;
            }
            set
            {
                this.Style.Font.FontSize = value;
            }
        }

        #endregion

        #region public bool FontBold

        /// <summary>
        /// Gets or sets the Font bold value of this component
        /// </summary>
        [PDFAttribute("font-bold", Const.PDFStylesNamespace)]
        public bool FontBold
        {
            get
            {
                PDFFontStyle b;
                if (this.HasStyle && this.Style.TryGetFont(out b))
                    return b.FontBold;
                else
                    return false;
            }
            set
            {
                this.Style.Font.FontBold = value;
            }
        }

        #endregion

        #region public bool FontItalic

        /// <summary>
        /// Gets or sets the Font italic value of this component
        /// </summary>
        [PDFAttribute("font-italic", Const.PDFStylesNamespace)]
        public bool FontItalic
        {
            get
            {
                PDFFontStyle b;
                if (this.HasStyle && this.Style.TryGetFont(out b))
                    return b.FontItalic;
                else
                    return false;
            }
            set
            {
                this.Style.Font.FontItalic = value;
            }
        }

        #endregion


        #region public HorizontalAlignment HorizontalAlignment

        /// <summary>
        /// Gets or sets the Horizontal alignment of this component
        /// </summary>
        [PDFAttribute("h-align", Const.PDFStylesNamespace)]
        public HorizontalAlignment HorizontalAlignment
        {
            get
            {
                PDFPositionStyle bg;
                if (this.HasStyle && this.Style.TryGetPosition(out bg))
                    return bg.HAlign;
                else
                    return HorizontalAlignment.Left;
            }
            set
            {
                this.Style.Position.HAlign = value;
            }
        }

        #endregion

        #region public VerticalAlignment VerticalAlignment

        /// <summary>
        /// Gets or sets the vertical alignment of this component
        /// </summary>
        [PDFAttribute("v-align", Const.PDFStylesNamespace)]
        public VerticalAlignment VerticalAlignment
        {
            get
            {
                PDFPositionStyle bg;
                if (this.HasStyle && this.Style.TryGetPosition(out bg))
                    return bg.VAlign;
                else
                    return VerticalAlignment.Top;
            }
            set
            {
                this.Style.Position.VAlign = value;
            }
        }

        #endregion

        #region public PositionMode PositionMode

        /// <summary>
        /// Gets or sets the position mode of this component (flow, relative, absolute)
        /// </summary>
        [PDFAttribute("position-mode", Const.PDFStylesNamespace)]
        public PositionMode PositionMode
        {
            get
            {
                PDFPositionStyle bg;
                if (this.HasStyle && this.Style.TryGetPosition(out bg))
                    return bg.PositionMode;
                else
                    return PositionMode.Block;
            }
            set
            {
                this.Style.Position.PositionMode = value;
            }
        }

        #endregion

        #region public bool FillWidth

        /// <summary>
        /// Gets or sets the fill-width style attribute
        /// </summary>
        [PDFAttribute("full-width",Const.PDFStylesNamespace)]
        public bool FullWidth
        {
            get
            {
                PDFPositionStyle pos;
                if (this.HasStyle && this.Style.TryGetPosition(out pos))
                    return pos.FullWidth;
                else
                    return false;
            }
            set
            {
                this.Style.Position.FullWidth = value;
            }
        }

        #endregion

        #region public int ColumnCount {get;set;}

        /// <summary>
        /// Gets or sets the column count in this visual component
        /// </summary>
        [PDFAttribute("column-count",Const.PDFStylesNamespace)]
        public int ColumnCount
        {
            get
            {
                PDFColumnsStyle cols;
                if (this.HasStyle && this.Style.TryGetColumns(out cols))
                    return cols.ColumnCount;
                else
                    return 1;
            }
            set { this.Style.Columns.ColumnCount = value; }
        }

        #endregion

        #region public PDFUnit AlleyWidth {get;set;}

        /// <summary>
        /// Gets or sets the column count in this visual component
        /// </summary>
        [PDFAttribute("alley-width", Const.PDFStylesNamespace)]
        public PDFUnit AlleyWidth
        {
            get
            {
                PDFColumnsStyle cols;
                if (this.HasStyle && this.Style.TryGetColumns(out cols))
                    return cols.AlleyWidth;
                else
                    return PDFColumnsStyle.DefaultAlleyWidth;
            }
            set { this.Style.Columns.AlleyWidth = value; }
        }

        #endregion

        #region public WordWrap TextWrapping

        /// <summary>
        /// Gets or sets the text wrapping mode of this component
        /// </summary>
        [PDFAttribute("text-wrap", Const.PDFStylesNamespace)]
        public Scryber.Text.WordWrap TextWrapping
        {
            get
            {
                PDFTextStyle bg;
                if (this.HasStyle && this.Style.TryGetTextStyle(out bg))
                    return bg.WrapText;
                else
                    return Text.WordWrap.Auto;
            }
            set
            {
                this.Style.Text.WrapText = value;
            }
        }

        #endregion

        #region public OverflowSplit OverflowSplit {get; set;}

        /// <summary>
        /// Gets or sets the split option for this visual component
        /// </summary>
        [PDFAttribute("overflow-split",Const.PDFStylesNamespace)]
        public OverflowSplit OverflowSplit
        {
            get
            {
                PDFOverflowStyle overflow;
                if (this.HasStyle && this.Style.TryGetOverflow(out overflow))
                    return overflow.Split;
                else
                    return Drawing.OverflowSplit.Any;
            }
            set
            {
                this.Style.Overflow.Split = value;
            }
        }

        #endregion

        #region public string OutlineTitle {get;set;}

        /// <summary>
        /// Gets or sets the title of the outline for this component.
        /// </summary>
        [PDFAttribute("outline-title")]
        public string OutlineTitle
        {
            get
            {
                if (this.HasOutline)
                    return this.Outline.Title;
                else
                    return string.Empty;
            }
            set
            {
                if (!this.HasOutline && string.IsNullOrEmpty(value))
                {
                    //Nothing to do as we don't have an outline and the value is empty
                    return;
                }
                this.Outline.Title = value;
            }
        }

        #endregion


        //
        // ctor
        //

        protected PDFVisualComponent(PDFObjectType type)
            : base(type)
        {
            
        }

        #region Databind()
        
        /// <summary>
        /// Inheritors should override this method to provide their own databing implementations.
        /// </summary>
        /// <param name="includeChildren">Flag to identifiy if children should be databound also</param>
        protected override void DoDataBind(PDFDataContext context, bool includeChildren)
        {
            if (includeChildren && this.HasStyle)
                this.Style.DataBind(context);
            base.DoDataBind(context, includeChildren);
        }

        #endregion

        

    }

    public class PDFVisualComponentList : PDFComponentWrappingList<PDFVisualComponent>
    {
        public PDFVisualComponentList(PDFComponentList innerList)
            : base(innerList)
        {
        }

    }
}
