﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using Scryber.Components;

namespace Scryber.Data
{

    /// <summary>
    /// Abstract base class for all components that extract and return data
    /// </summary>
    public abstract class PDFDataSourceBase : PDFComponent, IPDFDataSource
    {

        #region public override bool Visible {get;set;}

        /// <summary>
        /// Overrides the default value to return false, 
        /// as this is not a visible Component of the page
        /// </summary>
        public override bool Visible
        {
            get
            {
                return false;
            }
            set
            {
                
            }
        }

        #endregion

        #region public virtual bool SupportsDataSchema { get; }

        /// <summary>
        /// IPDFDataSource implementation that returns true if this 
        /// data source supports schema interrogation.
        /// </summary>
        /// <remarks>
        /// It is up to inheritors to provide their own implementation
        /// and override this property to return the correct value</remarks>
        public virtual bool SupportsDataSchema { get { return false; } }

        #endregion

        protected PDFDataSourceBase(PDFObjectType type)
            : base(type)
        {
        }


        #region IPDFDataSource Members

        /// <summary>
        /// Retrieves the data from this DataSource based on the path
        /// </summary>
        /// <param name="path">The path to source the data from - for an XML DataSource this is an XPath query</param>
        /// <returns>The required data (could be null)</returns>
        public object Select(string path, PDFDataContext context)
        {
            object data;
            if (path == null)
                path = string.Empty;

            try
            {
                data = this.DoSelectData(path, context);
            }
            catch (Exception ex)
            {
                throw RecordAndRaise.Data(ex, Errors.CouldNotSelectData);
            }

            return data;
        }

        /// <summary>
        /// Returns the PDFDataScema associated with this source. 
        /// NOTE: Check that this operation is supported via the SupportsDataSchema property before calling.
        /// </summary>
        /// <param name="path"></param>
        /// <param name="context"></param>
        /// <returns>The data schema associated with the source</returns>
        /// <exception cref="System.NotSupportedException" >Thrown if this data source does not support schema extraction</exception>
        public virtual PDFDataSchema GetDataSchema(string path, PDFDataContext context)
        {
            throw new NotSupportedException(Errors.DataSchemaExtractionIsNotSupported);
        }

        #endregion

        /// <summary>
        /// Abstract method that all inheritors must override to 
        /// return the appropriate data based on the defined path
        /// </summary>
        /// <param name="path"></param>
        /// <returns></returns>
        protected abstract object DoSelectData(string path, PDFDataContext context);

        
    }


    /// <summary>
    /// A list of PDFDataSource implementations
    /// </summary>
    public class PDFDataSourceList : System.Collections.ObjectModel.KeyedCollection<string, PDFDataSourceBase>
    {

        protected override string GetKeyForItem(PDFDataSourceBase item)
        {
            return item.ID;
        }

        public bool TryGetSourceWithID(string id, out PDFDataSourceBase source)
        {
            PDFDataSourceBase found;
            if (this.Count == 0 || !this.Dictionary.TryGetValue(id, out found))
            {
                source = null;
                return false;
            }
            else
            {
                source = found;
                return true;
            }

            //TODO: Support DataSourceList-Ref
        }
    }
}
