﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using Scryber.Components;

namespace Scryber.Data
{
    [PDFParsableComponent("ForEach")]
    public class PDFForEach : PDFDataTemplateComponent
    {
        private int _start, _count, _step;

        /// <summary>
        /// Gets or sets the (zero based) index of the first item to be generated
        /// </summary>
        [PDFAttribute("start-index")]
        public int StartIndex
        {
            get { return _start; }
            set { _start = value; }
        }

        /// <summary>
        /// Gets or sets the maximum number of items that will be generated.
        /// </summary>
        [PDFAttribute("count")]
        public int MaxCount
        {
            get { return _count; }
            set { _count = value; }
        }

        /// <summary>
        /// Gets or sets the step for the loop. Default is one for every item to be generated.
        /// </summary>
        [PDFAttribute("step")]
        public int Step
        {
            get { return _step; }
            set { _step = value; }
        }



        #region public virtual IPDFTemplate Template {get;}

        private IPDFTemplate _template;

        /// <summary>
        /// Gets or sets the IPDFTemplate used to instantiate child Components
        /// </summary>
        [PDFTemplate()]
        [PDFElement("Template")]
        public virtual IPDFTemplate Template
        {
            get { return _template; }
            set { _template = value; }
        }

        public bool HasTemplate
        {
            get { return null != this.Template; }
        }

        #endregion

        public PDFForEach()
            : this(PDFObjectTypes.NoOp)
        {
        }

        protected PDFForEach(PDFObjectType type)
            : base(type)
        {
            _start = 0;
            _count = int.MaxValue;
            _step = 1;
        }


        protected override IPDFTemplate GetTemplateForBinding(PDFDataContext context, int index, int count)
        {
            return this.Template;
        }

        protected override void DoDataBind(PDFDataContext context, bool includeChildren)
        {
            object data = null;
            bool hasdata = false;
            System.Xml.IXmlNamespaceResolver resolver = context.NamespaceResolver;

            if (this.HasCustomDataSource)
            {
                data = this.GetBindingData(context);
                context.DataStack.Push(data);
                hasdata = true;
            }
            
            base.DoDataBind(context, includeChildren);

            if (hasdata)
            {
                context.DataStack.Pop();
            }

            context.NamespaceResolver = resolver;
        }

        protected override IEnumerator CreateEnumerator(IEnumerable enumerable)
        {
            if (this.StartIndex == 0 && this.MaxCount == int.MaxValue && this.Step == 1)
                return base.CreateEnumerator(enumerable);
            else
                return new RangeEnumerator(enumerable.GetEnumerator(), this.StartIndex, this.MaxCount, this.Step);
        }

        #region private class RangeEnumerator : IEnumerator

        /// <summary>
        /// Implements enumerating over a range (with an optional step)
        /// </summary>
        private class RangeEnumerator : IEnumerator
        {
            private int _currindex;
            private IEnumerator _root;
            private int _start;
            private int _count;
            private int _step;

            

            public object Current
            {
                get { return this._root.Current; }
            }

            public RangeEnumerator(IEnumerator all, int start)
                : this(all, start, int.MaxValue)
            {
            }

            public RangeEnumerator(IEnumerator all, int start, int count)
                : this(all, start, count, 1)
            { }

            public RangeEnumerator(IEnumerator all, int start, int count, int step)
            {
                this._root = all;
                this._start = start;
                this._count = count;
                this._step = step;
                this._currindex = -1;
            }

            public bool MoveNext()
            {
                bool moved = false;
                do
                {
                    int stepped = 0;
                    do
                    {
                        moved = this._root.MoveNext();
                        stepped++;
                        this._currindex++;

                    } while ((_currindex > _start) && moved && stepped < this._step);



                } while (moved && (_currindex < _start));

                if (_currindex >= (_count + _start))
                    moved = false;
                return moved;
            }

            public void Reset()
            {
                this._currindex = -1;
                this._root.Reset();
            }

        }

        #endregion

    }
}
