﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using Scryber.Components;
using System.Xml;

namespace Scryber.Data
{
    /// <summary>
    /// Loads an XML file as a data source that can be cached and then enumerated over.
    /// </summary>
    [PDFParsableComponent("XMLDataSource")]
    public class PDFXMLDataSource : PDFDataSourceBase
    {

        #region public string SourcePath {get;set;}

        private string _source;

        /// <summary>
        /// The source data file for the xml.
        /// </summary>
        [PDFAttribute("source-path")]
        public string SourcePath
        {
            get { return _source; }
            set { _source = value; }
        }

        #endregion

        #region public int CacheDuration {get;set;}

        private int _cachedur = -1;

        /// <summary>
        /// The total number of minutes to cache the data source
        /// </summary>
        [PDFAttribute("cache-duration")]
        public int CacheDuration
        {
            get { return this._cachedur; }
            set { this._cachedur = value; }
        }

        #endregion

        #region public System.Xml.XPath.XPathNavigator XmlData

        /// <summary>
        /// Local copy - It may be that we have multiple references and to this source and we don't have any caching set.
        /// So keep a local copy just for this purpose.
        /// </summary>
        private System.Xml.XPath.XPathNavigator _nav = null;

        /// <summary>
        /// Gets or sets the XmlDocument data this datasource should use. Can be set in the actual databinding to a node set.
        /// </summary>
        [PDFAttribute("xml-data")]
        public System.Xml.XPath.XPathNavigator XmlData
        {
            get { return _nav; }
            set { _nav = value; }
        }

        #endregion


        private PDFXmlNamespaceCollection _nss;

        [PDFArray(typeof(PDFXmlNamespaceDeclaration))]
        [PDFElement("Namespaces")]
        public PDFXmlNamespaceCollection Namespaces
        {
            get
            {
                if (null == _nss)
                    _nss = new PDFXmlNamespaceCollection();
                return _nss;
            }
            
        }

        private PDFXmlTransformation _transform;

        /// <summary>
        /// Gets or set the optional XSLT data for this source
        /// </summary>
        [PDFElement("Transform")]
        public PDFXmlTransformation Transformer
        {
            get { return _transform; }
            set { _transform = value; }
        }

        /// <summary>
        /// Creates a new PDFXMLDatasource 
        /// </summary>
        public PDFXMLDataSource()
            : this(PDFObjectTypes.XmlData)
        {
        }

        protected PDFXMLDataSource(PDFObjectType type)
            : base(type)
        {
        }


        //
        // implementation
        //


        #region protected override object DoSelectData(string path, PDFDataContext context)

        /// <summary>
        /// Overrides the base implementation to load the actual data at the path
        /// </summary>
        /// <param name="path"></param>
        /// <returns></returns>
        protected override object DoSelectData(string path, PDFDataContext context)
        {
            if (_nav == null)
            {
                XmlDocument doc = this.LoadXMLDocument(this.SourcePath);
                
                this._nav = doc.CreateNavigator();
                System.Xml.IXmlNamespaceResolver resolver = this.Document.CreateNamespaceResolver(this.Namespaces, doc.NameTable);
                context.NamespaceResolver = resolver;
            }
            if (null != this.Transformer)
            {
                this._nav = this.Transformer.TransformData(this._nav, this.CacheDuration, this, context);
            }

            if (string.IsNullOrEmpty(path))
                return _nav.Select("*");
            else
                return _nav.Select(path, context.NamespaceResolver);
        }

        #endregion

        #region private System.Xml.XmlDocument LoadXMLDocument(string sourcepath)

        /// <summary>
        /// private implementation to load the data from the source.
        /// </summary>
        /// <param name="sourcepath"></param>
        /// <returns></returns>
        private System.Xml.XmlDocument LoadXMLDocument(string sourcepath)
        {
            if (string.IsNullOrEmpty(sourcepath))
                throw new ArgumentNullException(string.Format(Errors.NoSourcePathDefinedOnXmlDataSource, this.ID));


            PDFDocument pdf = this.Document;
            string path = this.MapPath(sourcepath);

            object obj = null;

            System.Xml.XmlDocument doc = null;

            try
            {

                if (this.CacheDuration > 0)
                {
                    IPDFCacheProvider cache = pdf.CacheProvider;

                    if (null != cache && cache.TryRetrieveFromCache(PDFObjectTypes.XmlData.ToString(), path, out obj))
                        doc = (System.Xml.XmlDocument)obj;
                    else
                    {
                        doc = new System.Xml.XmlDocument();
                        doc.Load(path);
                        cache.AddToCache(PDFObjectTypes.XmlData.ToString(), path, doc, new TimeSpan(0, this.CacheDuration, 0));
                    }
                }
                else //not using cacheing
                {
                    doc = new System.Xml.XmlDocument();
                    doc.Load(path);
                }
            }
            catch (Exception ex)
            {
                throw new PDFDataException(string.Format(Errors.CouldNotLoadTheDataFromTheSourcePath, "XML", sourcepath), ex);
            }
            return doc;
        }

        #endregion
    }
}
