﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Scryber.Data
{
    public class PDFXmlTransformation
    {

        private const string XSLTCacheType = "Scryber.PDF.XSLT";
        private const bool XSLTUseDebug = false;

        /// <summary>
        /// Gets or sets the path to the XSLT with which to transform the data
        /// </summary>
        [PDFAttribute("path")]
        public string XSLTPath { get; set; }


        private System.Xml.Xsl.XslCompiledTransform _transformer;

        /// <summary>
        /// Gets or sets the XSL compiled transformer for this 
        /// </summary>
        public System.Xml.Xsl.XslCompiledTransform Transformer
        {
            get { return _transformer; }
            set { _transformer = value; }
        }

        private PDFXsltArgumentList _args;

        [PDFArray(typeof(PDFXsltArgumentBase))]
        [PDFElement("")]
        public PDFXsltArgumentList Arguments
        {
            get
            {
                if (null == _args)
                    _args = new PDFXsltArgumentList();
                return _args;
            }
            set
            {
                _args = value;
            }
        }

        public bool HasArguments
        {
            get { return null != _args && _args.Count > 0; }
        }

        public System.Xml.XPath.XPathNavigator TransformData(System.Xml.XPath.XPathNavigator nav, int cacheduration, IPDFDataSource source, PDFDataContext context)
        {
            //Check we have something to use for a transformation.
            if (string.IsNullOrEmpty(this.XSLTPath) && null == this.Transformer)
                return nav;

            System.Xml.XmlWriter xmlWriter = null;
            System.IO.MemoryStream memoryStream = null;
            System.IO.StreamWriter streamWriter = null;
            System.Xml.XmlDocument result = null;
            
            System.Xml.XmlDocument output = new System.Xml.XmlDocument();
            try
            {
                System.Xml.Xsl.XslCompiledTransform trans = this.DoGetTransformer(cacheduration, source, context);
                System.Xml.Xsl.XsltArgumentList args = this.DoGetArguments(context);
                memoryStream = new System.IO.MemoryStream();
                streamWriter = new System.IO.StreamWriter(memoryStream, Encoding.UTF8);

                System.Xml.XmlWriterSettings writerSettings = CreateWriterSettings();
                xmlWriter = System.Xml.XmlWriter.Create(streamWriter, writerSettings);

                trans.Transform(nav, args, xmlWriter);
                xmlWriter.Flush();
                streamWriter.Flush();

                result = new System.Xml.XmlDocument();
                memoryStream.Position = 0;
                result.Load(memoryStream);

            }
            catch (Exception ex)
            {
                throw new PDFDataException(Errors.CouldNotTransformInputData, ex);
            }
            finally
            {
                if (null != xmlWriter)
                    xmlWriter.Close();
                if (null != streamWriter)
                    streamWriter.Dispose();
                if (null != memoryStream)
                    memoryStream.Dispose();
            }

            return result.CreateNavigator();
        }

        protected virtual System.Xml.Xsl.XslCompiledTransform DoGetTransformer(int cacheduration, IPDFDataSource source, PDFDataContext context)
        {

            if (null == _transformer)
            {
                string path = this.XSLTPath;
                if (string.IsNullOrEmpty(path))
                    throw new NullReferenceException(string.Format(Errors.XSLTPathOrTransformerNotSetOnInstance, source.ID));

                path = source.MapPath(path);
                IPDFCacheProvider cache = ((Scryber.Components.PDFDocument)source.Document).CacheProvider;
                System.Xml.Xsl.XslCompiledTransform transformer;

                object found;

                if (cacheduration > 0 && cache.TryRetrieveFromCache(XSLTCacheType, path, out found))
                    transformer = (System.Xml.Xsl.XslCompiledTransform)found;
                else
                {
                    
                    transformer = new System.Xml.Xsl.XslCompiledTransform(XSLTUseDebug);
                    try
                    {
                        transformer.Load(path);
                    }
                    catch (Exception ex)
                    {
                        throw new PDFDataException(string.Format(Errors.XSLTCouldNotBeLoadedFromPath, path), ex);
                    }

                    if (cacheduration > 0)
                        cache.AddToCache(XSLTCacheType, path, transformer, new TimeSpan(0, cacheduration, 0));
                }
                
                _transformer = transformer;
            }

            return _transformer;
        }


        protected virtual System.Xml.Xsl.XsltArgumentList DoGetArguments(PDFDataContext context)
        {
            System.Xml.Xsl.XsltArgumentList args = new System.Xml.Xsl.XsltArgumentList();
            if (this.HasArguments)
            {
                foreach (PDFXsltArgumentBase arg in this.Arguments)
                {
                    string name = arg.Name;
                    string ns = null == arg.Namespace ? string.Empty : arg.Namespace;
                    object value = arg.GetValue(context);
                    args.AddParam(name, ns, value);
                }
            }
            return args;
        }



        protected virtual System.Xml.XmlWriterSettings CreateWriterSettings()
        {
            System.Xml.XmlWriterSettings settings = new System.Xml.XmlWriterSettings();
            settings.ConformanceLevel = System.Xml.ConformanceLevel.Auto;
            settings.Indent = true;
            settings.NewLineHandling = System.Xml.NewLineHandling.Entitize;
            return settings;
        }
    }
}
