﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Scryber.Data
{
    
    public abstract class PDFXsltArgumentBase
    {
        [PDFAttribute("name")]
        public string Name { get; set; }

        [PDFAttribute("namespace")]
        public string Namespace { get; set; }


        /// <summary>
        /// Gets the fully qualified name of the argument including the Namespace if set.
        /// </summary>
        public string FullName
        {
            get
            {
                return GetFullName(this);
            }
        }

        public static string GetFullName(PDFXsltArgumentBase arg)
        {
            if (null == arg)
                throw new ArgumentNullException("arg");

            return GetFullName(arg.Namespace, arg.Name);
        }

        public static string GetFullName(string ns, string name)
        {
            string full;
            if (string.IsNullOrEmpty(ns) == false)
            {
                full = ns;
                full += "+";
                full += name;
            }
            else
                full = name;

            return full;
        }

        public abstract object GetValue(PDFDataContext context);

    }

    /// <summary>
    /// A list of defined cslt arguments
    /// </summary>
    public class PDFXsltArgumentList : System.Collections.ObjectModel.KeyedCollection<string, PDFXsltArgumentBase>
    {

        protected override string GetKeyForItem(PDFXsltArgumentBase arg)
        {
            return arg.FullName;
        }
    }


    //
    // Concrete argument classes
    //


    [PDFParsableComponent("XsltArgument")]
    public class PDFXsltArgument : PDFXsltArgumentBase
    {
        [PDFAttribute("value")]
        [PDFElement("")]
        public string Value
        {
            get;
            set;
        }

        public override object GetValue(PDFDataContext context)
        {
            return Value;
        }
    }


    [PDFParsableComponent("XsltQSArgument")]
    public class PDFXsltQueryStringArgument : PDFXsltArgumentBase
    {
        [PDFAttribute("default-value")]
        [PDFElement("")]
        public string DefaultValue
        {
            get;
            set;
        }

        /// <summary>
        /// Optional query string parameter name. If not set then the base name of 
        /// this parameter will be used.
        /// </summary>
        [PDFAttribute("param-name")]
        public string QSParamName
        {
            get;
            set;
        }

        public override object GetValue(PDFDataContext context)
        {
            string name = this.QSParamName;
            if (string.IsNullOrEmpty(name))
                name = this.Name;

            System.Web.HttpContext webcontext = System.Web.HttpContext.Current;
            if (null == context)
                throw new PDFDataException(Errors.NoWebContextAvailable);
            System.Collections.Specialized.NameValueCollection qs = webcontext.Request.QueryString;
            string value = qs[name];
            if (string.IsNullOrEmpty(value))
                value = this.DefaultValue;
            return value;
        }
    }

    [PDFParsableComponent("XsltItemArgument")]
    public class PDFXsltContextItemArgument : PDFXsltArgumentBase
    {
        [PDFAttribute("default-value")]
        [PDFElement("")]
        public string DefaultValue
        {
            get;
            set;
        }

        /// <summary>
        /// Optional context item name. If not set then the base name of 
        /// this parameter will be used.
        /// </summary>
        [PDFAttribute("item-name")]
        public string ItemName
        {
            get;
            set;
        }

        public override object GetValue(PDFDataContext context)
        {
            string name = this.ItemName;
            if (string.IsNullOrEmpty(name))
                name = this.Name;

            
            object value = context.Items[name];

            if (null == value)
                value = this.DefaultValue;

            return value;
        }
    }

    
}
