﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using Scryber.Components;
using Scryber.Drawing;
using Scryber.Native;

namespace Scryber
{
    public abstract class PDFAnnotationEntry : IArtefactEntry
    {
        public abstract PDFObjectRef RenderToPDF(PDFRenderContext context, PDFWriter writer);
    }

    /// <summary>
    /// An annotation that links to a component
    /// </summary>
    internal class PDFLinkAnnotationEntry : PDFAnnotationEntry
    {
        public PDFAction Action { get; set; }

        public PDFComponent Component { get; private set; }

        public string AlternateText { get; set; }

        public Styles.PDFStyle AnnotationStyle { get; private set; }

        public PDFLinkAnnotationEntry(PDFComponent component, Styles.PDFStyle style)
        {
            this.Component = component;
            this.AnnotationStyle = style;
        }

        public override PDFObjectRef RenderToPDF(PDFRenderContext context, PDFWriter writer)
        {
            //TODO: User multiple links areas for a text annotation as it flows over multiple lines.

            List<PDFObjectRef> all = new List<PDFObjectRef>();

            int pageindex = context.PageIndex;
            PDFComponentArrangement arrange = this.Component.GetFirstArrangement();
            int index = 1;
            while (null != arrange && arrange.PageIndex == pageindex)
            {
                PDFObjectRef annotref = writer.BeginObject();
                writer.BeginDictionary();
                writer.WriteDictionaryNameEntry("Type", "Annot");
                writer.WriteDictionaryNameEntry("Subtype", "Link");
                if (!string.IsNullOrEmpty(this.AlternateText))
                    writer.WriteDictionaryStringEntry("Contents", this.AlternateText);
                PDFRect bounds = arrange.RenderBounds;
                if (bounds != PDFRect.Empty)
                {
                    if (context.DrawingOrigin == DrawingOrigin.TopLeft)
                    {
                        //PDFs have origin at bottom so need to convert.
                        PDFReal value = context.Graphics.GetXPosition(bounds.X.RealValue);
                        bounds.X = new PDFUnit(value.Value, PageUnits.Points);
                        value = context.Graphics.GetYPosition(bounds.Y.RealValue);
                        bounds.Y = new PDFUnit(value.Value, PageUnits.Points);
                        bounds.Width = bounds.X + bounds.Width;
                        bounds.Height = bounds.Y - bounds.Height;
                    }
                    else
                    {
                        bounds.Width = bounds.X + bounds.Width;
                        bounds.Height = bounds.Y + bounds.Height;
                    }

                    writer.BeginDictionaryEntry("Rect");
                    writer.WriteArrayRealEntries(bounds.X.Value, bounds.Y.Value, bounds.Width.Value , bounds.Height.Value);
                    writer.EndDictionaryEntry();
                }
                writer.WriteDictionaryStringEntry("NM", this.Component.UniqueID + "_" + index.ToString());

                //Draw the border
                Styles.PDFBorderStyle bstyle;
                if (this.AnnotationStyle != null && this.AnnotationStyle.TryGetBorder(out bstyle) && bstyle.Width > 0)
                {
                    PDFUnit corner = bstyle.CornerRadius;
                    PDFUnit width = bstyle.Width;
                    PDFDash dash = bstyle.Dash;
                    PDFColor c = bstyle.Color;

                    writer.BeginDictionaryEntry("Border");
                    writer.WriteArrayRealEntries(corner.PointsValue, corner.PointsValue, width.PointsValue);
                    writer.EndDictionaryEntry();

                    writer.BeginDictionaryEntry("C");
                    if (c.ColorSpace == ColorSpace.G)
                        writer.WriteArrayRealEntries(c.Gray.Value);
                    else if (c.ColorSpace == ColorSpace.RGB)
                        writer.WriteArrayRealEntries(c.Red.Value, c.Green.Value, c.Blue.Value);
                    else if (context.Conformance == ConformanceMode.Strict)
                        RecordAndRaise.ArgumentOutOfRange("c", Errors.ColorValueIsNotCurrentlySupported, c.ColorSpace);
                    else
                        context.TraceLog.Add(TraceLevel.Error, "Annotation Entry", string.Format(Errors.ColorValueIsNotCurrentlySupported, c.ColorSpace));

                    writer.EndDictionaryEntry();


                }
                else
                {
                    writer.BeginDictionaryEntry("Border");
                    writer.WriteArrayRealEntries(0.0, 0.0, 0.0);
                    writer.EndDictionaryEntry();
                }

                writer.BeginDictionaryEntry("A");
                PDFObjectRef actionref = this.Action.RenderToPDF(context, writer);
                if (null != actionref)
                    writer.WriteObjectRefS(actionref);
                writer.EndDictionaryEntry();

                writer.EndDictionary();
                writer.EndObject();

                all.Add(annotref);

                //If we have more than one arrangement on the object then move to the next one
                if (arrange is PDFComponentMultiArrangement)
                    arrange = ((PDFComponentMultiArrangement)arrange).NextArrangement;
                index++;
            }

            if (all.Count == 0)
                return null;
            else if (all.Count == 1)
                return all[0];
            else
            {
                //More than one annotation entry was written
                //So output in an array and return a reference to that.
                PDFObjectRef array = writer.BeginObject();
                writer.BeginArray();
                foreach (PDFObjectRef inner in all)
                {
                    writer.BeginArrayEntry();
                    writer.WriteObjectRefS(inner);
                    writer.EndArrayEntry();
                }
                writer.EndArray();
                writer.EndObject();

                return array;
            }

        }

        protected virtual PDFRect GetComponentBounds(PDFComponent comp, PDFComponentArrangement arrange)
        {
            if (null != arrange)
                return arrange.RenderBounds;
            else
                return PDFRect.Empty;
        }

    }

    /// <summary>
    /// An Annotation that links to a text component (so that lines are tracked individually)
    /// </summary>
    internal class PDFLinkTextAnnotationEntry : PDFLinkAnnotationEntry
    {

        public PDFLinkTextAnnotationEntry(PDFComponent textcomp, Styles.PDFStyle style, int lineindex)
            : base(textcomp, style)
        {
            if (!(textcomp is IPDFTextComponent))
                throw RecordAndRaise.InvalidCast(Errors.CannotConvertObjectToType, textcomp.GetType(), typeof(IPDFTextComponent));
        }


        protected override PDFRect GetComponentBounds(PDFComponent comp, PDFComponentArrangement arrange)
        {
            //TODO:Split the text into lines and use those as rectangles
            PDFRect rect = base.GetComponentBounds(comp, arrange);
            IPDFTextComponent text = this.Component as IPDFTextComponent;
            rect.Y -= text.TextLayout.Size.Height;
            rect.Size = text.TextLayout.Size;
            return rect;
        }
    }
}
