﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using Scryber.Native;

namespace Scryber
{
    public class PDFWriter14 : PDFWriter
    {
        private const string TraceCategory = "PDFWriter";
        private const TraceLevel TraceDefaultLevel = TraceLevel.Debug;
        
        public override Version Version { get { return new Version(1, 5); } }

        internal protected static class Constants
        {
            public const string StartObject = "obj\r\n";
            public const string CommentStart = "%";
            public const string EndObject = "\r\nendobj\r\n";
            public const string StartStream = "\r\nstream\r\n";
            public const string EndStream = "\r\nendstream";
            public const string StartDictionary = "<< ";
            public const string EndDictionary = " >>";
            public const string StartName = "/";
            public const string StartArray = "[";
            public const string EndArray = "]";
            public const string ArrayEntrySeparator = " ";
            public const string DictionaryEntrySeparator = "\r\n";
            public const string DictionaryNameValueSeparator = " ";
            public const string CommentPrePend = "%";
            public const string StartString = "(";
            public const string EndString = ")";
            public const string StartHexString = "<";
            public const string EndHexString = ">";
            public const string WhiteSpace = " ";
            public const string NullString = "null";
            
        }

        private bool _finishedentry = false;

        protected bool FinishedEntry
        {
            get { return _finishedentry; }
            set { _finishedentry = value; }
        }

        private bool _usehex;

        protected bool UseHex
        {
            get { return _usehex; }
            set { _usehex = value; }
        }
        
        #region Constructors

        public PDFWriter14(Stream stream, PDFTraceLog log)
            : this(stream, 0, log)
        {
        }

        

                
        public PDFWriter14(Stream stream, int gen, PDFTraceLog log)
            : base(stream, new PDFXRefTable(gen), log)
        {
        }

        

        #endregion

        public override void Open()
        {
            
        }

        public override void Close(PDFDocumentID documentid)
        {
            this.BaseStream.Flush();
            this.Log(TraceLevel.Message, "Closing the writer and outputting indirect object data onto the underlying stream");

            this.BaseStream.WriteLine(String.Format("%PDF-{0}.{1}", this.Version.Major, this.Version.Minor));
            this.BaseStream.Write(Constants.CommentStart);
            this.BaseStream.Write("????");
            this.BaseStream.WriteLine();
            
            foreach (IIndirectObject pfo in this.XRefTable.References)
            {
                if (pfo.Deleted == false)
                {
                    this.Log("Writing indirect object data " + pfo.Number);
                    pfo.Offset = this.BaseStream.Position;
                    this.BaseStream.Write(pfo.Number.ToString());
                    this.BaseStream.Write(" ");
                    this.BaseStream.Write(pfo.Generation.ToString());
                    this.BaseStream.Write(" ");
                    this.BaseStream.Write(Constants.StartObject);

                    WriteIndirecObjectData(pfo);
                    if (pfo.HasStream)
                    {
                        WriteIndirectStreamData(pfo);
                    }
                    this.BaseStream.Write(Constants.EndObject);
                    this.BaseStream.WriteLine();
                    this.BaseStream.Flush();
                }
            }
            this.XRefTable.Offset = this.BaseStream.Position;
            this.WriteXRefTable(this.XRefTable);
            this.BaseStream.WriteLine();
            this.BaseStream.WriteLine();
            WriteTrailerObject(documentid);

            this.BaseStream.Flush();
            this.BaseStream.WriteLine("startxref");
            this.BaseStream.WriteLine(this.XRefTable.Offset.ToString());
            this.BaseStream.Write("%%EOF");
            this.Log("Written end of file marker");
            this.BaseStream.Flush();
        }

        /// <summary>
        /// Outputs the Indirect objects stream data
        /// </summary>
        /// <param name="pfo"></param>
        protected virtual void WriteIndirectStreamData(IIndirectObject pfo)
        {
            this.BaseStream.Write(Constants.StartStream);
            pfo.Stream.WriteTo(this.BaseStream);
            this.BaseStream.Write(Constants.EndStream);
        }

        /// <summary>
        /// Outputs the Indirect objects object data
        /// </summary>
        /// <param name="pfo"></param>
        protected virtual void WriteIndirecObjectData(IIndirectObject pfo)
        {
            pfo.ObjectData.WriteTo(this.BaseStream);
        }

        protected virtual void WriteTrailerObject(PDFDocumentID documentid)
        {
            this.BaseStream.WriteLine("trailer");
            this.BaseStream.Write(Constants.StartDictionary);
            WriteTrailerDictionary(documentid);
            this.BaseStream.WriteLine(Constants.EndDictionary);
        }

        protected virtual void WriteTrailerDictionary(PDFDocumentID documentid)
        {
            this.BaseStream.Write(Constants.StartName);
            this.BaseStream.Write("Size");
            this.BaseStream.Write(Constants.WhiteSpace);
            this.BaseStream.WriteLine(this.XRefTable.References.Count.ToString());

            PDFObjectRef oref;
            if (this.ObjectDictionary.TryGetValue("Catalog", out oref))
            {
                this.BaseStream.Write(Constants.StartName);
                this.BaseStream.Write("Root");
                this.BaseStream.Write(Constants.WhiteSpace);
                this.BaseStream.Write(oref.Number + Constants.WhiteSpace + oref.Generation + " R");
                this.BaseStream.WriteLine();
            }

            if (this.ObjectDictionary.TryGetValue("Info", out oref))
            {
                this.BaseStream.Write(Constants.StartName);
                this.BaseStream.Write("Info");
                this.BaseStream.Write(Constants.WhiteSpace);
                this.BaseStream.Write(oref.Number + Constants.WhiteSpace + oref.Generation + " R");

            }

            if (documentid != null)
            {
                this.BaseStream.WriteLine();
                this.BaseStream.Write(Constants.StartName);
                this.BaseStream.Write("ID");
                this.BaseStream.Write(Constants.WhiteSpace);
                this.BaseStream.Write(Constants.StartArray);
                
                this.WriteStringHex(documentid.One);
                this.BaseStream.Write(" ");
                this.WriteStringHex(documentid.Two);
                this.BaseStream.WriteLine(Constants.EndArray);
            }
        }

       
        private void WriteXRefTable(PDFXRefTable table)
        {
            this.Log("Outputting XRefTable onto the stream at position " + this.BaseStream.Position.ToString());
            this.BaseStream.WriteLine("xref");
            this.BaseStream.Write(table.References[0].Number.ToString());
            this.BaseStream.Write(" ");
            this.BaseStream.WriteLine(table.References.Count.ToString());
            foreach (IIndirectObject io in table.References)
            {
                this.BaseStream.Write(io.Offset.ToString("0000000000"));
                if (io.Deleted)
                {
                    this.BaseStream.WriteLine(" 65535 f");
                    //this.BaseStream.WriteLine(String.Format("{0:0000000000} 65535 f", io.Offset));
                }
                else if (io.Generation != table.Generation)
                {
                    this.BaseStream.Write(io.Generation.ToString("00000"));
                    this.BaseStream.WriteLine(" f");
                    //this.BaseStream.WriteLine(String.Format("{0:0000000000} {1:00000} f", io.Offset, io.Generation));
                }
                else
                {
                    this.BaseStream.Write(io.Generation.ToString("00000"));
                    this.BaseStream.WriteLine(" n");
                    //this.BaseStream.WriteLine(String.Format("{0:0000000000} {1:00000} n", io.Offset, io.Generation));
                }
            }
            this.BaseStream.Flush();
        }

        public override void WriteLine()
        {
            this.CurrentStream.WriteLine();
        }

        public override void WriteCommentLine()
        {
            this.CurrentStream.WriteLine(Constants.CommentPrePend);
        }
        public override void WriteCommentLine(string comment)
        {
            this.CurrentStream.Write(Constants.CommentPrePend);
            this.CurrentStream.WriteLine(comment);
        }

        public override void WriteComment(string comment)
        {
            this.CurrentStream.Write(Constants.CommentPrePend);
            this.CurrentStream.Write(comment);
        }

        public override void WriteCommentLine(string comment, params object[] args)
        {
            this.CurrentStream.Write(Constants.CommentPrePend);
            this.CurrentStream.WriteLine(String.Format(comment, args));
        }

        public override void WriteComment(string comment, params object[] args)
        {
            this.CurrentStream.Write(Constants.CommentPrePend);
            this.CurrentStream.Write(String.Format(comment, args));
        }

        public override PDFObjectRef BeginObject(string name)
        {
            PDFIndirectObject obj = CreateIndirectObject();
            PDFObjectRef oref = this.InitializeIndirectObject(name, obj);
            this.Log("Begun a new indirect object: " + oref);
            return oref;
        }

        protected PDFIndirectObject CreateIndirectObject()
        {
            return new PDFIndirectObject();
        }

        public override void BeginStream(PDFObjectRef onobject, IStreamFilter[] filters)
        {
            PDFIndirectObject pio = onobject.Reference as PDFIndirectObject;
            pio.InitStream(filters);
            this.Stack.Push(pio.Stream);
        }

        public override void EndObject()
        {
            PDFIndirectObject obj = this.Stack.Peek().IndirectObject;
            this.Log(String.Format("Ended indirect object {0} {1} R ", obj.Number, obj.Generation));
            
            this.ReleaseIndirectObject();
        }

        public override long EndStream()
        {
            PDFStream str = this.Stack.Pop();
            str.Flush();
            if (str.HasFilters)
                return str.FilteredLength;
            else
                return str.Length;
        }

        public override void BeginDictionary()
        {
            CurrentStream.Write(Constants.StartDictionary);
            this.FinishedEntry = false;
        }

        public override void BeginDictionaryEntry(string name)
        {
            if (FinishedEntry)
                CurrentStream.Write(Constants.DictionaryNameValueSeparator);
            FinishedEntry = false;
            CurrentStream.Write(Constants.StartName);
            CurrentStream.Write(ValidateName(name));
            CurrentStream.Write(Constants.WhiteSpace);
        }

        
        public override void EndDictionaryEntry()
        {
            this.CurrentStream.Write(Constants.DictionaryEntrySeparator);
        }

        public override void EndDictionary()
        {
            this.FinishedEntry = false;
            CurrentStream.Write(Constants.EndDictionary);
        }


        public override void BeginArray()
        {
            CurrentStream.Write(Constants.StartArray);
            this.FinishedEntry = false;
        }

        public override void BeginArrayEntry()
        {
            if (FinishedEntry)
                CurrentStream.Write(Constants.ArrayEntrySeparator);
            this.FinishedEntry = false;
        }

        public override void EndArrayEntry()
        {
            CurrentStream.Write(Constants.ArrayEntrySeparator);
        }

        
        public override void EndArray()
        {
            this.FinishedEntry = false;
            CurrentStream.Write(Constants.EndArray);
        }


        public override void WriteByteString(string byteString)
        {
            CurrentStream.Write(Constants.StartHexString);
            CurrentStream.Write(byteString);
            CurrentStream.Write(Constants.EndHexString);
        }

        public override void WriteStringLiteral(string value, Scryber.FontEncoding encoding)
        {
            if (UseHex)
            {
                //TODO Write encoding prefix
                CurrentStream.Write(Constants.StartHexString);
                value = ToHex(value, encoding);
                CurrentStream.Write(value);
                CurrentStream.Write(Constants.EndHexString);
            }
            else
            {
                Scryber.Text.PDFEncoding enc = Scryber.Text.PDFEncoding.GetEncoding(encoding);
                CurrentStream.Write(Constants.StartString);
                if (null != enc.Prefix && enc.Prefix.Length > 0)
                    CurrentStream.Write(enc.Prefix);
                CurrentStream.Write(enc.GetBytes(value));
                CurrentStream.Write(Constants.EndString);
            }
        }

        public override void WriteStringHex(byte[] value)
        {
            CurrentStream.Write(Constants.StartHexString);
            CurrentStream.Write(ToHex(value));
            CurrentStream.Write(Constants.EndHexString);
        }


        private string ToHex(string value, FontEncoding encoding)
        {
            Scryber.Text.PDFEncoding enc = Scryber.Text.PDFEncoding.GetEncoding(encoding);
            byte[] all = enc.GetBytes(value);
            return ToHex(all);
        }

        private static char[] lookup = new char[] { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

        private string ToHex(byte[] data)
        {
            StringBuilder sb = new StringBuilder((data.Length * 2) + 2);

            for (int buc = 0; buc < data.Length; buc++)
            {
                sb.Append(lookup[(data[buc] >> 4) & 0x0F]);
                sb.Append(lookup[data[buc] & 0x0F]);
            }

            return sb.ToString();
        }

        public override void WriteBoolean(bool value)
        {
            if (value)
                CurrentStream.Write("true");
            else
                CurrentStream.Write("false");
        }

        public override void WriteDate(DateTime value)
        {
            string d = DateToString(value);
            this.WriteStringLiteral(d);
        }

        private string DateToString(DateTime value)
        {
            StringBuilder sb = new StringBuilder(22);
            sb.Append(value.ToString("\\D:yyyyMMddhhmmss"));
            string offset = value.ToString("zzz");
            if (offset != "00:00")
            {
                offset = offset.Replace(':', '\'');
                sb.Append(offset);
                sb.Append('\'');
            }
            return sb.ToString();            
        }

        public override void WriteNumber(int value)
        {
            CurrentStream.Write(value.ToString());
        }
        public override void WriteNumber(long value)
        {
            CurrentStream.Write(value.ToString());
        }

        public override void WriteReal(double value)
        {
            CurrentStream.Write(value.ToString("F",System.Globalization.CultureInfo.InvariantCulture));
        }

        public override void WriteReal(decimal value)
        {
            CurrentStream.Write(value.ToString("F", System.Globalization.CultureInfo.InvariantCulture));
        }

        public override void WriteReal(float value)
        {
            CurrentStream.Write(value.ToString("F", System.Globalization.CultureInfo.InvariantCulture));
        }

        public override void WriteName(string name)
        {
            CurrentStream.Write(Constants.StartName);
            CurrentStream.Write(ValidateName(name));

        }

        public override void WriteObjectRef(int number, int generation)
        {
            this.CurrentStream.Write(number.ToString());
            this.CurrentStream.Write(" ");
            this.CurrentStream.Write(generation.ToString());
            this.CurrentStream.Write(" R");
        }

        public override void WriteNull()
        {
            this.CurrentStream.Write(Constants.NullString);
        }

        public override void WriteFileObject(IFileObject obj)
        {
            obj.WriteData(this);
        }
        
        public override void WriteRaw(string data)
        {
            this.CurrentStream.Write(data);
        }

        public override void WriteRaw(byte[] buffer, int offset, int length)
        {
            this.CurrentStream.Write(buffer, offset, length);
        }

        
        public override void WriteSpace()
        {
            this.CurrentStream.Write(Constants.WhiteSpace);
        }

        protected override string GetOpCode(PDFOpCode op)
        {
            string s = String.Empty;

            switch (op)
            {
                case PDFOpCode.ColorSetFillSpace:
                    s = "cs";
                    break;
                case PDFOpCode.ColorSetStrokeSpace:
                    s = "CS";
                    break;
                case PDFOpCode.ColorFillPattern:
                    s = "scn";
                    break;
                case PDFOpCode.ColorStrokePattern:
                    s = "SCN";
                    break;
                case PDFOpCode.ColorFillGrayscaleSpace:
                    s = "g";
                    break;
                case PDFOpCode.ColorStrokeGrayscaleSpace:
                    s = "G";
                    break;
                case PDFOpCode.ColorFillRGBSpace:
                    s = "rg";
                    break;
                case PDFOpCode.ColorStrokeRGBSpace:
                    s = "RG";
                    break;
                case PDFOpCode.ColorFillCMYK:
                    s = "k";
                    break;
                case PDFOpCode.ColorStrokeCMYK:
                    s = "K";
                    break;
                case PDFOpCode.ColorFillOpacity:
                    s = "ca";
                    break;
                case PDFOpCode.ColorStrokeOpacity:
                    s = "CA";
                    break;
                case PDFOpCode.SaveState:
                    s = "q";
                    break;
                case PDFOpCode.RestoreState:
                    s = "Q";
                    break;
                case PDFOpCode.XobjPaint:
                    s = "Do";
                    break;
                case PDFOpCode.XobjBegin:
                    s = "BI";
                    break;
                case PDFOpCode.XobjImageData:
                    s = "ID";
                    break;
                case PDFOpCode.XobjEndImage:
                    s = "EI";
                    break;
                case PDFOpCode.TxtBegin:
                    s = "BT";
                    break;
                case PDFOpCode.TxtEnd:
                    s = "ET";
                    break;
                case PDFOpCode.TxtCharSpacing:
                    s = "Tc";
                    break;
                case PDFOpCode.TxtWordSpacing:
                    s = "Tw";
                    break;
                case PDFOpCode.TxtHScaling:
                    s = "Tz";
                    break;
                case PDFOpCode.TxtLeading:
                    s = "TL";
                    break;
                case PDFOpCode.TxtFont:
                    s = "Tf";
                    break;
                case PDFOpCode.TxtPaint:
                    s = "Tj";
                    break;
                case PDFOpCode.TxtPaintArray:
                    s = "TJ";
                    break;
                case PDFOpCode.TxtRenderMode:
                    s = "Tr";
                    break;
                case PDFOpCode.TxtRise:
                    s = "Ts";
                    break;
                case PDFOpCode.TxtMoveNextOffset:
                    s = "Td";
                    break;
                case PDFOpCode.TxtTransformMatrix:
                    s = "Tm";
                    break;
                case PDFOpCode.TxtNextLine:
                    s = "T*";
                    break;
                case PDFOpCode.GraphTransformMatrix:
                    s = "cm";
                    break;
                case PDFOpCode.GraphLineWidth:
                    s = "w";
                    break;
                case PDFOpCode.GraphLineCap:
                    s = "J";
                    break;
                case PDFOpCode.GraphLineJoin:
                    s = "j";
                    break;
                case PDFOpCode.GraphMiterLimit:
                    s = "M";
                    break;
                case PDFOpCode.GraphDashPattern:
                    s = "d";
                    break;
                case PDFOpCode.GraphRenderingIntent:
                    s = "ri";
                    break;
                case PDFOpCode.GraphFlatness:
                    s = "i";
                    break;
                case PDFOpCode.GraphApplyState:
                    s = "gs";
                    break;
                case PDFOpCode.GraphMove:
                    s = "m";
                    break;
                case PDFOpCode.GraphLineTo:
                    s = "l";
                    break;
                case PDFOpCode.GraphCurve2Handle:
                    s = "c";
                    break;
                case PDFOpCode.GraphCurve1HandleEnd:
                    s = "v";
                    break;
                case PDFOpCode.GraphCurve1HandleBegin:
                    s = "y";
                    break;
                case PDFOpCode.GraphClose:
                    s = "h";
                    break;
                case PDFOpCode.GraphRect:
                    s = "re";
                    break;
                case PDFOpCode.GraphStrokePath:
                    s = "S";
                    break;
                case PDFOpCode.GraphCloseAndStroke:
                    s = "s";
                    break;
                case PDFOpCode.GraphFillPath:
                    s = "f";
                    break;
                case PDFOpCode.GraphFillPathEvenOdd:
                    s = "f*";
                    break;
                case PDFOpCode.GraphFillAndStroke:
                    s = "B";
                    break;
                case PDFOpCode.GraphFillAndStrokeEvenOdd:
                    s = "B*";
                    break;
                case PDFOpCode.GraphCloseFillStroke:
                    s = "b";
                    break;
                case PDFOpCode.GraphCloseFileStrokeEvenOdd:
                    s = "b*";
                    break;
                case PDFOpCode.GraphNoOp:
                case PDFOpCode.GraphEndPath:
                    s = "n";
                    break;
                case PDFOpCode.GraphSetClip:
                    s = "W";
                    break;
                default:
                    break;
            }
            return s;
        }
        

        #region Validation Methods

        private string ValidateName(string name)
        {
            //TODO: Increase validation support
            name = name.Replace(" ", "#20");
            return name;
        }

        private string ValidateLiteral(string value)
        {
            //TODO: Implement Literal Validation
            return value;
        }

        #endregion

        
        protected void Log(string message)
        {
            this.Log(TraceDefaultLevel, TraceCategory, message);
        }

        protected void Log(TraceLevel level, string message)
        {
            this.Log(level, TraceCategory, message);
        }

        protected void Log(TraceLevel level, string category, string message)
        {
            if (null != this.TraceLog)
                this.TraceLog.Add(level, category, message);
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                Log(TraceLevel.Debug, "Disposing PDF writer");
                foreach (IIndirectObject pfo in this.XRefTable.References)
                {
                    Log(TraceLevel.Debug, "Disposing indirect object '" + pfo.ToString());
                    pfo.Dispose();
                }
            }
            
            base.Dispose(disposing);
            
        }
    }
}
