﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Configuration;

namespace Scryber.Configuration
{

    /// <summary>
    /// The configuration section for the PDF generation. Contains the Extensions collection that maps an xml namespace to a runtime assembly and namespace
    /// </summary>
    /// <remarks>
    /// The generation section in web.config allows the definition of extensions. Components that exist in a different namespace
    /// and are defined in a separate dll and namespace. By adding a mapping here you can enable the XML Parser to  know about and understand
    /// the component definition.
    ///   <example>
    ///      &lt;generation&gt;
    ///      &lt;extensions&gt;
    ///        &lt;remove xml-namespace=&quot;http://www.scryber.co.uk/schemas/v08/Scryber.Styles.xsd&quot;/&gt;
    ///        &lt;add xml-namespace=&quot;http://www.scryber.co.uk/schemas/v08/Scryber.Styles.xsd&quot; 
    ///             runtime-namespace=&quot;Scryber.Styles&quot; runtime-assembly=&quot;Scryber.Styles, Version=0.8.0.0, Culture=neutral, PublicKeyToken=872cbeb81db952fe&quot; /&gt;
    ///      &lt;/extensions&gt;
    ///   </example>
    /// </remarks>
    public class GenerationConfigurationSection : ConfigurationSection
    {
        /// <summary>
        /// Xml local name of the collection 
        /// </summary>
        private const string ExtensionNameCollectionKey = "";

        /// <summary>
        /// local cached value of the collection for fast retrieval
        /// </summary>
        private GenerationExtensionCollection _col = null;


        /// <summary>
        /// Gets or sets the collection of defined extensions. 
        /// This collection is usually modified in the configuration file - register, remove and clear
        /// </summary>
        [ConfigurationCollection(typeof(FontMapping), CollectionType = ConfigurationElementCollectionType.AddRemoveClearMap,
            AddItemName = "register", RemoveItemName = "remove", ClearItemsName = "clear")]
        [ConfigurationProperty(ExtensionNameCollectionKey, IsDefaultCollection = true, IsRequired = false)]
        public GenerationExtensionCollection Extensions
        {
            get
            {
                if(null == _col)
                    _col = this[ExtensionNameCollectionKey] as GenerationExtensionCollection;
                if (null == _col)
                {
                    _col = new GenerationExtensionCollection();
                    _col.InitDefaults();
                }
                return _col;
            }
            set
            {
                this[ExtensionNameCollectionKey] = value;
                _col = null;
            }
        }

        /// <summary>
        /// Static default section
        /// </summary>
        private static GenerationConfigurationSection _def = new GenerationConfigurationSection();

        /// <summary>
        /// Gets the default Generation section if none has been defined in the configuration section.
        /// </summary>
        public static GenerationConfigurationSection Default
        {
            get
            {
                return _def;
            }
        }
    }
}
