﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using Scryber.Text;
using Scryber.Native;
using System.Drawing;
using Scryber.Resources;

namespace Scryber.Drawing
{
    public delegate object ResolveStyleCallback(string styleClass);

    public partial class PDFGraphics
    {

        #region public FontEncoding CurrentEncoding {get; protected set;}

        private FontEncoding _encoding = FontEncoding.WinAnsiEncoding;

        /// <summary>
        /// Gets the current encoding for text based on the current font
        /// </summary>
        public FontEncoding CurrentEncoding
        {
            get { return _encoding; }
            protected set { _encoding = value; }
        }

        #endregion

        #region protected SetCurrentFont(PDFFont font)

        public void SetCurrentFont(PDFFont font)
        {
            PDFResource rsrc = this.Container.Document.GetResource(PDFResource.FontDefnResourceType, font.FullName, true);
            if (null == rsrc)
                throw new NullReferenceException(String.Format(Errors.FontNotFound, font.FullName));

            PDFName name = this.Container.Register(rsrc);

            if (rsrc is PDFFontResource)
            {
                PDFFontDefinition defn = (rsrc as PDFFontResource).Definition;
                this.CurrentEncoding = defn.Encoding;
            }
            this.Writer.WriteOpCodeS(PDFOpCode.TxtFont, name, font.Size.RealValue);
            
        }

        #endregion

        #region protected RenderCurrentTextMode(TextRenderMode mode)

        protected void RenderCurrentTextMode(TextRenderMode mode)
        {
            this.Writer.WriteOpCodeS(PDFOpCode.TxtRenderMode, new PDFNumber((int)mode));
        }

        #endregion


        /// <summary>
        /// String format to stop the wrapping of text so a single string can be measured
        /// </summary>
        private static readonly System.Drawing.StringFormat NoWrappingWordFormat = InitStringFormat(StringTrimming.Word);

        private static readonly StringFormat NoWrappingCharacterFormant = InitStringFormat(StringTrimming.Character);

        private static System.Drawing.StringFormat InitStringFormat(StringTrimming triming)
        {

#if USE_NON_TYPOGRAPHIC

            //Shaves 20% off measure string time with less accuracy
            StringFormat format = StringFormat.GenericDefault.Clone() as StringFormat;
#else
            StringFormat format = StringFormat.GenericTypographic.Clone() as StringFormat;
#endif
            format.Alignment = StringAlignment.Near;
            format.FormatFlags = StringFormatFlags.NoWrap | StringFormatFlags.MeasureTrailingSpaces;
            format.Trimming = triming;
            return format;
        }

        private static StringFormat GetStringFormatting(PDFTextRenderOptions options)
        {
            StringFormat format;
            if (options.WrapText.HasValue && options.WrapText.Value == WordWrap.Character)
                format = NoWrappingCharacterFormant;
            else
                format = NoWrappingWordFormat;
            return format;
        }


        /// <summary>
        /// Measures a single line of text (no wrapping) and returns the size required, 
        /// and the number of characters fitted into that line - using GDI+, slow but does not require the font file
        /// </summary>
        /// <param name="chars"></param>
        /// <param name="available"></param>
        /// <param name="options"></param>
        /// <param name="charsfitted"></param>
        /// <returns></returns>
        protected PDFSize MeasureGraphicsString(string chars, PDFSize available, PDFTextRenderOptions options, out int charsfitted)
        {
            //PDFUnit h = options.Leading.HasValue? options.Leading.Value : options.Font.Size;
            //PDFUnit w = chars.Length * h.PointsValue;
            //charsfitted = chars.Length;
            //return new PDFSize(w,h);

            if(_wingraphics == null)
                _wingraphics = this.CreateWinGraphics();
            if (available.Width <= 0)
            {
                charsfitted = 0;
                return PDFSize.Empty;
            }
            else
            {
                StringFormat format = GetStringFormatting(options);

                System.Drawing.SizeF size = new SizeF((float)available.Width.PointsValue, (float)available.Height.PointsValue);
                System.Drawing.Font sysfont = options.SystemFont;
                int linesfitted;
                if (options.WrapText.HasValue && options.WrapText.Value == WordWrap.NoWrap)
                {
                    size = _wingraphics.MeasureString(chars, sysfont, PointF.Empty, format);
                    charsfitted = chars.Length;
                }
                else
                    size = _wingraphics.MeasureString(chars, sysfont, size, format, out charsfitted, out linesfitted);

                return new PDFSize(size.Width, size.Height);
            }

        }

        /// <summary>
        /// Measures a single line of text (no wrapping) and returns the size required, 
        /// and the number of characters fitted into that line. 
        /// </summary>
        /// <param name="chars"></param>
        /// <param name="startIndex"></param>
        /// <param name="available"></param>
        /// <param name="options"></param>
        /// <param name="charsfitted"></param>
        /// <returns></returns>
        /// <remarks>Uses the TTFFontFile if available, otherwise will default to using the GDI+ measure,
        /// which is slower but supports postscript fonts</remarks>
        public PDFSize MeasureString(string chars, int startIndex, PDFSize available, PDFTextRenderOptions options, out int charsfitted)
        {
            PDFFontDefinition defn = Scryber.Drawing.PDFFontFactory.GetFontDefinition(options.Font);
            if (defn.CanMeasureStrings)
            {
                bool trimtoword = true;
                if (options.WrapText.HasValue && options.WrapText.Value == WordWrap.Character)
                    trimtoword = false;

                return defn.MeasureStringWidth(chars, startIndex, options.Font.Size.PointsValue, available.Width.PointsValue, trimtoword, out charsfitted);
            }
            else
            {
                if (startIndex > 0)
                    chars = chars.Substring(startIndex);
                return this.MeasureGraphicsString(chars, available, options, out charsfitted);
            }
        }

        private double GetLineLeft(PDFSize size, PDFTextRenderOptions options)
        {
            if (options != null && options.FirstLineInset != PDFUnit.Empty)
                return options.FirstLineInset.Value.ToPoints().Value;
            else
                return 0.0;
        }

        private System.Drawing.Font GetSystemFont(PDFTextRenderOptions options)
        {
            if (options == null || options.Font == null)
                return null;
            else
                return options.Font.GetSystemFont();
            
        }

        private PDFTextRenderOptions ResolveTextClass(string classname)
        {
            throw new NotImplementedException("ResolveTextClass");
        }

        private void ReleaseTextRenderOptions(PDFTextRenderOptions options, PDFRect bounds)
        {
            if (options.FillBrush != null)
                options.FillBrush.ReleaseGraphics(this, bounds);
            if (options.Stroke != null)
                options.Stroke.ReleaseGraphics(this, bounds);
        }

        private TextRenderMode SetTextRenderOptions(PDFTextRenderOptions options, PDFRect bounds)
        {
            
            TextRenderMode mode = TextRenderMode.NoOp;
            if (options == null)
                throw new ArgumentNullException("options");
            
            if (options.FillBrush != null)
            {
                mode = TextRenderMode.Fill;
                options.FillBrush.SetUpGraphics(this, bounds);
            }
            if (options.Stroke != null)
            {
                if (mode == TextRenderMode.Fill)
                    mode = TextRenderMode.FillAndStroke;
                else
                    mode = TextRenderMode.Stroke;

                options.Stroke.SetUpGraphics(this, bounds);
            }
            if (options.Font != null)
                this.SetCurrentFont(options.Font);

            if (options.Leading.HasValue)
            {
                SetTextLeading(options);
            }
            SetTextRenderMode(mode);
            return mode;
        }

        public void SetTextLeading(PDFTextRenderOptions options)
        {
            this.Writer.WriteOpCodeS(PDFOpCode.TxtLeading, options.Leading.Value.RealValue);
        }

        public void SetTextRenderMode(TextRenderMode mode)
        {
            this.Writer.WriteOpCodeS(PDFOpCode.TxtRenderMode, (PDFNumber)(int)mode);
        }

        public void FillText(TextBlock block, PDFTextRenderOptions textoptions, PDFPoint location)
        {
            this.FillText(block, textoptions, new PDFRect(location, new PDFSize(block.Width,block.Height)));
        }

        public void FillText(TextBlock block, PDFTextRenderOptions textoptions, PDFRect bounds)
        {
            ResolveStyleCallback callback = new ResolveStyleCallback(this.ResolveTextClass);
            this.FillText(block, textoptions, bounds, callback);
        }

        public void FillText(TextBlock block, PDFTextRenderOptions textoptions, PDFRect bounds, ResolveStyleCallback callback)
        {
            this.DoFillText(block, textoptions, bounds, callback);
        }




        private void DoFillText(TextBlock block, PDFTextRenderOptions options, PDFRect bounds, ResolveStyleCallback callback)
        {
            PDFPoint pt = bounds.Location;
            pt.X += block.XOffset;
            pt.Y += block.YOffset;
            this.SaveGraphicsState();
            this.Writer.WriteOpCodeS(PDFOpCode.TxtBegin);

            try
            {
                
                TextRenderMode origmode = this.SetTextRenderOptions(options, bounds);
                //Stack<TextRenderMode> modes = null;

                List<int> linecounts = new List<int>();

                
                PDFUnit inset;
                PDFUnit yoffset = PDFUnit.Zero;
                PDFUnit xoffset = PDFUnit.Zero;

                int totallinecount = 0;
                int paracount = 0;


                foreach (Text.TextParagraph para in block.Paragraphs)
                {
                   
                    inset = para.FirstLineInset;
                    inset += para.XOffset;

                    if (para.ResetY)
                    {
                        this.BeginNewLine(inset.RealValue, para.YOffset);
                    }

                    if (para.LineCount > 0)
                    {
                        double lastlineheight = 0.0;
                        foreach (Text.TextLine line in para.Lines)
                        {
                            if (totallinecount == 0)
                            {
                                pt.Y += (PDFUnit)line.AscentHeight;
                                PDFUnit x = inset + (PDFUnit)line.LineInset;
                                this.BeginFirstLine(x.RealValue + pt.X.RealValue, pt.Y.RealValue);
                                xoffset = line.LineInset;
                            }
                            else
                            {
                                PDFUnit x = line.LineInset - xoffset;
                                this.BeginNewLine(x.RealValue, (PDFReal)lastlineheight);
                                xoffset = line.LineInset;
                            }


                            //this.BeginNewLine(reset, totallinecount, paralinecount, lastparalinecount, bounds, inset.RealValue, pt, (PDFReal)line.Height, (PDFReal)line.Width, lastlinewidth, halign);

                            foreach (Text.TextData data in line)
                            {
                                if (data is TextBeginStyle)
                                {
                                    //if (null == modes)
                                    //{
                                    //    modes = new Stack<TextRenderMode>();
                                    //    modes.Push(origmode);
                                    //}

                                    TextBeginStyle style = data as TextBeginStyle;
                                    ApplyTextStyle(bounds, style.RenderOptions);
                                }
                                else if (data is TextEndStyle)
                                {
                                    //if (null == modes || modes.Count < 2) //there should always be 1 for the root
                                    //    throw new IndexOutOfRangeException(Errors.UnbalancedTextStyle);

                                    TextEndStyle end = data as TextEndStyle;
                                    //modes.Pop();
                                    ApplyTextStyle(bounds, end.RenderOptions);
                                }
                                else if (data is TextSpan)
                                {
                                    TextSpan span = data as TextSpan;
                                    this.FillSpan(span);
                                }
                            }

                            totallinecount++;
                            lastlineheight = line.Height;
                        }
                        
                    }
                    paracount++;

                }
            }
            catch (Exception ex)
            {
                throw new InvalidOperationException(String.Format(Errors.CouldNotFillText, ex.Message), ex);
            }
            finally
            {
                this.Writer.WriteOpCodeS(PDFOpCode.TxtEnd);
                this.ReleaseTextRenderOptions(options, bounds);
                this.RestoreGraphicsState();
            }
        }

        private void ApplyTextStyle(PDFRect bounds, PDFTextRenderOptions style)
        {
            TextRenderMode mode = TextRenderMode.NoOp;
            //TextRenderMode prev = modes.Peek();
            if (null != style)
            {
                if (null != style.Font)
                    this.SetCurrentFont(style.Font);

                if (null != style.FillBrush)
                {
                    style.FillBrush.SetUpGraphics(this, bounds);
                    mode = TextRenderMode.Fill;
                }
                if (null != style.Stroke)
                {
                    style.Stroke.SetUpGraphics(this, bounds);
                    if (mode == TextRenderMode.Fill)
                        mode = TextRenderMode.FillAndStroke;
                    else
                        mode = TextRenderMode.Stroke;
                }
            }
            //if (mode != prev)
            //{
            //    if (prev == TextRenderMode.Stroke && mode == TextRenderMode.Fill)
            //        mode = TextRenderMode.FillAndStroke;
            //    else if (prev == TextRenderMode.Fill && mode == TextRenderMode.Stroke)
            //        mode = TextRenderMode.FillAndStroke;
            //}
            //modes.Push(mode);
            this.SetTextRenderMode(mode);
        }

        private void BeginFirstLine(PDFReal xoffset, PDFReal yoffset)
        {
            this.Writer.WriteOpCodeS(PDFOpCode.TxtMoveNextOffset, GetXPosition(xoffset), GetYPosition(yoffset));
        }

        private void BeginNewLine(PDFReal xoffset, PDFReal yoffset)
        {
            this.Writer.WriteOpCodeS(PDFOpCode.TxtMoveNextOffset, GetXPosition(xoffset), GetYOffset(yoffset));
        }

        private void BeginNewLine(bool reset, int totallinecount, int paralinecount, int lastparalinecount, PDFRect bounds, PDFReal inset, PDFPoint pt, PDFReal lineheight, PDFReal linewidth, PDFReal lastlinewidth, HorizontalAlignment halign)
        {
            if (halign == HorizontalAlignment.Left)
            {
                if (reset)
                    //was pt.X.RealValue + 
                    this.Writer.WriteOpCodeS(PDFOpCode.TxtMoveNextOffset, GetXPosition(inset), GetYOffset(pt.Y.RealValue + lineheight));

                else if (totallinecount == 0)
                    this.Writer.WriteOpCodeS(PDFOpCode.TxtMoveNextOffset, GetXPosition(pt.X.RealValue + inset), GetYPosition(pt.Y));

                else if (lastparalinecount > 1 && paralinecount == 0)
                    this.Writer.WriteOpCodeS(PDFOpCode.TxtMoveNextOffset, GetXOffset(inset), GetYOffset(lineheight));

                else if (paralinecount == 1)
                    this.Writer.WriteOpCodeS(PDFOpCode.TxtMoveNextOffset, GetXOffset(PDFReal.Zero - inset), GetYOffset(lineheight));

                else
                    this.Writer.WriteOpCodeS(PDFOpCode.TxtMoveNextOffset, GetXOffset(PDFReal.Zero), GetYOffset(lineheight));
            }
            else if (halign == HorizontalAlignment.Right)
            {
                if (totallinecount == 0)
                    this.Writer.WriteOpCodeS(PDFOpCode.TxtMoveNextOffset, GetXPosition((pt.X.RealValue + bounds.Width.RealValue) - linewidth), GetYPosition(pt.Y));
                else
                    this.Writer.WriteOpCodeS(PDFOpCode.TxtMoveNextOffset, GetXPosition(lastlinewidth - linewidth), GetYOffset(lineheight));
            }
            else if (halign == HorizontalAlignment.Center)
            {
                double middle = pt.X.RealValue.Value + (bounds.Width.RealValue.Value/2);
                double start = middle - (linewidth.Value/2);
                this.Writer.WriteOpCodeS(PDFOpCode.TxtMoveNextOffset, GetXPosition(start),GetYPosition(pt.Y));
            }

        }

        
        
        private void FillSpan(TextSpan span)
        {
            string chars = span.Characters;
            
            if (chars.IndexOfAny(TextConst.EscapeChars) > -1)
            {
                for (int i = 0; i < TextConst.EscapeStrings.Length; i++)
                {
                    chars = chars.Replace(TextConst.EscapeStrings[i], TextConst.ReplaceStrings[i]);
                }
            }
            this.Writer.WriteStringLiteralS(chars, this.CurrentEncoding);
            this.Writer.WriteOpCodeS(PDFOpCode.TxtPaint);
        }

        

    }
}
