﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using Scryber.Native;

namespace Scryber.Drawing
{
    public abstract class PDFImageData : PDFObject
    {

        //
        // properties
        //

        #region public string SourcePath {get;set;}

        private string _path;
        /// <summary>
        /// Gets the source path this image data was loaded from
        /// </summary>
        public string SourcePath
        {
            get { return _path; }
            protected set { _path = value; }
        }

        #endregion


        #region public int PixelWidth {get;set;}

        private int _w;
        /// <summary>
        /// Gets the total number of pixels in one row of the image
        /// </summary>
        public int PixelWidth
        {
            get { return _w; }
            protected set { _w = value; }
        }

        #endregion

        #region public int PixelHeight {get;set;}

        private int _h;

        public int PixelHeight
        {
            get { return _h; }
            protected set { _h = value; }
        }

        #endregion


        #region public byte[] Data {get;set;}

        private byte[] _data;

        /// <summary>
        /// Gets the byte array of image data that represent the image
        /// </summary>
        public byte[] Data
        {
            get { return _data; }
            internal protected set
            {
                _data = value;
                this.ResetFilterCache();
            }
        }

        #endregion


        #region  public int BitsPerColor {get;set;}

        private int _bitdepth;

        /// <summary>
        /// Gets or set the number of bits per single color sample
        /// </summary>
        public int BitsPerColor
        {
            get { return _bitdepth; }
            internal protected set { _bitdepth = value; }
        }

        #endregion

        #region public int BytesPerLine {get; protected internal set;}

        private int _bperline;

        /// <summary>
        /// Gets the total number of bytes in one line
        /// </summary>
        public int BytesPerLine
        {
            get { return _bperline; }
            protected internal set { _bperline = value; }
        }

        #endregion

        #region public ColorSpace ColorSpace {get;set;}

        private ColorSpace _cs = ColorSpace.RGB;

        public ColorSpace ColorSpace
        {
            get { return _cs; }
            internal protected set { _cs = value; }
        }

        #endregion

        #region public int ColorsPerSample {get; set;}

        private int _colspsample;

        /// <summary>
        /// Gets the numbers of colour values that make up an individual pixel.
        /// </summary>
        public int ColorsPerSample
        {
            get { return _colspsample; }
            protected internal set { _colspsample = value; }
        }

        #endregion

        #region public double HorizontalResolution {get;set;}

        private double _hres;
        /// <summary>
        /// Gets the horizontal resolution of the image (pixels per inch)
        /// </summary>
        public double HorizontalResolution
        {
            get { return _hres; }
            internal protected set { _hres = value; }
        }

        #endregion

        #region public double VerticalResolution {get;set;}

        private double _vres;
        /// <summary>
        /// Gets the vertical resolution of the image (pixels per inch)
        /// </summary>
        public double VerticalResolution
        {
            get { return _vres; }
            internal protected set { _vres = value; }
        }

        #endregion

        #region public PDFUnit DisplayHeight {get;}

        public PDFUnit DisplayHeight
        {
            get { return new PDFUnit(this.PixelHeight / VerticalResolution, PageUnits.Inches); }
        }

        #endregion

        #region public PDFUnit DisplayWidth {get;}

        public PDFUnit DisplayWidth
        {
            get { return new PDFUnit(this.PixelWidth / this.HorizontalResolution, PageUnits.Inches); }
        }

        #endregion

        #region public string Filter {get;set;}

        private IStreamFilter[] _filters;

        /// <summary>
        /// Gets the stream filter
        /// </summary>
        public IStreamFilter[] Filters
        {
            get { return _filters; }
            internal protected set
            {
                _filters = value;
                this.ResetFilterCache();
            }
        }

        #endregion

        #region  public bool HasFilter {get;}

        public bool HasFilter
        {
            get { return null != _filters && _filters.Length > 0; }
        }

        #endregion

        #region public virtual bool IsPrecompressedData {get;}

        /// <summary>
        /// If true then the raw image data is precompressed otherwise it is raw binary pixel data
        /// </summary>
        public virtual bool IsPrecompressedData
        {
            get { return false; }
        }

        #endregion

        //
        // ctor(s)
        //

        #region protected PDFImageData(string source, int w, int h)

        /// <summary>
        /// Protected constructor - accepts the source path, width and height (in pixels)
        /// </summary>
        /// <param name="source"></param>
        /// <param name="w"></param>
        /// <param name="h"></param>
        protected PDFImageData(string source, int w, int h)
            : base(PDFObjectTypes.ImageData)
        {
            this._path = source;
            this._w = w;
            this._h = h;
        }

        #endregion

        //
        // rendering

        public PDFSize GetSize()
        {
            double w = ((double)this.PixelWidth) / this.HorizontalResolution;
            double h = ((double)this.PixelHeight) / this.VerticalResolution;

            return new PDFSize(new PDFUnit(w, PageUnits.Inches), new PDFUnit(h, PageUnits.Inches));
        }

        internal PDFObjectRef Render(PDFName name, PDFContextBase context, PDFWriter writer)
        {

            context.TraceLog.Add(TraceLevel.Debug, "IMAGE", "Rendering image data");

            PDFObjectRef renderref = writer.BeginObject(name.Value);

            writer.BeginDictionaryS();
            writer.WriteDictionaryNameEntry("Name", name.Value);
            writer.WriteDictionaryNameEntry("Type", "XObject");
            writer.WriteDictionaryNameEntry("Subtype", "Image");
            
            RenderImageInformation(context, writer);
           
            writer.EndDictionary();
            writer.BeginStream(renderref);

            this.RenderImageStreamData(context, writer);
           
            writer.EndStream();

            writer.EndObject();

            if (context.TraceLog.ShouldLog(TraceLevel.Debug))
                context.TraceLog.Add(TraceLevel.Debug, "IMAGE", String.Format("Completed render of the image XObject for source '{0}'", name));



            return renderref;
        }

        protected virtual void RenderImageInformation(PDFContextBase context,  PDFWriter writer)
        {
            writer.WriteDictionaryNumberEntry("Width", this.PixelWidth);
            writer.WriteDictionaryNumberEntry("Height", this.PixelHeight);
            writer.WriteDictionaryNumberEntry("Length", this.Data.LongLength);

            if (this.ColorSpace == ColorSpace.G)
                writer.WriteDictionaryNameEntry("ColorSpace", "DeviceGray");
            else if (this.ColorSpace == ColorSpace.Custom)
                this.RenderCustomColorSpace(writer);
            else if (this.ColorSpace == Drawing.ColorSpace.CMYK)
                writer.WriteDictionaryNameEntry("ColorSpace", "DeviceCMYK");
            else
                writer.WriteDictionaryNameEntry("ColorSpace", "DeviceRGB");

            WriteFilterNames(context, writer);

            writer.WriteDictionaryNumberEntry("BitsPerComponent", this.BitsPerColor);

            
        }

        protected virtual bool ShouldApplyFilters(PDFContextBase context)
        {
            if (this.HasFilter)
            {
                if (context.Compression == OutputCompression.Compress || this.IsPrecompressedData)
                    return true;
            }
            return false;
        }

        protected void WriteFilterNames(PDFContextBase context, PDFWriter writer)
        {
            if (this.ShouldApplyFilters(context))
            {
                if (this.Filters.Length == 1)
                    writer.WriteDictionaryNameEntry("Filter", this.Filters[0].FilterName);
                else
                {
                    writer.BeginDictionaryEntry("Filter");
                    writer.BeginArray();
                    for (int i = 0; i < this.Filters.Length; i++)
                    {
                        writer.BeginArrayEntry();
                        writer.WriteName(this.Filters[i].FilterName);
                        writer.EndArrayEntry();
                    }
                    writer.EndArray();
                    writer.EndDictionaryEntry();
                }
            }
        }

        protected virtual void RenderCustomColorSpace(PDFWriter writer)
        {
            throw new NotSupportedException("Implementors must use their own custom color space rendering");
        }

        protected virtual void RenderImageStreamData(PDFContextBase context, PDFWriter writer)
        {
            if (this.Data.LongLength > (long)int.MaxValue)
                throw new ArgumentOutOfRangeException("This image is too large to be included in a PDF file");

            byte[] data = this.Data;
            

            if (this.ShouldApplyFilters(context))
            {
                byte[] filtered;
                filtered = this.GetCachedFilteredData(this.Filters, context);
                if (null == filtered)
                {
                    filtered = ApplyFiltersToData(data, context);
                    this.SetCachedFilteredData(this.Filters, filtered, context);
                }
                data = filtered;
            }

            writer.WriteRaw(data, 0, data.Length);
        }

        protected byte[] ApplyFiltersToData(byte[] data, PDFContextBase context)
        {
            if (this.ShouldApplyFilters(context))
            {
                int origlength = data.Length;
                if (context.TraceLog.ShouldLog(TraceLevel.Debug))
                    context.TraceLog.Add(TraceLevel.Debug,"ImageData", "Applying filters to image data for '" + this.SourcePath + "'. Original length = " + origlength);

                if (this.Filters.Length == 1)
                    data = this.Filters[0].FilterStream(data);
                else
                {
                    for (int i = 0; i < this.Filters.Length; i++)
                    {
                        if (context.TraceLog.ShouldLog(TraceLevel.Debug))
                            context.TraceLog.Add(TraceLevel.Debug, "ImageData", "Applying filter '" + this.Filters[i].FilterName + "' to image data.");

                        data = this.Filters[i].FilterStream(data);
                    }
                }

                int finallength = data.Length;

                if (context.TraceLog.ShouldLog(TraceLevel.Debug))
                    context.TraceLog.Add(TraceLevel.Debug, "ImageData", "Applied all filters to image data for '" + this.SourcePath + "'. Final length = " + finallength);
                else if (context.TraceLog.ShouldLog(TraceLevel.Message))
                    context.TraceLog.Add(TraceLevel.Message, "ImageData", "Filters applied to image data for '" + this.SourcePath + "', and size has gone from " + origlength + " to " + finallength);
            }
            else if (context.TraceLog.ShouldLog(TraceLevel.Debug))
                context.TraceLog.Add(TraceLevel.Debug, "ImageData", "Should not apply filters to image data for '" + this.SourcePath + "', returning original data");

            return data;
        }

        //
        // filter caching
        // 

        private string _filtercachekey = null;
        private byte[] _filterdata = null;

        public virtual void ResetFilterCache()
        {
            this._filtercachekey = null;
            this._filterdata = null;
        }

        protected byte[] GetCachedFilteredData(IStreamFilter[] filters, PDFContextBase context)
        {
            if (null == filters || filters.Length == 0)
                return null;
            else if (null == _filterdata || null == _filtercachekey)
            {
                return null;
            }
            else
            {
                //We have a filter cacke key and some data, so check that they are the same - and if so return the data.

                string full = GetFullFilterCacheKey(filters);
                if (string.Equals(full, _filtercachekey))
                    return _filterdata;
                else
                    return null;
            }
        }

        protected virtual string GetFullFilterCacheKey(IStreamFilter[] filters)
        {
            if (null == filters || filters.Length == 0)
                throw new ArgumentNullException("filters");
            
            string full;
            if (filters.Length == 1)
                full = filters[0].FilterName;
            else
            {
                StringBuilder sb = new StringBuilder();
                foreach (IStreamFilter filter in filters)
                {
                    if (sb.Length > 0)
                        sb.Append(" ");
                    sb.Append(filter.FilterName);
                }
                full = sb.ToString();
            }
            return full;
        }

        protected void SetCachedFilteredData(IStreamFilter[] filters, byte[] data, PDFContextBase context)
        {
            if (null == filters)
                throw new ArgumentNullException("filters");
            else if (filters.Length == 0)
                throw new ArgumentOutOfRangeException("filters");

            if(null == data)
                throw new ArgumentNullException("data");

            string key = GetFullFilterCacheKey(filters);

            this._filtercachekey = key;
            this._filterdata = data;

        }


        //
        // static methods
        //

        public static PDFImageData LoadImageFromURI(string uri)
        {
             IPDFImageDataFactory factory;
             if (Scryber.Configuration.ImagingConfiguration.IsFactoryRegisterd(uri, out factory))
             {
                 PDFImageData img = factory.LoadImageData(uri);
                 return img;
             }
             else
             {
                 System.Net.WebClient wc = new System.Net.WebClient();
                 using (wc.OpenRead(uri))
                 {
                     PDFImageData img;
                     byte[] data = wc.DownloadData(uri);
                     img = InitImageData(uri, data);
                     return img;
                 }
             }
        }

        public static PDFImageData LoadImageFromLocalFile(string path)
        {
            IPDFImageDataFactory factory;
            if (Scryber.Configuration.ImagingConfiguration.IsFactoryRegisterd(path, out factory))
            {
                PDFImageData img = factory.LoadImageData(path);
                return img;
            }
            else
            {
                System.IO.FileInfo fi = new System.IO.FileInfo(path);
                if (fi.Exists == false)
                    throw new ArgumentNullException("path", "The file at the path '" + path + "' does not exist.");

                using (System.Drawing.Image bmp = System.Drawing.Image.FromFile(path))
                {
                    PDFImageData img;
                    img = InitImageData(path, bmp);
                    return img;
                }
            }
        }

        /// <summary>
        /// Creates a new PDFImageData instance from the specified bitmap.
        /// The sourceKey is a unique reference to this image data in the document
        /// </summary>
        /// <param name="sourcekey"></param>
        /// <param name="bitmap"></param>
        /// <returns></returns>
        public static PDFImageData LoadImageFromBitmap(string sourcekey, System.Drawing.Bitmap bitmap)
        {
            if (null == bitmap)
                throw new ArgumentNullException("bitmap");
            PDFImageData data = InitImageData(sourcekey, bitmap);
            
            return data;
        }

        private static PDFImageData InitImageData(string uri, byte[] data)
        {
            PDFImageData img;
            using (System.IO.MemoryStream ms = new System.IO.MemoryStream(data))
            {
                System.Drawing.Bitmap bmp = new System.Drawing.Bitmap(ms, false);
                img = InitImageData(uri, bmp);
                bmp.Dispose();
            }
            return img;
        }

        private static PDFImageData InitImageData(string uri, System.Drawing.Image bmp)
        {
            PDFImageData imgdata;

            try
            {
                Imaging.ImageParser parser = Imaging.ImageFormatParser.GetParser(bmp);
                imgdata = parser(uri, bmp);
            }
            catch (Exception ex)
            {
                throw new ArgumentException(String.Format(Errors.CouldNotParseTheImageAtPath, uri, ex.Message), "bmp", ex);
            }

            return imgdata;
        }

        

        
        

        

        

       


        

    }
}
