﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using Scryber.Drawing;
using Scryber;

namespace Scryber.Text
{
    /// <summary>
    /// Defines a root layout of a text block
    /// </summary>
    public class PDFTextLayout
    {

        #region public List<TextBlock> AllBlocks {get;}

        private List<TextBlock> _all = null;

        /// <summary>
        /// Gets the list of blocks in this layout
        /// </summary>
        public List<TextBlock> AllBlocks
        {
            get { return this._all; }
        }

        #endregion

        #region public PDFUnit Width {get;}

        /// <summary>
        /// Gets the total width of this layout block
        /// </summary>
        public PDFUnit Width
        {
            get { return (PDFUnit)this._innerblock.Width; }
        }

        #endregion

        #region public PDFUnit Height {get;}

        /// <summary>
        /// Gets the total height of this layout block
        /// </summary>
        public PDFUnit Height
        {
            get { return (PDFUnit)this._innerblock.Height; }
        }

        #endregion

        #region public PDFSize Size {get;}

        /// <summary>
        /// Gets the total size of this layout block
        /// </summary>
        public PDFSize Size
        {
            get { return new PDFSize(this.Width, this.Height); }
        }

        #endregion

        #region  public PDFSize MeasuredSize

        private PDFSize _measuredSize;

        /// <summary>
        /// Gets or sets the size of the measured space when this text was originally laid out
        /// </summary>
        public PDFSize MeasuredSize
        {
            get { return _measuredSize; }
            internal set { _measuredSize = value; }
        }

        #endregion

        #region public TextBlock InnerBlock

        private TextBlock _innerblock;

        /// <summary>
        /// Gets or sets the inner text block for this 
        /// </summary>
        public TextBlock InnerBlock
        {
            get { return this._innerblock; }
        }

        #endregion

        #region public PDFTextRenderOptions RootTextOptions {get;set;}

        private PDFTextRenderOptions _rootoptions;

        /// <summary>
        /// Gets or sets the root text rendering options
        /// </summary>
        public PDFTextRenderOptions RootTextOptions
        {
            get { return _rootoptions; }
        }

        #endregion

        #region public PDFPositionOptions RootPositionOptions

        private PDFPositionOptions _rootpos;

        /// <summary>
        /// Gets the root position options for this text block
        /// </summary>
        public PDFPositionOptions RootPositionOptions
        {
            get { return _rootpos; }
        }

        #endregion

        #region public PDFTextRenderOptions CurrentTextOptions

        private Stack<PDFTextRenderOptions> _options = null;

        /// <summary>
        /// Gets the current text options
        /// </summary>
        public PDFTextRenderOptions CurrentTextOptions
        {
            get
            {
                if (null == _options || _options.Count == 0)
                    return this.RootTextOptions;
                else
                    return _options.Peek();
            }
        }

        #endregion


        //
        // ctor
        //

        #region public PDFTextLayout(TextBlock innerblock)

        /// <summary>
        /// Creates a new PDFTextLayout instance with the specified inner block
        /// </summary>
        /// <param name="innerblock"></param>
        public PDFTextLayout(PDFTextRenderOptions roottext, PDFPositionOptions rootpos)
        {
            if (null == roottext)
                roottext = new PDFTextRenderOptions();

            if (null == rootpos)
                rootpos = new PDFPositionOptions();
            _all = new List<TextBlock>(1);
            _rootoptions = roottext;
            _rootpos = rootpos;
        }

        #endregion

        //
        // methods
        //

        #region public TextBlock BeginNewBlock(int pgIndex)

        /// <summary>
        /// Starts a new block on the specified page. 
        /// It is not supported to have more than one block on a single page
        /// </summary>
        /// <param name="pgIndex"></param>
        /// <returns></returns>
        public TextBlock BeginNewBlock(int pgIndex)
        {
            TextBlock block = new TextBlock(pgIndex);
            block.PageIndex = pgIndex;
            _all.Add(block);
            _innerblock = block;
            return block;
        }

        #endregion

        #region public TextBlock GetBlockForPage(int pageindex)

        /// <summary>
        /// Returns any block associated with that page index
        /// </summary>
        /// <param name="pageindex"></param>
        /// <returns></returns>
        public TextBlock GetBlockForPage(int pageindex)
        {
            foreach (TextBlock block in this.AllBlocks)
            {
                if (block.PageIndex == pageindex)
                    return block;
            }

            return null;
        }

        #endregion

        #region public void PushTextRenderOptions(PDFTextRenderOptions options)

        /// <summary>
        /// Pushes a new text options onto the stack making them current
        /// </summary>
        /// <param name="options">The current line options</param>
        public void PushTextRenderOptions(PDFTextRenderOptions options)
        {
            if (null == _options)
                _options = new Stack<PDFTextRenderOptions>();
            _options.Push(options);
        }

        #endregion

        #region public void PopTextRenderOptions()

        /// <summary>
        /// Pops the last text render options from the stack, making the previous options current
        /// </summary>
        public void PopTextRenderOptions()
        {
            if (null == _options || _options.Count == 0)
                throw new IndexOutOfRangeException(Errors.UnbalancedTextStyle);
            _options.Pop();
        }

        #endregion


        //
        // static factory methods
        //


        /// <summary>
        /// Initializes and returns a new text layout with the first block, paragraph and line created along with the layout variables initialized
        /// </summary>
        /// <param name="pgindex"></param>
        /// <param name="textOpts"></param>
        /// <param name="posOpts"></param>
        /// <returns></returns>
        public static PDFTextLayout InitLayout(int pgindex, double yoffset, double firstlineinset, PDFTextRenderOptions textOpts, PDFPositionOptions posOpts)
        {
            PDFTextLayout layout = new PDFTextLayout(textOpts, posOpts);
            Text.TextBlock textblock = layout.BeginNewBlock(pgindex);
            textblock.YOffset = yoffset;

            Text.TextParagraph textPara = new Text.TextParagraph();
            textblock.AddParagraph(textPara);

            Text.TextLine textline = textblock.LastParagraph.AddLine();
            textline.LineInset = firstlineinset;

            return layout;
        }
    }

}
