﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;

namespace Scryber.Text
{
    public class TextBlock : TextData, IEnumerable<TextParagraph>
    {
        #region public override double Width {get; set;}
       
        /// <summary>
        /// Gets or sets the with of this text block. 
        /// If not set this is calculated from the contained paragraphs
        /// </summary>
        /// <remarks>To clear any assigned or previously calculated values call the ResetDimensions() method</remarks>
        public override double Width
        {
            get
            {
                if (this._paras == null || this._paras.Count == 0)
                    return 0.0;

                double w = base.Width;
                if (w < 0.0)
                {
                    w = 0.0;
                    foreach (TextParagraph para in this._paras)
                    {
                        w = Math.Max(w, para.Width);
                    }
                    base.Width = w;
                }
                return w;
            }
            set
            {
                base.Width = value;
            }
        }

        #endregion

        #region public override double Height {get;set;}

        /// <summary>
        /// Gets or sets the height of this text block. 
        /// If not set this is calculated from the contained paragraphs
        /// </summary>
        /// <remarks>To clear any assigned or previously calculated values call the ResetDimensions() method</remarks>
        public override double Height
        {
            get
            {
                if (this._paras == null || this._paras.Count == 0)
                    return 0.0;
                double h = base.Height;
                if (h < 0.0)
                {
                    h = 0.0;
                    foreach (TextParagraph para in this._paras)
                    {
                        h += para.Height;
                    }
                    base.Height = h;
                }
                return h;
            }
            set
            {
                base.Height = value;
            }
        }

        #endregion

        #region public double XOffset {get;set;}

        private double _x = 0.0;

        /// <summary>
        /// Gets or sets the XOffset from the layout (in points)
        /// </summary>
        public double XOffset
        {
            get { return _x; }
            set { _x = value; }
        }

        #endregion

        #region public double YOffset {get;set;}

        private double _y = 0.0;

        /// <summary>
        /// Gets or sets the YOffset from the layout (in points)
        /// </summary>
        public double YOffset
        {
            get { return _y; }
            set { _y = value; }
        }

        #endregion

        #region public int PageIndex {get;set;}

        private int _pgIndex;

        /// <summary>
        /// Gets or sets the page index this text block sits on
        /// </summary>
        public int PageIndex
        {
            get { return _pgIndex; }
            set { _pgIndex = value; }
        }

        #endregion

        #region public TextParagraph[] Paragraphs {get;}

        private List<TextParagraph> _paras;
        /// <summary>
        /// Gets the array of paragraphs in this block 
        /// (TextBlock also implements IEnumerable on the contained collection 
        /// of paragraphs).
        /// </summary>
        public TextParagraph[] Paragraphs
        {
            get { return (_paras == null) ? (new TextParagraph[] { }) : _paras.ToArray(); }
        }

        #endregion

        #region public int ParagraphCount {get;}

        /// <summary>
        /// Gets the number of paragraphs in this TextBlock
        /// </summary>
        public int ParagraphCount
        {
            get
            {
                if (null == _paras)
                    return 0;
                else
                    return _paras.Count;
            }

        }

        #endregion

        #region public TextParagraph LastParagraph {get;}

        /// <summary>
        /// Gets the last paragraph in this text block.
        /// </summary>
        /// <exception cref="ArgumentOutOfRangeException" >Thrown if there are no paragraphs in this block</exception>
        public TextParagraph LastParagraph
        {
            get
            {
                if (null == _paras || _paras.Count == 0)
                    throw new ArgumentOutOfRangeException("index", Errors.NoParagraphsInBlock);
                else
                    return _paras[_paras.Count - 1];
            }
        }

        #endregion

        //
        // .ctor
        //

        #region public TextBlock()
        
        /// <summary>
        /// Creates a new TextBlock - collection of paragrpaphs
        /// </summary>
        public TextBlock(int pgIndex)
        {
            this._pgIndex = pgIndex;
            this.ResetDimensions();
        }

        #endregion

        //
        // methods
        //

        #region public void AddParagraph(TextParagraph para)

        /// <summary>
        /// Adds the specified TextParagraph to this block
        /// </summary>
        /// <param name="para"></param>
        ///  <remarks>This will also reset any dimensions set for the width and height of the block</remarks>
        public void AddParagraph(TextParagraph para)
        {
            if (this._paras == null)
                this._paras = new List<TextParagraph>(1);
            this._paras.Add(para);
            this.ResetDimensions();
        }

        /// <summary>
        /// Creates a new paragraph and adds it to this block
        /// </summary>
        /// <returns></returns>
        public TextParagraph AddParagraph()
        {
            TextParagraph para = new TextParagraph();
            this.AddParagraph(para);
            return para;
        }

        #endregion

        #region public void ClearParagraphs()

        /// <summary>
        /// Removes all paragraphs from this TextBlock
        /// </summary>
        ///  <remarks>This will also reset any dimensions set for the width and height of the paragraph</remarks>
        public void ClearParagraphs()
        {
            if (this._paras != null)
                this._paras.Clear();
            this.ResetDimensions();
        }

        #endregion

        /// <summary>
        /// Removes the last paragraph from the block
        /// </summary>
        public void RemoveLastParagraph()
        {
            if (null != this._paras && this._paras.Count > 0)
                this._paras.RemoveAt(this._paras.Count - 1);
        }


        #region public virtual void ResetDimensions()

        /// <summary>
        /// Clears any cached values for the width and height of this text block. 
        /// Ensuring that they will be re-calculated the next time they are retrieved
        /// </summary>
        public virtual void ResetDimensions()
        {
            base.Width = -1;
            base.Height = -1;
        }

        #endregion



        #region IEnumerable<TextParagraph> Members

        private static readonly List<TextParagraph> _empty = new List<TextParagraph>(0);

        /// <summary>
        /// Gets an emumerator for this TextBlock that will enumerate over the paragraphs in this block.
        /// </summary>
        /// <returns></returns>
        public IEnumerator<TextParagraph> GetEnumerator()
        {
            if (null == this._paras)
                return _empty.GetEnumerator();
            else
                return _paras.GetEnumerator();
        }

        #endregion

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return this.GetEnumerator();
        }

        #endregion



        public override string ToString()
        {
            if (_paras == null || _paras.Count == 0)
                return "//";
            else
            {
                StringBuilder sb = new StringBuilder();
                sb.Append("/b");
                foreach (TextParagraph para in this._paras)
                {
                    sb.Append(para.ToString().Replace('\n',' '));
                }
                sb.Append("/!b");
                return sb.ToString();
            }
        }

        public string ToString(int truncateToLength)
        {
            string full = this.ToString();

            if (full.Length > truncateToLength)
            {
                int lefthalf = truncateToLength / 2;
                int righthalf = truncateToLength - lefthalf - 3;
                string left = full.Substring(0, lefthalf);
                string right = "...";
                if (righthalf > 0)
                    right += full.Substring(full.Length - righthalf);
                full = left + right;
            }

            return full;
        }
    }
}
