﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;

namespace Scryber.Text
{
    /// <summary>
    /// Represents a line of textual data (and style changes) in a text block;
    /// </summary>
    /// <remarks>The TextLine contains TextSpan instances and TextStyle instances that can be enumerated over.
    /// You can also add entries using the AddSpan and AddStyle methods</remarks>
    public class TextLine : TextData, IEnumerable<TextData>
    {

        #region public TextData[] LineContent

        private List<TextData> _spans;

        /// <summary>
        /// Gets an array of Textdata items in this line
        /// </summary>
        public TextData[] LineContent
        {
            get { return (_spans == null) ? (new TextData[] { }) : (_spans.ToArray()); }
        }

        #endregion

        #region public int ContentCount {get;}

        /// <summary>
        /// Gets the number of items in this line
        /// </summary>
        public int ContentCount
        {
            get
            {
                return this._spans == null ? 0 : this._spans.Count;
            }
        }

        #endregion

        #region public override double Width {get;set}

        /// <summary>
        /// Gets the total Width of the line. 
        /// Setting the value will throw an exception.
        /// </summary>
        public override double Width
        {
            get
            {
                double w = base.Width;
                if (w < 0.0)
                {
                    if (this._spans == null || _spans.Count == 0)
                        return 0.0;

                    w = 0.0;
                    foreach (TextData data in this._spans)
                    {
                        //we know that TextStlye instances have a width of Zero so we are OK to do this
                        w += data.Width;
                    }
                    base.Width = w;
                }
                return w;
            }
            set
            {
                base.Width = value;
            }
        }

        #endregion

        #region public override double Height {get;set;}
        
        /// <summary>
        /// Gets or sets the height of this line
        /// </summary>
        public override double Height
        {
            get
            {
                double h = base.Height;
                if (h < 0.0)
                {
                    if (this._spans == null || _spans.Count == 0)
                        return 0.0;

                    h = 0.0;
                    foreach (TextData data in this._spans)
                    {
                        //we know that TextStlye instances have a height of Zero so we are OK to do this
                        h = Math.Max(h, data.Height);
                    }
                    base.Height = h;
                }
                return h;
            }
            set
            {
                base.Height = value;
            }
        }

        #endregion

        #region public double LineInset

        private double _lineinset;

        /// <summary>
        /// Gets or sets the inset of this line from the x offset of the block
        /// </summary>
        public double LineInset
        {
            get { return _lineinset; }
            set { _lineinset = value; }
        }

        #endregion

        #region public double AscentHeight {get;set;}

        private double _assenth;

        /// <summary>
        /// Gets or sets the assent height of the line.
        /// </summary>
        /// <remarks>This is different to the actual height, 
        /// as it is used to calculate the rendering baseline from the height of a text line</remarks>
        public double AscentHeight
        {
            get { return _assenth; }
            set { _assenth = value; }
        }

        #endregion

        //
        // methods
        //

        #region public void AddSpan(TextSpan span)

        /// <summary>
        /// Adds a span to the line
        /// </summary>
        /// <param name="span"></param>
        public void AddSpan(TextSpan span)
        {
            if (this._spans == null)
                this._spans = new List<TextData>();

            this._spans.Add(span);
            this.ResetDimensions();
        }

        #endregion

        #region public void AddStyle(TextBeginStyle style) + 1 overload

        /// <summary>
        /// Adds a style end change to the line
        /// </summary>
        /// <param name="style"></param>
        public void AddStyle(TextBeginStyle style)
        {
            if (this._spans == null)
                this._spans = new List<TextData>();

            this._spans.Add(style);
            //Don't need to reset the width as Styles have a width of Zero.
        }


        /// <summary>
        /// Adds a style end change to the line
        /// </summary>
        /// <param name="style"></param>
        public void AddStyle(TextEndStyle style)
        {
            if (this._spans == null)
                this._spans = new List<TextData>();

            this._spans.Add(style);
            //Don't need to reset the width as Styles have a width of Zero.
        }

        #endregion

        #region public void ClearContents()

        /// <summary>
        /// Removes all content from this line
        /// </summary>
        ///  <remarks>This will also reset any dimensions set for the width and height of the paragraph</remarks>
        public void ClearContents()
        {
            if (null != _spans)
                _spans.Clear();
            this.ResetDimensions();
        }

        #endregion

        #region public void ResetDimensions()

        /// <summary>
        /// Clears any cached dimensions for this lines width or height
        /// </summary>
        public void ResetDimensions()
        {
            this.Width = -1.0;
            this.Height = -1.0;
        }

        #endregion

        public int TextContentCount()
        {
            int count = 0;
            foreach (TextData data in this)
            {
                if (data is TextSpan)
                    count++;
            }
            return count;
        }

        #region IEnumerable<TextData> Members

        private static List<TextData> _empty = new List<TextData>(0);

        public IEnumerator<TextData> GetEnumerator()
        {
            if (null == _spans)
                return _empty.GetEnumerator();
            else
                return _spans.GetEnumerator();
        }

        #endregion

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            throw new NotImplementedException();
        }

        #endregion

        public override string ToString()
        {
            if (null == _spans || _spans.Count == 0)
                return "/r";
            else
            {
                StringBuilder sb = new StringBuilder();
                foreach (TextData data in this._spans)
                {
                    sb.Append(data.ToString());
                }
                sb.Append("/r");
                return sb.ToString();
            }

        }
    }
}
