﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using Scryber.Drawing;

namespace Scryber.Text
{
    /// <summary>
    /// Represents a text paragraph containing a collection of lines. 
    /// </summary>
    public class TextParagraph : TextData, IEnumerable<TextLine>
    {

        //properties

        #region public TextLine[] Lines {get;set;}

        private List<TextLine> _lines = null;

        /// <summary>
        /// Gets an array of lines in this paragraph
        /// </summary>
        public TextLine[] Lines
        {
            get { return (_lines == null) ? (new TextLine[] { }) : (_lines.ToArray()); }
        }

        #endregion

        #region public int LineCount {get;}

        /// <summary>
        /// Gets the number of lines in this paragraph
        /// </summary>
        public int LineCount
        {
            get
            {
                if (this._lines == null)
                    return 0;
                else
                    return _lines.Count;
            }
        }

        #endregion

        #region public TextLine LastLine {get;}

        /// <summary>
        /// Gets the last line in this paragraph.
        /// </summary>
        /// <exception cref="ArgumentOutOfRange" >Thrown if there are no lines in this paragraph</exception>
        public TextLine LastLine
        {
            get
            {
                if (null == _lines || _lines.Count == 0)
                    throw new ArgumentOutOfRangeException("index",Errors.NoLinesInParagraph);
                else
                    return _lines[_lines.Count - 1];
            }
        }

        #endregion



        #region public PDFUnit FirstLineInset {get;set;}

        private PDFUnit _firstinset;

        /// <summary>
        /// Gets or sets the inset of the first line from the start of the block
        /// </summary>
        public PDFUnit FirstLineInset
        {
            get
            {
                return _firstinset;
            }
            set
            {
                _firstinset = value;
            }
        }

        #endregion

        #region public override double Width {get;}

        /// <summary>
        /// Gets the total width of this paragraph (re-calculated)
        /// </summary>
        public override double Width
        {
            get
            {
                if (this.LineCount == 0)
                    return 0.0;
                double w = base.Width;

                if (w < 0.0)
                {
                    w = 0.0;
                    foreach (TextLine line in this._lines)
                    {
                        w = Math.Max(w, line.Width + line.LineInset);
                    }
                    base.Width = w;
                }
                return w;
            }
        }

        #endregion

        #region public override double Height {get;}

        /// <summary>
        /// Gets the height of this paragraph calculated based on the contained lines
        /// </summary>
        public override double Height
        {
            get
            {
                if (this.LineCount == 0)
                    return 0.0;

                double h = base.Height;
                if (h < 0.0)
                {
                    h = 0.0;
                    foreach (TextLine line in this._lines)
                    {
                        h += line.Height;
                    }
                    base.Height = h;
                }
                return h;
            }
        }

        #endregion

        /// <summary>
        /// Gets or sets the XOffset relative to the previous Paragraph
        /// </summary>
        public double XOffset
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the YOffset relative to the previous Paragraph
        /// </summary>
        public double YOffset
        {
            get;
            set;
        }

        /// <summary>
        /// True if the Y value should reset to to the top of the block
        /// </summary>
        public bool ResetY
        {
            get;
            set;
        }

        //.ctor

        #region public TextParagraph()

        /// <summary>
        /// Creates a new TextParagraph
        /// </summary>
        public TextParagraph()
        {
            this.Width = -1;
            this.Height = -1;
        }

        #endregion

        // methods

        #region public void AddLine(TextLine textLine)

        /// <summary>
        /// Adds the line to this paragraph.
        /// </summary>
        /// <param name="textLine"></param>
        /// <remarks>This will also reset any dimensions set for the width and height of the paragraph</remarks>
        public void AddLine(TextLine textLine)
        {
            if (_lines == null)
                _lines = new List<TextLine>(1);
            _lines.Add(textLine);
            this.ResetDimensions();
        }

        #endregion

        #region public TextLine AddLine()

        /// <summary>
        /// Adds a new empty line to this paragraph and returns it so that it can be added to.
        /// </summary>
        /// <returns></returns>
        /// <remarks>This will also reset any dimensions set for the width and height of the paragraph</remarks>
        public TextLine AddLine()
        {
            TextLine line = new TextLine();
            this.AddLine(line);
            this.ResetDimensions();
            return line;
        }

        #endregion

        #region public void ClearLines()

        /// <summary>
        /// Removes all the lines from this paragraph
        /// </summary>
        ///  <remarks>This will also reset any dimensions set for the width and height of the paragraph</remarks>
        public void ClearLines()
        {
            if (null != _lines)
                _lines.Clear();
            this.ResetDimensions();
        }

        #endregion

        #region public virtual void ResetDimensions()

        /// <summary>
        /// Clears any cached values for the width and height of this paragraph. 
        /// Ensuring that they will be re-calculated the next time they are retrieved
        /// </summary>
        public virtual void ResetDimensions()
        {
            base.Width = -1;
            base.Height = -1;
        }

        #endregion

        #region IEnumerable<TextLine> Members

        private static List<TextLine> _empty = new List<TextLine>(0);

        /// <summary>
        /// Creates and returns an enumerator for the collection of lines in this paragraph.
        /// </summary>
        /// <returns></returns>
        public IEnumerator<TextLine> GetEnumerator()
        {
            if (null == this._lines || _lines.Count == 0)
                return _empty.GetEnumerator();
            else
                return _lines.GetEnumerator();
        }

        #endregion

        #region IEnumerable Members

        /// <summary>
        /// Creates and returns a non-strongly typed enumerator
        /// </summary>
        /// <returns></returns>
        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return this.GetEnumerator();
        }

        #endregion

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("/p");
            foreach (TextLine line in this._lines)
            {
                sb.Append(line.ToString());
                sb.Append("\r\n");
            }
            sb.Append("/!p");
            return sb.ToString();
        }
    }
}
