﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Xml;

namespace Scryber.Generation
{
    public class PDFXMLParser : IPDFParser
    {
        private static readonly string[] TextElements = new string[] { "BR" };

        private const string PDFXProcessingInstructionsName = "scryber";
        private const string ParserLogCategory = "XML PARSER";
        private const string InheritsAttributeName = "inherits";
        private const string CodeBehindAttributeName = "code-file";
        private const string FilePathAttributeName = "source";
        private const string SelectAttributeName = "select";
        private const string XmlSchemaNamespace = "http://www.w3.org/XML/1998/namespace";
        private object _root;

        public object RootComponent
        {
            get { return _root; }
            set { _root = value; }
        }

        private PDFGeneratorSettings _settings;

        public PDFGeneratorSettings Settings
        {
            get { return _settings; }
        }

        
        public ConformanceMode Mode
        {
            get
            {
                return _settings.ConformanceMode;
            }
            set
            {
                _settings.ConformanceMode = value;
            }
        }

        #region internal bool DoLog

        private bool _dolog = true;

        /// <summary>
        /// The log setting for the parser - can be updated from the processing instruction and then becomes independant of the configured value
        /// </summary>
        internal bool DoWriteParseLog
        {
            get { return _dolog; }
        }

        #endregion

        public PDFXMLParser(PDFGeneratorSettings settings)
        {
            this._settings = settings;
        }

        public IPDFComponent Parse(string source, Stream stream, bool istemplate)
        {
            using (XmlReader reader = XmlReader.Create(stream))
            {
                return Parse(source, reader, istemplate);
            }
        }

        public IPDFComponent Parse(string source, TextReader reader, bool istemplate)
        {
            using (XmlReader xreader = XmlReader.Create(reader))
            {
                return this.Parse(source, xreader, istemplate);
            }
        }

        public virtual IPDFComponent Parse(string source, XmlReader reader, bool istemplate)
        {
            IPDFComponent parsed = null;
            while (reader.Read())
            {
                if (reader.NodeType == XmlNodeType.ProcessingInstruction)
                {
                    string name = reader.Name;
                    if (name == PDFXProcessingInstructionsName)
                        this.ParseProcessingInstructions(reader);
                }
                if (reader.NodeType == XmlNodeType.Element)
                {
                    if (istemplate)
                        LogBegin(TraceLevel.Debug, "Beginning parsing tempalate");
                    else
                        LogBegin(TraceLevel.Message, "Beginning parse of XML source '{0}'", source);

                    object value = ParseComponent(reader, true);

                    // must return a value frim this parse methos
                    if (!(value is IPDFComponent))
                        throw new InvalidCastException(String.Format(Errors.CannotConvertObjectToType, value.GetType(), typeof(IPDFComponent)));

                    parsed = (IPDFComponent)value;

                    if (istemplate)
                        LogEnd(TraceLevel.Debug, "Completed parsing template and generated component '{0}'", parsed.GetType(), source);
                    else
                        LogEnd(TraceLevel.Message, "Completed parse of XML source and generated component '{0}'", parsed.GetType(), source);
                    break;
                }
            }
            if (parsed is IPDFLoadableComponent)
            {
                IPDFLoadableComponent comp = parsed as IPDFLoadableComponent;
                comp.LoadedSource = source;
                comp.LoadType = ComponentLoadType.ReflectiveParser;
            }
            return parsed;
        }

        private object ParseComponent(XmlReader reader, bool isroot)
        {
            if (reader.NodeType != XmlNodeType.Element)
                throw new PDFParserException(Errors.CanOnlyParseComponentAsElement);
            
            bool isremote;
            ParserClassDefinition cdef;
            if (this.Mode == ConformanceMode.Strict)
                cdef = AssertGetClassDefinition(reader, out isremote);
            else if (!TryGetClassDefinition(reader, out cdef, out isremote))
            {
                //No Class definition - so skip this element
                SkipOverCurrentElement(reader);
                return null;
            }

            this.CheckFrameworkIsSupported(reader, cdef, this.Mode);

            if (isremote)
                return this.ParseRemoteComponent(cdef, reader);
            else
                return ParseComplexComponent(reader, cdef, isroot);
        }

        private void CheckFrameworkIsSupported(XmlReader reader, ParserClassDefinition cdef, ConformanceMode conformanceMode)
        {
            if (cdef.IsMinFrameworkSupported == false)
            {
                string message = string.Format(Errors.ComponentCannotBeUsedBasedOnMinVersion, reader.Name, cdef.MinRequiredFramework, PDFFramework.CurrentVersion);
                if (conformanceMode == ConformanceMode.Strict)
                    throw new PDFParserException(message);
                else
                {
                    LogAdd(TraceLevel.Error, message);
                    SkipOverCurrentElement(reader);
                }
            }
            if (cdef.IsMaxFrameworkSupported == false)
            {
                string message = string.Format(Errors.ComponentCannotBeUsedBasedOnMaxVersion, reader.Name, cdef.MaxSupportedFramework, PDFFramework.CurrentVersion);
                if (conformanceMode == ConformanceMode.Strict)
                    throw new PDFParserException(message);
                else
                {
                    LogAdd(TraceLevel.Error, message);
                    SkipOverCurrentElement(reader);
                }

            }
        }

        


        private void SkipOverCurrentElement(XmlReader reader)
        {
            if (reader.IsEmptyElement)
            {
                reader.ReadInnerXml();
            }
            else
            {
                string endname = reader.Name;
                do
                {
                    reader.Read();
                }
                while (!reader.EOF && ((reader.Name != endname) && reader.NodeType == XmlNodeType.EndElement));
            }
        }

        /// <summary>
        /// Resolves and loads a remote referenced definition
        /// </summary>
        /// <param name="cdef"></param>
        /// <param name="resolver"></param>
        /// <param name="reader"></param>
        /// <returns></returns>
        private object ParseRemoteComponent(ParserClassDefinition cdef, XmlReader reader)
        {
            string element = reader.LocalName;
            string ns = reader.NamespaceURI;
            bool empty = reader.IsEmptyElement;

            if (!reader.MoveToAttribute(FilePathAttributeName))
                throw new PDFParserException(String.Format(Errors.RequiredAttributeNoFoundOnElement, FilePathAttributeName, element));

            string path = reader.Value;
            string select = string.Empty;

            if (reader.MoveToAttribute(SelectAttributeName))
                select = reader.Value;

            object complex = this.Settings.Resolver(path, select);

           if (reader.HasAttributes && reader.AttributeCount > 1)
                this.ParseAttributes(complex, true, reader, cdef);

            if (!empty)
                this.ParseContents(complex, reader, element, ns, cdef);

            //this.ParseComplexComponent(complex, reader, cdef, resolver);
            return complex;
        }

        private object ParseComplexComponent(XmlReader reader, ParserClassDefinition cdef, bool isroot)
        {
            object value;
            LogBegin(TraceLevel.Debug, "Parsing found component '{0}' of type '{1}'", reader.Name, cdef.ClassType);

            if (reader.MoveToAttribute(InheritsAttributeName))
            {
                string typename = reader.Value;
                value = CreateInheritedInstance(typename, cdef);
                if (null == value) //fallback to default
                    value = this.CreateInstance(cdef);

                reader.MoveToFirstAttribute();
            }
            else
            {
                value = this.CreateInstance(cdef);
            }
            if (isroot && null == this.RootComponent)
                this.RootComponent = value;

            ParseComplexComponent(value, reader, cdef);

            LogEnd(TraceLevel.Debug, "Completed parsing of component of type '{0}'", cdef.ClassType);

            return value;
        }

        private object CreateInheritedInstance(string typename, ParserClassDefinition cdef)
        {
            LogAdd(TraceLevel.Debug, "Creating instance of inherited type '{0}'", typename);

            object created = null;

            if (string.IsNullOrEmpty(typename) == false)
            {
                Type found;
                try
                {
                    found = Type.GetType(typename, false);
                }
                catch (Exception ex)
                {
                    if (this.Mode == ConformanceMode.Strict)
                        throw new PDFParserException(String.Format(Errors.CannotCreateInstanceOfType, typename, ex.Message), ex);
                    else
                        LogAdd(TraceLevel.Error, Errors.CannotCreateInstanceOfType, typename, ex.Message);
                    //Default to returning an instance of the base type
                    return this.CreateInstance(cdef);
                }

                LogAdd(TraceLevel.Debug, "Loaded type from name");

                if (null == found || !cdef.ClassType.IsAssignableFrom(found))
                {
                    if (this.Mode == ConformanceMode.Strict)
                        throw new PDFParserException(String.Format(Errors.CannotCreateInstanceOfType, typename, "Cannot create or cast an object of type '" + typename + "'"));
                    else
                        LogAdd(TraceLevel.Error, Errors.CannotCreateInstanceOfType, typename, "Cannot create or cast an object of type '" + typename + "'");

                }
                else
                {
                    LogAdd(TraceLevel.Debug, "Type can be converted to base type '" + cdef.ClassType.FullName);

                    try
                    {
                        created = System.Activator.CreateInstance(found);
                    }
                    catch (Exception ex)
                    {
                        if (this.Mode == ConformanceMode.Strict)
                            throw new PDFParserException(String.Format(Errors.CannotCreateInstanceOfType, typename, ex.Message), ex);
                        else
                            LogAdd(TraceLevel.Error, Errors.CannotCreateInstanceOfType, typename, ex.Message);
                        //Default to returning an instance of the base type
                    }

                }
            }

            return created;

        }

        private void ParseComplexComponent(object component, XmlReader reader, ParserClassDefinition cdef)
        {
            string name = reader.LocalName;
            string ns = reader.NamespaceURI;

            bool isempty = reader.IsEmptyElement;

            if (reader.HasAttributes)
                ParseAttributes(component, false, reader, cdef);
            if (!isempty)
            {
                ParseContents(component, reader, name, ns, cdef);
            }
        }

        private void ParseAttributes(object container, bool isremotecomponent, XmlReader reader, ParserClassDefinition cdef)
        {
            if (!reader.MoveToFirstAttribute())
                return;
            do
            {
                string name = reader.LocalName;
                string ns = reader.NamespaceURI;

                ParserPropertyDefinition attr;
                ParserEventDefinition evt;
                if (IsSpecialAttribute(reader, isremotecomponent))
                {
                    //skip the xmlns attributes
                }
                else if (cdef.Attributes.TryGetPropertyDefinition(name, ns, out attr))
                {
                    string expression;
                    BindingType bindingtype = ParserHelper.IsBindingExpression(reader.Value, out expression);
                    if (bindingtype == BindingType.XPath)
                    {
                        GenerateXPathBindingExpression(container, cdef, attr, expression, bindingtype);
                    }
                    else if (bindingtype == BindingType.Code)
                    {
                        throw new NotSupportedException("No code type supported");
                    }
                    else
                    {
                        object actualValue = attr.GetValue(reader, this.Settings);
                        this.SetValue(container, actualValue, attr);
                    }
                }
                else if (cdef.Events.TryGetPropertyDefinition(name, out evt))
                {
                    string expression;
                    BindingType bindingtype = ParserHelper.IsBindingExpression(reader.Value, out expression);
                    if (bindingtype != BindingType.None)
                    {
                        if (this.Mode == ConformanceMode.Strict)
                            throw new PDFParserException(String.Format(Errors.CannotSpecifyBindingExpressionsOnEvents, expression, name));
                        else
                            LogAdd(TraceLevel.Error, Errors.CannotSpecifyBindingExpressionsOnEvents, expression, name);
                    }
                    else
                    {
                        AttachEventHandler(container, reader.Value, evt, cdef);
                    }
                }
                else if (this.Mode == ConformanceMode.Strict)
                    throw new PDFParserException(String.Format(Errors.ParsedTypeDoesNotContainDefinitionFor, cdef.ClassType, name, "attribute"));
                else
                    LogAdd(TraceLevel.Error, "Skipping unknown attribute '{0}' on type {1}", name, cdef.ClassType);

            } while (reader.MoveToNextAttribute());
        }

        private const string EventInvokeMethodName = "Invoke";

        private void AttachEventHandler(object container, string methodname, ParserEventDefinition evt, ParserClassDefinition cdef)
        {
            //The event handlers must be implemented on the root component
            Type roottype = this.RootComponent.GetType();

            Type evtType = evt.Event.EventHandlerType;
            System.Reflection.MethodInfo evtsignature = evtType.GetMethod(EventInvokeMethodName);

            List<Type> param = new List<Type>(2);
            foreach (System.Reflection.ParameterInfo pi in evtsignature.GetParameters())
            {
                param.Add(pi.ParameterType);
            }

            System.Reflection.MethodInfo declared = roottype.GetMethod(methodname,
                System.Reflection.BindingFlags.Instance | System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.NonPublic, null,
                param.ToArray(), null);

            if (null == declared)
            {
                if (this.Mode == ConformanceMode.Strict)
                    throw new PDFParserException(String.Format(Errors.ParsedTypeDoesNotContainDefinitionFor, roottype.FullName, methodname, "event handler"));
                else
                    LogAdd(TraceLevel.Error, Errors.ParsedTypeDoesNotContainDefinitionFor, roottype.FullName, methodname, "event handler method");
            }
            else
            {
                Delegate del = Delegate.CreateDelegate(evtType, this.RootComponent, declared);
                evt.Event.AddEventHandler(container, del);
            }

        }

        private bool IsSpecialAttribute(XmlReader reader, bool isremotecomponent)
        {
            if (reader.Depth == 1)
            {
                if (reader.LocalName == InheritsAttributeName)
                    return true;
                else if (reader.LocalName == CodeBehindAttributeName)
                    return true;
                //check code-behind and inherits
            }
            if (reader.Name.ToLower() == "xmlns" || reader.Name.ToLower().StartsWith("xmlns:"))
                return true;
            else if (isremotecomponent && (reader.LocalName.ToLower() == FilePathAttributeName))
                return true;
            else if (isremotecomponent && (reader.LocalName.ToLower() == SelectAttributeName))
                return true;
            else if (reader.NamespaceURI == XmlSchemaNamespace) //is an attribute in the XML namespace
                return true;
            else
                return false;
        }

        private void GenerateXPathBindingExpression(object container, ParserClassDefinition cdef, ParserPropertyDefinition attr, string expression, BindingType bindingtype)
        {
            //throw new NotImplementedException();
            if (container is IPDFBindableComponent)
            {
                try
                {
                    IPDFBindableComponent bindable = (IPDFBindableComponent)container;
                    PDFValueConverter conv = this.GetValueConverter(attr);
                    ParserXPathBinding binding = ParserXPathBinding.Create(expression, conv, attr.PropertyInfo);
                    bindable.DataBinding += new PDFDataBindEventHandler(binding.BindComponent);
                }
                catch (Exception ex)
                {
                    throw new InvalidOperationException("A new binding expression could not be created on container " + container.ToString(), ex);
                }
            }
            else
                throw new NotSupportedException(String.Format(Errors.DatabindingIsNotSupportedOnType, cdef.ClassType.FullName));
        }

        private PDFValueConverter GetValueConverter(ParserPropertyDefinition attr)
        {
            PDFValueConverter valueconverter;
            if (ParserDefintionFactory.IsSimpleObjectType(attr.ValueType, out valueconverter))
            {
                return valueconverter;
            }
            else if (ParserDefintionFactory.IsCustomParsableObjectType(attr.ValueType, out valueconverter))
            {
                return valueconverter;
            }
            else
                throw new NotSupportedException(String.Format(Errors.ParserAttributeMustBeSimpleOrCustomParsableType, attr.Name, attr.ValueType));

        }

        
        private void ParseContents(object container, XmlReader reader, string element, string ns, ParserClassDefinition cdef)
        {
            StringBuilder textcontent = null;

            LogAdd(TraceLevel.Debug, "Parsing container contents");
            while (reader.Read())
            {
                if (reader.NodeType == XmlNodeType.EndElement)
                {
                    if (reader.LocalName == element && reader.NamespaceURI == ns)
                        break;
                }
                else if (reader.NodeType == XmlNodeType.Element)
                {
                    ParserPropertyDefinition prop;
                    if (cdef.Elements.TryGetPropertyDefinition(reader.LocalName, reader.NamespaceURI, out prop))
                    {

                        if (prop.ParseType == DeclaredParseType.SimpleElement)
                            this.ParseSimpleElement(container, reader, prop);
                        else if (prop.ParseType == DeclaredParseType.ArrayElement)
                            this.ParseInnerCollection(container, reader, prop);
                        else if (prop.ParseType == DeclaredParseType.ComplexElement)
                            this.ParseComplexType(container, reader, prop);
                        else if (prop.ParseType == DeclaredParseType.TempateElement)
                            this.ParseTemplateContent(container, reader, prop);
                        else if (this.Mode == ConformanceMode.Strict)
                            throw new PDFParserException(String.Format(Errors.ParsedTypeDoesNotContainDefinitionFor, cdef.ClassType, reader.LocalName, "element"));
                        else
                            reader.Skip();
                    }
                    else if (cdef.DefaultElement != null)
                    {
                        if (cdef.DefaultElement.Name == reader.LocalName && ns == reader.NamespaceURI)
                            this.ParseComplexType(container, reader, cdef.DefaultElement);
                        else if (cdef.DefaultElement.ParseType == DeclaredParseType.ArrayElement)
                        {
                            
                            ParserArrayDefinition arry = (ParserArrayDefinition)cdef.DefaultElement;
                            object collection = InitArrayCollection(container, arry);
                            if (IsTextElement(reader))
                            {
                                if (null == textcontent)
                                    textcontent = new StringBuilder();
                                this.AppendTextElement(reader, textcontent);
                                this.AppendTextToCollection(textcontent, arry, collection);
                                textcontent.Length = 0;
                            }
                            else
                            {
                                object component = ParseComponent(reader, false);
                                if (null != component)
                                {
                                    LogAdd(TraceLevel.Debug, "Adding component '{0}' to default collection in property {1} of type '{2}'", component, arry.PropertyInfo.Name, cdef.ClassType);
                                    arry.AddToCollection(collection, component);
                                }
                            }

                        }
                        else if (reader.NamespaceURI == XmlSchemaNamespace)
                        {
                            //Ignore xml namespace elements
                        }
                        else if (this.Mode == ConformanceMode.Strict)
                            throw new PDFParserException(String.Format(Errors.ParsedTypeDoesNotContainDefinitionFor, cdef.ClassType, reader.LocalName, "element"));
                        else
                        {
                            LogAdd(TraceLevel.Error, "Skipping unknown element '{0}'", reader.Name);
                            reader.Skip();
                        }
                    }
                    else if(reader.NamespaceURI == XmlSchemaNamespace)
                    { 
                        //Ignore xml namespace elements
                    }
                    else if (this.Mode == ConformanceMode.Strict)
                        throw new PDFParserException(String.Format(Errors.ParsedTypeDoesNotContainDefinitionFor, cdef.ClassType, reader.LocalName, "element"));
                    else
                    {
                        LogAdd(TraceLevel.Error, "Skipping unknown element '{0}'", reader.Name);
                        reader.Skip();
                    }

                }
                else if (reader.NodeType == XmlNodeType.Text)
                {
                    if (cdef.DefaultElement != null && string.IsNullOrEmpty(cdef.DefaultElement.Name))
                    {
                        if (cdef.DefaultElement.ParseType == DeclaredParseType.SimpleElement)
                        {
                            ParserSimpleElementDefinition se = (ParserSimpleElementDefinition)cdef.DefaultElement;
                            object value = se.GetValue(reader, this.Settings);
                            this.SetValue(container, value, se);
                            break; //we have read past the end of the content of this container
                        }
                        else if (cdef.DefaultElement.ParseType == DeclaredParseType.ArrayElement)
                        {
                            //We are already positioned on the current content so it should be 
                            //included until we end the element
                            bool usecurrenttextnode = true;
                            ParseInnerCollection(container, element, ns, usecurrenttextnode, reader, cdef.DefaultElement);
                            //If we have parsed all the content of the element then we need to break off
                            if (reader.NodeType == XmlNodeType.EndElement)
                            {
                                if (reader.LocalName == element && reader.NamespaceURI == ns)
                                    break;
                            }
                        }
                        else if (this.Mode == ConformanceMode.Strict)
                            throw new PDFParserException(String.Format(Errors.ParserAttributeMustBeSimpleOrCustomParsableType, cdef.DefaultElement.PropertyInfo.Name, cdef.ClassType));
                        else
                            LogAdd(TraceLevel.Message, "Skipping text content of property '" + cdef.DefaultElement.PropertyInfo.Name + "' on class '" + cdef.ClassType + ", because it is not a simple or parsable type");
                    }
                    else if (reader.NamespaceURI == XmlSchemaNamespace)
                    {
                        //Ignore xml namespace elements
                    }
                    else if (this.Mode == ConformanceMode.Strict)
                        throw new PDFParserException(String.Format(Errors.ParsedTypeDoesNotContainDefinitionFor, cdef.ClassType, reader.LocalName, "element"));
                    else
                    {
                        LogAdd(TraceLevel.Message, "Skipping unknown element '{0}'", reader.Name);
                        reader.Skip();
                    }

                }
            }
        }


        private string ReadAllContent(XmlReader reader, string endelement, string endns)
        {
            return reader.ReadContentAsString();
        }

        private void ParseSimpleElement(object container, XmlReader reader, ParserPropertyDefinition prop)
        {
            LogAdd(TraceLevel.Debug, "Parsing simple element '{0}' for property '{2}' on class '{3}'", reader.Name, prop.Name, prop.PropertyInfo.Name, prop.PropertyInfo.DeclaringType);

            object converted = prop.GetValue(reader, this.Settings);
            this.SetValue(container, converted, prop);
        }

        private void ParseInnerCollection(object container, XmlReader reader, ParserPropertyDefinition prop)
        {
            //Extract the end element names from the current node
            string endname = reader.LocalName;
            string endns = reader.NamespaceURI;

            //we want to read the inner content
            bool parseCurrentNode = false;

            ParseInnerCollection(container, endname, endns, parseCurrentNode, reader, prop);
        }
    
        private void ParseInnerCollection(object container, string endname, string endns, bool parsecurrentNode, XmlReader reader, ParserPropertyDefinition prop)
        {
            LogAdd(TraceLevel.Debug, "Parsing inner collection from element '{0}' for property '{2}' on class '{3}'", reader.Name, prop.Name, prop.PropertyInfo.Name, prop.PropertyInfo.DeclaringType);
            bool lastwastext = false;
            StringBuilder textString = new StringBuilder();

            ParserArrayDefinition arraydefn = (ParserArrayDefinition)prop;
            

            object collection = InitArrayCollection(container, arraydefn);

            while (parsecurrentNode || reader.Read())
            {
                if (reader.NodeType == XmlNodeType.Element)
                {
                    if (IsTextElement(reader))
                    {
                        this.AppendTextElement(reader, textString);
                        lastwastext = true;
                    }
                    else
                    {
                        if (lastwastext)
                        {
                            AppendTextToCollection(textString, arraydefn, collection);
                            textString.Length = 0;
                            lastwastext = false;
                        }
                        object inner = this.ParseComponent(reader, false);
                        if (null != inner)
                        {
                            arraydefn.AddToCollection(collection, inner);
                        }
                    }
                }
                else if (reader.NodeType == XmlNodeType.EndElement)
                {
                    if (IsTextElement(reader))
                    {
                        this.AppendTextElement(reader, textString);
                        lastwastext = true;
                    }
                    if (reader.LocalName == endname && reader.NamespaceURI == endns)
                        break;
                }
                else if (reader.NodeType == XmlNodeType.Text)
                {
                    string val = reader.Value;
                    val = System.Security.SecurityElement.Escape(val);

                    textString.Append(val);
                    lastwastext = true;
                }
                parsecurrentNode = false;
            }

            if (lastwastext && textString.Length > 0)
            {
                AppendTextToCollection(textString, arraydefn, collection);
            }
        }

        private System.ComponentModel.PropertyDescriptor _textPropertyFromLiteral = null;
        private System.ComponentModel.PropertyDescriptor _formatPropertyFromLiteral = null;

        private void AppendTextToCollection(StringBuilder textString, ParserArrayDefinition arraydefn, object collection)
        {
            TextFormat format = TextFormat.XML;
            Type literaltype = this.Settings.TextLiteralType;
            object obj = CreateInstance(literaltype);
            if (!(obj is IPDFTextLiteral))
                throw new InvalidCastException(String.Format(Errors.CannotConvertObjectToType, literaltype, typeof(IPDFTextLiteral)));

            IPDFTextLiteral literal = (IPDFTextLiteral)obj;
            string textSubString = textString.ToString();
            literal.Text = textSubString;
            literal.ReaderFormat = format;

            if (textSubString.Length > 60)
                textSubString = textSubString.Substring(0, 30) + "..." + textSubString.Substring(textString.Length - 30);

            this.LogAdd(TraceLevel.Debug, "Generated text literal component with text '{0}' in format {1}", textSubString, format);
            arraydefn.AddToCollection(collection, literal);
        }

        

        private bool IsTextElement(XmlReader reader)
        {
            if (Array.IndexOf<string>(TextElements, reader.LocalName) > -1)
                    return true;
            else
                return false;
        }

        private void AppendTextElement(XmlReader reader, StringBuilder content)
        {
            if (reader.NodeType == XmlNodeType.Element)
            {
                content.Append("<");
                content.Append(reader.LocalName);
                if (reader.IsEmptyElement)
                {
                    content.Append("/>");
                }
                else
                {
                    content.Append(">");
                }
            }
            else if (reader.NodeType == XmlNodeType.EndElement)
            {
                content.Append("</");
                content.Append(reader.LocalName);
                content.Append(">");
            }
        }

        private object InitArrayCollection(object container, ParserArrayDefinition arraydefn)
        {
            object collection = arraydefn.PropertyInfo.GetValue(container, null);
            if (null == collection)
            {
                LogAdd(TraceLevel.Debug, "Collection from property is null - creating instance of type '{0}'", arraydefn.ValueType);
                collection = arraydefn.CreateInstance();
                this.SetValue(container, collection, arraydefn);
            }
            return collection;
        }

        private void ParseComplexType(object container, XmlReader reader, ParserPropertyDefinition prop)
        {
            ParserClassDefinition cdef = ParserDefintionFactory.GetClassDefinition(prop.ValueType);

            bool created = false;
            object value = this.GetValue(container, prop);
            if (null == value)
            {
                value = this.CreateInstance(cdef);
                created = true;
            }
            this.ParseComplexComponent(value, reader, cdef);

            if(created)
                this.SetValue(container, value, prop);
        }

        private void ParseTemplateContent(object container, XmlReader reader, ParserPropertyDefinition prop)
        {
            IPDFComponent comp;
            if (container is IPDFComponent)
                comp = container as IPDFComponent;
            else
                throw new InvalidCastException(Errors.TemplateComponentParentMustBeContainer);

            string all = reader.ReadInnerXml();

            object gen = CreateInstance(this.Settings.TempateGeneratorType);
            IPDFTemplateGenerator tempgen = (IPDFTemplateGenerator)gen;
            tempgen.InitTemplate(comp, all, new XmlNamespaceManager(reader.NameTable));

            prop.PropertyInfo.SetValue(container, gen, null);
        }

        private bool TryGetClassDefinition(XmlReader reader, out ParserClassDefinition cdef, out bool isremote)
        {
            LogAdd(TraceLevel.Debug, "Looking for PDFComponent declared with local name '{0}' and namespace '{1}'", reader.LocalName, reader.NamespaceURI);

            cdef = ParserDefintionFactory.GetClassDefinition(reader.LocalName, reader.NamespaceURI, false, out isremote);

            if (null == cdef)
            {
                LogAdd(TraceLevel.Error, Errors.NoTypeFoundWithPDFComponentNameInNamespace, reader.LocalName, reader.NamespaceURI);
                return false;
            }
            else
                return true;

        }

        private ParserClassDefinition AssertGetClassDefinition(XmlReader reader, out bool isremote)
        {
            LogAdd(TraceLevel.Debug, "Looking for PDFComponent declared with local name '{0}' and namespace '{1}'", reader.LocalName, reader.NamespaceURI);
            
            ParserClassDefinition cdef = ParserDefintionFactory.GetClassDefinition(reader.LocalName, reader.NamespaceURI, true, out isremote);

            if (null == cdef)
                throw new PDFParserException(String.Format(Errors.NoTypeFoundWithPDFComponentNameInNamespace, reader.LocalName, reader.NamespaceURI));
            return cdef;
        }

        /// <summary>
        /// Gets the current value for the property of the specified instance
        /// </summary>
        /// <param name="instance"></param>
        /// <param name="property"></param>
        /// <returns></returns>
        private object GetValue(object instance, ParserPropertyDefinition property)
        {
            return property.PropertyInfo.GetValue(instance, null);
        }

        private void SetValue(object instance, object value, ParserPropertyDefinition property)
        {
            if (property.PropertyInfo.CanWrite)
            {
                LogAdd(TraceLevel.Debug, "Setting value of property '{0}' on class {1} to value '{2}'", property.Name, property.PropertyInfo.DeclaringType, value);
                property.PropertyInfo.SetValue(instance, value, null);
            }
            else
                throw new PDFParserException(String.Format("The property '{0}' on class '{1}' cannot be written to. It does not have a set method.", property.Name, property.PropertyInfo.DeclaringType.Name));
            
        }

        //
        // Logging
        //

        private void LogAdd(TraceLevel level, string message, params object[] args)
        {
            if (this.DoWriteParseLog && this.Settings.Log.ShouldLog(level))
            {
                this.Settings.Log.Add(level, ParserLogCategory, SafeFormat(message, args));
            }
        }

        private void LogBegin(TraceLevel level, string message, params object[] args)
        {
            if (this.DoWriteParseLog && this.Settings.Log.ShouldLog(level))
            {
                this.Settings.Log.Begin(level, ParserLogCategory, SafeFormat(message, args));
            }
        }

        private void LogEnd(TraceLevel level, string message, params object[] args)
        {
            if (this.DoWriteParseLog && this.Settings.Log.ShouldLog(level))
            {
                this.Settings.Log.End(level, ParserLogCategory, SafeFormat(message, args));
            }
        }

        private string SafeFormat(string message, object[] args)
        {
            if (string.IsNullOrEmpty(message))
                return string.Empty;
            else if (null == args || args.Length == 0)
                return message;
            else
            {
                try
                {
                    string formatted = string.Format(message, args);
                    message = formatted;
                }
                catch (Exception)
                {
                    //Intentional sinking of exception
                }
                return message;
            }
        }

        //
        // processing instructions
        //

        public void ParseProcessingInstructions(XmlReader reader)
        {
            string value = reader.Value;
            
            
            string mode = GetProcessingValue("parser-mode", value);
            if (!string.IsNullOrEmpty(mode) && Enum.IsDefined(typeof(ConformanceMode), mode))
            {
                this.Mode = (ConformanceMode)Enum.Parse(typeof(ConformanceMode), mode);
            }

            
            string level = GetProcessingValue("parser-log", value);
            bool dolog;
            if (!string.IsNullOrEmpty(level) && bool.TryParse(level, out dolog))
            {
                this._dolog = dolog;
            }

            LogAdd(TraceLevel.Debug, "Parsed processing instructions '{0}'. Parser mode = {1}, and logging = {2}", value, this.Settings.ConformanceMode, this.DoWriteParseLog);
        }


        private string GetProcessingValue(string attributename, string full)
        {
            if (string.IsNullOrEmpty(attributename))
                return string.Empty;

            int index = full.IndexOf(attributename + "=");
            
            if (index == 0 || (index > 0 && full[index-1] == ' '))//make sure we are at a boundary
            {
                index += attributename.Length + 1;

                char separator = full[index];
                int start, end;

                if (separator == '\'' || separator == '"')
                {
                    start = index + 1;
                    end = full.IndexOf(separator, start);
                }
                else
                {
                    start = index;
                    end = full.IndexOf(' ', start);
                    if (end < start)
                        end = full.Length; 
                }

                return full.Substring(start, end - start);
            }
            return string.Empty;
        }

        internal object CreateInstance(ParserClassDefinition cdef)
        {
            if (null == cdef)
                throw new ArgumentNullException("cdef");

            return CreateInstance(cdef.ClassType);
        }

        internal object CreateInstance(Type type)
        {
            if (null == type)
                throw new ArgumentNullException("type");
            try
            {
                return Activator.CreateInstance(type);
            }
            catch (Exception ex)
            {
                throw new PDFParserException(String.Format(Errors.CannotCreateInstanceOfType, type, ex.Message), ex);
            }
        }
    }
}
