﻿/*  Copyright 2012 PerceiveIT Limited
 *  This file is part of the Scryber library.
 *
 *  You can redistribute Scryber and/or modify 
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Scryber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Scryber source code in the COPYING.txt file.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;

namespace Scryber.Generation
{
    internal abstract class ParserXPathBinding
    {
        private string _expr;
        private System.Xml.XPath.XPathExpression _compiled;
        private PDFValueConverter _converter;
        private System.Reflection.PropertyInfo _property;

        protected PDFValueConverter Converter
        {
            get { return _converter; }
        }

        protected System.Reflection.PropertyInfo Property
        {
            get { return _property; }
        }


        

        protected ParserXPathBinding()
        {
            
        }


        internal abstract void BindComponent(object sender, PDFDataBindEventArgs args);
        

        protected System.Xml.XPath.XPathExpression GetExpression(object data, PDFDataContext context)
        {
            if (null == _compiled)
            {
                _compiled = System.Xml.XPath.XPathExpression.Compile(this._expr, context.NamespaceResolver);
            }
            else
                _compiled.SetContext(context.NamespaceResolver);

            return _compiled;
        }

        internal static ParserXPathBinding Create(string expr, PDFValueConverter convert, System.Reflection.PropertyInfo property)
        {
            if (null == convert)
                throw new ArgumentNullException("convert");
            if (null == property)
                throw new ArgumentNullException("property");
            if (string.IsNullOrEmpty(expr))
                throw new ArgumentNullException("expr");

            //Run a quick compile to validat the return type
            System.Xml.XPath.XPathExpression compiled = System.Xml.XPath.XPathExpression.Compile(expr);
            ParserXPathBinding binding = null;
            switch (compiled.ReturnType)
            {

                case System.Xml.XPath.XPathResultType.Boolean:
                    binding = new ParserXPathBooleanBinding();
                    break;
                
                case System.Xml.XPath.XPathResultType.NodeSet:
                    binding = new ParserXPathNodeSetBinding();
                    break;

                case System.Xml.XPath.XPathResultType.Number:
                case System.Xml.XPath.XPathResultType.String:
                    binding = new ParserXPathValueBinding();
                    break;

                case System.Xml.XPath.XPathResultType.Error:
                    throw new PDFParserException(String.Format(Errors.InvalidXPathExpression, expr));

                case System.Xml.XPath.XPathResultType.Any:
                default:
                    throw new PDFParserException(String.Format(Errors.ReturnTypeOfXPathExpressionCouldNotBeDetermined, expr));

            }

            binding._converter = convert;
            binding._expr = expr;
            binding._compiled = compiled;
            binding._property = property;

            return binding;
        }
    }

    internal class ParserXPathNodeSetBinding : ParserXPathBinding
    {
        
        internal override void BindComponent(object sender, PDFDataBindEventArgs args)
        {
            object data = args.Context.DataStack.Current;
            System.Xml.XPath.XPathExpression expr = this.GetExpression(data, args.Context);

            System.Xml.XPath.XPathNavigator nav = (System.Xml.XPath.XPathNavigator)data;

            nav = nav.SelectSingleNode(expr.Expression, args.Context.NamespaceResolver);
            if (null != nav)
            {
                string result = nav.Value;
                object value = this.Converter(result, this.Property.PropertyType);
                this.Property.SetValue(sender, value, null);
            }
        }
    }

    internal class ParserXPathValueBinding : ParserXPathBinding
    {
        internal override void BindComponent(object sender, PDFDataBindEventArgs args)
        {
            object data = args.Context.DataStack.Current;
            System.Xml.XPath.XPathExpression expr = this.GetExpression(data, args.Context);
            System.Xml.XPath.XPathNavigator nav = (System.Xml.XPath.XPathNavigator)data;
            
            object value = nav.Evaluate(expr);
            if (null != value)
            {
                string result = value.ToString();
                value = this.Converter(result, this.Property.PropertyType);
                this.Property.SetValue(sender, value, null);
            }
        }
    }

    internal class ParserXPathBooleanBinding : ParserXPathBinding
    {
        internal override void BindComponent(object sender, PDFDataBindEventArgs args)
        {
            object data = args.Context.DataStack.Current;
            System.Xml.XPath.XPathExpression expr = this.GetExpression(data, args.Context);

            System.Xml.XPath.XPathNavigator nav = (System.Xml.XPath.XPathNavigator)data;

            bool value = (bool)nav.Evaluate(expr);
            this.Property.SetValue(sender, value, null);
            
        }
    }

    internal class ParserXPathNavigatorBinding : ParserXPathBinding
    {
        internal override void BindComponent(object sender, PDFDataBindEventArgs args)
        {
            object data = args.Context.DataStack.Current;
            System.Xml.XPath.XPathExpression expr = this.GetExpression(data, args.Context);

            System.Xml.XPath.XPathNavigator nav = (System.Xml.XPath.XPathNavigator)data;

            System.Xml.XPath.XPathNodeIterator itter = nav.Select(expr);
            if (itter.CurrentPosition < 0)
                itter.MoveNext();
            
            string value = itter.Current.Value;
            object converted = this.Converter(value, this.Property.PropertyType);

            this.Property.SetValue(sender, converted, null);

        }
    }
}
